@@ -1,19 +1,24 @@
 /* Copyright (C) 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of Aladdin Ghostscript.
+  This file is part of GNU Ghostscript.
   
-  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
-  or distributor accepts any responsibility for the consequences of using it,
-  or for whether it serves any particular purpose or works at all, unless he
-  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
-  License (the "License") for full details.
+  GNU Ghostscript is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
+  anyone for the consequences of using it or for whether it serves any
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
-  Every copy of Aladdin Ghostscript must include a copy of the License,
-  normally in a plain ASCII text file named PUBLIC.  The License grants you
-  the right to copy, modify and redistribute Aladdin Ghostscript, but only
-  under certain conditions described in the License.  Among other things, the
-  License requires that the copyright notice and this notice be preserved on
-  all copies.
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gdevbbox.c */
@@ -40,11 +45,7 @@ public_st_device_bbox();
 /* Device procedures */
 private dev_proc_open_device(bbox_open_device);
 private dev_proc_close_device(gx_forward_close_device);	/* see below */
-#ifdef TEST
 private dev_proc_output_page(bbox_output_page);
-#else
-#define bbox_output_page gx_forward_output_page
-#endif
 private dev_proc_fill_rectangle(bbox_fill_rectangle);
 private dev_proc_copy_mono(bbox_copy_mono);
 private dev_proc_copy_color(bbox_copy_color);
@@ -69,10 +70,18 @@ private dev_proc_strip_copy_rop(bbox_strip_copy_rop);
 #ifndef TEST
 private const
 #endif
-/* We initialize with a very high resolution, to prevent limitchecks */
-/* if the resolution is changed to any reasonable value. */
-/* We initialize the width and height to "infinite" values, */
-/* leaving a little room for stroke widths, rounding, etc. */
+/*
+ * The bbox device sets the resolution to some value R (currently 4000), and
+ * the page size in device pixels to slightly smaller than the largest
+ * representable values (around 500K), leaving a little room for stroke
+ * widths, rounding, etc.  If an input file (or the command line) resets the
+ * resolution to a value R' > R, the page size in pixels will get multiplied
+ * by R'/R, and will thereby exceed the representable range, causing a
+ * limitcheck.  That is why the bbox device must set the resolution to a
+ * value larger than that of any real device.  A consequence of this is that
+ * the page size in inches is limited to the maximum representable pixel
+ * size divided by R, which gives a limit of about 120" in each dimension.
+ */
 #define max_coord (min(max_int, fixed2int(max_fixed)) - 1000)
 #define max_resolution 4000
 gx_device_bbox far_data gs_bbox_device = {
@@ -234,18 +243,23 @@ bbox_open_device(gx_device *dev)
 	}
 }
 
-#ifdef TEST
 private int 
 bbox_output_page(gx_device *dev, int num_copies, int flush)
-{	gs_rect bbox;
+{
+#ifdef TEST
+	gs_rect bbox;
 
 	/* Print the page bounding box. */
 	gx_device_bbox_bbox((gx_device_bbox *)dev, &bbox);
 	dprintf2("[gdevbbox] lower left  = %f %f\n", bbox.p.x, bbox.p.y);
 	dprintf2("[gdevbbox] upper right = %f %f\n", bbox.q.x, bbox.q.y);
+#endif
+	/* Propagate the PageCount to the target, */
+	/* since it changes every time gs_output_page is called. */
+	if ( bdev->target )
+	  bdev->target->PageCount = dev->PageCount;
 	return gx_forward_output_page(dev, num_copies, flush);
 }
-#endif
 
 /* ---------------- Low-level drawing ---------------- */
 
@@ -648,7 +662,7 @@ bbox_begin_image(gx_device *dev,
 	if ( (code = gs_matrix_invert(&pim->ImageMatrix, &mat)) < 0 ||
 	     (code = gs_matrix_multiply(&mat, &ctm_only(pis), &mat)) < 0
 	   )
-	  return_error(gs_error_rangecheck);
+	  return code;
 	pbe = gs_alloc_struct(memory, bbox_image_enum, &st_bbox_image_enum,
 			      "bbox_begin_image");
 	if ( pbe == 0 )
