@@ -1,13 +1,24 @@
-/* Copyright (C) 1989, 1990, 1991, 1993, 1994 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1990, 1991, 1993, 1994, 1996 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gdevbgi.c */
@@ -77,7 +88,7 @@ private gx_device_procs bgi_procs = {
 	NULL,
 	bgi_get_xfont_procs
 };
-gx_device_bgi gs_bgi_device = {
+gx_device_bgi far_data gs_bgi_device = {
 	std_device_std_body(gx_device_bgi, &bgi_procs, "bgi",
 	  0, 0,		/* width and height are set in bgi_open */
 	  1, 1		/* density is set in bgi_open */
