@@ -1,4 +1,4 @@
-/* Copyright (C) 1990, 1995 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1990, 1995, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -209,8 +209,7 @@ bj10e_print_page(gx_device_printer *pdev, FILE *prn_stream)
 	int limit = last_row - bits_per_column;
 
 	if ( in == 0 || out == 0 )
-	{	code = gs_error_VMerror;
-		gs_note_error(code);
+	{	code = gs_note_error(gs_error_VMerror);
 		goto fin;
 	}
 
