@@ -1,4 +1,4 @@
-/* Copyright (C) 1995, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -26,6 +26,9 @@
 #include "gx.h"
 #include "gpcheck.h"
 #include "gserrors.h"
+#include "gsbittab.h"
+#include "gsropt.h"
+#include "gxdcolor.h"
 #include "gxdevice.h"
 #include "gxdevmem.h"
 #include "gxcpath.h"
@@ -80,6 +83,7 @@ gx_device_fill_in_procs(register gx_device *dev)
 	fill_dev_proc(dev, end_image, gx_default_end_image);
 	fill_dev_proc(dev, strip_tile_rectangle, gx_default_strip_tile_rectangle);
 	fill_dev_proc(dev, strip_copy_rop, gx_default_strip_copy_rop_proc);
+	fill_dev_proc(dev, get_clipping_box, gx_default_get_clipping_box);
 }
 
 int
@@ -147,45 +151,28 @@ int
 gx_default_copy_mono(gx_device *dev, const byte *data,
   int dx, int raster, gx_bitmap_id id, int x, int y, int w, int h,
   gx_color_index zero, gx_color_index one)
-{	byte invert;
+{	bool invert;
 	gx_color_index color;
-	dev_proc_fill_rectangle((*fill));
-	const byte *row;
-	int iy;
+	gx_device_color devc;
 
 	fit_copy(dev, data, dx, raster, id, x, y, w, h);
-	fill = dev_proc(dev, fill_rectangle);
 	if ( one != gx_no_color_index )
-	  { invert = 0;
+	  { invert = false;
 	    color = one;
 	    if ( zero != gx_no_color_index )
-	      { int code = (*fill)(dev, x, y, w, h, zero);
+	      { int code = (*dev_proc(dev, fill_rectangle))
+		  (dev, x, y, w, h, zero);
 		if ( code < 0 )
 		  return code;
 	      }
 	  }
 	else
-	  { invert = 0xff;
+	  { invert = true;
 	    color = zero;
 	  }
-	for ( row = data, iy = 0; iy < h; row += raster, iy++ )
-	  { int ix;
-	    for ( ix = 0; ix < w; )
-	      { int i0;
-#define row_bit(i) ((row[(i) >> 3] ^ invert) & (0x80 >> ((i) & 7)))
-		while ( ix < w && row_bit(ix + dx) == 0 )
-		  ix++;
-		for ( i0 = ix; ix < w && row_bit(ix + dx) != 0; )
-		  ix++;
-		if ( ix > i0 )
-		  { int code = (*fill)(dev, i0 + x, iy + y, ix - i0, 1, color);
-		    if ( code < 0 )
-		      return code;
-		  }
-#undef row_bit
-	      }
-	  }
-	return 0;
+	color_set_pure(&devc, color);
+	return gx_dc_default_fill_masked
+	  (&devc, data, dx, raster, id, x, y, w, h, dev, rop3_T, invert);
 }
 
 /* Implement copy_color by filling lots of small rectangles. */
@@ -478,12 +465,21 @@ gx_default_fill_mask(gx_device *orig_dev,
 	if ( lop != lop_default )
 	  { gx_color_index scolors[2];
 
-	    scolors[0] = 1;
-	    scolors[1] = 0;
+	    scolors[0] =	/* white */
+	      (*dev_proc(dev, map_rgb_color))
+	        (dev, gx_max_color_value, gx_max_color_value, gx_max_color_value);
+	    scolors[1] =	/* black */
+	      (*dev_proc(dev, map_rgb_color))
+	        (dev, (gx_color_value)0, (gx_color_value)0, (gx_color_value)0);
+	    if ( tile == 0 )
+	      colors[0] = colors[1];	/* pure color */
+	    /* We want to write only where the mask is a 1, */
+	    /* so enable source transparency. */
 	    return (*dev_proc(dev, strip_copy_rop))
 	      (dev, data, dx, raster, id, scolors, tile, colors,
 	       x, y, w, h,
-	       gx_dc_phase(pdcolor).x, gx_dc_phase(pdcolor).y, lop);
+	       gx_dc_phase(pdcolor).x, gx_dc_phase(pdcolor).y,
+	       lop | lop_S_transparent);
 	  }
 	if ( tile == 0 )
 	  { return (*dev_proc(dev, copy_mono))
@@ -498,29 +494,51 @@ gx_default_fill_mask(gx_device *orig_dev,
 	fit_copy(orig_dev, data, dx, raster, id, x, y, w, h);
 	{ dev_proc_strip_tile_rectangle((*tile_proc)) =
 	    dev_proc(dev, strip_tile_rectangle);
-	  const byte *row;
+	  const byte *row = data + (dx >> 3);
+	  int dx_bit = dx & 7;
+	  int wdx = w + dx_bit;
 	  int iy;
 
 	  for ( row = data, iy = 0; iy < h; row += raster, iy++ )
 	    { int ix;
-	      for ( ix = 0; ix < w; )
+	      for ( ix = dx_bit; ix < wdx; )
 		{ int i0;
-#define row_bit(i) (row[(i) >> 3] & (0x80 >> ((i) & 7)))
-		  while ( ix < w && row_bit(ix + dx) == 0 )
-		    ix++;
-		  for ( i0 = ix; ix < w && row_bit(ix + dx) != 0; )
-		    ix++;
-		  if ( ix > i0 )
-		    { int code = (*tile_proc)
-			(dev, tile, i0 + x, iy + y, ix - i0, 1,
-			 colors[0], colors[1],
-			 gx_dc_phase(pdcolor).x, gx_dc_phase(pdcolor).y);
-		      if ( code < 0 )
-			return code;
+		  uint b;
+		  uint len;
+		  int code;
+
+		  /* Skip 0-bits. */
+		  b = row[ix >> 3];
+		  len = byte_bit_run_length[ix & 7][b ^ 0xff];
+		  if ( len )
+		    { ix += ((len - 1) & 7) + 1;
+		      continue;
 		    }
+		  /* Scan 1-bits. */
+		  i0 = ix;
+		  for ( ; ; )
+		    { b = row[ix >> 3];
+		      len = byte_bit_run_length[ix & 7][b];
+		      if ( !len )
+			break;
+		      ix += ((len - 1) & 7) + 1;
+		      if ( ix >= wdx )
+			{ ix = wdx;
+			  break;
+			}
+		      if ( len < 8 )
+			break;
+		    }
+		  /* Now color the run from i0 to ix. */
+		  code = (*tile_proc)
+		    (dev, tile, i0 - dx_bit + x, iy + y, ix - i0, 1,
+		     colors[0], colors[1],
+		     gx_dc_phase(pdcolor).x, gx_dc_phase(pdcolor).y);
+		  if ( code < 0 )
+		    return code;
 #undef row_bit
 		}
-	  }
+	    }
 	}
 	return 0;
 }
@@ -537,21 +555,8 @@ gx_default_strip_tile_rectangle(gx_device *dev, const gx_strip_bitmap *tiles,
 	int rwidth = tiles->rep_width;
 	int rheight = tiles->rep_height;
 	int shift = tiles->shift;
-	int xoff =
-	  (shift == 0 ? px :
-	   px + (y + py) / rheight * tiles->rep_shift);
-	int irx = ((rwidth & (rwidth - 1)) == 0 ?	/* power of 2 */
-		   (x + xoff) & (rwidth - 1) :
-		   (x + xoff) % rwidth);
-	int ry = ((rheight & (rheight - 1)) == 0 ?	/* power of 2 */
-		  (y + py) & (rheight - 1) :
-		  (y + py) % rheight);
-	int icw = width - irx;
-	int ch = height - ry;
-	byte *row = tiles->data + ry * raster;
-	dev_proc_copy_mono((*proc_mono));
-	dev_proc_copy_color((*proc_color));
-	int code;
+
+	fit_fill_xy(dev, x, y, w, h);
 
 #ifdef DEBUG
 if ( gs_debug_c('t') )
@@ -570,12 +575,6 @@ if ( gs_debug_c('t') )
    }
 #endif
 
-	/*
-	 * We should really make the following check before doing
-	 * all the computations above, but since we expect devices
-	 * to implement strip_tile_rectangle rather than tile_rectangle,
-	 * the check will rarely succeed.
-	 */
 	if ( dev_proc(dev, tile_rectangle) != gx_default_tile_rectangle )
 	  { if ( shift == 0 )
 	      { /*
@@ -599,10 +598,30 @@ if ( gs_debug_c('t') )
 	    /* of window systems, but we don't yet. */
 	  }
 
-	if ( color0 == gx_no_color_index && color1 == gx_no_color_index )
-	  proc_color = dev_proc(dev, copy_color);
-	else
-	  proc_color = 0, proc_mono = dev_proc(dev, copy_mono);
+	{ /*
+	   * Note: we can't do the following computations until after
+	   * the fit_fill_xy.
+	   */
+	  int xoff =
+	    (shift == 0 ? px :
+	     px + (y + py) / rheight * tiles->rep_shift);
+	  int irx = ((rwidth & (rwidth - 1)) == 0 ?	/* power of 2 */
+		     (x + xoff) & (rwidth - 1) :
+		     (x + xoff) % rwidth);
+	  int ry = ((rheight & (rheight - 1)) == 0 ?	/* power of 2 */
+		    (y + py) & (rheight - 1) :
+		    (y + py) % rheight);
+	  int icw = width - irx;
+	  int ch = height - ry;
+	  byte *row = tiles->data + ry * raster;
+	  dev_proc_copy_mono((*proc_mono));
+	  dev_proc_copy_color((*proc_color));
+	  int code;
+
+	  if ( color0 == gx_no_color_index && color1 == gx_no_color_index )
+	    proc_color = dev_proc(dev, copy_color);
+	  else
+	    proc_color = 0, proc_mono = dev_proc(dev, copy_mono);
 
 /****** SHOULD ALSO PASS id IF COPYING A FULL TILE ******/
 #define real_copy_tile(srcx, tx, ty, tw, th)\
@@ -691,6 +710,7 @@ if ( gs_debug_c('t') )
 	   }
 #undef copy_tile
 #undef real_copy_tile
+	}
 	return 0;
 }
 
@@ -715,6 +735,21 @@ gx_default_strip_copy_rop(gx_device *dev,
 	   x, y, width, height, phase_x, phase_y, lop);
 }
 
+void
+gx_default_get_clipping_box(gx_device *dev, gs_fixed_rect *pbox)
+{	pbox->p.x = 0;
+	pbox->p.y = 0;
+	pbox->q.x = int2fixed(dev->width);
+	pbox->q.y = int2fixed(dev->height);
+}
+void
+gx_get_largest_clipping_box(gx_device *dev, gs_fixed_rect *pbox)
+{	pbox->p.x = min_fixed;
+	pbox->p.y = min_fixed;
+	pbox->q.x = max_fixed;
+	pbox->q.y = max_fixed;
+}
+
 /* The following is not really a device procedure.  See gxdevice.h. */
 
 /* Create an ordinary memory device for page or band buffering. */
@@ -755,18 +790,16 @@ gx_default_end_page(gx_device *dev, int reason, gs_state *pgs)
 
 /*
  * Implementing unaligned operations in terms of the standard aligned
- * operations requires adjusting the bitmap origin and/or the raster
- * to be aligned.  Adjusting the origin is simple, but non-portable,
- * since there is no portable way to determine the alignment of a pointer.
- * Adjusting the raster requires doing the operation one scan line at a time.
+ * operations requires adjusting the bitmap origin and/or the raster to be
+ * aligned.  Adjusting the origin is simple; adjusting the raster requires
+ * doing the operation one scan line at a time.
  */
-
 int
 gx_copy_mono_unaligned(gx_device *dev, const byte *data,
   int dx, int raster, gx_bitmap_id id, int x, int y, int w, int h,
   gx_color_index zero, gx_color_index one)
 {	dev_proc_copy_mono((*copy_mono)) = dev_proc(dev, copy_mono);
-	uint offset = (uint)(data - (const byte *)0) & (align_bitmap_mod - 1);
+	uint offset = alignment_mod(data, align_bitmap_mod);
 	int step = raster & (align_bitmap_mod - 1);
 
 	/* Adjust the origin. */
