@@ -1,13 +1,24 @@
-/* Copyright (C) 1989, 1995 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1995, 1996 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gdevm1.c */
@@ -18,6 +29,17 @@
 #include "gxdevmem.h"			/* semi-public definitions */
 #include "gdevmem.h"			/* private definitions */
 
+extern dev_proc_strip_copy_rop(mem_mono_strip_copy_rop);  /* in gdevmrop.c */
+
+/* Optionally, use the slow RasterOp implementations for testing. */
+/*#define USE_COPY_ROP*/
+
+#ifdef USE_COPY_ROP
+#include "gsrop.h"
+#endif
+
+/* ================ Standard (byte-oriented) device ================ */
+
 /* We went to a lot of trouble to optimize mem_mono_tile_rectangle. */
 /* It has a substantial effect on the total time at high resolutions. */
 /* However, it takes quite a lot of code, so we omit it on 16-bit systems. */
@@ -29,41 +51,51 @@ private dev_proc_map_color_rgb(mem_mono_map_color_rgb);
 private dev_proc_copy_mono(mem_mono_copy_mono);
 private dev_proc_fill_rectangle(mem_mono_fill_rectangle);
 #if OPTIMIZE_TILE
-private dev_proc_tile_rectangle(mem_mono_tile_rectangle);
+private dev_proc_strip_tile_rectangle(mem_mono_strip_tile_rectangle);
 #else
-#  define mem_mono_tile_rectangle gx_default_tile_rectangle
+#  define mem_mono_strip_tile_rectangle gx_default_strip_tile_rectangle
 #endif
 
 /* The device descriptor. */
 /* The instance is public. */
 const gx_device_memory far_data mem_mono_device =
-  mem_full_device("image(mono)", 0, 1, mem_open,
+  mem_full_alpha_device("image1", 0, 1, mem_open,
     mem_mono_map_rgb_color, mem_mono_map_color_rgb,
     mem_mono_copy_mono, gx_default_copy_color, mem_mono_fill_rectangle,
-    mem_mono_tile_rectangle, mem_get_bits, gx_default_map_cmyk_color);
+    mem_get_bits, gx_default_map_cmyk_color, gx_no_copy_alpha,
+    mem_mono_strip_tile_rectangle, mem_mono_strip_copy_rop);
 
-/* Map color to/from RGB.  This depends on the 'inverted' flag. */
+/* Map color to/from RGB.  This may be inverted. */
 private gx_color_index
 mem_mono_map_rgb_color(gx_device *dev, gx_color_value r, gx_color_value g,
   gx_color_value b)
-{	return gx_default_w_b_map_rgb_color(dev, r, g, b) ^
-	  (mdev->inverted & 1);
+{	return (gx_default_w_b_map_rgb_color(dev, r, g, b) ^
+		mdev->palette.data[0]) & 1;
 }
 private int
 mem_mono_map_color_rgb(gx_device *dev, gx_color_index color,
   gx_color_value prgb[3])
-{	return (gx_default_w_b_map_color_rgb(dev, color ^ (mdev->inverted & 1),
-					     prgb));
+{	return gx_default_w_b_map_color_rgb(dev,
+			(color ^ mdev->palette.data[0]) & 1,
+			prgb);
 }
 
 /* Fill a rectangle with a color. */
 private int
 mem_mono_fill_rectangle(gx_device *dev, int x, int y, int w, int h,
   gx_color_index color)
-{	fit_fill(dev, x, y, w, h);
+{
+#ifdef USE_COPY_ROP
+	return mem_mono_copy_rop(dev, NULL, 0, 0, gx_no_bitmap_id, NULL,
+				 NULL, NULL,
+				 x, y, w, h, 0, 0,
+				 (color ? rop3_1 : rop3_0));
+#else
+	fit_fill(dev, x, y, w, h);
 	bits_fill_rectangle(scan_line_base(mdev, y), x, mdev->raster,
 			    -(mono_fill_chunk)color, w, h);
 	return 0;
+#endif
 }
 
 /* Convert x coordinate to byte offset in scan line. */
@@ -88,7 +120,7 @@ mem_mono_fill_rectangle(gx_device *dev, int x, int y, int w, int h,
 /* Fetch a chunk that straddles a chunk boundary. */
 #  define cfetch2(cptr, cskew, skew)\
     (cfetch_left(cptr, cskew, skew) +\
-     cfetch_right((chunk *)(cptr) + 1, skew, cskew))
+     cfetch_right((const chunk *)(cptr) + 1, skew, cskew))
 #else				/* little-endian */
 #  define chunk bits16
 private const bits16 right_masks2[9] = {
@@ -97,12 +129,12 @@ private const bits16 right_masks2[9] = {
 private const bits16 left_masks2[9] = {
 	0xffff, 0xfefe, 0xfcfc, 0xf8f8, 0xf0f0, 0xe0e0, 0xc0c0, 0x8080, 0x0000
 };
-#  define ccont(cptr, off) (((chunk *)(cptr))[off])
+#  define ccont(cptr, off) (((const chunk *)(cptr))[off])
 #  define cfetch_right(cptr, shift, cshift)\
 	((shift) < 8 ?\
 	 ((ccont(cptr, 0) >> (shift)) & right_masks2[shift]) +\
 	  (ccont(cptr, 0) << (cshift)) :\
-	 ((chunk)*(byte *)(cptr) << (cshift)) & 0xff00)
+	 ((chunk)*(const byte *)(cptr) << (cshift)) & 0xff00)
 #  define cfetch_left(cptr, shift, cshift)\
 	((shift) < 8 ?\
 	 ((ccont(cptr, 0) << (shift)) & left_masks2[shift]) +\
@@ -115,7 +147,7 @@ private const bits16 left_masks2[9] = {
 	((cskew) < 8 ?\
 	 ((ccont(cptr, 0) << (cskew)) & left_masks2[cskew]) +\
 	  (ccont(cptr, 0) >> (skew)) +\
-	  (((chunk)(((byte *)(cptr))[2]) << (cskew)) & 0xff00) :\
+	  (((chunk)(((const byte *)(cptr))[2]) << (cskew)) & 0xff00) :\
 	 (((ccont(cptr, 0) & 0xff00) >> (skew)) & 0xff) +\
 	  ((ccont(cptr, 1) >> (skew)) & right_masks2[skew]) +\
 	   (ccont(cptr, 1) << (cskew)))
@@ -123,7 +155,7 @@ private const bits16 left_masks2[9] = {
 /* Since source and destination are both always big-endian, */
 /* fetching an aligned chunk never requires byte swapping. */
 #  define cfetch_aligned(cptr)\
-	(*(chunk *)(cptr))
+	(*(const chunk *)(cptr))
 
 /* copy_function and copy_shift get added together for dispatch */
 typedef enum {
@@ -154,7 +186,17 @@ private int
 mem_mono_copy_mono(gx_device *dev,
   const byte *base, int sourcex, int sraster, gx_bitmap_id id,
   int x, int y, int w, int h, gx_color_index zero, gx_color_index one)
-{	register const byte *bptr;		/* actually chunk * */
+{
+#ifdef USE_COPY_ROP
+	return mem_mono_copy_rop(dev, base, sourcex, sraster, id, NULL,
+				 NULL, NULL,
+				 x, y, w, h, 0, 0,
+				 ((zero == gx_no_color_index ? rop3_D :
+				   zero == 0 ? rop3_0 : rop3_1) & ~rop3_S) |
+				 ((one ==  gx_no_color_index ? rop3_D :
+				   one == 0 ? rop3_0 : rop3_1) & rop3_S));
+#else				/* !USE_COPY_ROP */
+	register const byte *bptr;		/* actually chunk * */
 	int dbit, wleft;
 	uint mask;
 	copy_mode mode;
@@ -375,20 +417,33 @@ uand:				    write_unaligned(write_and, write_and_masked);
 #undef next_x_chunk
 	return 0;
 	/* Handle the funny cases that aren't supposed to happen. */
-funny:	return (invert ? -1 : mem_mono_fill_rectangle(dev, x, y, w, h, zero));
+funny:	return (invert ? gs_note_error(-1) :
+		mem_mono_fill_rectangle(dev, x, y, w, h, zero));
 #undef optr
+#endif				/* !USE_COPY_ROP */
 }
 
 #if OPTIMIZE_TILE		/**************** ****************/
 
-/* Tile with a monochrome halftone. */
+/* Strip-tile with a monochrome halftone. */
 /* This is a performance bottleneck for monochrome devices, */
 /* so we re-implement it, even though it takes a lot of code. */
 private int
-mem_mono_tile_rectangle(gx_device *dev, register const gx_tile_bitmap *tile,
+mem_mono_strip_tile_rectangle(gx_device *dev,
+  register const gx_strip_bitmap *tiles,
   int tx, int y, int tw, int th, gx_color_index color0, gx_color_index color1,
   int px, int py)
-{	register uint invert;
+{
+#ifdef USE_COPY_ROP
+	return mem_mono_strip_copy_rop(dev, NULL, 0, 0, tile->id, NULL,
+				tiles, NULL,
+				tx, y, tw, th, px, py,
+				((color0 == gx_no_color_index ? rop3_D :
+				  color0 == 0 ? rop3_0 : rop3_1) & ~rop3_T) |
+				((color1 ==  gx_no_color_index ? rop3_D :
+				  color1 == 0 ? rop3_0 : rop3_1) & rop3_T));
+#else				/* !USE_COPY_ROP */
+	register uint invert;
 	int sraster;
 	uint tile_bits_size;
 	const byte *base;
@@ -402,15 +457,16 @@ mem_mono_tile_rectangle(gx_device *dev, register const gx_tile_bitmap *tile,
 #define optr ((chunk *)dbptr)
 	register int skew;
 
-	if ( color0 != (color1 ^ 1) )
-	  return gx_default_tile_rectangle(dev, tile, tx, y, tw, th,
-					   color0, color1, px, py);
+	/* This implementation doesn't handle strips yet. */
+	if ( color0 != (color1 ^ 1) || tiles->shift != 0 )
+	  return gx_default_strip_tile_rectangle(dev, tiles, tx, y, tw, th,
+						 color0, color1, px, py);
 	fit_fill(dev, tx, y, tw, th);
 	invert = -(uint)color0;
-	sraster = tile->raster;
-	base = tile->data + ((y + py) % tile->rep_height) * sraster;
-	tile_bits_size = tile->size.y * sraster;
-	end = tile->data + tile_bits_size;
+	sraster = tiles->raster;
+	base = tiles->data + ((y + py) % tiles->rep_height) * sraster;
+	tile_bits_size = tiles->size.y * sraster;
+	end = tiles->data + tile_bits_size;
 #undef end_y_loop
 #define end_y_loop(sdelta, ddelta)\
   if ( --h == 0 ) break;\
@@ -426,8 +482,8 @@ mem_mono_tile_rectangle(gx_device *dev, register const gx_tile_bitmap *tile,
 	 * copy of the tile.  Note that all iterations except the first
 	 * have sourcex = 0.
 	 */
-	{ int sourcex = (x + px) % tile->rep_width;
-	  w = tile->size.x - sourcex;
+	{ int sourcex = (x + px) % tiles->rep_width;
+	  w = tiles->size.x - sourcex;
 	  bptr = base + ((sourcex & ~chunk_align_bit_mask) >> 3);
 	  dbit = x & chunk_align_bit_mask;
 	  skew = dbit - (sourcex & chunk_align_bit_mask);
@@ -503,13 +559,19 @@ outer:	if ( w > rw )
 #define next_x_chunk\
   bptr += chunk_bytes; dbptr += chunk_bytes
 
-		set_mono_left_mask(mask, dbit);
 		set_mono_right_mask(rmask, wleft & chunk_bit_mask);
 		if ( skew == 0 )	/* optimize the aligned case */
-		   {	for ( ; ; )
+		   {	if ( dbit == 0 )
+			  mask = 0;
+			else
+			  set_mono_left_mask(mask, dbit);
+			for ( ; ; )
 			  { int count = wleft;
 			    /* Do first partial chunk. */
-			    write_store_masked(cfetch_aligned(bptr), mask, 0);
+			    if ( mask )
+			      write_store_masked(cfetch_aligned(bptr), mask, 0);
+			    else
+			      write_store(cfetch_aligned(bptr));
 			    /* Do full chunks. */
 			    while ( (count -= chunk_bits) >= 0 )
 			      { next_x_chunk;
@@ -527,7 +589,9 @@ outer:	if ( w > rw )
 			  (skew >= 0 ? true :
 			   ((bptr += chunk_bytes), false));
 			int cskew = -skew & chunk_bit_mask;
+
 			skew &= chunk_bit_mask;
+			set_mono_left_mask(mask, dbit);
 			for ( ; ; )
 			   {	int count = wleft;
 				if ( case_right )
@@ -558,12 +622,69 @@ outer:	if ( w > rw )
 #undef optr
 	if ( (rw -= w) > 0 )
 	  {	x += w;
-		w = tile->size.x;
+		w = tiles->size.x;
 		bptr = base;
 		skew = dbit = x & chunk_align_bit_mask;
 		goto outer;
 	  }
 	return 0;
+#endif				/* !USE_COPY_ROP */
 }
 
 #endif				/**************** ****************/
+
+/* ================ "Word"-oriented device ================ */
+
+/* Note that on a big-endian machine, this is the same as the */
+/* standard byte-oriented-device. */
+
+#if !arch_is_big_endian
+
+/* Procedures */
+private dev_proc_copy_mono(mem1_word_copy_mono);
+private dev_proc_fill_rectangle(mem1_word_fill_rectangle);
+#define mem1_word_strip_tile_rectangle gx_default_strip_tile_rectangle
+
+/* Here is the device descriptor. */
+const gx_device_memory far_data mem_mono_word_device =
+  mem_full_alpha_device("image1w", 0, 1, mem_open,
+    mem_mono_map_rgb_color, mem_mono_map_color_rgb,
+    mem1_word_copy_mono, gx_default_copy_color, mem1_word_fill_rectangle,
+    mem_word_get_bits, gx_default_map_cmyk_color, gx_no_copy_alpha,
+    mem1_word_strip_tile_rectangle, gx_no_strip_copy_rop);
+
+/* Fill a rectangle with a color. */
+private int
+mem1_word_fill_rectangle(gx_device *dev, int x, int y, int w, int h,
+  gx_color_index color)
+{	byte *base;
+	uint raster;
+	fit_fill(dev, x, y, w, h);
+	base = scan_line_base(mdev, y);
+	raster = mdev->raster;
+	mem_swap_byte_rect(base, raster, x, w, h, true);
+	bits_fill_rectangle(base, x, raster, -(mono_fill_chunk)color, w, h);
+	mem_swap_byte_rect(base, raster, x, w, h, true);
+	return 0;
+}
+
+/* Copy a bitmap. */
+private int
+mem1_word_copy_mono(gx_device *dev,
+  const byte *base, int sourcex, int sraster, gx_bitmap_id id,
+  int x, int y, int w, int h, gx_color_index zero, gx_color_index one)
+{	byte *row;
+	uint raster;
+	bool store;
+	fit_copy(dev, base, sourcex, sraster, id, x, y, w, h);
+	row = scan_line_base(mdev, y);
+	raster = mdev->raster;
+	store = (zero != gx_no_color_index && one != gx_no_color_index);
+	mem_swap_byte_rect(row, raster, x, w, h, store);
+	mem_mono_copy_mono(dev, base, sourcex, sraster, id,
+			   x, y, w, h, zero, one);
+	mem_swap_byte_rect(row, raster, x, w, h, false);
+	return 0;
+}
+
+#endif				/* !arch_is_big_endian */
