@@ -1,4 +1,4 @@
-/* Copyright (C) 1989, 1992, 1993 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1992, 1993, 1994 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -26,6 +26,10 @@
 
 #define MGR_RESERVEDCOLORS 16
 
+/* Color mapping routines for 8-bit color (with a fixed palette). */
+dev_proc_map_rgb_color(mgr_8bit_map_rgb_color);
+dev_proc_map_color_rgb(mgr_8bit_map_color_rgb);
+
 
 /* extract from dump.h */
 
