@@ -35,7 +35,9 @@
 #include "gxdevrop.h"
 #include "gdevmrop.h"
 
-#define mdev ((gx_device_memory *)dev)
+/* Define whether we implement transparency correctly, or whether we */
+/* implement it as documented in the H-P manuals. */
+#define TRANSPARENCY_PER_H_P
 
 /**************** NOTE: ****************
  * The 2- and 4-bit cases don't handle transparency right.
@@ -43,9 +45,10 @@
  * The 16- and 32-bit cases aren't implemented.
  ***************** ****************/
 
+#define mdev ((gx_device_memory *)dev)
+
 /* Forward references */
-private gs_rop3_t gs_transparent_rop(P3(gs_rop3_t rop, bool source_transparent,
-  bool pattern_transparent));
+private gs_rop3_t gs_transparent_rop(P1(gs_logical_operation_t lop));
 
 #define chunk byte
 
@@ -119,7 +122,7 @@ mem_mono_strip_copy_rop(gx_device *dev,
   const gx_strip_bitmap *textures, const gx_color_index *tcolors,
   int x, int y, int width, int height,
   int phase_x, int phase_y, gs_logical_operation_t lop)
-{	gs_rop3_t rop = lop & lop_rop_mask;
+{	gs_rop3_t rop = gs_transparent_rop(lop); /* handle transparency */
 	gx_strip_bitmap no_texture;
 	bool invert;
 	uint draster = mdev->raster;
@@ -150,10 +153,6 @@ mem_mono_strip_copy_rop(gx_device *dev,
 			    height, "initial dest bits");
 #endif
 
-	/* Handle source and destination transparency. */
-	rop = gs_transparent_rop(rop, lop & lop_S_transparent,
-				 lop & lop_T_transparent);
-
 	/*
 	 * RasterOp is defined as operating in RGB space; in the monobit
 	 * case, this means black = 0, white = 1.  However, most monobit
@@ -165,6 +164,11 @@ mem_mono_strip_copy_rop(gx_device *dev,
 	if ( invert )
 	  rop = byte_reverse_bits[rop] ^ 0xff;
 
+	/*
+	 * From this point on, rop works in terms of device pixel values,
+	 * not RGB-space values.
+	 */
+
 	/* Modify the raster operation according to the source palette. */
 	if ( scolors != 0 )
 	{	/* Source with palette. */
@@ -189,28 +193,87 @@ mem_mono_strip_copy_rop(gx_device *dev,
 		}
 	}
 
-	/* Handle constant source and/or texture. */
-	if ( rop3_uses_S(rop) )
-	{	fit_copy(dev, sdata, sourcex, sraster, id,
-			 x, y, width, height);
-	}
-	else
-	{	/* Source is not used; sdata et al may be garbage. */
-		sdata = mdev->base;	/* arbitrary, as long as all */
+	/* Handle constant source and/or texture, and other special cases. */
+	{ gx_color_index color0, color1;
+
+	switch ( rop_usage_table[rop] )
+	  {
+	  case rop_usage_none:
+	    /* We're just filling with a constant. */
+	    return (*dev_proc(dev, fill_rectangle))
+	      (dev, x, y, width, height, (gx_color_index)(rop & 1));
+	  case rop_usage_D:
+	    /* This is either D (no-op) or ~D. */
+	    if ( rop == rop3_D )
+	      return 0;
+	    /* Code no_S inline, then finish with no_T. */
+	    fit_fill(dev, x, y, width, height);
+	    sdata = mdev->base;
+	    sourcex = x;
+	    sraster = 0;
+	    goto no_T;
+	  case rop_usage_S:
+	    /* This is either S or ~S, which copy_mono can handle. */
+	    if ( rop == rop3_S )
+	      color0 = 0, color1 = 1;
+	    else
+	      color0 = 1, color1 = 0;
+do_copy:    return (*dev_proc(dev, copy_mono))
+	      (dev, sdata, sourcex, sraster, id, x, y, width, height,
+	       color0, color1);
+	  case rop_usage_DS:
+	    /* This might be a case that copy_mono can handle. */
+#define copy_case(c0, c1) color0 = c0, color1 = c1; goto do_copy;
+	    switch ( (uint)rop ) /* cast shuts up picky compilers */
+	      {
+	      case rop3_D & rop3_not(rop3_S): copy_case(gx_no_color_index, 0);
+	      case rop3_D | rop3_S: copy_case(gx_no_color_index, 1);
+	      case rop3_D & rop3_S: copy_case(0, gx_no_color_index);
+	      case rop3_D | rop3_not(rop3_S): copy_case(1, gx_no_color_index);
+	      default: ;
+	      }
+#undef copy_case
+	    fit_copy(dev, sdata, sourcex, sraster, id, x, y, width, height);
+no_T:	    /* Texture is not used; textures may be garbage. */
+	    no_texture.data = mdev->base;	/* arbitrary */
+	    no_texture.raster = 0;
+	    no_texture.size.x = width;
+	    no_texture.size.y = height;
+	    no_texture.rep_width = no_texture.rep_height = 1;
+	    no_texture.rep_shift = no_texture.shift = 0;
+	    textures = &no_texture;
+	    break;
+	  case rop_usage_T:
+	    /* This is either T or ~T, which tile_rectangle can handle. */
+	    if ( rop == rop3_T )
+	      color0 = 0, color1 = 1;
+	    else
+	      color0 = 1, color1 = 0;
+do_tile:    return (*dev_proc(dev, strip_tile_rectangle))
+	      (dev, textures, x, y, width, height, color0, color1,
+	       phase_x, phase_y);
+	  case rop_usage_DT:
+	    /* This might be a case that tile_rectangle can handle. */
+#define tile_case(c0, c1) color0 = c0, color1 = c1; goto do_tile;
+	    switch ( (uint)rop ) /* cast shuts up picky compilers */
+	      {
+	      case rop3_D & rop3_not(rop3_T): tile_case(gx_no_color_index, 0);
+	      case rop3_D | rop3_T: tile_case(gx_no_color_index, 1);
+	      case rop3_D & rop3_T: tile_case(0, gx_no_color_index);
+	      case rop3_D | rop3_not(rop3_T): tile_case(1, gx_no_color_index);
+	      default: ;
+	      }
+#undef tile_case
+	    fit_fill(dev, x, y, width, height);
+	    /* Source is not used; sdata et al may be garbage. */
+	    sdata = mdev->base;		/* arbitrary, as long as all */
 					/* accesses are valid */
-		sourcex = x;		/* guarantee no source skew */
-		sraster = 0;
-		fit_fill(dev, x, y, width, height);
-	}
-	if ( !rop3_uses_T(rop) )
-	{	/* Texture is not used; texture may be garbage. */
-		no_texture.data = mdev->base;	/* arbitrary */
-		no_texture.raster = 0;
-		no_texture.size.x = width;
-		no_texture.size.y = height;
-		no_texture.rep_width = no_texture.rep_height = 1;
-		no_texture.rep_shift = no_texture.shift = 0;
-		textures = &no_texture;
+	    sourcex = x;		/* guarantee no source skew */
+	    sraster = 0;
+	    break;
+	  default:		/* rop_usage_[D]ST */
+	    fit_copy(dev, sdata, sourcex, sraster, id, x, y, width, height);
+	  }
 	}
 
 #ifdef DEBUG
@@ -226,15 +289,15 @@ mem_mono_strip_copy_rop(gx_device *dev,
 
 	/* Loop over scan lines. */
 	for ( ; line_count-- > 0; drow += draster, srow += sraster, ++ty )
-	{	/* Loop over copies of the tile. */
-		int sx = sourcex;
+	{	int sx = sourcex;
 		int dx = x;
 		int w = width;
 		const byte *trow =
-		  textures->data + (ty % textures->size.y) * traster;
+		  textures->data + (ty % textures->rep_height) * traster;
 		int xoff = x_offset(phase_x, ty, textures);
 		int nw;
 
+		/* Loop over (norizontal) copies of the tile. */
 		for ( ; w > 0; sx += nw, dx += nw, w -= nw )
 		{	int dbit = dx & 7;
 			int sbit = sx & 7;
@@ -266,7 +329,7 @@ mem_mono_strip_copy_rop(gx_device *dev,
 				byte tbyte = fetch1(tptr, tskew);
 #undef fetch1
 				byte result =
-				  (*rop_proc_tab[rop])(dbyte, sbyte, tbyte);
+				  (*rop_proc_table[rop])(dbyte, sbyte, tbyte);
 				if ( left <= nx )
 				  mask &= rmask;
 				*dptr = (mask == 0xff ? result :
@@ -367,7 +430,7 @@ mem_gray8_rgb24_strip_copy_rop(gx_device *dev,
   const gx_strip_bitmap *textures, const gx_color_index *tcolors,
   int x, int y, int width, int height,
   int phase_x, int phase_y, gs_logical_operation_t lop)
-{	gs_rop3_t rop = lop & lop_rop_mask;
+{	gs_rop3_t rop = lop_rop(lop);
 	gx_color_index const_source = gx_no_color_index;
 	gx_color_index const_texture = gx_no_color_index;
 	uint draster = mdev->raster;
@@ -437,7 +500,7 @@ mem_gray8_rgb24_strip_copy_rop(gx_device *dev,
        t_pixel == ttrans	/* Po = 0, p_tr = 1 */\
      )\
     continue;\
-  *dptr = (*rop_proc_tab[rop])(*dptr, s_pixel, t_pixel)
+  *dptr = (*rop_proc_table[rop])(*dptr, s_pixel, t_pixel)
 /* 24-bit */
 #define get24(ptr)\
   (((gx_color_index)(ptr)[0] << 16) | ((gx_color_index)(ptr)[1] << 8) | (ptr)[2])
@@ -453,7 +516,7 @@ mem_gray8_rgb24_strip_copy_rop(gx_device *dev,
      )\
     continue;\
   { gx_color_index d_pixel = get24(dptr);\
-    d_pixel = (*rop_proc_tab[rop])(d_pixel, s_pixel, t_pixel);\
+    d_pixel = (*rop_proc_table[rop])(d_pixel, s_pixel, t_pixel);\
     put24(dptr, d_pixel);\
   }
 
@@ -684,12 +747,11 @@ gx_real_default_strip_copy_rop(gx_device *dev,
 	 * pixels, the memory device implementation to do the operation,
 	 * and copy_color to write the pixels back.
 	 */
-	gs_rop3_t rop = lop & lop_rop_mask;
 	int depth = dev->color_info.depth;
 	const gx_device_memory *mdproto = gdev_mem_device_for_bits(depth);
 	gx_device_memory mdev;
 	uint draster = gx_device_raster(dev, true);
-	bool uses_d = rop3_uses_D(rop);				  
+	bool uses_d = rop3_uses_D(gs_transparent_rop(lop));
 	byte *row;
 	int code;
 	int py;
@@ -703,6 +765,12 @@ gx_real_default_strip_copy_rop(gx_device *dev,
 #endif
 	if ( mdproto == 0 )
 	  return_error(gs_error_rangecheck);
+	if ( sdata == 0 )
+	  { fit_fill(dev, x, y, width, height);
+	  }
+	else
+	  { fit_copy(dev, sdata, sourcex, sraster, id, x, y, width, height);
+	  }
 	gs_make_mem_device(&mdev, mdproto, 0, -1, dev);
 	mdev.width = width;
 	mdev.height = 1;
@@ -872,30 +940,43 @@ private dev_proc_copy_color(rop_texture_copy_color);
 private const gx_device_rop_texture far_data gs_rop_texture_device =
 {	std_device_std_body(gx_device_rop_texture, 0, "rop source",
 	  0, 0, 1, 1),
-	{	gx_default_open_device,
-		NULL,		/* get_initial_matrix */
-		gx_default_sync_output,
-		gx_default_output_page,
-		gx_default_close_device,
-		NULL,		/* map_rgb_color */
-		NULL,		/* map_color_rgb */
+	{	NULL,		/* open_device */
+		gx_forward_get_initial_matrix,
+		NULL,		/* default_sync_output */
+		NULL,		/* output_page */
+		NULL,		/* close_device */
+		gx_forward_map_rgb_color,
+		gx_forward_map_color_rgb,
 		rop_texture_fill_rectangle,
-		gx_default_tile_rectangle,
+		NULL,		/* tile_rectangle */
 		rop_texture_copy_mono,
 		rop_texture_copy_color,
-		gx_default_draw_line,
+		NULL,		/* draw_line */
 		NULL,		/* get_bits */
-		NULL,		/* get_params */
-		NULL,		/* put_params */
-		NULL,		/* map_cmyk_color */
-		NULL,		/* get_xfont_procs */
-		NULL,		/* get_xfont_device */
-		NULL,		/* map_rgb_alpha_color */
-		NULL,		/* get_page_device */
-		gx_default_get_alpha_bits,	/* (no alpha) */
+		gx_forward_get_params,
+		gx_forward_put_params,
+		gx_forward_map_cmyk_color,
+		gx_forward_get_xfont_procs,
+		gx_forward_get_xfont_device,
+		gx_forward_map_rgb_alpha_color,
+		gx_forward_get_page_device,
+		NULL,		/* get_alpha_bits (no alpha) */
 		gx_no_copy_alpha,		/* shouldn't be called */
-		NULL,		/* get_band */
-		gx_no_copy_rop			/* shouldn't be called */
+		gx_forward_get_band,
+		gx_no_copy_rop,			/* shouldn't be called */
+		NULL,		/* fill_path */
+		NULL,		/* stroke_path */
+		NULL,		/* fill_mask */
+		NULL,		/* fill_trapezoid */
+		NULL,		/* fill_parallelogram */
+		NULL,		/* fill_triangle */
+		NULL,		/* draw_thin_line */
+		NULL,		/* begin_image */
+		NULL,		/* image_data */
+		NULL,		/* end_image */
+		NULL,		/* strip_tile_rectangle */
+		NULL,		/* strip_copy_rop */
+		gx_forward_get_clipping_box
 	},
 	0,			/* target */
 	lop_default,		/* log_op */
@@ -908,7 +989,8 @@ void
 gx_make_rop_texture_device(gx_device_rop_texture *dev, gx_device *target,
   gs_logical_operation_t log_op, const gx_device_color *texture)
 {	*dev = gs_rop_texture_device;
-	gx_device_forward_fill_in_procs((gx_device_forward *)dev);
+	/* Drawing operations are defaulted, non-drawing are forwarded. */
+	gx_device_fill_in_procs((gx_device *)dev);
 	dev->color_info = target->color_info;
 	dev->target = target;
 	dev->log_op = log_op;
@@ -926,6 +1008,7 @@ rop_texture_fill_rectangle(gx_device *dev, int x, int y, int w, int h,
 	source.sraster = 0;
 	source.id = gx_no_bitmap_id;
 	source.scolors[0] = source.scolors[1] = color;
+	source.use_scolors = true;
 	return gx_device_color_fill_rectangle(rtdev->texture,
 				x, y, w, h, rtdev->target,
 				rtdev->log_op, &source);
@@ -946,6 +1029,7 @@ rop_texture_copy_mono(gx_device *dev,
 	source.id = id;
 	source.scolors[0] = color0;
 	source.scolors[1] = color1;
+	source.use_scolors = true;
 	/* Adjust the logical operation per transparent colors. */
 	if ( color0 == gx_no_color_index )
 	  lop = rop3_use_D_when_S_0(lop);
@@ -968,6 +1052,7 @@ rop_texture_copy_color(gx_device *dev,
 	source.sraster = raster;
 	source.id = id;
 	source.scolors[0] = source.scolors[1] = gx_no_color_index;
+	source.use_scolors = true;
 	return gx_device_color_fill_rectangle(rtdev->texture,
 				x, y, w, h, rtdev->target,
 				rtdev->log_op, &source);
@@ -978,11 +1063,14 @@ rop_texture_copy_color(gx_device *dev,
 /* Compute the effective RasterOp for the 1-bit case, */
 /* taking transparency into account. */
 private gs_rop3_t
-gs_transparent_rop(gs_rop3_t rop, bool source_transparent,
-  bool pattern_transparent)
-{	/*
+gs_transparent_rop(gs_logical_operation_t lop)
+{	gs_rop3_t rop = lop_rop(lop);
+	/*
 	 * The algorithm for computing an effective RasterOp is presented,
 	 * albeit obfuscated, in the H-P PCL5 technical documentation.
+	 * Define So ("source opaque") and Po ("pattern opaque") as masks
+	 * that have 1-bits precisely where the source or pattern
+	 * respectively are not white (transparent).
 	 * One applies the original RasterOp to compute an intermediate
 	 * result R, and then computes the final result as
 	 * (R & M) | (D & ~M) where M depends on transparencies as follows:
@@ -991,23 +1079,30 @@ gs_transparent_rop(gs_rop3_t rop, bool source_transparent,
 	 *	 0	 1	~So | Po (? Po ?)
 	 *	 1	 0	So
 	 *	 1	 1	So & Po
-	 * or equivalently
-	 *	So	Po	M
-	 *	 0	 0	~s_tr (? ~s_tr & ~p_tr ?)
-	 *	 0	 1	~s_tr
-	 *	 1	 0	~p_tr
-	 *	 1	 1	1
 	 * The s_tr = 0, p_tr = 1 case seems wrong, but it's clearly
 	 * specified that way in the "PCL 5 Color Technical Reference
-	 * Manual."  So and Po are "source opaque" and "pattern opaque";
-	 * in the uninverted 1-bit case with black = 0, these are
-	 * equivalent to ~S and ~P.
+	 * Manual."
+	 *
+	 * In the 1-bit case, So = ~S and Po = ~P, so we can apply the
+	 * above table directly.
 	 */
 #define So rop3_not(rop3_S)
 #define Po rop3_not(rop3_T)
+#ifdef TRANSPARENCY_PER_H_P
+#  define MPo (rop3_uses_S(rop) ? rop3_not(So) | Po : Po)
+#else
+#  define MPo Po
+#endif
+	/*
+	 * If the operation doesn't use S or T, we must disregard the
+	 * corresponding transparency flag.
+	 */
+#define source_transparent ((lop & lop_S_transparent) && rop3_uses_S(rop))
+#define pattern_transparent ((lop & lop_T_transparent) && rop3_uses_T(rop))
 	gs_rop3_t mask =
 	  (source_transparent ?
 	   (pattern_transparent ? So & Po : So) :
-	   (pattern_transparent ? ~So | Po : rop3_1));
+	   (pattern_transparent ? MPo : rop3_1));
+#undef MPo
 	return (rop & mask) | (rop3_D & ~mask);
 }
