@@ -1,4 +1,4 @@
-/* Copyright (C) 1992, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1992, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -71,6 +71,7 @@ struct gx_device_pbm_s {
 					/* 2 or 3 if colored (PPM only) */
 	int alpha_text;			/* # of alpha bits for text (1,2,4) */
 	int alpha_graphics;		/* ditto for graphics (1,2,4) */
+	dev_proc_copy_alpha((*save_copy_alpha));
 };
 typedef struct gx_device_pbm_s gx_device_pbm;
 
@@ -95,7 +96,7 @@ typedef struct gx_device_pbm_s gx_device_pbm;
 	 { 0 },\
 	is_raw,\
 	optimize,\
-	0, 1, 1\
+	0, 1, 1, 0\
 }
 
 /* For all but PBM, we need our own color mapping and alpha procedures. */
@@ -107,6 +108,7 @@ private dev_proc_map_cmyk_color(pkm_map_cmyk_color);
 private dev_proc_map_color_rgb(pkm_map_color_rgb);
 private dev_proc_put_params(ppm_put_params);
 private dev_proc_get_alpha_bits(ppm_get_alpha_bits);
+private dev_proc_copy_alpha(pnm_copy_alpha);
 
 /* We need to initialize uses_color when opening the device, */
 /* and after each showpage. */
@@ -134,11 +136,11 @@ private gx_device_procs pbm_procs =
 }
 
 private gx_device_procs pgm_procs =
-    pgpm_procs(pgm_map_rgb_color, pgm_map_color_rgb, NULL);
+  pgpm_procs(pgm_map_rgb_color, pgm_map_color_rgb, NULL);
 private gx_device_procs ppm_procs =
-    pgpm_procs(ppm_map_rgb_color, ppm_map_color_rgb, NULL);
+  pgpm_procs(ppm_map_rgb_color, ppm_map_color_rgb, NULL);
 private gx_device_procs pkm_procs =
-    pgpm_procs(NULL, pkm_map_color_rgb, pkm_map_cmyk_color);
+  pgpm_procs(NULL, pkm_map_color_rgb, pkm_map_cmyk_color);
 
 /* The device descriptors themselves */
 gx_device_pbm far_data gs_pbm_device =
@@ -180,16 +182,32 @@ gx_device_pbm far_data gs_pkmraw_device =
 
 /* ------ Initialization ------ */
 
+/* Set the copy_alpha procedure if necessary. */
+private void
+ppm_set_copy_alpha(gx_device *pdev)
+{	if ( pdev->std_procs.copy_alpha != pnm_copy_alpha )
+	  { bdev->save_copy_alpha = pdev->std_procs.copy_alpha;
+	    if ( pdev->color_info.depth > 4 )
+	      pdev->std_procs.copy_alpha = pnm_copy_alpha;
+	  }
+}
+
 private int
 ppm_open(gx_device *pdev)
-{	bdev->uses_color = 0;
-	return gdev_prn_open(pdev);
+{	int code = gdev_prn_open(pdev);
+
+	if ( code < 0 )
+	  return code;
+	bdev->uses_color = 0;
+	ppm_set_copy_alpha(pdev);
+	return code;
 }
 
 /* Print a page, and reset uses_color if this is a showpage. */
 private int
 ppm_output_page(gx_device *pdev, int num_copies, int flush)
 {	int code = gdev_prn_output_page(pdev, num_copies, flush);
+
 	if ( code < 0 )
 	  return code;
 	if ( flush )
@@ -386,6 +404,7 @@ ppm_put_params(gx_device *pdev, gs_param_list *plist)
 	    bdev->alpha_graphics = agraphics;
 	    pdev->color_info = save_info;
 	  }
+	ppm_set_copy_alpha(pdev);
 	return code;
 }
 
@@ -395,6 +414,22 @@ ppm_get_alpha_bits(gx_device *pdev, graphics_object_type type)
 {	return (type == go_text ? bdev->alpha_text : bdev->alpha_graphics);
 }
 
+/* Copy an alpha map, noting whether we may generate some non-black/white */
+/* colors through blending. */
+private int
+pnm_copy_alpha(gx_device *pdev, const byte *data, int data_x,
+  int raster, gx_bitmap_id id, int x, int y, int width, int height,
+  gx_color_index color, int depth)
+{	if ( pdev->color_info.depth < 24 ||
+	     (color >> 8) == (color & 0xffff)
+	   )
+	  bdev->uses_color |= 1;
+	else
+	  bdev->uses_color |= 2;
+	return (*bdev->save_copy_alpha)(pdev, data, data_x, raster, id,
+					x, y, width, height, color, depth);
+}
+
 /* ------ Internal routines ------ */
 
 /* Print a page using a given row printing routine. */
