@@ -1,4 +1,4 @@
-/* Copyright (C) 1992, 1993, 1994, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1992, 1993, 1994, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -1092,7 +1092,7 @@ pm_alloc_bitmap(gx_device_pm *pmdev, gx_device *param_dev)
 	}
 	/* Shared memory can't be decommitted */
 #ifdef __DLL__
-	if (needed < pmdev->committed) {
+	if (pmdev->dll && (needed < pmdev->committed)) {
 		/* decommit memory */
 		if (rc = DosSetMem(pmdev->bitmap + needed, 
 			pmdev->committed - needed, 
