@@ -1,19 +1,24 @@
 /* Copyright (C) 1990, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of Aladdin Ghostscript.
+  This file is part of GNU Ghostscript.
   
-  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
-  or distributor accepts any responsibility for the consequences of using it,
-  or for whether it serves any particular purpose or works at all, unless he
-  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
-  License (the "License") for full details.
+  GNU Ghostscript is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
+  anyone for the consequences of using it or for whether it serves any
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
-  Every copy of Aladdin Ghostscript must include a copy of the License,
-  normally in a plain ASCII text file named PUBLIC.  The License grants you
-  the right to copy, modify and redistribute Aladdin Ghostscript, but only
-  under certain conditions described in the License.  Among other things, the
-  License requires that the copyright notice and this notice be preserved on
-  all copies.
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gdevprn.c */
@@ -131,7 +136,7 @@ open_c:		pcldev->data = base;
 		code = (*ppdev->printer_procs.make_buffer_device)
 		  (pmemdev, pdev, pdev->memory, false);
 		if ( code < 0 )
-		  { gs_free(base, space, 1, "command list buffer");
+		  { gs_free(base, (uint)mem_space, 1, "printer buffer");
 		    return code;
 		  }
 		pmemdev->base = base;
@@ -615,36 +620,18 @@ gdev_prn_print_scan_lines(gx_device *pdev)
 
 /* Open the current page for printing. */
 int
-gdev_prn_open_printer(gx_device *pdev, int binary_mode)
-{	char *fname = ppdev->fname;
-	char pfname[prn_fname_sizeof + 20];
-	char *fchar = fname;
-	long count1 = ppdev->PageCount + 1;
-
-	while ( (fchar = strchr(fchar, '%')) != NULL )
-	{	if ( *++fchar == '%' )
-		  {	++fchar;
-			continue;
-		  }
-		while ( !isalpha(*fchar) )
-		  ++fchar;
-		sprintf(pfname, fname,
-			(*fchar == 'l' ? count1 : (int)count1));
-		fname = pfname;
-		break;
+gdev_prn_open_printer(gx_device *pdev, bool binary_mode)
+{	if ( ppdev->file != 0 )
+	  { ppdev->file_is_new = false;
+	    return 0;
+	  }
+	{ int code = gx_device_open_output_file(pdev, ppdev->fname,
+						binary_mode, false,
+						&ppdev->file);
+	  if ( code < 0 )
+	    return code;
 	}
-	if ( ppdev->file == NULL )
-	{	if ( !strcmp(fname, "-") )
-			ppdev->file = stdout;
-		else
-		{	ppdev->file = gp_open_printer(fname, binary_mode);
-			if ( ppdev->file == NULL )
-				return_error(gs_error_invalidfileaccess);
-		}
-		ppdev->file_is_new = true;
-	}
-	else
-		ppdev->file_is_new = false;
+	ppdev->file_is_new = true;
 	return 0;
 }
 
