@@ -1,19 +1,24 @@
 /* Copyright (C) 1989, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of Aladdin Ghostscript.
+  This file is part of GNU Ghostscript.
   
-  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
-  or distributor accepts any responsibility for the consequences of using it,
-  or for whether it serves any particular purpose or works at all, unless he
-  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
-  License (the "License") for full details.
+  GNU Ghostscript is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
+  anyone for the consequences of using it or for whether it serves any
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
-  Every copy of Aladdin Ghostscript must include a copy of the License,
-  normally in a plain ASCII text file named PUBLIC.  The License grants you
-  the right to copy, modify and redistribute Aladdin Ghostscript, but only
-  under certain conditions described in the License.  Among other things, the
-  License requires that the copyright notice and this notice be preserved on
-  all copies.
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gdevprn.h */
@@ -244,8 +249,8 @@ extern gx_device_procs prn_std_procs;
 /* This accounts for the short parameter names here and below. */
 #define prn_device_margins_body(dtype, procs, dname, w10, h10, xdpi, ydpi, lo, to, lm, bm, rm, tm, ncomp, depth, mg, mc, dg, dc, print_page)\
 	std_device_full_body(dtype, &procs, dname,\
-	  (int)((long)w10 * xdpi / 10),\
-	  (int)((long)h10 * ydpi / 10),\
+	  (int)((long)(w10) * (xdpi) / 10),\
+	  (int)((long)(h10) * (ydpi) / 10),\
 	  xdpi, ydpi,\
 	  ncomp, depth, mg, mc, dg, dc,\
 	  -(lo) * (xdpi), -(to) * (ydpi),\
@@ -260,8 +265,8 @@ extern gx_device_procs prn_std_procs;
 
 #define prn_device_std_margins_body(dtype, procs, dname, w10, h10, xdpi, ydpi, lo, to, lm, bm, rm, tm, color_bits, print_page)\
 	std_device_std_color_full_body(dtype, &procs, dname,\
-	  (int)((long)w10 * xdpi / 10),\
-	  (int)((long)h10 * ydpi / 10),\
+	  (int)((long)(w10) * (xdpi) / 10),\
+	  (int)((long)(h10) * (ydpi) / 10),\
 	  xdpi, ydpi, color_bits,\
 	  -(lo) * (xdpi), -(to) * (ydpi),\
 	  (lm) * 72.0, (bm) * 72.0,\
@@ -285,7 +290,7 @@ extern gx_device_procs prn_std_procs;
 /* ------ Utilities ------ */
 /* These are defined in gdevprn.c. */
 
-int gdev_prn_open_printer(P2(gx_device *dev, int binary_mode));
+int gdev_prn_open_printer(P2(gx_device *dev, bool binary_mode));
 #define gdev_prn_file_is_new(pdev) ((pdev)->file_is_new)
 #define gdev_prn_raster(pdev) gx_device_raster((gx_device *)(pdev), 0)
 int gdev_prn_get_bits(P4(gx_device_printer *, int, byte *, byte **));
