@@ -1,19 +1,24 @@
 /* Copyright (C) 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of Aladdin Ghostscript.
+  This file is part of GNU Ghostscript.
   
-  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
-  or distributor accepts any responsibility for the consequences of using it,
-  or for whether it serves any particular purpose or works at all, unless he
-  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
-  License (the "License") for full details.
+  GNU Ghostscript is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
+  anyone for the consequences of using it or for whether it serves any
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
-  Every copy of Aladdin Ghostscript must include a copy of the License,
-  normally in a plain ASCII text file named PUBLIC.  The License grants you
-  the right to copy, modify and redistribute Aladdin Ghostscript, but only
-  under certain conditions described in the License.  Among other things, the
-  License requires that the copyright notice and this notice be preserved on
-  all copies.
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gdevvec.c */
@@ -176,14 +181,12 @@ gdev_vector_reset(gx_device_vector *vdev)
 int
 gdev_vector_open_file_bbox(gx_device_vector *vdev, uint strmbuf_size,
   bool bbox)
-{	char fmode[4];
+{	/* Open the file as positionable if possible. */
+	int code = gx_device_open_output_file((gx_device *)vdev, vdev->fname,
+					      true, true, &vdev->file);
 
-	/* Open the file as positionable if possible. */
-	strcpy(fmode, "w+");
-	strcat(fmode, gp_fmode_binary_suffix);
-	vdev->file = gp_fopen(vdev->fname, fmode);
-	if ( vdev->file == 0 )
-	  return_error(gs_error_invalidfileaccess);
+	if ( code < 0 )
+	  return code;
 	if ( (vdev->strmbuf = gs_alloc_bytes(vdev->v_memory, strmbuf_size,
 					     "vector_open(strmbuf)")) == 0 ||
 	     (vdev->strm = s_alloc(vdev->v_memory,
@@ -722,25 +725,21 @@ int
 gdev_vector_fill_path(gx_device *dev, const gs_imager_state *pis,
   gx_path *ppath, const gx_fill_params *params,
   const gx_device_color *pdevc, const gx_clip_path *pcpath)
-{	int code = gdev_vector_prepare_fill(vdev, pis, params, pdevc);
+{	int code;
 
-	if ( code < 0 )
-	  return code;
-	code = gdev_vector_update_clip_path(vdev, pcpath);
-	if ( code < 0 )
-	  return code;
-	if ( vdev->bbox_device )
-	  { code = (*dev_proc(vdev->bbox_device, fill_path))
-	      ((gx_device *)vdev->bbox_device, pis, ppath, params,
-	       pdevc, pcpath);
-	    if ( code < 0 )
-	      return code;
-	  }
-	return (*vdev_proc(vdev, dopath))(vdev, ppath,
-					  (params->rule > 0 ?
-					   gx_path_type_even_odd :
-					   gx_path_type_winding_number) |
-					  gx_path_type_fill);
+	if ( (code = gdev_vector_prepare_fill(vdev, pis, params, pdevc)) < 0 ||
+	     (code = gdev_vector_update_clip_path(vdev, pcpath)) < 0 ||
+	     (vdev->bbox_device &&
+	      (code = (*dev_proc(vdev->bbox_device, fill_path))
+	       ((gx_device *)vdev->bbox_device, pis, ppath, params,
+		pdevc, pcpath)) < 0) || 
+	     (code = (*vdev_proc(vdev, dopath))
+	      (vdev, ppath,
+	       (params->rule > 0 ? gx_path_type_even_odd :
+		gx_path_type_winding_number) | gx_path_type_fill)) < 0
+	   )
+	  return gx_default_fill_path(dev, pis, ppath, params, pdevc, pcpath);
+	return code;
 }
 
 int
@@ -750,21 +749,18 @@ gdev_vector_stroke_path(gx_device *dev, const gs_imager_state *pis,
 {	int code;
 
 	/****** HANDLE SCALE ******/
-	code = gdev_vector_prepare_stroke(vdev, pis, params, pdcolor,
-					  dev->HWResolution[0]);
-	if ( code < 0 )
-	  return code;
-	code = gdev_vector_update_clip_path(vdev, pcpath);
-	if ( code < 0 )
-	  return code;
-	if ( vdev->bbox_device )
-	  { code = (*dev_proc(vdev->bbox_device, stroke_path))
-	      ((gx_device *)vdev->bbox_device, pis, ppath, params,
-	       pdcolor, pcpath);
-	    if ( code < 0 )
-	      return code;
-	  }
-	return (*vdev_proc(vdev, dopath))(vdev, ppath, gx_path_type_stroke);
+	if ( (code = gdev_vector_prepare_stroke(vdev, pis, params, pdcolor,
+						dev->HWResolution[0])) < 0 ||
+	     (code = gdev_vector_update_clip_path(vdev, pcpath)) < 0 ||
+	     (vdev->bbox_device &&
+	      (code = (*dev_proc(vdev->bbox_device, stroke_path))
+	       ((gx_device *)vdev->bbox_device, pis, ppath, params,
+		pdcolor, pcpath)) < 0) ||
+	     (code = (*vdev_proc(vdev, dopath))
+	      (vdev, ppath, gx_path_type_stroke)) < 0
+	   )
+	  return gx_default_stroke_path(dev, pis, ppath, params, pdcolor, pcpath);
+	return code;
 }
 
 int
