@@ -32,7 +32,24 @@
 /* This driver uses the printer default size and resolution and
  * ignores page size and resolution set using -gWIDTHxHEIGHT and
  * -rXxY.  You must still set the correct PageSize to get the
- * correct clipping path
+ * correct clipping path.
+ * The code in win_pr2_getdc() does try to set the printer page
+ * size from the PostScript PageSize, but it isn't working
+ * reliably at the moment.
+ * 
+ * This driver doesn't work with some Windows printer drivers.
+ * The reason is unknown.  All printers to which I have access
+ * work. 
+ *
+ * rjl 1997-11-20
+ */
+
+/* Supported printer parameters are
+ *  -dBitsPerPixel=n
+ *     Override what the Window printer driver returns.
+ *  -dNoCancel
+ *     Don't display cancel dialog box.  Useful for unattended or
+ *     console EXE operation.
  */
 
 #include "gdevprn.h"
@@ -59,6 +76,7 @@ private dev_proc_close_device(win_pr2_close);
 private dev_proc_print_page(win_pr2_print_page);
 private dev_proc_map_rgb_color(win_pr2_map_rgb_color);
 private dev_proc_map_color_rgb(win_pr2_map_color_rgb);
+private dev_proc_get_params(win_pr2_get_params);
 private dev_proc_put_params(win_pr2_put_params);
 
 private void win_pr2_set_bpp(gx_device *dev, int depth);
@@ -66,7 +84,7 @@ private void win_pr2_set_bpp(gx_device *dev, int depth);
 private gx_device_procs win_pr2_procs =
   prn_color_params_procs(win_pr2_open, gdev_prn_output_page, win_pr2_close,
    win_pr2_map_rgb_color, win_pr2_map_color_rgb, 
-   gdev_prn_get_params, win_pr2_put_params);
+   win_pr2_get_params, win_pr2_put_params);
 
 
 /* The device descriptor */
@@ -75,17 +93,20 @@ struct gx_device_win_pr2_s {
 	gx_device_common;
 	gx_prn_device_common;
 	HDC hdcprn;
+	bool nocancel;
 	DLGPROC lpfnAbortProc;
 	DLGPROC lpfnCancelProc;
 };
 
 gx_device_win_pr2 far_data gs_mswinpr2_device = {
   prn_device_std_body(gx_device_win_pr2, win_pr2_procs, "mswinpr2",
-  DEFAULT_WIDTH_10THS, DEFAULT_HEIGHT_10THS, 72, 72,
+  DEFAULT_WIDTH_10THS, DEFAULT_HEIGHT_10THS, 72.0, 72.0,
   0, 0, 0, 0,
   0, win_pr2_print_page), /* depth = 0 */
   0,	/* hdcprn */
-  NULL	/* lpfnAbortProc */
+  0,	/* nocancel */
+  NULL,	/* lpfnAbortProc */
+  NULL	/* lpfnCancelProc */
 };
 
 private int win_pr2_getdc(gx_device_win_pr2 *);
@@ -101,8 +122,9 @@ win_pr2_open(gx_device *dev)
 	float m[4];
 	FILE *pfile;
 
-	if (hDlgModeless) {
+	if ((!wdev->nocancel) && (hDlgModeless)) {
 	    /* device cannot opened twice since only one hDlgModeless */
+	    fprintf(stderr, "Can't open mswinpr2 device twice\n");
 	    return gs_error_limitcheck;
 	}
 
@@ -141,6 +163,7 @@ win_pr2_open(gx_device *dev)
 #endif
 	Escape(wdev->hdcprn,SETABORTPROC,0,(LPSTR)wdev->lpfnAbortProc,NULL);  
 	if (Escape(wdev->hdcprn, STARTDOC, lstrlen(szAppName), szAppName, NULL) <= 0) {
+	    fprintf(stderr, "Printer Escape STARTDOC failed\n");
 #if !defined(__WIN32__) && !defined(__DLL__)
 	    FreeProcInstance((FARPROC)wdev->lpfnAbortProc);
 #endif
@@ -148,11 +171,11 @@ win_pr2_open(gx_device *dev)
 	    return gs_error_limitcheck;
 	}
 
-	dev->x_pixels_per_inch = GetDeviceCaps(wdev->hdcprn, LOGPIXELSX);
-	dev->y_pixels_per_inch = GetDeviceCaps(wdev->hdcprn, LOGPIXELSY);
-	Escape(wdev->hdcprn, GETPHYSPAGESIZE, NULL, NULL, (LPPOINT)&size);
+	dev->x_pixels_per_inch = (float)GetDeviceCaps(wdev->hdcprn, LOGPIXELSX);
+	dev->y_pixels_per_inch = (float)GetDeviceCaps(wdev->hdcprn, LOGPIXELSY);
+	Escape(wdev->hdcprn, GETPHYSPAGESIZE, 0, NULL, (LPPOINT)&size);
 	gx_device_set_width_height(dev, (int)size.x, (int)size.y);
-	Escape(wdev->hdcprn, GETPRINTINGOFFSET, NULL, NULL, (LPPOINT)&offset);
+	Escape(wdev->hdcprn, GETPRINTINGOFFSET, 0, NULL, (LPPOINT)&offset);
 	/* m[] gives margins in inches */
 	m[0] /*left*/ = offset.x / dev->x_pixels_per_inch;
 	m[3] /*top*/ = offset.y / dev->y_pixels_per_inch;
@@ -161,8 +184,7 @@ win_pr2_open(gx_device *dev)
 		 / dev->x_pixels_per_inch;
 	m[1] /*bottom*/ =
 		(size.y - offset.y - GetDeviceCaps(wdev->hdcprn, VERTRES))
-		 / dev->y_pixels_per_inch
-		+ 0.15;  /* hack to add a bit more margin for deskjet printer */
+		 / dev->y_pixels_per_inch;
 	gx_device_set_margins(dev, m, true);
 
 	depth = dev->color_info.depth;
@@ -183,7 +205,8 @@ win_pr2_open(gx_device *dev)
 	/* delete unwanted temporary file */
 	unlink(wdev->fname);
 
-	/* inform user of progress with dialog box and allow cancel */
+	if (!wdev->nocancel) {
+	    /* inform user of progress with dialog box and allow cancel */
 #ifdef __WIN32__
 	wdev->lpfnCancelProc = (DLGPROC)CancelDlgProc;
 #else
@@ -193,8 +216,10 @@ win_pr2_open(gx_device *dev)
 	wdev->lpfnCancelProc = (DLGPROC)MakeProcInstance((FARPROC)CancelDlgProc, phInstance);
 #endif
 #endif
-	hDlgModeless = CreateDialog(phInstance, "CancelDlgBox", hwndtext, wdev->lpfnCancelProc);
-	ShowWindow(hDlgModeless, SW_HIDE);
+	    hDlgModeless = CreateDialog(phInstance, "CancelDlgBox", 
+		hwndtext, wdev->lpfnCancelProc);
+	    ShowWindow(hDlgModeless, SW_HIDE);
+	}
 
 	return code;
 };
@@ -206,13 +231,16 @@ win_pr2_close(gx_device *dev)
 	int aborted = FALSE;
 	/* Free resources */
 
-	if (!hDlgModeless)
-	    aborted = TRUE;
-	DestroyWindow(hDlgModeless);
-	hDlgModeless = 0;
+	if (!wdev->nocancel) {
+	    if (!hDlgModeless)
+		aborted = TRUE;
+	    else
+		DestroyWindow(hDlgModeless);
+	    hDlgModeless = 0;
 #if !defined(__WIN32__) && !defined(__DLL__)
-	FreeProcInstance((FARPROC)wdev->lpfnCancelProc);
+	    FreeProcInstance((FARPROC)wdev->lpfnCancelProc);
 #endif
+	}
 
 	if (aborted)
 	    Escape(wdev->hdcprn,ABORTDOC,0,NULL,NULL);
@@ -310,9 +338,11 @@ win_pr2_print_page(gx_device_printer *pdev, FILE *file)
 		bmi.h.biClrImportant = 0;
 	}
 
-	sprintf(dlgtext, "Printing page %d", (int)(pdev->PageCount)+1);
-	SetWindowText(GetDlgItem(hDlgModeless, CANCEL_PRINTING), dlgtext);
-	ShowWindow(hDlgModeless, SW_SHOW);
+	if (!wdev->nocancel) {
+	    sprintf(dlgtext, "Printing page %d", (int)(pdev->PageCount)+1);
+	    SetWindowText(GetDlgItem(hDlgModeless, CANCEL_PRINTING), dlgtext);
+	    ShowWindow(hDlgModeless, SW_SHOW);
+	}
 
 	for ( y = 0; y < scan_lines; ) {
 	    /* copy slice to row buffer */
@@ -329,29 +359,36 @@ win_pr2_print_page(gx_device_printer *pdev, FILE *file)
 		(BITMAPINFO FAR *)&bmi, DIB_RGB_COLORS);
 	    y += lines;
 
-	    /* inform user of progress */
-	    sprintf(dlgtext, "%d%% done", (int)(y * 100L / scan_lines));
-	    SetWindowText(GetDlgItem(hDlgModeless, CANCEL_PCDONE), dlgtext);
+	    if (!wdev->nocancel) {
+		/* inform user of progress */
+		sprintf(dlgtext, "%d%% done", (int)(y * 100L / scan_lines));
+		SetWindowText(GetDlgItem(hDlgModeless, CANCEL_PCDONE), dlgtext);
+	    }
+
 	    /* process message loop */
 	    while (PeekMessage(&msg, hDlgModeless, 0, 0, PM_REMOVE)) {
-		if ((hDlgModeless == 0) || !IsDialogMessage(hDlgModeless, &msg)) {
+		if ((hDlgModeless == 0) || !IsDialogMessage(hDlgModeless, &msg))
+		{
 		    TranslateMessage(&msg);
 		    DispatchMessage(&msg);
 		}
 	    }
-	    if (hDlgModeless == 0) {
+	    if ((!wdev->nocancel) && (hDlgModeless == 0)) {
 	        /* user pressed cancel button */
 		break;
 	    }
 	}
 
-	if (hDlgModeless == 0)
+	if ((!wdev->nocancel) && (hDlgModeless == 0))
 	    code = gs_error_Fatal;	/* exit Ghostscript cleanly */
 	else {
 	    /* push out the page */
-	    SetWindowText(GetDlgItem(hDlgModeless, CANCEL_PCDONE), "Ejecting page...");
+	    if (!wdev->nocancel)
+		SetWindowText(GetDlgItem(hDlgModeless, CANCEL_PCDONE), 
+			"Ejecting page...");
 	    Escape(wdev->hdcprn,NEWFRAME,0,NULL,NULL);
-	    ShowWindow(hDlgModeless, SW_HIDE);
+	    if (!wdev->nocancel)
+		ShowWindow(hDlgModeless, SW_HIDE);
 	}
 
 bmp_done:
@@ -442,6 +479,15 @@ win_pr2_set_bpp(gx_device *dev, int depth)
 	}
 }
 
+/* Get device parameters */
+int
+win_pr2_get_params(gx_device *dev, gs_param_list *plist)
+{	int code = gdev_prn_get_params(dev, plist);
+	if (code >=0)
+	   code = param_write_bool(plist, "NoCancel", 
+		&(((gx_device_win_pr2 *)dev)->nocancel));
+	return code;
+}
 
 /* We implement this ourselves so that we can change BitsPerPixel */
 /* before the device is opened */
@@ -450,6 +496,7 @@ win_pr2_put_params(gx_device *dev, gs_param_list *plist)
 {	int ecode = 0, code;
 	int old_bpp = dev->color_info.depth;
 	int bpp = old_bpp;
+	bool nocancel = ((gx_device_win_pr2 *)dev)->nocancel;
 
 	switch ( code = param_read_int(plist, "BitsPerPixel", &bpp) )
 	{
@@ -469,6 +516,24 @@ bppe:		param_signal_error(plist, "BitsPerPixel", ecode);
 		break;
 	}
 
+	switch ( code = param_read_bool(plist, "NoCancel", &nocancel) )
+	{
+	case 0:
+		if ( dev->is_open )
+		  ecode = gs_error_rangecheck;
+		else
+		  {
+			((gx_device_win_pr2 *)dev)->nocancel = nocancel;
+			break;
+		  }
+		goto nocancele;
+	default:
+		ecode = code;
+nocancele:	param_signal_error(plist, "NoCancel", ecode);
+	case 1:
+		break;
+	}
+
 	if ( ecode >= 0 )
 		ecode = gdev_prn_put_params(dev, plist);
 	return ecode;
@@ -606,8 +671,8 @@ HANDLE hprinter;
     if ( (devcap = gs_malloc(devcapsize, 1, "win_pr2_getdc")) == (LPBYTE)NULL )
         return FALSE;
     n = pfnDeviceCapabilities(device, output, DC_PAPERSIZE, devcap, NULL);
-    paperwidth = wdev->MediaSize[0] / 72 * 254; 
-    paperheight = wdev->MediaSize[1] / 72 * 254; 
+    paperwidth = wdev->MediaSize[0] * 254 / 72; 
+    paperheight = wdev->MediaSize[1] * 254 / 72; 
     papername[0] = '\0';
     papersize = 0;
     paperindex = -1;
@@ -683,6 +748,7 @@ HANDLE hprinter;
     if (!is_win32s) {
 	/* change the page size by changing the form */
 	/* WinNT only */
+	/*  at present, changing the page size isn't working under Win NT */
 	/* Win95 returns FALSE to GetForm */
 	LPBYTE lpbForm;
 	FORM_INFO_1 *fi1;
@@ -690,7 +756,8 @@ HANDLE hprinter;
 	DWORD dwNeeded;
 	dwNeeded = 0;
 	dwBuf = 1024;
-        if ( (lpbForm = gs_malloc(dwBuf, 1, "win_pr2_getdc")) == (LPBYTE)NULL ) {
+        if ( (lpbForm = gs_malloc(dwBuf, 1, "win_pr2_getdc")) == (LPBYTE)NULL )
+	{
 	    gs_free(podevmode, size, 1, "win_pr2_getdc");
 	    gs_free(pidevmode, size, 1, "win_pr2_getdc");
 	    ClosePrinter(hprinter);
