@@ -183,12 +183,12 @@ win_prn_open(gx_device *dev)
 	unlink(wdev->mfname);
 	wdev->hdcmf = CreateMetaFile(wdev->mfname);
 
-	dev->x_pixels_per_inch = GetDeviceCaps(wdev->hdcprn, LOGPIXELSX);
-	dev->y_pixels_per_inch = GetDeviceCaps(wdev->hdcprn, LOGPIXELSY);
-	Escape(wdev->hdcprn, GETPHYSPAGESIZE, NULL, NULL, (LPPOINT)&size);
+	dev->x_pixels_per_inch = (float)GetDeviceCaps(wdev->hdcprn, LOGPIXELSX);
+	dev->y_pixels_per_inch = (float)GetDeviceCaps(wdev->hdcprn, LOGPIXELSY);
+	Escape(wdev->hdcprn, GETPHYSPAGESIZE, 0, NULL, (LPPOINT)&size);
 	dev->width = size.x;
 	dev->height = size.y;
-	Escape(wdev->hdcprn, GETPRINTINGOFFSET, NULL, NULL, (LPPOINT)&offset);
+	Escape(wdev->hdcprn, GETPRINTINGOFFSET, 0, NULL, (LPPOINT)&offset);
 	m[0] /*left*/ = offset.x / dev->x_pixels_per_inch;
 	m[3] /*top*/ = offset.y / dev->y_pixels_per_inch;
 	m[2] /*right*/ =
@@ -252,7 +252,7 @@ win_prn_open(gx_device *dev)
 	}
 	SetMapMode(wdev->hdcmono, GetMapMode(wdev->hdcprn));
 	SelectObject(wdev->hdcmono, wdev->hbmmono);
-	(void) SelectPalette(wdev->hdcmf,wdev->himgpalette,NULL);
+	(void) SelectPalette(wdev->hdcmf,wdev->himgpalette,FALSE);
 	RealizePalette(wdev->hdcmf);
 	win_prn_maketools(wdev,wdev->hdcmf);
 	wdev->bm_id = gx_no_bitmap_id;
@@ -301,16 +301,16 @@ RECT rect;
 HMETAFILE hmf;
 	hmf = CloseMetaFile(wdev->hdcmf);
 	
-	Escape(wdev->hdcprn, NEXTBAND, NULL, NULL, (LPRECT)&rect);
+	Escape(wdev->hdcprn, NEXTBAND, 0, NULL, (LPRECT)&rect);
 	while (!IsRectEmpty(&rect)) {
 	    PlayMetaFile(wdev->hdcprn, hmf);
-	    if (Escape(wdev->hdcprn, NEXTBAND, NULL, NULL, (LPRECT)&rect) <= 0)
+	    if (Escape(wdev->hdcprn, NEXTBAND, 0, NULL, (LPRECT)&rect) <= 0)
 		break;
 	}
 	DeleteMetaFile(hmf);
 	unlink(wdev->mfname);
 	wdev->hdcmf = CreateMetaFile(wdev->mfname);
-	(void) SelectPalette(wdev->hdcmf,wdev->himgpalette,NULL);
+	(void) SelectPalette(wdev->hdcmf,wdev->himgpalette,FALSE);
 	RealizePalette(wdev->hdcmf);
 	SelectObject(wdev->hdcmf,wdev->hpen);
 	SelectObject(wdev->hdcmf,wdev->hbrush);
@@ -326,7 +326,7 @@ win_prn_map_rgb_color(gx_device *dev, gx_color_value r, gx_color_value g,
 {	int i = wdev->nColors;
 	gx_color_index color = win_map_rgb_color(dev, r, g, b);
 	if ( color != i ) return color;
-	(void) SelectPalette(wdev->hdcmf,wdev->himgpalette,NULL);
+	(void) SelectPalette(wdev->hdcmf,wdev->himgpalette,FALSE);
 	RealizePalette(wdev->hdcmf);
 	win_prn_addtool(wdev, i);
 
