@@ -1,4 +1,4 @@
-/* Copyright (C) 1989, 1995, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -114,17 +114,9 @@ private gx_device_procs x_procs = {
 };
 
 /* The instance is public. */
-#ifdef A4
-#  define DEFAULT_WIDTH_10THS 83/10
-#  define DEFAULT_HEIGHT_10THS 117/10
-#else
-#  define DEFAULT_WIDTH_10THS 85/10
-#  define DEFAULT_HEIGHT_10THS 11
-#endif
-#  define DEFAULT_
 gx_device_X far_data gs_x11_device = {
 	std_device_color_body(gx_device_X, &x_procs, "x11",
-	  FAKE_RES*DEFAULT_WIDTH_10THS, FAKE_RES*DEFAULT_HEIGHT_10THS, /* x and y extent (nominal) */
+	  FAKE_RES*DEFAULT_WIDTH_10THS/10, FAKE_RES*DEFAULT_HEIGHT_10THS/10, /* x and y extent (nominal) */
 	  FAKE_RES, FAKE_RES,	/* x and y density (nominal) */
 	  /*dci_color(*/24, 255, 256/*)*/),
 	{ 0 },			/* std_procs */
@@ -162,7 +154,7 @@ gx_device_X far_data gs_x11_device = {
 	(Atom)0, (Atom)0, (Atom)0,	/* Atoms: NEXT, PAGE, DONE */
 	 { 0, 0, 0, 0 }, 0, 0,	/* update, up_area, up_count */
 	(Pixmap)0,		/* dest */
-	0L, ~0L,		/* colors_or, colors_and */
+	0L, (ulong)~0L,		/* colors_or, colors_and */
 	 { /* cp */
 	   (Pixmap)0,		/* pixmap */
 	   NULL,		/* gc */
@@ -198,12 +190,17 @@ gx_device_X far_data gs_x11_device = {
 };
 
 /* If XPutImage doesn't work, do it ourselves. */
-private void alt_put_image(P11(gx_device *dev, Display *dpy, Drawable win,
+private int alt_put_image(P11(gx_device *dev, Display *dpy, Drawable win,
   GC gc, XImage *pi, int sx, int sy, int dx, int dy, unsigned w, unsigned h));
 #define put_image(dpy,win,gc,im,sx,sy,x,y,w,h)\
-  if ( xdev->useXPutImage) XPutImage(dpy,win,gc,im,sx,sy,x,y,w,h);\
-  else alt_put_image(dev,dpy,win,gc,im,sx,sy,x,y,w,h)
-
+  do {\
+    if ( xdev->useXPutImage ) {\
+      XPutImage(dpy,win,gc,im,sx,sy,x,y,w,h);\
+    } else {\
+      int code_ = alt_put_image(dev,dpy,win,gc,im,sx,sy,x,y,w,h);\
+      if ( code_ < 0 ) return code_;\
+    }\
+  } while (0)
 
 /* Open the device.  Most of the code is in gdevxini.c. */
 private int
@@ -647,7 +644,7 @@ x_copy_mono(register gx_device *dev,
 
     fit_copy(dev, base, sourcex, raster, id, x, y, w, h);
 
-    xdev->image.width = raster << 3;
+    xdev->image.width = sourcex + w;
     xdev->image.height = h;
     xdev->image.data = (char *)base;
     xdev->image.bytes_per_line = raster;
@@ -708,12 +705,12 @@ hard:	/* Handle the hard 1-color case. */
 	    XCreatePixmap(xdev->dpy, xdev->win, raster << 3, h, 1);
 	if (xdev->cp.pixmap == (Pixmap) 0) {
 	    lprintf("x_copy_mono: can't allocate pixmap\n");
-	    exit(1);
+	    return_error(gs_error_VMerror);
 	}
 	xdev->cp.gc = XCreateGC(xdev->dpy, xdev->cp.pixmap, 0, 0);
 	if (xdev->cp.gc == (GC) 0) {
 	    lprintf("x_copy_mono: can't allocate GC\n");
-	    exit(1);
+	    return_error(gs_error_VMerror);
 	}
 	xdev->cp.raster = raster;
 	xdev->cp.height = h;
@@ -804,16 +801,16 @@ x_copy_color(register gx_device *dev,
 		XDrawPoint(xdev->dpy, xdev->dest, xdev->gc, x, y);
 	  }
 	else
-	  {    xdev->image.width = raster << 3;
-	       xdev->image.height = h;
-	       xdev->image.format = ZPixmap;
-	       xdev->image.data = (char *)base;
-	       xdev->image.depth = depth;
-	       xdev->image.bytes_per_line = raster;
-	       xdev->image.bits_per_pixel = depth;
-	       XPutImage(xdev->dpy, xdev->dest, xdev->gc, &xdev->image,
-			 sourcex, 0, x, y, w, h);
-	       xdev->image.depth = xdev->image.bits_per_pixel = 1;
+	  {	xdev->image.width = sourcex + w;
+		xdev->image.height = h;
+		xdev->image.format = ZPixmap;
+		xdev->image.data = (char *)base;
+		xdev->image.depth = depth;
+		xdev->image.bytes_per_line = raster;
+		xdev->image.bits_per_pixel = depth;
+		XPutImage(xdev->dpy, xdev->dest, xdev->gc, &xdev->image,
+			  sourcex, 0, x, y, w, h);
+		xdev->image.depth = xdev->image.bits_per_pixel = 1;
 	  }
 	if (xdev->bpixmap != (Pixmap) 0)
 	  x_update_add(dev, x, y, w, h);
@@ -829,16 +826,22 @@ private int
 x_get_bits(gx_device *dev, int y, byte *str, byte **actual_data)
 {	gx_device_X *xdev = (gx_device_X *)dev;
 	int depth = dev->color_info.depth;
-	uint raster = gx_device_raster(dev, true);
+	uint width_bytes = gx_device_raster(dev, false);
+	XImage *image;
+
+	/* Make sure the frame buffer is up to date. */
+	if ( xdev->up_area != 0 &&
+	     y >= xdev->update.yo && y < xdev->update.ye
+	   )
+	  update_do_flush(dev);
 	/*
 	 * The X library doesn't provide any way to specify the desired
 	 * bit or byte ordering for the result, so we just hope for the best
 	 * (big-endian).
 	 */
-	XImage *image = XGetImage(xdev->dpy, xdev->dest,
-				  0, y, dev->width, 1,
-				  (1L << depth) - 1, ZPixmap);
-	memcpy(str, image->data, raster);
+	image = XGetImage(xdev->dpy, xdev->dest, 0, y, dev->width, 1,
+			  (1L << depth) - 1, ZPixmap);
+	memcpy(str, image->data, width_bytes);
 	XDestroyImage(image);
 	*actual_data = str;
 	return 0;
@@ -1180,7 +1183,7 @@ x_update_add(gx_device *dev, int xo, int yo, int w, int h)
 /* This is a total hack to get around an apparent bug */
 /* in some X servers.  It only works with the specific */
 /* parameters (bit/byte order, padding) used above. */
-private void
+private int
 alt_put_image(gx_device *dev, Display *dpy, Drawable win, GC gc,
 	  XImage *pi, int sx, int sy, int dx, int dy, unsigned w, unsigned h)
 {
@@ -1214,7 +1217,7 @@ alt_put_image(gx_device *dev, Display *dpy, Drawable win, GC gc,
 	}
     } else {
 	lprintf("alt_put_image: unimplemented function.\n");
-	exit(1);
+	return_error(gs_error_rangecheck);
     }
 
     for (yi = 0; yi < h; yi++, data += raster) {
@@ -1227,8 +1230,8 @@ alt_put_image(gx_device *dev, Display *dpy, Drawable win, GC gc,
 		int xleft = xi;
 
 		if (rp == &rects[nrects]) {
-		    XFillRectangles(dpy, win, gc, rects, nrects);
-		    rp = rects;
+		  XFillRectangles(dpy, win, gc, rects, nrects);
+		  rp = rects;
 		}
 		/* Scan over a run of 1-bits */
 		rp->x = dx + xi, rp->y = dy + yi;
@@ -1248,4 +1251,5 @@ alt_put_image(gx_device *dev, Display *dpy, Drawable win, GC gc,
     }
     XFillRectangles(dpy, win, gc, rects, rp - rects);
     if (invert) XSetForeground(dpy, gc, gcv.foreground);
+    return 0;
 }
