@@ -1,4 +1,4 @@
-/* Copyright (C) 1989, 1995, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -23,10 +23,11 @@
 
 /* gdevxini.c */
 /* X Windows driver initialization for Ghostscript library */
-#include "gx.h"			/* for gx_bitmap; includes std.h */
-#include "memory_.h"
 #include "math_.h"
+#include "memory_.h"
 #include "x_.h"
+#include "gx.h"
+#include "gserrors.h"
 #include "gxdevice.h"
 #include "gdevx.h"
 
@@ -144,7 +145,7 @@ typedef enum {
 } orientation;
 
 /* Forward references */
-private void gdev_x_setup_colors(P1(gx_device_X *));
+private int gdev_x_setup_colors(P1(gx_device_X *));
 private void gdev_x_setup_fontmap(P1(gx_device_X *));
 
 /* Catch the alloc error when there is not enough resources for the
@@ -187,6 +188,7 @@ gdev_x_open(register gx_device_X *xdev)
     XColor xc;
     int zero = 0;
     int xid_height, xid_width;
+    int code;
 
 #ifdef DEBUG
 # ifdef have_Xdebug
@@ -200,16 +202,16 @@ gdev_x_open(register gx_device_X *xdev)
     if (!(xdev->dpy = XOpenDisplay((char *)NULL))) {
 	char *dispname = getenv("DISPLAY");
 
-	eprintf1("gs: Cannot open X display `%s'.\n",
+	eprintf1("Cannot open X display `%s'.\n",
 		 (dispname == NULL ? "(null)" : dispname));
-	exit(1);
+	return_error(gs_error_ioerror);
     }
     xdev->dest = 0;
     if ((window_id = getenv("GHOSTVIEW"))) {
 	if (!(xdev->ghostview = sscanf(window_id, "%ld %ld",
 				       &(xdev->win), &(xdev->dest)))) {
-	    eprintf("gs: Cannot get Window ID from ghostview.\n");
-	    exit(1);
+	    eprintf("Cannot get Window ID from ghostview.\n");
+	    return_error(gs_error_ioerror);
 	}
     }
     if (xdev->pwin != (Window)None) {/* pick up the destination window parameters if specified */
@@ -268,12 +270,12 @@ gdev_x_open(register gx_device_X *xdev)
 			    &left_margin, &bottom_margin,
 			    &right_margin, &top_margin);
 	    if (!(nitems == 8 || nitems == 12)) {
-		eprintf("gs: Cannot get ghostview property.\n");
-		exit(1);
+		eprintf("Cannot get ghostview property.\n");
+		return_error(gs_error_ioerror);
 	    }
 	    if (xdev->dest && xdev->bpixmap) {
-		eprintf("gs: Both destination and backing pixmap specified.\n");
-		exit(1);
+		eprintf("Both destination and backing pixmap specified.\n");
+		return_error(gs_error_rangecheck);
 	    }
 	    if (xdev->dest) {
 		Window root;
@@ -336,8 +338,8 @@ gdev_x_open(register gx_device_X *xdev)
 	    xdev->ImagingBBox_set = true;
 
 	} else if (xdev->pwin == (Window)None) {
-	    eprintf("gs: Cannot get ghostview property.\n");
-	    exit(1);
+	    eprintf("Cannot get ghostview property.\n");
+	    return_error(gs_error_ioerror);
 	}
     } else {
 	Screen *scr = DefaultScreenOfDisplay(xdev->dpy);
@@ -350,8 +352,8 @@ gdev_x_open(register gx_device_X *xdev)
     xvinfo.visualid = XVisualIDFromVisual(xvinfo.visual);
     xdev->vinfo = XGetVisualInfo(xdev->dpy, VisualIDMask, &xvinfo, &nitems);
     if (xdev->vinfo == NULL) {
-	eprintf("gs: Cannot get XVisualInfo.\n");
-	exit(1);
+	eprintf("Cannot get XVisualInfo.\n");
+	return_error(gs_error_ioerror);
     }
 
     /* Buggy X servers may cause a Bad Access on XFreeColors. */
@@ -376,7 +378,11 @@ gdev_x_open(register gx_device_X *xdev)
     XQueryColor(xdev->dpy, xdev->cmap, &xc);
     XAllocColor(xdev->dpy, xdev->cmap, &xc);
 
-    gdev_x_setup_colors(xdev);
+    code = gdev_x_setup_colors(xdev);
+    if ( code < 0 )
+      { XCloseDisplay(xdev->dpy);
+        return code;
+      }
     gdev_x_setup_fontmap(xdev);
 
     if (!xdev->ghostview) {
@@ -547,6 +553,11 @@ gdev_x_open(register gx_device_X *xdev)
 	if (xdev->pwin == (Window)None) {	/* there isn't a next event for existing windows */
 	    XNextEvent(xdev->dpy, &event);
 	}
+	/* Now turn off graphics exposure events so they don't queue up */
+	/* indefinitely.  Also, since we can't do anything about real */
+	/* Expose events, mask them out. */
+	XSetGraphicsExposures(xdev->dpy, xdev->gc, False);
+	XSelectInput(xdev->dpy, xdev->win, NoEventMask);
     } else {
 	/* Create an unmapped window, that the window manager will ignore.
 	 * This invisible window will be used to receive "next page"
@@ -725,7 +736,7 @@ setup_cube(gx_device_X *xdev, int ramp_size, bool colors)
 }
 
 /* Setup color mapping. */
-private void
+private int
 gdev_x_setup_colors(gx_device_X *xdev)
 {
     char palette;
@@ -750,8 +761,8 @@ gdev_x_setup_colors(gx_device_X *xdev)
 	    nitems = sscanf(buf, "%*s %ld %ld", &(xdev->foreground),
 			    		        &(xdev->background));
 	    if (nitems != 2 || (*buf != 'M' && *buf != 'G' && *buf != 'C')) {
-		eprintf("gs: Malformed ghostview color property.\n");
-		exit(1);
+		eprintf("Malformed ghostview color property.\n");
+		return_error(gs_error_rangecheck);
 	    }
 	    palette = max(palette, *buf);
 	}
@@ -876,9 +887,10 @@ monochrome:
 	xdev->color_info.max_gray = 1;
 	xdev->color_info.dither_grays = 2;
     } else {
-	eprintf1("gs: Unknown palette: %s\n", xdev->palette);
-	exit(1);
+	eprintf1("Unknown palette: %s\n", xdev->palette);
+	return_error(gs_error_rangecheck);
     }
+    return 0;
 }
 
 /* ------ Initialize font mapping ------ */
