@@ -51,7 +51,7 @@ mrealloc(void *old_ptr, size_t old_size, size_t new_size)
  * (3) the VMS shell is radically different from all others.
  *
  * Usage:
- *	genconf [-Z] [-n [prefix | -]] [@]xxx.dev*
+ *	genconf [-Z] [-n [name_prefix | -]] [@]xxx.dev*
  *	  [-f gconfigf.h] [-h gconfig.h]
  *	  [-p[l|L][u][e] pattern] [-l|o|lo|ol out.tr]
  * & in the pattern acts as an escape character as follows:
@@ -62,6 +62,7 @@ mrealloc(void *old_ptr, size_t old_size, size_t new_size)
  *	&x, for any other character x, is an error.
  */
 
+/* DEFAULT_PREFIX should be DEFAULT_NAME_PREFIX. */
 #define DEFAULT_PREFIX "gs_"
 
 /* Structures for accumulating information. */
@@ -91,7 +92,8 @@ typedef struct string_pattern_s {
 } string_pattern;
 typedef struct config_s {
 	int debug;
-	const char *prefix;
+	const char *name_prefix;
+	const char *file_prefix;
 	/* file_names and file_contents are special.... */
 	string_list file_names;
 	string_list file_contents;
@@ -110,7 +112,8 @@ typedef struct config_s {
 /* small allocations. */
 static const config init_config = {
 	0,			/* debug */
-	DEFAULT_PREFIX,		/* prefix */
+	DEFAULT_PREFIX,		/* name_prefix */
+	"",			/* file_prefix */
 	{ 200 },		/* file_names */
 	{ 200 },		/* file_contents */
 	{ 100, uniq_first },	/* resources */
@@ -172,8 +175,13 @@ main(int argc, char *argv[])
 		}
 		switch ( arg[1] )
 		{
+		case 'C':	/* change directory, by analogy with make */
+			conf.file_prefix =
+			  (argv[i + 1][0] == '-' ? "" : argv[i + 1]);
+			++i;
+			continue;
 		case 'n':
-			conf.prefix =
+			conf.name_prefix =
 			  (argv[i + 1][0] == '-' ? "" : argv[i + 1]);
 			++i;
 			continue;
@@ -245,7 +253,7 @@ pbreak:			if ( pat == &conf.obj_p )
 			{ char template[80];
 			  sprintf(template,
 				  "font_(\"0.font_%%s\",%sf_%%s,zf_%%s)\n",
-				  conf.prefix);
+				  conf.name_prefix);
 			  write_list(out, &conf.fonts, template);
 			}
 			break;
@@ -253,7 +261,7 @@ pbreak:			if ( pat == &conf.obj_p )
 			fputs("/* This file was generated automatically by genconf.c. */\n", out);
 			{ char template[80];
 			  sprintf(template, "device_(%s%%s_device)\n",
-				  conf.prefix);
+				  conf.name_prefix);
 			  write_list(out, &conf.devs, template);
 			}
 			sort_uniq(&conf.resources);
@@ -303,22 +311,32 @@ alloc_list(string_list *list)
 /* of the uniq_modes of all (direct and indirect) items in the file. */
 string_item *
 read_file(config *pconf, const char *fname)
-{	int i;
+{	char *cname = malloc(strlen(fname) + strlen(pconf->file_prefix) + 1);
+	int i;
 	FILE *in;
 	int end, nread;
 	char *cont;
 	string_item *item;
 
+	if ( cname == 0 )
+	  { fprintf(stderr, "Can't allocate space for file name %s%s.\n",
+		    pconf->file_prefix, fname);
+	    exit(1);
+	  }
+	strcpy(cname, pconf->file_prefix);
+	strcat(cname, fname);
 	for ( i = 0; i < pconf->file_names.count; ++i )
-	  if ( !strcmp(pconf->file_names.items[i].str, fname) )
-	    return &pconf->file_contents.items[i];
+	  if ( !strcmp(pconf->file_names.items[i].str, cname) )
+	    { free(cname);
+	      return &pconf->file_contents.items[i];
+	    }
 	/* Try to open the file in binary mode, to avoid the overhead */
 	/* of unnecessary EOL conversion in the C library. */
-	in = fopen(fname, "rb");
+	in = fopen(cname, "rb");
 	if ( in == 0 )
-	{	in = fopen(fname, "r");
+	{	in = fopen(cname, "r");
 		if ( in == 0 )
-		{	fprintf(stderr, "Can't read %s.\n", fname);
+		{	fprintf(stderr, "Can't read %s.\n", cname);
 			exit(1);
 		}
 	}
@@ -327,7 +345,7 @@ read_file(config *pconf, const char *fname)
 	cont = malloc(end + 1);
 	if ( cont == 0 )
 	{	fprintf(stderr, "Can't allocate %d bytes to read %s.\n",
-			end + 1, fname);
+			end + 1, cname);
 		exit(1);
 	}
 	rewind(in);
@@ -335,8 +353,8 @@ read_file(config *pconf, const char *fname)
 	fclose(in);
 	cont[nread] = 0;
 	if ( pconf->debug )
-	  printf("File %s = %d bytes.\n", fname, nread);
-	add_item(&pconf->file_names, fname);
+	  printf("File %s = %d bytes.\n", cname, nread);
+	add_item(&pconf->file_names, cname);
 	item = add_item(&pconf->file_contents, cont);
 	item->index = 0;		/* union of uniq_modes */
 	return item;
@@ -458,7 +476,7 @@ add_entry(config *pconf, char *category, const char *item)
 			  { pat = "io_device_(%siodev_%%s)"; }
 			else
 			  goto err;
-			sprintf(template, pat, pconf->prefix);
+			sprintf(template, pat, pconf->name_prefix);
 			pat = template;
 			break;
 		case 'l':
@@ -469,7 +487,12 @@ add_entry(config *pconf, char *category, const char *item)
 			goto err;
 		case 'o':
 			if ( is_cat("obj") )
-			  { list = &pconf->objs; break; }
+			  { list = &pconf->objs;
+			    strcpy(template, pconf->file_prefix);
+			    strcat(template, "%s");
+			    pat = template;
+			    break;
+			  }
 			if ( is_cat("oper") )
 			  { pat = "oper_(%s_op_defs)"; break; }
 			goto err;
