@@ -1,4 +1,4 @@
-/* Copyright (C) 1995, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -56,6 +56,7 @@ private RELOC_PTRS_BEGIN(ref_memory_reloc_ptrs) {
 } RELOC_PTRS_END
 
 /* Forward references */
+private ulong compute_free_objects(P1(gs_ref_memory_t *));
 private obj_header_t *alloc_obj(P5(gs_ref_memory_t *, ulong, gs_memory_type_ptr_t, bool, client_name_t));
 private chunk_t *alloc_add_chunk(P4(gs_ref_memory_t *, ulong, bool, client_name_t));
 void alloc_close_chunk(P1(gs_ref_memory_t *));
@@ -86,7 +87,8 @@ private gs_memory_proc_register_root(i_register_root);
 private gs_memory_proc_unregister_root(i_unregister_root);
 private gs_memory_proc_status(i_status);
 private gs_memory_proc_enable_free(i_enable_free);
-private const gs_memory_procs_t imemory_procs = {
+/* We export the procedures for subclasses. */
+const gs_memory_procs_t gs_ref_memory_procs = {
 	i_alloc_bytes,
 	i_alloc_bytes_immovable,
 	i_alloc_struct,
@@ -120,7 +122,7 @@ ialloc_alloc_state(gs_memory_t *parent, uint chunk_size)
 
 	if ( iimem == 0 )
 	  return 0;
-	iimem->procs = imemory_procs;
+	iimem->procs = gs_ref_memory_procs;
 	iimem->parent = parent;
 	iimem->chunk_size = chunk_size;
 	iimem->large_size = ((chunk_size / 4) & -obj_align_mod) + 1;
@@ -187,7 +189,9 @@ void
 ialloc_reset_free(gs_ref_memory_t *mem)
 {	int i;
 	obj_header_t **p;
-	mem->freed_lost = 0;
+	mem->lost.objects = 0;
+	mem->lost.refs = 0;
+	mem->lost.strings = 0;
 	mem->cfreed.cp = 0;
 	for ( i = 0, p = &mem->freelists[0]; i < num_freelists; i++, p++ )
 	  *p = 0;
@@ -412,9 +416,10 @@ i_resize_object(gs_memory_t *mem, void *obj, uint new_num_elements,
 	void *new_obj;
 
 	if ( (byte *)obj + obj_align_round(old_size) == imem->cc.cbot &&
-	     imem->cc.ctop - imem->cc.cbot < new_size + obj_align_mod
+	     imem->cc.ctop - imem->cc.cbot > new_size + obj_align_mod
 	   )
 	  { imem->cc.cbot = (byte *)obj + obj_align_round(new_size);
+	    pp->o_size = new_size;
 	    if_debug8('A', "[a%d:%c%c ]%s %s(%lu=>%lu) 0x%lx\n",
 		      imem->space,
 		      (new_size > old_size ? '>' : '<'),
@@ -444,11 +449,40 @@ i_free_object(gs_memory_t *mem, void *ptr, client_name_t cname)
 	pp = (obj_header_t *)ptr - 1;
 #ifdef DEBUG
 	if ( gs_debug_c('?') )
-	  {	if ( pp->o_type == &st_free )
+	  {	chunk_locator_t cld;
+
+		if ( pp->o_type == &st_free )
 		  {	lprintf2("%s: object 0x%lx already free!\n",
-				 client_name_string(cname), (ulong)pp);
+				 client_name_string(cname), (ulong)ptr);
 			return;/*gs_abort();*/
 		  }
+		/* Check that this allocator owns the object being freed. */
+		cld.memory = imem;
+		while ( (cld.cp = cld.memory->clast),
+			!chunk_locate_ptr(ptr, &cld)
+		      )
+		  { if ( !cld.memory->saved )
+		      { lprintf3("%s: freeing 0x%lx, not owned by memory 0x%lx!\n",
+				 client_name_string(cname), (ulong)ptr,
+				 (ulong)mem);
+		        return;/*gs_abort();*/
+		      }
+		  /****** HACK: we know the saved state is the first ******
+		   ****** member of an alloc_save_t. ******/
+		    cld.memory = (gs_ref_memory_t *)cld.memory->saved;
+		  }
+		/* Check that the object is in the allocated region. */
+		if ( cld.memory == imem && cld.cp == imem->pcc )
+		  cld.cp = &imem->cc;
+		if ( !(ptr_between((const byte *)pp, cld.cp->cbase,
+				   cld.cp->cbot))
+		   )
+		  { lprintf5("%s: freeing 0x%lx,\n\toutside chunk 0x%lx cbase=0x%lx, cbot=0x%lx!\n",
+			     client_name_string(cname), (ulong)ptr,
+			     (ulong)cld.cp, (ulong)cld.cp->cbase,
+			     (ulong)cld.cp->cbot);
+		    return;/*gs_abort();*/
+		  }
 	  }
 #endif
 	size = pre_obj_contents_size(pp);
@@ -461,7 +495,7 @@ i_free_object(gs_memory_t *mem, void *ptr, client_name_t cname)
 	  }
 	if ( (byte *)ptr + obj_align_round(size) == imem->cc.cbot )
 	{	if_debug4('A', "[a%d:-o ]%s(%u) 0x%lx\n", imem->space,
-			  client_name_string(cname), size, (ulong)pp);
+			  client_name_string(cname), size, (ulong)ptr);
 		gs_alloc_fill(ptr, gs_alloc_fill_free, size);
 		imem->cc.cbot = (byte *)pp;
 		return;
@@ -479,7 +513,7 @@ i_free_object(gs_memory_t *mem, void *ptr, client_name_t cname)
 		  cld.cp = 0;
 		  if_debug5('a', "[a%d:-o%c]%s(%u) 0x%lx\n", imem->space,
 			    (chunk_locate_ptr(ptr, &cld) ? 'L' : '~'),
-			    client_name_string(cname), size, (ulong)pp);
+			    client_name_string(cname), size, (ulong)ptr);
 		}
 #endif
 		cl.memory = imem;
@@ -491,7 +525,7 @@ i_free_object(gs_memory_t *mem, void *ptr, client_name_t cname)
 		/* Don't overwrite even if gs_alloc_debug is set. */
 	}
 	if ( size <= max_freelist_size &&
-	     size >= sizeof(obj_header_t *)
+	     obj_align_round(size) >= sizeof(obj_header_t *)
 	   )
 	  {	/*
 		 * Put the object on a freelist, unless it belongs to
@@ -509,7 +543,7 @@ i_free_object(gs_memory_t *mem, void *ptr, client_name_t cname)
 			*pfl = (obj_header_t *)ptr;
 			if_debug4('A', "[a%d:-oF]%s(%u) 0x%lx\n",
 				  imem->space, client_name_string(cname),
-				  size, (ulong)pp);
+				  size, (ulong)ptr);
 			return;
 		  }
 		/* Don't overwrite even if gs_alloc_debug is set. */
@@ -519,8 +553,8 @@ i_free_object(gs_memory_t *mem, void *ptr, client_name_t cname)
 		gs_alloc_fill(ptr, gs_alloc_fill_free, size);
 	  }
 	if_debug4('A', "[a%d:-o#]%s(%u) 0x%lx\n", imem->space,
-		  client_name_string(cname), size, (ulong)pp);
-	imem->freed_lost += obj_size_round(size);
+		  client_name_string(cname), size, (ulong)ptr);
+	imem->lost.objects += obj_size_round(size);
 }
 private byte *
 i_alloc_string(gs_memory_t *mem, uint nbytes, client_name_t cname)
@@ -553,6 +587,8 @@ top:	if ( imem->cc.ctop - imem->cc.cbot > nbytes )
 		alloc_close_chunk(imem);
 		imem->pcc = cp;
 		imem->cc = *imem->pcc;
+		gs_alloc_fill(imem->cc.cbase, gs_alloc_fill_free,
+			      imem->cc.climit - imem->cc.cbase);
 		goto top;
 	}
 }
@@ -580,17 +616,20 @@ i_resize_string(gs_memory_t *mem, byte *data, uint old_num, uint new_num,
 		imem->cc.ctop - imem->cc.cbot > new_num - old_num)
 	   )
 	{	/* Resize in place. */
-		register uint i;
 		ptr = data + old_num - new_num;
 		if_debug6('A', "[a%d:%c> ]%s(%u->%u) 0x%lx\n",
 			  imem->space, (new_num > old_num ? '>' : '<'),
 			  client_name_string(cname), old_num, new_num,
 			  (ulong)ptr);
 		imem->cc.ctop = ptr;
-		if ( new_num < old_num )
-		  for ( i = new_num; i > 0; ) --i, ptr[i] = data[i];
+		memmove(ptr, data, min(old_num, new_num));
+#ifdef DEBUG
+		if ( new_num > old_num )
+		  gs_alloc_fill(ptr + old_num, gs_alloc_fill_alloc,
+				new_num - old_num);
 		else
-		  for ( i = 0; i < old_num; ) ptr[i] = data[i], i++;
+		  gs_alloc_fill(data, gs_alloc_fill_free, old_num - new_num);
+#endif
 	}
 	else
 	{	/* Punt. */
@@ -613,13 +652,15 @@ i_free_string(gs_memory_t *mem, byte *data, uint nbytes,
 	else
 	{	if_debug4('A', "[a%d:->#]%s(%u) 0x%lx\n", imem->space,
 			  client_name_string(cname), nbytes, (ulong)data);
-		imem->freed_lost += nbytes;
+		imem->lost.strings += nbytes;
 	}
+	gs_alloc_fill(data, gs_alloc_fill_free, nbytes);
 }
 private void
 i_status(gs_memory_t *mem, gs_memory_status_t *pstat)
-{	ulong unused = imem->freed_lost;
+{	ulong unused = imem->lost.refs + imem->lost.strings;
 	ulong inner = 0;
+
 	alloc_close_chunk(imem);
 	/* Add up unallocated space within each chunk. */
 	/* Also keep track of space allocated to inner chunks, */
@@ -632,18 +673,7 @@ i_status(gs_memory_t *mem, gs_memory_status_t *pstat)
 			cp = cp->cnext;
 		  }
 	  }
-	/* Add up space on free lists. */
-	  {	int i;
-		const obj_header_t *pfree;
-		for ( i = 0; i < num_freelists; i++ )
-		  {	uint free_size =
-			  (i << log2_obj_align_mod) + sizeof(obj_header_t);
-			for ( pfree = imem->freelists[i]; pfree != 0;
-			      pfree = *(const obj_header_t * const *)pfree
-			    )
-			  unused += free_size;
-		  }
-	  }
+	unused += compute_free_objects(imem);
 	pstat->used = imem->allocated + inner - unused +
 	  imem->previous_status.used;
 	pstat->allocated = imem->allocated +
@@ -661,6 +691,26 @@ i_enable_free(gs_memory_t *mem, bool enable)
 
 /* ------ Internal procedures ------ */
 
+/* Compute the amount of free object space by scanning free lists. */
+private ulong
+compute_free_objects(gs_ref_memory_t *mem)
+{	ulong unused = mem->lost.objects;
+	int i;
+
+	/* Add up space on free lists. */
+	for ( i = 0; i < num_freelists; i++ )
+	  {	uint free_size =
+		  (i << log2_obj_align_mod) + sizeof(obj_header_t);
+		const obj_header_t *pfree;
+
+		for ( pfree = mem->freelists[i]; pfree != 0;
+		      pfree = *(const obj_header_t * const *)pfree
+		    )
+		  unused += free_size;
+	  }
+	return unused;
+}
+
 /* Allocate an object.  This handles all but the fastest, simplest case. */
 private obj_header_t *
 alloc_obj(gs_ref_memory_t *mem, ulong lsize, gs_memory_type_ptr_t pstype,
@@ -678,9 +728,6 @@ alloc_obj(gs_ref_memory_t *mem, ulong lsize, gs_memory_type_ptr_t pstype,
 			return 0;
 		ptr = (obj_header_t *)cp->cbot;
 		cp->cbot += asize;
-		if_debug5('a', "[a%d%c+ L]%s(%lu) = 0x%lx\n",
-			  imem->space, (immovable ? '|' : ':'),
-			  client_name_string(cname), lsize, (ulong)ptr);
 		ptr->o_large = 1;
 		pre_obj_set_large_size(ptr, lsize);
 	}
@@ -698,6 +745,8 @@ alloc_obj(gs_ref_memory_t *mem, ulong lsize, gs_memory_type_ptr_t pstype,
 			alloc_close_chunk(mem);
 			mem->pcc = cp;
 			mem->cc = *mem->pcc;
+			gs_alloc_fill(mem->cc.cbase, gs_alloc_fill_free,
+				      mem->cc.climit - mem->cc.cbase);
 		}
 		mem->cc.cbot = (byte *)ptr + asize;
 		ptr->o_large = 0;
@@ -759,7 +808,7 @@ alloc_link_chunk(chunk_t *cp, gs_ref_memory_t *mem)
 	if ( prev == 0 )
 		imem->cfirst = cp;
 	else
-		prev->cnext= cp;
+		prev->cnext = cp;
 	if ( imem->pcc != 0 )
 	{	imem->cc.cnext = imem->pcc->cnext;
 		imem->cc.cprev = imem->pcc->cprev;
@@ -840,15 +889,28 @@ alloc_init_chunk(chunk_t *cp, byte *bot, byte *top, bool has_strings,
 		cp->smark_size = string_quanta_mark_size(nquanta);
 		cp->sreloc =
 		  (string_reloc_offset *)(cp->smark + cp->smark_size);
+		cp->sfree1 = (ushort *)cp->sreloc;
 	}
 	else
 	{	/* No strings, don't need the string GC tables. */
 		cp->climit = cp->cend;
+		cp->sfree1 = 0;
 		cp->smark = 0;
 		cp->smark_size = 0;
 		cp->sreloc = 0;
 	}
 	cp->ctop = cp->climit;
+	alloc_init_free_strings(cp);
+}
+
+/* Initialize the string freelists in a chunk. */
+void
+alloc_init_free_strings(chunk_t *cp)
+{	if ( cp->sfree1 )
+	  memset(cp->sfree1, 0,
+		 ((cp->climit - csbase(cp) + 255) >> 8) *
+		   sizeof(*cp->sfree1));
+	cp->sfree = 0;
 }
 
 /* Close up the current chunk. */
@@ -973,10 +1035,197 @@ chunk_locate_ptr(const void *vptr, chunk_locator_t *clp)
 
 /* ------ Debugging printout ------ */
 
+/*
+ * All of this code should be in a separate file, but we added it just
+ * before a release, and it would have been too disruptive to add a new
+ * file at this point.
+ */
+
 #ifdef DEBUG
 
+#include "string_.h"
+
+/*
+ * Define the options for a memory dump.  These may be or'ed together.
+ */
+typedef enum {
+  dump_do_default = 0,		/* pro forma */
+  dump_do_strings = 1,
+  dump_do_type_addresses = 2,
+  dump_do_no_types = 4,
+  dump_do_pointers = 8,
+  dump_do_pointed_strings = 16,	/* only if do_pointers also set */
+  dump_do_contents = 32,
+  dump_do_marks = 64
+} dump_options_t;
+
+/*
+ * Define all the parameters controlling what gets dumped.
+ */
+typedef struct dump_control_s {
+  dump_options_t options;
+  const byte *bottom;
+  const byte *top;
+} dump_control_t;
+#define obj_in_control_region(obot, otop, pdc)\
+  ( ((pdc)->bottom == NULL || (const byte *)(otop) > (pdc)->bottom) &&\
+    ((pdc)->top == NULL || (const byte *)(obot) < (pdc)->top) )
+const dump_control_t dump_control_default = {
+  dump_do_default, NULL, NULL
+};
+const dump_control_t dump_control_all = {
+  dump_do_strings | dump_do_type_addresses | dump_do_pointers |
+  dump_do_pointed_strings | dump_do_contents, NULL, NULL
+};
+
+/*
+ * Internal procedure to dump a block of memory, in hex and optionally
+ * also as characters.
+ */
+private void
+debug_indent(int indent)
+{	int i;
+	for ( i = indent; i > 0; --i )
+	  dputc(' ');
+}
+private void
+debug_dump_contents(const byte *bot, const byte *top, int indent,
+  bool as_chars)
+{	const byte *block;
+#define block_size 16
+
+	if ( bot >= top )
+	  return;
+	for ( block = bot - ((bot - (byte *)0) & (block_size - 1));
+	      block < top; block += block_size
+	    ) {
+	  int i;
+	  char label[12];
+
+	  /* Check for repeated blocks. */
+	  if ( block >= bot + block_size &&
+	       block <= top - (block_size * 2) &&
+	       !memcmp(block, block - block_size, block_size) &&
+	       !memcmp(block, block + block_size, block_size)
+	     ) {
+	    if ( block < bot + block_size * 2 ||
+		 memcmp(block, block - block_size * 2, block_size)
+	       ) {
+	      debug_indent(indent);
+	      dputs("  ...\n");
+	    }
+	    continue;
+	  }
+	  sprintf(label, "0x%lx:", (ulong)block);
+	  debug_indent(indent);
+	  dputs(label);
+	  for ( i = 0; i < block_size; ++i ) {
+	    const char *sepr = ((i & 3) == 0 && i != 0 ? "  " : " ");
+
+	    dputs(sepr);
+	    if ( block + i >= bot && block + i < top )
+	      dprintf1("%02x", block[i]);
+	    else
+	      dputs("  ");
+	  }
+	  dputc('\n');
+	  if ( as_chars ) {
+	    debug_indent(indent + strlen(label));
+	    for ( i = 0; i < block_size; ++i ) {
+	      byte ch;
+	      if ( (i & 3) == 0 && i != 0 )
+		dputc(' ');
+	      if ( block + i >= bot && block + i < top &&
+		   (ch = block[i]) >= 32 && ch <= 126
+		 )
+		dprintf1("  %c", ch);
+	      else
+		dputs("   ");
+	    }
+	    dputc('\n');
+	  }
+	}
+#undef block_size
+}
+
+/* Print one object with the given options. */
+/* Relevant options: type_addresses, no_types, pointers, pointed_strings, */
+/* contents. */
 void
-debug_print_chunk(const chunk_t *cp)
+debug_print_object(const void *obj, const dump_control_t *control)
+{	const obj_header_t *pre = ((const obj_header_t *)obj) - 1;
+	ulong size = pre_obj_contents_size(pre);
+	const gs_memory_struct_type_t *type = pre->o_type;
+	dump_options_t options = control->options;
+
+	dprintf3("  pre=0x%lx(obj=0x%lx) size=%lu", (ulong)pre, (ulong)obj,
+		 size);
+	switch ( options & (dump_do_type_addresses | dump_do_no_types) )
+	  {
+	  case dump_do_type_addresses + dump_do_no_types: /* addresses only */
+	    dprintf1(" type=0x%lx", (ulong)type);
+	    break;
+	  case dump_do_type_addresses: /* addresses & names */
+	    dprintf2(" type=%s(0x%lx)", struct_type_name_string(type),
+		     (ulong)type);
+	    break;
+	  case 0:		/* names only */
+	    dprintf1(" type=%s", struct_type_name_string(type));
+	  case dump_do_no_types: /* nothing */
+	    ;
+	  }
+	if ( options & dump_do_marks ) {
+	  if ( pre->o_large )
+	    dprintf1(" lmark=%d", pre->o_lmark);
+	  else
+	    dprintf2(" smark/back=%u (0x%x)", pre->o_smark, pre->o_smark);
+	}
+	dputc('\n');
+	if ( type == &st_free )
+	  return;
+	if ( options & dump_do_pointers ) {
+	  struct_proc_enum_ptrs((*proc)) = type->enum_ptrs;
+	  uint index = 0;
+	  const void *ptr;
+	  gs_ptr_type_t ptype;
+
+	  /*
+	   * NOTE: the following cast should be unnecessary, but that
+	   * will require adding 'const' to the first prototype argument
+	   * of struct_proc_enum_ptrs.
+	   */
+	  if ( proc != 0 )
+	    for ( ; (ptype = (*proc)((obj_header_t *)pre + 1, size, index, &ptr)) != 0;
+		  ++index
+		) {
+	      dprintf1("    ptr %u: ", index);
+	      if ( ptype == ptr_string_type || ptype == ptr_const_string_type ) {
+		const gs_const_string *str = (const gs_const_string *)ptr;
+
+		dprintf2("0x%lx(%u)", (ulong)str->data, str->size);
+		if ( options & dump_do_pointed_strings ) {
+		  dputs(" =>\n");
+		  debug_dump_contents(str->data, str->data + str->size, 6,
+				      true);
+		} else {
+		  dputc('\n');
+		}
+	      } else {
+		dprintf1((ptr_between(ptr, obj, (const byte *)obj + size) ?
+			  "(0x%lx)\n" : "0x%lx\n"), (ulong)ptr);
+	      }
+	    }
+	}
+	if ( options & dump_do_contents ) {
+	  debug_dump_contents((const byte *)obj, (const byte *)obj + size,
+			      0, false);
+	}
+}
+
+/* Print the contents of a chunk with the given options. */
+/* Relevant options: all. */
+void
+debug_dump_chunk(const chunk_t *cp, const dump_control_t *control)
 {	dprintf1("chunk at 0x%lx:\n", (ulong)cp);
 	dprintf3("   chead=0x%lx  cbase=0x%lx sbase=0x%lx\n",
 		 (ulong)cp->chead, (ulong)cp->cbase, (ulong)cp->sbase);
@@ -991,20 +1240,47 @@ debug_print_chunk(const chunk_t *cp)
 		 (ulong)cp->cprev, (ulong)cp->cnext, (ulong)cp->outer,
 		 cp->inner_count, (cp->has_refs? "true" : "false"));
 
+	dprintf2("  sfree1=0x%lx   sfree=0x%x\n",
+		 (ulong)cp->sfree1, cp->sfree);
+	if ( control->options & dump_do_strings ) {
+	  debug_dump_contents((control->bottom == 0 ? cp->ctop :
+			       max(control->bottom, cp->ctop)),
+			      (control->top == 0 ? cp->climit :
+			       min(control->top, cp->climit)),
+			      0, true);
+	}
 	SCAN_CHUNK_OBJECTS(cp)
 	  DO_ALL
-	    dprintf3("    pre=0x%lx type=0x%lx size=%lu  ",
-		     (ulong)pre, (ulong)pre->o_type, size);
-	    if ( pre->o_large )
-	      dprintf2("lmark=%d large size=%u\n",
-		       pre->o_lmark, pre->o_lsize);
-	    else
-	      dprintf2("smark/back=%u (0x%x)\n", pre->o_smark, pre->o_smark);
+	    if ( obj_in_control_region(pre + 1,
+				       (const byte *)(pre + 1) + size,
+				       control)
+	       )
+	      debug_print_object(pre + 1, control);
 /* Temporarily redefine gs_exit so a chunk parsing error */
 /* won't actually exit. */
 #define gs_exit(n) DO_NOTHING
 	END_OBJECTS_SCAN
 #undef gs_exit
 }
+void debug_print_chunk(const chunk_t *cp)
+{	dump_control_t control;
+
+	control = dump_control_default;
+	debug_dump_chunk(cp, &control);
+}
+
+/* Print the contents of all chunks managed by an allocator. */
+/* Relevant options: all. */
+void
+debug_dump_memory(const gs_ref_memory_t *mem, const dump_control_t *control)
+{	const chunk_t *mcp;
+
+	for ( mcp = mem->cfirst; mcp != 0; mcp = mcp->cnext ) {
+	  const chunk_t *cp = (mcp == mem->pcc ? &mem->cc : mcp);
+
+	  if ( obj_in_control_region(cp->cbase, cp->cend, control) )
+	    debug_dump_chunk(cp, control);
+	}
+}
 
 #endif					/* DEBUG */
