@@ -1,4 +1,4 @@
-/* Copyright (C) 1992, 1995, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1992, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -91,16 +91,6 @@ private void near cie_matrix_init(P1(gs_matrix3 *));
 
 /* Allocator structure types */
 private_st_joint_caches();
-private_st_const_string();
-public_st_const_string_element();
-#define sptr ((gs_const_string *)vptr)
-private ENUM_PTRS_BEGIN(const_string_enum_ptrs) return 0;
-	case 0: *pep = (void *)sptr; return ptr_const_string_type;
-ENUM_PTRS_END
-private RELOC_PTRS_BEGIN(const_string_reloc_ptrs) {
-	gs_reloc_const_string(sptr, gcst);
-} RELOC_PTRS_END
-#undef sptr
 
 /* Define the CIE color space types. */
 /* We use CIExxx rather than CIEBasedxxx in some places because */
@@ -256,27 +246,27 @@ const gs_cie_render_table_procs RenderTableT_default =
 /* ------ Adjust reference counts for a CIE color space ------ */
 
 private void
-gx_adjust_cspace_CIEDEFG(const gs_color_space *pcs, gs_state *pgs, int delta)
-{	rc_adjust_const(pcs->params.defg, delta, pgs->memory,
-			"gx_adjust_cspace_CIEDEFG");
+gx_adjust_cspace_CIEDEFG(const gs_color_space *pcs, gs_memory_t *mem,
+  int delta)
+{	rc_adjust_const(pcs->params.defg, delta, "gx_adjust_cspace_CIEDEFG");
 }
 
 private void
-gx_adjust_cspace_CIEDEF(const gs_color_space *pcs, gs_state *pgs, int delta)
-{	rc_adjust_const(pcs->params.def, delta, pgs->memory,
-			"gx_adjust_cspace_CIEDEF");
+gx_adjust_cspace_CIEDEF(const gs_color_space *pcs, gs_memory_t *mem,
+  int delta)
+{	rc_adjust_const(pcs->params.def, delta, "gx_adjust_cspace_CIEDEF");
 }
 
 private void
-gx_adjust_cspace_CIEABC(const gs_color_space *pcs, gs_state *pgs, int delta)
-{	rc_adjust_const(pcs->params.abc, delta, pgs->memory,
-			"gx_adjust_cspace_CIEABC");
+gx_adjust_cspace_CIEABC(const gs_color_space *pcs, gs_memory_t *mem,
+  int delta)
+{	rc_adjust_const(pcs->params.abc, delta, "gx_adjust_cspace_CIEABC");
 }
 
 private void
-gx_adjust_cspace_CIEA(const gs_color_space *pcs, gs_state *pgs, int delta)
-{	rc_adjust_const(pcs->params.a, delta, pgs->memory,
-			"gx_adjust_cspace_CIEA");
+gx_adjust_cspace_CIEA(const gs_color_space *pcs, gs_memory_t *mem,
+  int delta)
+{	rc_adjust_const(pcs->params.a, delta, "gx_adjust_cspace_CIEA");
 }
 
 /* ================ Table setup ================ */
@@ -404,8 +394,7 @@ gs_setcolorrendering(gs_state *pgs, gs_cie_render *pcie)
 {	int code = gs_cie_render_init(pcie);
 	if ( code < 0 )
 	  return code;
-	rc_assign(pgs->cie_render, pcie, pgs->memory,
-		  "gs_setcolorrendering");
+	rc_assign(pgs->cie_render, pcie, "gs_setcolorrendering");
 	/* Load the caches. */
 	CIE_LOAD_CACHE3_BODY(pcie->caches.EncodeLMN, pcie->DomainLMN.ranges,
 			     &pcie->EncodeLMN, pcie, "EncodeLMN");
@@ -765,8 +754,9 @@ cie_joint_caches_complete(gx_cie_joint_caches *pjc,
 /* ================ Color rendering (using the caches) ================ */
 
 private int near cie_remap_finish(P3(const cie_cached_vector3 *,
-  frac *, const gs_state *));
-private void near cie_lookup_mult3(P2(cie_cached_vector3 *, const gx_cie_vector_cache *));
+				     frac *, const gs_imager_state *));
+private void near cie_lookup_mult3(P2(cie_cached_vector3 *,
+				      const gx_cie_vector_cache *));
 #ifdef DEBUG
 private void near
 cie_lookup_map3(cie_cached_vector3 *pvec,
@@ -785,25 +775,21 @@ cie_lookup_map3(cie_cached_vector3 *pvec,
 #endif
 
 /* Determine the concrete space underlying a CIEBased space. */
-private const gs_color_space cie_rgb_space =
-	{ &gs_color_space_type_DeviceRGB };
-private const gs_color_space cie_cmyk_space =
-	{ &gs_color_space_type_DeviceCMYK };
 private const gs_color_space *
-gx_concrete_space_CIE(const gs_color_space *pcs, const gs_state *pgs)
-{	const gs_cie_render *pcie = pgs->cie_render;
+gx_concrete_space_CIE(const gs_color_space *pcs, const gs_imager_state *pis)
+{	const gs_cie_render *pcie = pis->cie_render;
 	if ( pcie == 0 || pcie->RenderTable.lookup.table == 0 ||
 	     pcie->RenderTable.lookup.m == 3
 	   )
-	  return &cie_rgb_space;
+	  return gs_color_space_DeviceRGB();
 	else				/* pcie->RenderTable.lookup.m == 4 */
-	  return &cie_cmyk_space;
+	  return gs_color_space_DeviceCMYK();
 }
 
 /* Render a CIEBasedDEFG color. */
 private int
 gx_concretize_CIEDEFG(const gs_client_color *pc, const gs_color_space *pcs,
-  frac *pconc, const gs_state *pgs)
+  frac *pconc, const gs_imager_state *pis)
 {	/****** NOT IMPLEMENTED YET ******/
 	return_error(gs_error_undefined);
 }
@@ -811,7 +797,7 @@ gx_concretize_CIEDEFG(const gs_client_color *pc, const gs_color_space *pcs,
 /* Render a CIEBasedDEF color. */
 private int
 gx_concretize_CIEDEF(const gs_client_color *pc, const gs_color_space *pcs,
-  frac *pconc, const gs_state *pgs)
+  frac *pconc, const gs_imager_state *pis)
 {	/****** NOT IMPLEMENTED YET ******/
 	return_error(gs_error_undefined);
 }
@@ -821,7 +807,8 @@ gx_concretize_CIEDEF(const gs_client_color *pc, const gs_color_space *pcs,
 /* needs to be efficient. */
 private int
 gx_remap_CIEABC(const gs_client_color *pc, const gs_color_space *pcs,
-  gx_device_color *pdc, const gs_state *pgs)
+  gx_device_color *pdc, const gs_imager_state *pis, gx_device *dev,
+  gs_color_select_t select)
 {	frac conc[4];
 	const gs_cie_abc *pcie = pcs->params.abc;
 	cie_cached_vector3 vec3;
@@ -840,20 +827,21 @@ gx_remap_CIEABC(const gs_client_color *pc, const gs_color_space *pcs,
 	  cie_lookup_map3(&vabc /*&vlmn*/, &pcie->caches.DecodeABC[0],
 			  "Decode/MatrixABC");
 #undef vabc
-	switch ( cie_remap_finish(&vlmn, conc, pgs) )
+	switch ( cie_remap_finish(&vlmn, conc, pis) )
 	  {
 	  case 3:
 		if_debug3('c', "[c]=RGB [%g %g %g]\n",
 			  frac2float(conc[0]), frac2float(conc[1]),
   			  frac2float(conc[2]));
-		gx_remap_concrete_rgb(conc[0], conc[1], conc[2], pdc, pgs);
+		gx_remap_concrete_rgb(conc[0], conc[1], conc[2], pdc, pis,
+				      dev, select);
 		return 0;
 	  case 4:
 		if_debug4('c', "[c]=CMYK [%g %g %g %g]\n",
 			  frac2float(conc[0]), frac2float(conc[1]),
   			  frac2float(conc[2]), frac2float(conc[3]));
 		gx_remap_concrete_cmyk(conc[0], conc[1], conc[2], conc[3],
-				       pdc, pgs);
+				       pdc, pis, dev, select);
 		return 0;
 	  }
 	/* Can't happen. */
@@ -862,7 +850,7 @@ gx_remap_CIEABC(const gs_client_color *pc, const gs_color_space *pcs,
 }
 private int
 gx_concretize_CIEABC(const gs_client_color *pc, const gs_color_space *pcs,
-  frac *pconc, const gs_state *pgs)
+  frac *pconc, const gs_imager_state *pis)
 {	const gs_cie_abc *pcie = pcs->params.abc;
 	cie_cached_vector3 vec3;
 
@@ -878,7 +866,7 @@ gx_concretize_CIEABC(const gs_client_color *pc, const gs_color_space *pcs,
 	  cie_lookup_map3(&vabc /*&vlmn*/, &pcie->caches.DecodeABC[0],
 			  "Decode/MatrixABC");
 #undef vabc
-	cie_remap_finish(&vlmn, pconc, pgs);
+	cie_remap_finish(&vlmn, pconc, pis);
 #undef vlmn
 	return 0;
 }
@@ -886,7 +874,7 @@ gx_concretize_CIEABC(const gs_client_color *pc, const gs_color_space *pcs,
 /* Render a CIEBasedA color. */
 private int
 gx_concretize_CIEA(const gs_client_color *pc, const gs_color_space *pcs,
-  frac *pconc, const gs_state *pgs)
+  frac *pconc, const gs_imager_state *pis)
 {	const gs_cie_a *pcie = pcs->params.a;
 	cie_cached_value a = float2cie_cached(pc->paint.values[0]);
 	cie_cached_vector3 vlmn;
@@ -895,16 +883,16 @@ gx_concretize_CIEA(const gs_client_color *pc, const gs_color_space *pcs,
 
 	/* Apply DecodeA and MatrixA. */
 	vlmn = lookup_value(a, &pcie->caches.DecodeA);
-	return cie_remap_finish(&vlmn, pconc, pgs);
+	return cie_remap_finish(&vlmn, pconc, pis);
 }
 
 /* Common rendering code. */
 /* Return 3 if RGB, 4 if CMYK. */
 private int near
 cie_remap_finish(const cie_cached_vector3 *plmn, frac *pconc,
-  const gs_state *pgs)
-{	const gs_cie_render *pcie = pgs->cie_render;
-	const gx_cie_joint_caches *pjc = pgs->cie_joint_caches;
+  const gs_imager_state *pis)
+{	const gs_cie_render *pcie = pis->cie_render;
+	const gx_cie_joint_caches *pjc = pis->cie_joint_caches;
 	const gs_const_string *table;
 	cie_cached_vector3 vec3;
 	int tabc[3];		/* indices for final EncodeABC lookup */
