@@ -1,4 +1,4 @@
-/* Copyright (C) 1992, 1993 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1992, 1993, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -37,8 +37,6 @@ const gs_color_space *gs_currentcolorspace(P1(const gs_state *));
 int	gs_setcolorspace(P2(gs_state *, gs_color_space *));
 const gs_client_color *gs_currentcolor(P1(const gs_state *));
 int	gs_setcolor(P2(gs_state *, const gs_client_color *));
-bool	gs_currentoverprint(P1(const gs_state *));
-void	gs_setoverprint(P2(gs_state *, bool));
 
 /* CIE-specific routines */
 #ifndef gs_cie_render_DEFINED
@@ -50,7 +48,7 @@ int	gs_setcolorrendering(P2(gs_state *, gs_cie_render *));
 
 /* Pattern template */
 typedef struct gs_client_pattern_s {
-	gs_uid uid;		/* XUID or nothing */
+	gs_uid uid;
 	int PaintType;
 	int TilingType;
 	gs_rect BBox;
