@@ -1,4 +1,4 @@
-/* Copyright (C) 1994 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1994, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -24,6 +24,9 @@
 /* gserror.h */
 /* Error return macros */
 
+#ifndef gserror_INCLUDED
+#  define gserror_INCLUDED
+
 #ifdef DEBUG
 int gs_log_error(P3(int, const char _ds *, int));
 #else
@@ -31,3 +34,7 @@ int gs_log_error(P3(int, const char _ds *, int));
 #endif
 #define gs_note_error(err) gs_log_error(err, __FILE__, __LINE__)
 #define return_error(err) return gs_note_error(err)
+#define return_if_error(expr)\
+  do { int code_ = (expr); if ( code_ < 0 ) return code; } while ( 0 )
+
+#endif					/* gserror_INCLUDED */
