@@ -1,19 +1,24 @@
 /* Copyright (C) 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of Aladdin Ghostscript.
+  This file is part of GNU Ghostscript.
   
-  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
-  or distributor accepts any responsibility for the consequences of using it,
-  or for whether it serves any particular purpose or works at all, unless he
-  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
-  License (the "License") for full details.
+  GNU Ghostscript is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
+  anyone for the consequences of using it or for whether it serves any
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
-  Every copy of Aladdin Ghostscript must include a copy of the License,
-  normally in a plain ASCII text file named PUBLIC.  The License grants you
-  the right to copy, modify and redistribute Aladdin Ghostscript, but only
-  under certain conditions described in the License.  Among other things, the
-  License requires that the copyright notice and this notice be preserved on
-  all copies.
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gstype2.c */
@@ -79,12 +84,23 @@ type2_sbw(gs_type1_state *pcis, cs_ptr csp, cs_ptr cstack, ip_state *ipsp,
 	/* Save the interpreter state. */
 	pcis->os_count = csp + 1 - cstack;
 	pcis->ips_count = ipsp - &pcis->ipstack[0] + 1;
+	memcpy(pcis->ostack, cstack, pcis->os_count * sizeof(cstack[0]));
 	if ( pcis->init_done < 0 )
 	  {	/* Finish init when we return. */
 		pcis->init_done = 0;
 	  }
 	return type1_result_sbw;
 }
+private int near
+type2_vstem(gs_type1_state *pcis, cs_ptr csp, cs_ptr cstack)
+{	fixed x = 0;
+	cs_ptr ap;
+
+	apply_path_hints(pcis, false);
+	for ( ap = cstack; ap + 1 <= csp; x += ap[1], ap += 2 )
+	  type1_vstem(pcis, x += ap[0], ap[1]);
+	return 0;
+}
 
 /* ------ Main interpreter ------ */
 
@@ -346,16 +362,30 @@ hstem:		check_first_operator(!((csp - cstack) & 1));
 		}
 		cnext;
 	case c2_hintmask:
+		/*
+		 * A hintmask at the beginning of the CharString is
+		 * equivalent to vstemhm + hintmask.  For simplicity, we use
+		 * this interpretation everywhere.
+		 */
+		pcis->have_hintmask = true;
+		check_first_operator(!((csp - cstack) & 1));
+		type2_vstem(pcis, csp, cstack);
+		clear;
 		/* (falls through) */
 	case c2_cntrmask:
 		{ byte mask[max_total_stem_hints / 8];
 		  int i;
 
+		  if_debug1('1', "[1]mask(%d)",
+			    pcis->vstem_hints.count + pcis->hstem_hints.count);
 		  for ( i = 0;
 			i < pcis->vstem_hints.count + pcis->hstem_hints.count;
 			++cip, i += 8
 		      )
-		    charstring_next(*cip, state, mask[i >> 3], encrypted);
+		    { charstring_next(*cip, state, mask[i >> 3], encrypted);
+		      if_debug1('1', " 0x%02x", mask[i >> 3]);
+		    }
+		  if_debug0('1', "\n");
 		  ipsp->ip = cip;
 		  ipsp->dstate = state;
 		}
@@ -364,12 +394,7 @@ hstem:		check_first_operator(!((csp - cstack) & 1));
 	case c2_vstemhm:
 		pcis->have_hintmask = true;
 vstem:		check_first_operator(!((csp - cstack) & 1));
-		apply_path_hints(pcis, false);
-		{ fixed x = 0;
-
-		  for ( ap = cstack; ap + 1 <= csp; x += ap[1], ap += 2 )
-		    type1_vstem(pcis, x += ap[0], ap[1]);
-		}
+		type2_vstem(pcis, csp, cstack);
 		cnext;
 	case c2_rcurveline:
 		for ( ap = cstack; ap + 5 <= csp; ap += 6 )
