@@ -1,4 +1,4 @@
-/* Copyright (C) 1995, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -31,6 +31,7 @@
 #include "gxdevice.h"
 #include "gxdevmem.h"			/* must precede gxcldev.h */
 #include "gxcldev.h"
+#include "gxfmap.h"
 
 /*
  * Define when, if ever, to write character bitmaps in all bands.
@@ -44,6 +45,71 @@
 
 /* ------ Writing ------ */
 
+/*
+ * Determine the (possibly unpadded) width in bytes for writing a bitmap,
+ * per the algorithm in gxcldev.h.  If compression_mask has any of the
+ * cmd_mask_compress_any bits set, we assume the bitmap will be compressed.
+ * Return the total size of the bitmap.
+ */
+uint
+clist_bitmap_bytes(uint width_bits, uint height, int compression_mask,
+  uint *width_bytes, uint *raster)
+{	uint full_raster = *raster = bitmap_raster(width_bits);
+	uint short_raster = (width_bits + 7) >> 3;
+	uint width_bytes_last;
+
+	if ( compression_mask & cmd_mask_compress_any )
+	  *width_bytes = width_bytes_last = full_raster;
+	else if ( short_raster <= cmd_max_short_width_bytes ||
+		  height <= 1 ||
+		  (compression_mask & decompress_spread) != 0
+		)
+	  *width_bytes = width_bytes_last = short_raster;
+	else
+	  *width_bytes = full_raster, width_bytes_last = short_raster;
+	return
+	  (height == 0 ? 0 : *width_bytes * (height - 1) + width_bytes_last);
+}
+
+/*
+ * Compress a bitmap, skipping extra padding bytes at the end of each row if
+ * necessary.  We require height >= 1, raster >= bitmap_raster(width_bits).
+ */
+private int
+cmd_compress_bitmap(stream_state *st, const byte *data, uint width_bits,
+  uint raster, uint height, stream_cursor_write *pw)
+{	uint width_bytes = bitmap_raster(width_bits);
+	int status = 0;
+	stream_cursor_read r;
+
+	r.ptr = data - 1;
+	if ( raster == width_bytes )
+	  { r.limit = r.ptr + raster * height;
+	    status = (*st->template->process)(st, &r, pw, true);
+	  }
+	else
+	  { /* Compress row-by-row. */
+	    uint y;
+	    for ( y = 1; (r.limit = r.ptr + width_bytes), y < height; ++y )
+	      { status = (*st->template->process)(st, &r, pw, false);
+		if ( status )
+		  break;
+		if ( r.ptr != r.limit )
+		  { /* We don't attempt to handle compressors that */
+		    /* require >1 input byte to make progress. */
+		    status = -1;
+		    break;
+		  }
+		r.ptr += raster - width_bytes;
+	      }
+	    if ( status == 0 )
+	      status = (*st->template->process)(st, &r, pw, true);
+	  }
+	if ( st->template->release )
+	  (*st->template->release)(st);
+	return status;
+}
+
 /*
  * Put a bitmap in the buffer, compressing if appropriate.
  * pcls == 0 means put the bitmap in all bands.
@@ -57,34 +123,35 @@ int
 cmd_put_bits(gx_device_clist_writer *cldev, gx_clist_state *pcls,
   const byte *data, uint width_bits, uint height, uint raster, int op_size,
   int compression_mask, byte **pdp, uint *psize)
-{	uint full_raster = bitmap_raster(width_bits);
-	uint short_raster = (width_bits + 7) >> 3;
-	uint full_size =
-	  (height <= 1 ? short_raster * height :
-	   full_raster * (height - 1) + short_raster);
-	uint uncompressed_size =
-	  (height <= 1 ? full_size : full_raster * height);
+{	uint short_raster, full_raster;
 	uint short_size =
-	  (short_raster <= cmd_max_short_width_bytes ?
-	   short_raster * height : full_size);
+	  clist_bitmap_bytes(width_bits, height,
+			     compression_mask & ~cmd_mask_compress_any,
+			     &short_raster, &full_raster);
+	uint uncompressed_raster;
+	uint uncompressed_size =
+	  clist_bitmap_bytes(width_bits, height, compression_mask,
+			     &uncompressed_raster, &full_raster);
+	uint max_size = cbuf_size - op_size;
 	byte *dp;
 	int compress = 0;
 
-	/* See if compressing the bits is possible and worthwhile. */
-	/* Currently we can't compress if the compressed data */
-	/* won't fit in the command reading buffer, or if the */
-	/* decompressed data won't fit in the buffer and */
-	/* decompress_elsewhere isn't set. */
+	/*
+	 * See if compressing the bits is possible and worthwhile.
+	 * Currently we can't compress if the compressed data won't fit in
+	 * the command reading buffer, or if the decompressed data won't fit
+	 * in the buffer and decompress_elsewhere isn't set.
+	 */
 	if ( short_size >= 50 &&
-	     (compression_mask & ~decompress_elsewhere) != 0 &&
-	     (uncompressed_size <= cbuf_size ||
+	     (compression_mask & cmd_mask_compress_any) != 0 &&
+	     (uncompressed_size <= max_size ||
 	      (compression_mask & decompress_elsewhere) != 0)
 	   )
-	  {	stream_cursor_read r;
-		byte *wbase;
-		stream_cursor_write w;
-		int status;
-		uint wcount;
+	  {	union ss_ {
+		  stream_state ss;
+		  stream_CFE_state cf;
+		  stream_RLE_state rl;
+		} sstate;
 
 		*psize = op_size + uncompressed_size;
 		if ( pcls != 0 )
@@ -94,57 +161,66 @@ cmd_put_bits(gx_device_clist_writer *cldev, gx_clist_state *pcls,
 		  { set_cmd_put_all_op(dp, cldev, 0, *psize);
 		  }
 		cmd_uncount_op(0, *psize);
-		wbase = dp + (op_size - 1);
-		r.ptr = data - 1;
-		r.limit = r.ptr + uncompressed_size;
-		w.ptr = wbase;
-		w.limit = w.ptr + uncompressed_size;
+		/*
+		 * Note that we currently keep all the padding if we are
+		 * compressing.  This is ridiculous, but it's too hard to
+		 * change right now.
+		 */
 		if ( compression_mask & (1 << cmd_compress_cfe) )
 		  { /* Try CCITTFax compression. */
-		    stream_CFE_state sstate;
-		    clist_cfe_init(&sstate, width_bits);
-		    status =
-		      (*s_CFE_template.process)
-			((stream_state *)&sstate, &r, &w, true);
-		    (*s_CFE_template.release)((stream_state *)&sstate);
+		    clist_cfe_init(&sstate.cf,
+				   uncompressed_raster << 3 /*width_bits*/);
+		    sstate.ss.template = &s_CFE_template;
 		    compress = cmd_compress_cfe;
 		  }
 		else if ( compression_mask & (1 << cmd_compress_rle) )
 		  { /* Try RLE compression. */
-		    stream_RLE_state sstate;
-		    clist_rle_init(&sstate);
-		    status =
-		      (*s_RLE_template.process)
-			((stream_state *)&sstate, &r, &w, true);
+		    clist_rle_init(&sstate.rl);
+		    sstate.ss.template = &s_RLE_template;
 		    compress = cmd_compress_rle;
 		  }
-		if ( compress != 0 && status == 0 &&
-		     (wcount = w.ptr - wbase) <= short_size >> 1 &&
-		     wcount <= cbuf_size
-		   )
-		  {	/* Use compressed representation. */
+		if ( compress )
+		  { byte *wbase = dp + (op_size - 1);
+		    stream_cursor_write w;
+		    /*
+		     * We can give up on compressing if we generate too much
+		     * output to fit in the command reading buffer, or too
+		     * much to make compression worthwhile.
+		     */
+		    uint wmax = min(uncompressed_size, max_size);
+		    int status;
+
+		    w.ptr = wbase;
+		    w.limit = w.ptr + min(wmax, short_size >> 1);
+		    status = cmd_compress_bitmap((stream_state *)&sstate, data,
+						 uncompressed_raster << 3 /*width_bits*/,
+						 raster, height, &w);
+		    if ( status == 0 )
+		      { /* Use compressed representation. */
+			uint wcount = w.ptr - wbase;
 			cmd_shorten_list_op(cldev,
-				(pcls ? &pcls->list : &cldev->all_band_list),
+				(pcls ? &pcls->list : &cldev->band_range_list),
 				uncompressed_size - wcount);
 			*psize = op_size + wcount;
 			goto out;
+		      }
 		  }
-		if ( uncompressed_size > cbuf_size )
+		if ( uncompressed_size > max_size )
 		  { cmd_shorten_list_op(cldev,
-				(pcls ? &pcls->list : &cldev->all_band_list),
+				(pcls ? &pcls->list : &cldev->band_range_list),
 				*psize);
-		    return gs_error_limitcheck;
+		    return_error(gs_error_limitcheck);
 		  }
 		if ( uncompressed_size != short_size )
 		  {	cmd_shorten_list_op(cldev,
-				(pcls ? &pcls->list : &cldev->all_band_list),
+				(pcls ? &pcls->list : &cldev->band_range_list),
 				uncompressed_size - short_size);
 			*psize = op_size + short_size;
 		  }
 		compress = 0;
 	  }
-	else if ( full_size > cbuf_size )
-	  return gs_error_limitcheck;
+	else if ( uncompressed_size > max_size )
+	  return_error(gs_error_limitcheck);
 	else
 	  {	*psize = op_size + short_size;
 		if ( pcls != 0 )
@@ -155,10 +231,6 @@ cmd_put_bits(gx_device_clist_writer *cldev, gx_clist_state *pcls,
 		  }
 		cmd_uncount_op(0, *psize);
 	  }
-	if ( short_raster > cmd_max_short_width_bytes &&
-	     !(compression_mask & decompress_spread)
-	   )
-	  short_raster = full_raster;		/* = short_size / height */
 	bytes_copy_rectangle(dp + op_size, short_raster, data, raster,
 			     short_raster, height);
 out:	*pdp = dp;
@@ -166,91 +238,81 @@ out:	*pdp = dp;
 }
 
 /* Add a command to set the tile size and depth. */
-private int
-cmd_put_tile_params(gx_device_clist_writer *cldev, const gx_strip_bitmap *tile,
-  int depth)
-{	int tcsize = 2 + cmd_size_w(tile->size.x) + cmd_size_w(tile->size.y) +
-	  cmd_size_w(tile->rep_width) + cmd_size_w(tile->rep_height) +
-	    cmd_size_w(tile->rep_shift);
-	byte *dp;
+private uint
+cmd_size_tile_params(const gx_strip_bitmap *tile)
+{	return 2 + cmd_size_w(tile->rep_width) + cmd_size_w(tile->rep_height) +
+	  (tile->rep_width == tile->size.x ? 0 :
+	   cmd_size_w(tile->size.x / tile->rep_width)) +
+	  (tile->rep_height == tile->size.y ? 0 :
+	   cmd_size_w(tile->size.y / tile->rep_height)) +
+	  (tile->rep_shift == 0 ? 0 : cmd_size_w(tile->rep_shift));
+}
+private void
+cmd_store_tile_params(byte *dp, const gx_strip_bitmap *tile, int depth,
+  uint csize)
+{	byte *p = dp + 2;
+	byte bd = depth - 1;
 
-	set_cmd_put_all_op(dp, cldev, cmd_opv_set_tile_size, tcsize);
-	dp[1] = depth;
-	dp += 2;
-	dp = cmd_put_w(tile->size.x, dp);
-	dp = cmd_put_w(tile->size.y, dp);
-	dp = cmd_put_w(tile->rep_width, dp);
-	dp = cmd_put_w(tile->rep_height, dp);
-	cmd_put_w(tile->rep_shift, dp);
-	return 0;
+	*dp = cmd_count_op(cmd_opv_set_tile_size, csize);
+	p = cmd_put_w(tile->rep_width, p);
+	p = cmd_put_w(tile->rep_height, p);
+	if ( tile->rep_width != tile->size.x ) {
+	  p = cmd_put_w(tile->size.x / tile->rep_width, p);
+	  bd |= 0x20;
+	}
+	if ( tile->rep_height != tile->size.y ) {
+	  p = cmd_put_w(tile->size.y / tile->rep_height, p);
+	  bd |= 0x40;
+	}
+	if ( tile->rep_shift != 0 ) {
+	  cmd_put_w(tile->rep_shift, p);
+	  bd |= 0x80;
+	}
+	dp[1] = bd;
 }
 
 /* Add a command to set the tile index. */
 /* This is a relatively high-frequency operation, so we make it a macro. */
 /* Note that it may do a 'return' with an error code. */
-#define cmd_put_tile_index_inline(cldev, pcls, index)\
- { int idelta = (index) - (pcls)->tile_index + 8; byte *dp;\
+/* Some compilers try to substitute macro args in string literals! */
+#define cmd_put_tile_index_inline(cldev, pcls, indx)\
+ { int idelta = (indx) - (pcls)->tile_index + 8; byte *dp;\
    if ( !(idelta & ~15) )\
      { set_cmd_put_op(dp, cldev, pcls, cmd_op_delta_tile_index + idelta, 1); }\
    else\
-     { set_cmd_put_op(dp, cldev, pcls, cmd_op_set_tile_index + ((index) >> 8), 2);\
-       dp[1] = (index) & 0xff;\
+     { set_cmd_put_op(dp, cldev, pcls, cmd_op_set_tile_index + ((indx) >> 8), 2);\
+       dp[1] = (indx) & 0xff;\
        if_debug2('L', "[L]writing index=%u, offset=%lu\n",\
-		 index, cldev->tile_table[index].offset);\
+		 indx, cldev->tile_table[indx].offset);\
      }\
  }
 
-/* Add commands to represent a halftone order. */
-/****** NOT USED YET ******/
+/* If necessary, write out data for a single color map. */
 int
-cmd_put_ht_order(gx_device_clist_writer *cldev, const gx_ht_order *porder)
-{	long offset;
-	byte command[max(cmd_largest_size - 1, 100)];
-	byte *cp;
-	uint len;
-	byte *dp;
-	uint i, n;
+cmd_put_color_map(gx_device_clist_writer *cldev, cmd_map_index map_index,
+  const gx_transfer_map *map, gs_id *pid)
+{	byte *dp;
 
-	/* Put out the order parameters. */
-	/****** SET offset ******/
-	offset = 0; /****** bogus, to pacify compilers ******/
-	cp = cmd_put_w(offset, command);	/****** WRONG ******/
-	cp = cmd_put_w(porder->width, cp);
-	cp = cmd_put_w(porder->height, cp);
-	cp = cmd_put_w(porder->raster, cp);
-	cp = cmd_put_w(porder->shift, cp);
-	cp = cmd_put_w(porder->num_levels, cp);
-	cp = cmd_put_w(porder->num_bits, cp);
-	len = cp - command;
-	set_cmd_put_all_op(dp, cldev, cmd_opv_set_ht_size, len + 1);
-	memcpy(dp + 1, command, len);
-
-	/* Put out the levels array. */
-#define nlevels ((sizeof(command) - 2) / sizeof(*porder->levels))
-	for ( i = 0; i < porder->num_levels; i += n )
-	  { n = porder->num_levels - i;
-	    if ( n > nlevels )
-	      n = nlevels;
-	    set_cmd_put_all_op(dp, cldev, cmd_opv_set_ht_data,
-			       2 + n * sizeof(*porder->levels));
-	    dp[1] = n;
-	    memcpy(dp + 2, porder->levels + i, n * sizeof(*porder->levels));
+	if ( map == 0 )
+	  { if ( pid )
+	      { if ( *pid == gs_no_id )
+		  return 0;		/* no need to write */
+	        *pid = gs_no_id;
+	      }
+	    set_cmd_put_all_op(dp, cldev, cmd_opv_set_misc, 2);
+	    dp[1] = cmd_set_misc_map + map_index;
 	  }
-#undef nlevels
-
-	/* Put out the bits array. */
-#define nbits ((sizeof(command) - 2) / sizeof(*porder->bits))
-	for ( i = 0; i < porder->num_bits; i += n )
-	  { n = porder->num_bits - i;
-	    if ( n > nbits )
-	      n = nbits;
-	    set_cmd_put_all_op(dp, cldev, cmd_opv_set_ht_data,
-			       2 + n * sizeof(*porder->bits));
-	    dp[1] = n;
-	    memcpy(dp + 2, porder->bits + i, n * sizeof(*porder->bits));
+	else
+	  { if ( pid )
+	      { if ( map->id == *pid )
+		  return 0;		/* no need to write */
+	        *pid = map->id;
+	      }
+	    set_cmd_put_all_op(dp, cldev, cmd_opv_set_misc,
+			       2 + sizeof(map->values));
+	    dp[1] = cmd_set_misc_map + 0x20 + map_index;
+	    memcpy(dp + 2, map->values, sizeof(map->values));
 	  }
-#undef nbits
-
 	return 0;
 }
 
@@ -375,7 +437,8 @@ clist_add_tile(gx_device_clist_writer *cldev, const gx_strip_bitmap *tiles,
 	slot->cb_raster = raster;
 	slot->width = tiles->rep_width;
 	slot->height = tiles->rep_height;
-	slot->shift = tiles->rep_shift;
+	slot->shift = slot->rep_shift = tiles->rep_shift;
+	slot->x_reps = slot->y_reps = 1;
 	slot->id = tiles->id;
 	memset(ts_mask(slot), 0, cldev->tile_band_mask_size);
 	bytes_copy_rectangle(ts_bits(cldev, slot), raster,
@@ -403,6 +466,60 @@ clist_add_tile(gx_device_clist_writer *cldev, const gx_strip_bitmap *tiles,
 
 /* ------ Driver procedure support ------ */
 
+/* Change the tile parameters (size and depth). */
+/* Currently we do this for all bands at once. */
+private void
+clist_new_tile_params(gx_strip_bitmap *new_tile, const gx_strip_bitmap *tiles,
+  int depth, const gx_device_clist_writer *cldev)
+{	/*
+	 * Adjust the replication factors.  If we can, we replicate
+	 * the tile in X up to 32 bytes, and then in Y up to 4 copies,
+	 * as long as we don't exceed a total tile size of 256 bytes,
+	 * or more than 255 repetitions in X or Y, or make the tile so
+	 * large that not all possible tiles will fit in the cache.
+	 * Also, don't attempt Y replication if shifting is required. 
+	 */
+#define max_tile_reps_x 255
+#define max_tile_bytes_x 32
+#define max_tile_reps_y 4
+#define max_tile_bytes 256
+	uint rep_width = tiles->rep_width;
+	uint rep_height = tiles->rep_height;
+	uint rep_width_bits = rep_width * depth;
+	uint tile_overhead =
+	  sizeof(tile_slot) + cldev->tile_band_mask_size;
+	uint max_bytes = cldev->chunk.size / (rep_width_bits * rep_height);
+
+	max_bytes -= min(max_bytes, tile_overhead);
+	if ( max_bytes > max_tile_bytes )
+	  max_bytes = max_tile_bytes;
+	*new_tile = *tiles;
+	{ uint max_bits_x = max_bytes * 8 / rep_height;
+	  uint reps_x =
+	    min(max_bits_x, max_tile_bytes_x * 8) / rep_width_bits;
+	  uint reps_y;
+
+	  while ( reps_x > max_tile_reps_x )
+	    reps_x >>= 1;
+	  new_tile->size.x = max(reps_x, 1) * rep_width;
+	  new_tile->raster = bitmap_raster(new_tile->size.x * depth);
+	  if ( tiles->shift != 0 )
+	    reps_y = 1;
+	  else
+	    { reps_y = max_bytes / (new_tile->raster * rep_height);
+	      if ( reps_y > max_tile_reps_y )
+		reps_y = max_tile_reps_y;
+	      else if ( reps_y < 1 )
+		reps_y = 1;
+	    }
+	  new_tile->size.y = reps_y * rep_height;
+	}
+#undef max_tile_reps_x
+#undef max_tile_bytes_x
+#undef max_tile_reps_y
+#undef max_tile_bytes
+}
+
 /* Change tile for clist_tile_rectangle. */
 int
 clist_change_tile(gx_device_clist_writer *cldev, gx_clist_state *pcls,
@@ -410,10 +527,16 @@ clist_change_tile(gx_device_clist_writer *cldev, gx_clist_state *pcls,
 {	tile_loc loc;
 	int code;
 
+#define tile_params_differ(cldev, tiles, depth)\
+  ((tiles)->rep_width != (cldev)->tile_params.rep_width ||\
+   (tiles)->rep_height != (cldev)->tile_params.rep_height ||\
+   (tiles)->rep_shift != (cldev)->tile_params.rep_shift ||\
+   (depth) != (cldev)->tile_depth)
+
 top:	if ( clist_find_bits(cldev, tiles->id, &loc) )
 	  { /* The bitmap is in the cache.  Check whether this band */
 	    /* knows about it. */
-	    uint band_index = pcls - cldev->states;
+	    int band_index = pcls - cldev->states;
 	    byte *bptr = ts_mask(loc.tile) + (band_index >> 3);
 	    byte bmask = 1 << (band_index & 7);
 
@@ -424,100 +547,106 @@ top:	if ( clist_find_bits(cldev, tiles->id, &loc) )
 		cmd_put_tile_index_inline(cldev, pcls, loc.index);
 	      }
 	    else
-	      { /*
-		 * Not known yet.  Output the bits.  Note that the offset we
-		 * write is the one used by the reading phase, not the
-		 * writing phase.  Note also that the size of the cached and
-		 * written tile may differ from that of the client's tile.
-		 * Finally, note that this tile's size parameters are
-		 * guaranteed to be compatible with those stored in the
-		 * device (cldev->tile_params).
-		 */
-		ulong offset = (byte *)loc.tile - cldev->chunk.data;
-		uint rsize = 1 + cmd_size_w(loc.index) + cmd_size_w(offset);
-		byte *dp;
-		uint csize;
-		int code =
-		  cmd_put_bits(cldev, pcls, ts_bits(cldev, loc.tile),
-			       tiles->rep_width * depth, tiles->rep_height,
-			       loc.tile->cb_raster, rsize,
-			       (cldev->tile_params.size.x > tiles->rep_width ?
-				decompress_elsewhere | decompress_spread :
-				decompress_elsewhere),
-			       &dp, &csize);
+	      { uint extra = 0;
+	        if tile_params_differ(cldev, tiles, depth)
+		  { /*
+		     * We have a cached tile whose parameters differ from
+		     * the current ones.  Because of the way tile IDs are
+		     * managed, this is currently only possible when mixing
+		     * Patterns and halftones, but if we didn't generate new
+		     * IDs each time the main halftone cache needed to be
+		     * refreshed, this could also happen simply from
+		     * switching screens.
+		     */
+		    int band;
 
-		if ( code < 0 )
-		  return code;
-		*dp = cmd_count_op(cmd_opv_set_tile_bits, csize);
-		dp++;
-		dp = cmd_put_w(loc.index, dp);
-		cmd_put_w(offset, dp);
-		*bptr |= bmask;
-		loc.tile->num_bands++;
+		    clist_new_tile_params(&cldev->tile_params, tiles, depth,
+					  cldev);
+		    cldev->tile_depth = depth;
+		    /* No band knows about the new parameters. */
+		    for ( band = cldev->tile_known_min;
+			  band <= cldev->tile_known_max;
+			  ++band
+			)
+		      cldev->states[band].known &= ~tile_params_known;
+		    cldev->tile_known_min = cldev->nbands;
+		    cldev->tile_known_max = -1;
+		  }
+		if ( !(pcls->known & tile_params_known) )
+		  { /* We're going to have to write the tile parameters. */
+		    extra = cmd_size_tile_params(&cldev->tile_params);
+		  }
+	        { /*
+		   * This band doesn't know this tile yet, so output the
+		   * bits.  Note that the offset we write is the one used by
+		   * the reading phase, not the writing phase.  Note also
+		   * that the size of the cached and written tile may differ
+		   * from that of the client's tile.  Finally, note that
+		   * this tile's size parameters are guaranteed to be
+		   * compatible with those stored in the device
+		   * (cldev->tile_params).
+		   */
+		  ulong offset = (byte *)loc.tile - cldev->chunk.data;
+		  uint rsize =
+		    extra + 1 + cmd_size_w(loc.index) + cmd_size_w(offset);
+		  byte *dp;
+		  uint csize;
+		  int code =
+		    cmd_put_bits(cldev, pcls, ts_bits(cldev, loc.tile),
+				 tiles->rep_width * depth, tiles->rep_height,
+				 loc.tile->cb_raster, rsize,
+				 (cldev->tile_params.size.x > tiles->rep_width ?
+				  decompress_elsewhere | decompress_spread :
+				  decompress_elsewhere),
+				 &dp, &csize);
+
+		  if ( code < 0 )
+		    return code;
+		  if ( extra )
+		    { /* Write the tile parameters before writing the bits. */
+		      cmd_store_tile_params(dp, &cldev->tile_params, depth,
+					    extra);
+		      dp += extra;
+		      /* This band now knows the parameters. */
+		      pcls->known |= tile_params_known;
+		      if ( band_index < cldev->tile_known_min )
+			cldev->tile_known_min = band_index;
+		      if ( band_index > cldev->tile_known_max )
+			cldev->tile_known_max = band_index;
+		    }
+		  *dp = cmd_count_op(cmd_opv_set_tile_bits, csize - extra);
+		  dp++;
+		  dp = cmd_put_w(loc.index, dp);
+		  cmd_put_w(offset, dp);
+		  *bptr |= bmask;
+		  loc.tile->num_bands++;
+		}
 	      }
 	    pcls->tile_index = loc.index;
 	    pcls->tile_id = loc.tile->id;
 	    return 0;
 	  }
-	/* The tile is not in the cache. */
-	/* Ensure that the tile size is compatible. */
-	if ( tiles->rep_width != cldev->tile_params.rep_width ||
-	     tiles->rep_height != cldev->tile_params.rep_height ||
-	     tiles->rep_shift != cldev->tile_params.rep_shift ||
-	     depth != cldev->tile_depth
-	   )
-	  { /* Change the tile dimensions. We do this for all bands at once. */
-	    gx_strip_bitmap new_tile;
-	    uint rep_width_bits = tiles->rep_width * depth;
+	/* The tile is not in the cache, add it. */
+	{ gx_strip_bitmap new_tile;
+	  gx_strip_bitmap *ptile;
 
-	    /*
-	     * Adjust the replication factors.  If we can, we replicate
-	     * the tile in X up to 32 bytes, and then in Y up to 4 copies,
-	     * as long as we don't exceed a total tile size of 256 bytes.
-	     * However, don't attempt Y replication if shifting is required. 
-	     */
-#define max_tile_bytes_x 32
-#define max_tile_reps_y 4
-#define max_tile_bytes 256
-	    new_tile = *tiles;
-	    { uint max_bits_x = max_tile_bytes * 8 / new_tile.rep_height;
-	      uint reps_x =
-		min(max_bits_x, max_tile_bytes_x * 8) / rep_width_bits;
-	      uint reps_y;
-
-	      new_tile.size.x = max(reps_x, 1) * new_tile.rep_width;
-	      new_tile.raster = bitmap_raster(new_tile.size.x * depth);
-	      if ( tiles->shift != 0 )
-		reps_y = 1;
-	      else
-		{ reps_y =
-		    max_tile_bytes / (new_tile.raster * new_tile.rep_height);
-		  if ( reps_y > max_tile_reps_y )
-		    reps_y = max_tile_reps_y;
-		  else if ( reps_y < 1 )
-		    reps_y = 1;
-		}
-	      new_tile.size.y = reps_y * new_tile.rep_height;
+	  /* Ensure that the tile size is compatible. */
+	  if ( tile_params_differ(cldev, tiles, depth) )
+	    { /* We'll reset cldev->tile_params when we write the bits. */
+	      clist_new_tile_params(&new_tile, tiles, depth, cldev);
+	      ptile = &new_tile;
 	    }
-#undef max_tile_bytes_x
-#undef max_tile_reps_y
-#undef max_tile_bytes
-	    code = cmd_put_tile_params(cldev, &new_tile, depth);
-	    if ( code < 0 )
-	      return code;
-	    cldev->tile_params = new_tile;
-	    cldev->tile_depth = depth;
-	  }
-	else
-	  { cldev->tile_params.id = tiles->id;
-	    cldev->tile_params.data = tiles->data;
-	  }
-
-	code = clist_add_tile(cldev, &cldev->tile_params,
-			      tiles->raster, depth);
-	if ( code < 0 )
-	  return code;
+	  else
+	    { cldev->tile_params.id = tiles->id;
+	      cldev->tile_params.data = tiles->data;
+	      ptile = &cldev->tile_params;
+	    }
+	  code = clist_add_tile(cldev, ptile, tiles->raster, depth);
+	  if ( code < 0 )
+	    return code;
+	}
 	goto top;
+#undef tile_params_differ
 }
 
 /* Change "tile" for clist_copy_*.  tiles->[rep_]shift must be zero. */
