@@ -1,4 +1,4 @@
-/* Copyright (C) 1995, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -89,23 +89,29 @@
 typedef enum {
 	cmd_op_misc = 0x00,		/* (see below) */
 	  cmd_opv_end_run = 0x00,	/* (nothing) */
-	  cmd_opv_set_tile_size = 0x01,	/* depth, width#, height#, */
+	  cmd_opv_set_tile_size = 0x01,	/* rs?(1)nry?(1)nrx?(1)depth-1(5), */
 					/* rep_width#, rep_height#, */
-					/* rep_shift# */
+					/* [, nreps_x#][, nreps_y #] */
+					/* [, rep_shift#] */
 	  cmd_opv_set_tile_phase = 0x02, /* x#, y# */
 	  cmd_opv_set_tile_bits = 0x03, /* index#, offset#, <bits> */
 	  cmd_opv_set_bits = 0x04,	/* depth*4+compress, width#, height#, */
 					/* index#, offset#, <bits> */
 	  cmd_opv_set_tile_color = 0x05,  /* (nothing; next set/delta_color */
 					/* refers to tile) */
-	  cmd_opv_set_lop = 0x06,	/* lop# */
+	  cmd_opv_set_misc = 0x06,
+#define cmd_set_misc_lop (0 << 6)	/* 00: lop_lsb(6), lop_msb# */
+#define cmd_set_misc_data_x (1 << 6)	/* 01: more(1)dx_lsb(5)[, dx_msb#] */
+#define cmd_set_misc_map (2 << 6)	/* 10: non-0(1)map_index(5) */
+					/*   [, n x frac] */
+#define cmd_set_misc_halftone (3 << 6)	/* 11: type(6), num_comp# */
 	  cmd_opv_enable_lop = 0x07,	/* (nothing) */
 	  cmd_opv_disable_lop = 0x08,	/* (nothing) */
-	  cmd_opv_set_ht_size = 0x09,	/* offset#, */
+	  cmd_opv_set_ht_order = 0x09,	/* component+1#[, cname#], */
 					/* width#, height#, raster#, */
 					/* shift#, num_levels#, num_bits# */
 	  cmd_opv_set_ht_data = 0x0a,	/* n, n x (uint|gx_ht_bit) */
-	  cmd_opv_next_data_x = 0x0b,	/* data_x */
+	  cmd_opv_end_page = 0x0b,	/* (nothing) */
 	  cmd_opv_delta2_color0 = 0x0c,	/* dr5dg6db5 or dc4dm4dy4dk4 */
 #define cmd_delta2_24_bias 0x00102010
 #define cmd_delta2_24_mask 0x001f3f1f
@@ -129,11 +135,11 @@ typedef enum {
 	cmd_op_tile_rect = 0x60,	/* +dy2dh2, x#, w# | +0, rect# */
 	cmd_op_tile_rect_short = 0x70,	/* +dh, dx, dw | +0, rect_short */
 	cmd_op_tile_rect_tiny = 0x80,	/* +dw+0, rect_tiny | +dw+8 */
-	cmd_op_copy_mono = 0x90,	/* +compress, x#, y#, w#, h#, */
-					/*   <bits> | */
+	cmd_op_copy_mono = 0x90,	/* +compress, x#, y#, (w+data_x)#, */
+					/* h#, <bits> | */
 #define cmd_copy_ht_color 4
-					/* +4+compress, x#, y#, w#, h#, */
-					/*   <bits> | */
+					/* +4+compress, x#, y#, (w+data_x)#, */
+					/* h#, <bits> | */
 #define cmd_copy_use_tile 8
 					/* +8 (use tile), x#, y# | */
 					/* +12 (use tile), x#, y# */
@@ -153,8 +159,8 @@ typedef enum {
 
 #define cmd_misc_op_name_strings\
   "end_run", "set_tile_size", "set_tile_phase", "set_tile_bits",\
-  "set_bits", "set_tile_color", "set_lop", "enable_lop",\
-  "disable_lop", "set_ht_size", "set_ht_data", "next_data_x",\
+  "set_bits", "set_tile_color", "set_misc", "enable_lop",\
+  "disable_lop", "set_ht_order", "set_ht_data", "end_page",\
   "delta2_color0", "delta2_color1", "set_copy_color", "set_copy_alpha",
 
 #ifdef DEBUG
@@ -197,29 +203,36 @@ typedef struct {
 
 /*
  * When we write bitmaps, we remove raster padding selectively:
+ *	- If the bitmap is compressed, we don't remove any padding;
  *	- If the width is <= 6 bytes, we remove all the padding;
  *	- If the bitmap is only 1 scan line high, we remove the padding;
- *	- If the bitmap is uncompressed, we remove the padding from the
- *	last scan line;
- *	- Otherwise, we don't remove any padding.
- * Currently, this information is embodied in the code that writes bitmaps
- * (cmd_put_bits) and in the code that reads bitmaps (set_[tile_]bits and
- * copy_xxx opcodes).
- *
+ *	- If the bitmap is going to be replicated horizontally (see the
+ *	definition of decompress_spread below), we remove the padding;
+ *	- Otherwise, we remove the padding only from the last scan line.
+ */
+#define cmd_max_short_width_bytes 6
+#define cmd_max_short_width_bits (cmd_max_short_width_bytes * 8)
+/*
+ * Determine the (possibly unpadded) width in bytes for writing a bitmap,
+ * per the algorithm just outlined.  If compression_mask has any of the
+ * cmd_mask_compress_any bits set, we assume the bitmap will be compressed.
+ * Return the total size of the bitmap.
+ */
+uint clist_bitmap_bytes(P5(uint width_bits, uint height,
+			   int compression_mask,
+			   uint *width_bytes, uint *raster));
+/*
  * For halftone cells, we always write an unreplicated bitmap, but we
  * reserve cache space for the reading pass based on the replicated size.
  * See the clist_change_tile procedure for the algorithm that chooses the
  * replication factors.
  */
-#define cmd_max_short_width_bytes 6
-#define cmd_max_short_width_bits (cmd_max_short_width_bytes * 8)
 
 /* ---------------- Block file entries ---------------- */
 
 typedef struct cmd_block_s {
-	int band;
-#define cmd_band_all (-1)		/* do in all bands */
-#define cmd_band_end (-2)		/* end of band file */
+	int band_min, band_max;
+#define cmd_band_end (-1)		/* end of band file */
 	long pos;			/* starting position in cfile */
 } cmd_block;
 
@@ -248,20 +261,14 @@ struct gx_clist_state_s {
 	ushort color_is_alpha;		/* (Boolean) for copy_color_alpha */
 	ushort known;			/* flags for whether this band */
 					/* knows various misc. parameters */
-/****** THESE SHOULD BE MOVED TO gxclpath.h ******/
-#define flatness_known		(1<<0)
-#define fill_adjust_known	(1<<1)
-#define ctm_known		(1<<2)
-#define line_width_known	(1<<3)
-#define miter_limit_known	(1<<4)
-#define misc_known		(1<<5)
-#define dash_known		(1<<6)
-#define clip_path_known		(1<<7)
-#define stroke_all_known	((1<<8)-1)
-#define color_space_known	(1<<8)
-#define all_known		((1<<9)-1)
+	/* We assign 'known' flags here from the high end; */
+	/* gxclpath.h assigns them from the low end. */
+#define tile_params_known (1<<15)
+#define initial_known 0x7fff		/* exclude tile params */
 		/* Following are only used when writing */
 	cmd_list list;			/* list of commands for band */
+		/* Following is set when writing, read when reading */
+	ulong cost;			/* cost of rendering the band */
 };
 
 /* The initial values for a band state */
@@ -270,8 +277,8 @@ struct gx_clist_state_s {
 	 { gx_no_color_index, gx_no_color_index },\
 	0, gx_no_bitmap_id,\
 	 { 0, 0 }, { gx_no_color_index, gx_no_color_index },\
-	 { 0, 0, 0, 0 }, lop_default, 0, 0, 0, all_known,\
-	 { 0, 0 }
+	 { 0, 0, 0, 0 }, lop_default, 0, 0, 0, initial_known,\
+	 { 0, 0 }, 0
 
 /* Define the size of the command buffer used for reading. */
 /* This is needed to split up operations with a large amount of data, */
@@ -313,15 +320,20 @@ byte *cmd_put_op(P3(gx_device_clist_writer *cldev, gx_clist_state *pcls, uint si
     *dp = cmd_count_op(op, csize);\
   } while ( 0 )
 
-/* Add a command for all bands. */
-byte *cmd_put_all_op(P2(gx_device_clist_writer *cldev, uint size));
-/* Call cmd_put_all_op and return properly if an error occurs. */
-#define set_cmd_put_all_op(dp, cldev, op, csize)\
+/* Add a command for all bands or a range of bands. */
+byte *cmd_put_range_op(P4(gx_device_clist_writer *cldev, int band_min,
+			  int band_max, uint size));
+#define cmd_put_all_op(cldev, size)\
+  cmd_put_range_op(cldev, 0, (cldev)->nbands - 1, size)
+/* Call cmd_put_all/range_op and return properly if an error occurs. */
+#define set_cmd_put_range_op(dp, cldev, op, bmin, bmax, csize)\
   do {\
-    if ( (dp = cmd_put_all_op(cldev, csize)) == 0 )\
+    if ( (dp = cmd_put_range_op(cldev, bmin, bmax, csize)) == 0 )\
       return (cldev)->error_code;\
     *dp = cmd_count_op(op, csize);\
   } while ( 0 )
+#define set_cmd_put_all_op(dp, cldev, op, csize)\
+  set_cmd_put_range_op(dp, cldev, op, 0, (cldev)->nbands - 1, csize)
 
 /* Shorten the last allocated command. */
 /* Note that this does not adjust the statistics. */
@@ -330,8 +342,8 @@ byte *cmd_put_all_op(P2(gx_device_clist_writer *cldev, uint size));
 #define cmd_shorten_op(cldev, pcls, delta)\
   cmd_shorten_list_op(cldev, &(pcls)->list, delta)
 
-/* Flush the buffer. */
-int clist_flush_buffer(P1(gx_device_clist_writer *));
+/* End a page by flushing the buffer and terminating the command list. */
+int clist_end_page(P1(gx_device_clist_writer *));
 
 /* Compute the # of bytes required to represent a variable-size integer. */
 /* (This works for negative integers also; they are written as though */
@@ -345,7 +357,6 @@ int cmd_size_w(P1(uint));
   (w1byte((wx) | (wy)) ? 2 :\
    cmd_size_w((uint)(wx)) + cmd_size_w((uint)(wy)))
 #define cmd_sizexy(xy) cmd_size2w((xy).x, (xy).y)
-int cmd_size_rect(P1(const gx_cmd_rect *));
 #define cmd_sizew_max ((sizeof(uint) * 8 + 6) / 7)
 
 /* Put a variable-size integer in the buffer. */
@@ -413,10 +424,6 @@ int cmd_put_enable_clip(P3(gx_device_clist_writer *, gx_clist_state *, int));
 int cmd_set_lop(P3(gx_device_clist_writer *, gx_clist_state *,
   gs_logical_operation_t));
 
-/* Put out a fill or tile rectangle command. */
-int cmd_write_rect_cmd(P7(gx_device_clist_writer *cldev, gx_clist_state *pcls,
-  int op, int x, int y, int width, int height));
-
 /*
  * Define macros for dividing up an operation into bands.
  * Note that BEGIN_RECT resets y and height.  It is OK for the code that
@@ -427,7 +434,7 @@ int cmd_write_rect_cmd(P7(gx_device_clist_writer *cldev, gx_clist_state *pcls,
  */
 #define BEGIN_RECT\
    {	int yend = y + height;\
-	int band_height = cdev->band_height;\
+	int band_height = cdev->page_band_height;\
 	do\
 	   {	int band = y / band_height;\
 		gx_clist_state *pcls = cdev->states + band;\
@@ -440,6 +447,12 @@ int cmd_write_rect_cmd(P7(gx_device_clist_writer *cldev, gx_clist_state *pcls,
 	while ( (y += height) < yend );\
    }
 
+/* ------ Exported by gxclrect.c ------ */
+
+/* Put out a fill or tile rectangle command. */
+int cmd_write_rect_cmd(P7(gx_device_clist_writer *cldev, gx_clist_state *pcls,
+  int op, int x, int y, int width, int height));
+
 /* ------ Exported by gxclbits.c ------ */
 
 /*
@@ -469,6 +482,25 @@ int cmd_put_bits(P10(gx_device_clist_writer *cldev, gx_clist_state *pcls,
 		     uint raster, int op_size, int compression_mask,
 		     byte **pdp, uint *psize));
 
+/*
+ * Put out commands for a color map (transfer function, black generation, or
+ * undercolor removal).  If pid != 0, write the map only if its ID differs
+ * from the current one, and update the saved ID in the case.
+ */
+typedef enum {
+  cmd_map_transfer = 0,		/* all transfer functions */
+  cmd_map_transfer_0,		/* transfer[0] */
+  cmd_map_transfer_1,		/* transfer[1] */
+  cmd_map_transfer_2,		/* transfer[2] */
+  cmd_map_transfer_3,		/* transfer[3] */
+  cmd_map_ht_transfer,		/* transfer fn of most recent halftone order */
+  cmd_map_black_generation,
+  cmd_map_undercolor_removal
+} cmd_map_index;
+int cmd_put_color_map(P4(gx_device_clist_writer *cldev,
+			 cmd_map_index map_index,
+			 const gx_transfer_map *map, gs_id *pid));
+
 /*
  * Change tiles for clist_tile_rectangle.  (We make this a separate
  * procedure primarily for readability.)
@@ -482,3 +514,14 @@ int clist_change_tile(P4(gx_device_clist_writer *cldev, gx_clist_state *pcls,
  */
 int clist_change_bits(P4(gx_device_clist_writer *cldev, gx_clist_state *pcls,
   const gx_strip_bitmap *tiles, int depth));
+
+/* ------ Exported by gxclimag.c ------ */
+
+/*
+ * Add commands to represent a full (device) halftone.
+ * (This routine should probably be in some other module.)
+ * Note: in the FUTURE, the type parameter is unnecessary, because device
+ * halftones record the type.
+ */
+int cmd_put_halftone(P3(gx_device_clist_writer *cldev,
+  const gx_device_halftone *pdht, gs_halftone_type type));
