@@ -1,4 +1,4 @@
-/* Copyright (C) 1995, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -23,6 +23,7 @@
 
 /* gxclmem.c */
 /* RAM-based command list implementation */
+#include "memory_.h"
 #include "gx.h"
 #include "gserrors.h"
 #include "gxclmem.h"
@@ -38,9 +39,7 @@
    a couple of the more likely to change "tuning" parameters given in the
    two macros below -- NEED_TO_COMPRESS and GET_NUM_RAW_BUFFERS. Usually
    the NEED_TO_COMPRESS decision will be deferred as long as possible based
-   on some total system free RAM space remaining. The "need_to_compress"
-   counter is really only to force compression at a pre-defined file size
-   for use during debuging.
+   on some total system free RAM space remaining.
 
    The data structures are in "memfile.h", and the primary 'tuning' parameter
    is MEMFILE_DATA_SIZE. This should not be too small to keep the overhead
@@ -59,15 +58,15 @@
 
 COMPRESSION.
 
-   When compression is triggered for a file (which should be avoided for the
-   'b' or "index" file) then all of the blocks except the last are compressed.
-   Compression will result in a physical block that holds data for more than
-   one logical block. Each logical block now points to the start of compressed
-   data in a physical block with the 'phys_pdata' pointer. The 'data_limit'
-   pointer in the physical block is where the compression logic stopped storing
-   data (as stream data compressors are allowed to do). The data for the
-   logical block may span to the next physical block. Once physical blocks are
-   compressed, they are chained together using the 'link' field.
+   When compression is triggered for a file then all of the blocks except
+   the last are compressed.  Compression will result in a physical block
+   that holds data for more than one logical block. Each logical block now
+   points to the start of compressed data in a physical block with the
+   'phys_pdata' pointer. The 'data_limit' pointer in the physical block is
+   where the compression logic stopped storing data (as stream data
+   compressors are allowed to do). The data for the logical block may span
+   to the next physical block. Once physical blocks are compressed, they are
+   chained together using the 'link' field.
 
    The 'f->phys_curr' points to the block being filled by compression, with
    the 'f->wt.ptr' pointing to the last byte filled in the block. These are
@@ -134,28 +133,42 @@ LIMITATIONS.
 
 ********************************************************************************/
 
-/* The need to compress should be conditional on the amount of available */
-/* memory, but we don't have a way to communicate this to these routines. */
-#define NEED_TO_COMPRESS(f) ((f)->ok_to_compress)
+/*
+   The need to compress should be conditional on the amount of available
+   memory, but we don't have a way to communicate this to these routines.
+   Instead, we simply start compressing when we've allocated more than
+   COMPRESSION_THRESHOLD amount of data.  The threshold should be at
+   least as large as the fixed overhead of the compressor plus the
+   decompressor, plus the expected compressed size of a block that size.
+*/
+private const long COMPRESSION_THRESHOLD = 300000;
+#define NEED_TO_COMPRESS(f)\
+  ((f)->ok_to_compress && (f)->total_space > COMPRESSION_THRESHOLD)
 
    /* FOR NOW ALLOCATE 1 raw buffer for every 32 blocks (at least 8)	*/
 #define GET_NUM_RAW_BUFFERS( f ) 					\
 	 max(f->log_length/MEMFILE_DATA_SIZE/32, 8)
 
-/* Define malloc and free to use the C heap directly, because		*/
-/* we don't want to bother making all the memfile data structures	*/
-/* work properly with the garbage collector.				*/
 #define MALLOC(f, siz, cname)\
-  (void *)gs_alloc_bytes(&gs_memory_default, siz, cname)
+  (void *)gs_alloc_bytes((f)->data_memory, siz, cname)
+#define ALLOCATE(vptr, f, cname, msg)\
+  vptr = MALLOC(f, sizeof(*(vptr)), cname);\
+  if ( !(vptr) ) {\
+    eprintf(msg);\
+    return_error(gs_error_VMerror);\
+  }\
+  (f)->total_space += sizeof(*(vptr))
 #define FREE(f, obj, cname)\
-  gs_free_object(&gs_memory_default, obj, cname)
+  (gs_free_object((f)->data_memory, obj, cname),\
+   (f)->total_space -= sizeof(*(obj)))
+
+/* Structure descriptor for GC */
+private_st_MEMFILE();
 
 	/* forward references */
 private void	memfile_free_mem(P1(MEMFILE *f));
 private int	memfile_init_empty(P1(MEMFILE *f));
 
-int  need_to_compress;
-
 /************************************************/
 /*   #define DEBUG	/- force statistics -/	*/
 /************************************************/
@@ -166,6 +179,10 @@ long tot_raw;
 long tot_cache_miss;
 long tot_cache_hits;
 long tot_swap_out;
+/*
+   The following pointers are here only for helping with a dumb debugger
+   that can't inspect local variables!
+*/
 byte		*decomp_wt_ptr0, *decomp_wt_limit0;
 const byte 	*decomp_rd_ptr0, *decomp_rd_limit0;
 byte		*decomp_wt_ptr1, *decomp_wt_limit1;
@@ -175,25 +192,67 @@ const byte 	*decomp_rd_ptr1, *decomp_rd_limit1;
 /* ---------------- Open/close/unlink ---------------- */
 
 int
-memfile_open_scratch(char *fname, clist_file_ptr /*MEMFILE **/ *pf,
-  gs_memory_t *mem, bool ok_to_compress)
+memfile_fopen(char *fname, const char *fmode,
+  clist_file_ptr /*MEMFILE **/ *pf, gs_memory_t *mem, bool ok_to_compress)
 {
    MEMFILE	*f;
    int		code;
 
+   /* We don't implement reopening an existing file. */
+   if ( fname[0] != 0 || fmode[0] != 'w' )
+     return_error(gs_error_invalidfileaccess);
+
 	/* There is no need to set fname in this implementation, */
 	/* but we do it anyway. */
    fname[0] = (ok_to_compress ? 'a' : 'b');
    fname[1] = 0;
 
-   f = (MEMFILE *) gs_alloc_bytes(mem, sizeof(MEMFILE),
-				  "memfile_open_scratch" );
+   f = gs_alloc_struct(mem, MEMFILE, &st_MEMFILE,
+		       "memfile_open_scratch(MEMFILE)");
    if ( f == NULL ) {
-      eprintf1("memfile_open_scratch(%s): gs_alloc_bytes failed\n", fname);
+      eprintf1("memfile_open_scratch(%s): gs_alloc_struct failed\n", fname);
       return_error(gs_error_VMerror);
    }
    f->memory = mem;
-   f->ok_to_compress = ok_to_compress;
+#ifdef USE_C_HEAP_FOR_DATA
+   f->data_memory = &gs_memory_default;
+#else
+   f->data_memory = mem;
+#endif
+   /*
+    * Disregard the ok_to_compress flag, since the size threshold gives us
+    * a much better criterion for deciding when compression is appropriate.
+    */
+   f->ok_to_compress = /*ok_to_compress*/ true;
+   f->compress_state = 0;	/* make clean for GC */
+   f->decompress_state = 0;
+   if ( f->ok_to_compress ) {
+     const stream_state *compress_proto = clist_compressor_state(NULL);
+     const stream_state *decompress_proto = clist_decompressor_state(NULL);
+     const stream_template *compress_template = compress_proto->template;
+     const stream_template *decompress_template = decompress_proto->template;
+     f->compress_state =
+       gs_alloc_struct(mem, stream_state, compress_template->stype,
+		       "memfile_open_scratch(compress_state)");
+     f->decompress_state =
+       gs_alloc_struct(mem, stream_state, decompress_template->stype,
+		       "memfile_open_scratch(decompress_state)");
+     if ( f->compress_state == 0 || f->decompress_state == 0 ) {
+       eprintf1("memfile_open_scratch(%s): gs_alloc_struct failed\n", fname);
+       return_error(gs_error_VMerror);
+     }
+     memcpy(f->compress_state, compress_proto,
+	    gs_struct_type_size(compress_template->stype));
+     f->compress_state->memory = mem;
+     memcpy(f->decompress_state, decompress_proto,
+	    gs_struct_type_size(decompress_template->stype));
+     f->decompress_state->memory = mem;
+     if ( compress_template->set_defaults )
+       (*compress_template->set_defaults)(f->compress_state);
+     if ( decompress_template->set_defaults )
+       (*decompress_template->set_defaults)(f->decompress_state);
+   }
+   f->total_space = 0;
 	/* init an empty file		*/
    if( (code = memfile_init_empty(f)) != 0 )
       return_error(code);
@@ -214,13 +273,31 @@ memfile_open_scratch(char *fname, clist_file_ptr /*MEMFILE **/ *pf,
    return 0;
 }
 
-void
-memfile_fclose_and_unlink(clist_file_ptr cf, const char *fname)
+int
+memfile_fclose(clist_file_ptr cf, const char *fname, bool delete)
 {
    MEMFILE *f = (MEMFILE *)cf;
 
-   memfile_rewind( f, true );
-   gs_free_object(f->memory, f, "memfile_close_and_unlink");
+   /* We don't implement closing without deletion. */
+   if ( !delete )
+     return_error(gs_error_invalidfileaccess);
+   memfile_free_mem(f);
+   gs_free_object(f->memory, f->decompress_state,
+		  "memfile_close_and_unlink(decompress_state)");
+   gs_free_object(f->memory, f->compress_state,
+		  "memfile_close_and_unlink(compress_state)");
+   gs_free_object(f->memory, f, "memfile_close_and_unlink(MEMFILE)");
+   return 0;
+}
+
+int
+memfile_unlink(const char *fname)
+{
+  /*
+   * Since we have no way to represent a memfile other than by the
+   * pointer, we don't (can't) implement unlinking.
+   */
+  return_error(gs_error_invalidfileaccess);
 }
 
 /* ---------------- Writing ---------------- */
@@ -239,24 +316,20 @@ compress_log_blk( MEMFILE *f, LOG_MEMFILE_BLK *bp )
 
    bp->phys_blk = f->phys_curr;
    bp->phys_pdata = (char *)(f->wt.ptr) + 1;
-   if ( COMPRESS_TEMPLATE.reinit != 0 )
-     (*COMPRESS_TEMPLATE.reinit)((stream_state *)&(f->filter_state));
+   if ( f->compress_state->template->reinit != 0 )
+     (*f->compress_state->template->reinit)(f->compress_state);
    compressed_size = 0;
 
    start_ptr = f->wt.ptr;
-   status = (*COMPRESS_TEMPLATE.process)( (stream_state *)&(f->filter_state),
+   status = (*f->compress_state->template->process)(f->compress_state,
 			&(f->rd), &(f->wt), true );
    bp->phys_blk->data_limit = (char *)(f->wt.ptr);
 
    if( status == 1 ) {	/* More output space needed (see strimpl.h)	*/
       /* allocate another physical block, then compress remainder	*/
       compressed_size = f->wt.limit - start_ptr;
-      newphys = (PHYS_MEMFILE_BLK *) MALLOC(f, sizeof(PHYS_MEMFILE_BLK),
-					    "memfile newphys");
-      if( ! newphys ) {
-	 eprintf("compress_log_blk : MALLOC for 'newphys' failed\n");
-	 return_error(gs_error_VMerror);
-      }
+      ALLOCATE(newphys, f, "memfile newphys",
+	       "compress_log_blk : MALLOC for 'newphys' failed\n");
       newphys->link = NULL;
       bp->phys_blk->link = newphys;
       f->phys_curr = newphys;
@@ -264,7 +337,7 @@ compress_log_blk( MEMFILE *f, LOG_MEMFILE_BLK *bp )
       f->wt.limit = f->wt.ptr + MEMFILE_DATA_SIZE;
 
       start_ptr = f->wt.ptr;
-      status = (*COMPRESS_TEMPLATE.process)( (stream_state *)&(f->filter_state),
+      status = (*f->compress_state->template->process)(f->compress_state,
 			   &(f->rd), &(f->wt), true );
       if( status != 0 ) {
 	 eprintf("Compression required more than one full block!\n");
@@ -294,21 +367,13 @@ memfile_next_blk( MEMFILE *f )
 
    if( f->phys_curr == NULL ) {	/* means NOT compressing 		*/
       /* allocate a new block						*/
-      newphys = (PHYS_MEMFILE_BLK *) MALLOC(f, sizeof(PHYS_MEMFILE_BLK),
-					    "memfile newphys");
-      if( ! newphys ) {
-	 eprintf("memfile_next_blk: MALLOC 1 for 'newphys' failed\n");
-	 return_error(gs_error_VMerror);
-      }
+      ALLOCATE(newphys, f, "memfile newphys",
+	       "memfile_next_blk: MALLOC 1 for 'newphys' failed\n");
       newphys->link = NULL;
       newphys->data_limit = NULL;	/* raw				*/
 
-      newbp = (LOG_MEMFILE_BLK *) MALLOC(f, sizeof(LOG_MEMFILE_BLK),
-					 "memfile newbp");
-      if( newbp == NULL ) {
-	 eprintf("memfile_next_blk: MALLOC 1 for 'newbp' failed\n");
-	 return_error(gs_error_VMerror);
-      }
+      ALLOCATE(newbp, f, "memfile newbp",
+	       "memfile_next_blk: MALLOC 1 for 'newbp' failed\n");
       bp->link = newbp;
       newbp->link = NULL;
       newbp->raw_block = NULL;
@@ -323,10 +388,13 @@ memfile_next_blk( MEMFILE *f )
          /* compress the entire file up to this point 			*/
 	 if ( !f->compressor_initialized ) {
 	    int code = 0;
-	    f->filter_state.template = &COMPRESS_TEMPLATE;
-	    f->filter_state.memory = f->memory;
-	    if ( COMPRESS_TEMPLATE.init != 0 )
-	       code = (*COMPRESS_TEMPLATE.init)((stream_state *)&f->filter_state);
+	    if ( f->compress_state->template->init != 0 )
+	       code = (*f->compress_state->template->init)(f->compress_state);
+	    if ( code < 0 )
+	       return_error(gs_error_VMerror);	/****** BOGUS ******/
+	    if ( f->decompress_state->template->init != 0 )
+	      code = (*f->decompress_state->template->init)
+	        (f->decompress_state);
 	    if ( code < 0 )
 	       return_error(gs_error_VMerror);	/****** BOGUS ******/
 	    f->compressor_initialized = true;
@@ -347,12 +415,8 @@ memfile_next_blk( MEMFILE *f )
 	    bp = bp->link;
 	 }	/* end while( ) compress loop 				*/
 	 /* Allocate a physical block for this (last) logical block 	*/
-	 newphys = (PHYS_MEMFILE_BLK *) MALLOC(f, sizeof(PHYS_MEMFILE_BLK),
-					       "memfile newphys");
-	 if( ! newphys ) {
-	    eprintf("memfile_next_blk: MALLOC 2 for 'newphys' failed\n");
-	    return_error(gs_error_VMerror);
-	 }
+	 ALLOCATE(newphys, f, "memfile newphys",
+		  "memfile_next_blk: MALLOC 2 for 'newphys' failed\n");
 	 newphys->link = NULL;
 	 newphys->data_limit = NULL;		/* raw			*/
 
@@ -370,12 +434,8 @@ memfile_next_blk( MEMFILE *f )
  	 /* compresses bp on phys list	*/
       if( (code = compress_log_blk(f,bp)) != 0 )
 	 return_error( code );
-      newbp = (LOG_MEMFILE_BLK *) MALLOC(f, sizeof(LOG_MEMFILE_BLK),
-					 "memfile newbp");
-      if( ! newbp ) {
-	 eprintf("memfile_next_blk: MALLOC 2 for 'newbp' failed\n");
-	 return_error(gs_error_VMerror);
-      }
+      ALLOCATE(newbp, f, "memfile newbp",
+	       "memfile_next_blk: MALLOC 2 for 'newbp' failed\n");
       bp->link = newbp;
       newbp->link = NULL;
       newbp->raw_block = NULL;
@@ -394,7 +454,7 @@ memfile_fwrite_chars(const void *data, uint len, clist_file_ptr cf)
 {
    const char *str = (const char *)data;
    MEMFILE *f = (MEMFILE *)cf;
-   long count = len;
+   uint count = len;
    int status;
 
    /* check if we are writing to the start of the file.  If so, then	*/
@@ -408,14 +468,21 @@ memfile_fwrite_chars(const void *data, uint len, clist_file_ptr cf)
       eprintf(" Write file truncate -- need to free physical blocks.\n");
    }
 
-   while( count-- ) {
-      if( f->pdata == f->pdata_end )
+   while( count ) {
+     uint move_count = f->pdata_end - f->pdata;
+     if ( move_count == 0 ) {
 	 if( (status = memfile_next_blk( f )) != 0 ) {
 	    f->error_code = status;
 	    return( 0 );
 	 }
-
-      *(f->pdata++) = *str++;
+     } else {
+       if ( move_count > count )
+	 move_count = count;
+       memmove(f->pdata, str, move_count);
+       f->pdata += move_count;
+       str += move_count;
+       count -= move_count;
+     }
    }
    f->log_curr_pos += len;
    f->log_length = f->log_curr_pos;	/* truncate length to here	*/
@@ -454,12 +521,8 @@ memfile_get_pdata( MEMFILE *f )
       if( f->raw_head == NULL ) {
 	 /* need to allocate the raw buffer pool			*/
 	 num_raw_buffers = GET_NUM_RAW_BUFFERS( f );
-	 f->raw_head = (RAW_BUFFER *) MALLOC(f, sizeof(RAW_BUFFER),
-					     "memfile raw buffer");
-	 if( ! f->raw_head ) {
-	    eprintf("memfile_get_pdata: MALLOC for 'raw_head' failed\n");
-	    return_error(gs_error_VMerror);
-	 }
+	 ALLOCATE(f->raw_head, f, "memfile raw buffer",
+		  "memfile_get_pdata: MALLOC for 'raw_head' failed\n");
 	 f->raw_head->back = NULL;
 	 f->raw_tail = f->raw_head;
 	 f->raw_tail->log_blk = NULL;
@@ -468,6 +531,7 @@ memfile_get_pdata( MEMFILE *f )
 						     "memfile raw buffer");
 	       /* if MALLOC fails, then just stop allocating 		*/
 	    if( ! f->raw_tail->fwd ) break;
+	    f->total_space += sizeof(RAW_BUFFER);
 	    f->raw_tail->fwd->back = f->raw_tail;
 	    f->raw_tail = f->raw_tail->fwd;
 	    f->raw_tail->log_blk = NULL;
@@ -503,10 +567,8 @@ memfile_get_pdata( MEMFILE *f )
 
 	 /* Decompress the data into this raw block			*/
 	    /* Initialize the decompressor 				*/
-	 if ( DECOMPRESS_TEMPLATE.set_defaults != 0 )
-	   (*DECOMPRESS_TEMPLATE.set_defaults)((stream_state *)&(f->filter_state));
-	 if ( DECOMPRESS_TEMPLATE.reinit != 0 )
-	   (*DECOMPRESS_TEMPLATE.reinit)((stream_state *)&(f->filter_state));
+	 if ( f->decompress_state->template->reinit != 0 )
+	   (*f->decompress_state->template->reinit)(f->decompress_state);
 	    /* Set pointers and call the decompress routine		*/
 	 f->wt.ptr = (byte *)(f->raw_head->data) - 1;
 	 f->wt.limit = f->wt.ptr + MEMFILE_DATA_SIZE;
@@ -518,9 +580,8 @@ memfile_get_pdata( MEMFILE *f )
 	 decomp_rd_ptr0 = f->rd.ptr;
 	 decomp_rd_limit0 = f->rd.limit;
 #endif
-	 status = (*DECOMPRESS_TEMPLATE.process)(
-			(stream_state *)&(f->filter_state),
-			&(f->rd), &(f->wt), true );
+	 status = (*f->decompress_state->template->process)
+	   (f->decompress_state, &(f->rd), &(f->wt), true );
 	 if( status == 0 ) {	/* More input data needed */
 	    /* switch to next block and continue decompress 		*/
 	    int back_up = 0;		/* adjust pointer backwards	*/
@@ -538,9 +599,8 @@ memfile_get_pdata( MEMFILE *f )
 	    decomp_rd_ptr1 = f->rd.ptr;
 	    decomp_rd_limit1 = f->rd.limit;
 #endif
-	    status = (*DECOMPRESS_TEMPLATE.process)(
-			(stream_state *)&(f->filter_state),
-			&(f->rd), &(f->wt), true );
+	    status = (*f->decompress_state->template->process)
+	      (f->decompress_state, &(f->rd), &(f->wt), true );
 	    if( status == 0 ) {
 	       eprintf("Decompression required more than one full block!\n");
 	       return_error(gs_error_Fatal);
@@ -598,13 +658,14 @@ memfile_fread_chars(void *data, uint len, clist_file_ptr cf)
 	 f->log_curr_blk = (f->log_curr_blk)->link;
          memfile_get_pdata( f );
       }
-      move_count = count;
-      if( move_count > (f->pdata_end - f->pdata) )
-	 move_count = f->pdata_end - f->pdata;
-      count -= move_count;
+      move_count = f->pdata_end - f->pdata;
+      if ( move_count > count )
+	move_count = count;
       f->log_curr_pos += move_count - 1;	/* new position 	*/
-      while( move_count-- )
-	 *str++ = *(f->pdata++);
+      memmove(str, f->pdata, move_count);
+      str += move_count;
+      f->pdata += move_count;
+      count -= move_count;
    }
 
    return( num_read );
@@ -625,18 +686,23 @@ memfile_ftell(clist_file_ptr cf)
 }
 
 void
-memfile_rewind(clist_file_ptr cf, bool discard_data)
+memfile_rewind(clist_file_ptr cf, bool discard_data, const char *ignore_fname)
 {
    MEMFILE *f = (MEMFILE *)cf;
 
-   f->log_curr_blk = f->log_head;
-   f->log_curr_pos = 0;
-   memfile_get_pdata( f );
-   /**************** memfile_free_mem (f) if discard_data ****************/
+   if ( discard_data ) {
+     memfile_free_mem(f);
+     /* We have to call memfile_init_empty to preserve invariants. */
+     memfile_init_empty(f);
+   } else {
+     f->log_curr_blk = f->log_head;
+     f->log_curr_pos = 0;
+     memfile_get_pdata( f );
+   }
 }
 
 int
-memfile_fseek(clist_file_ptr cf, long offset, int mode)
+memfile_fseek(clist_file_ptr cf, long offset, int mode, const char *ignore_fname)
 {
    MEMFILE *f = (MEMFILE *)cf;
    long i, block_num, new_pos;
@@ -685,8 +751,6 @@ private void
 memfile_free_mem( MEMFILE *f )
 {
    LOG_MEMFILE_BLK *bp, *tmpbp;
-   PHYS_MEMFILE_BLK *pphys, *tmpphys;
-   RAW_BUFFER	    *tmpraw;
 
 #ifdef DEBUG
    /* output some diagnostics about the effectiveness			*/
@@ -709,16 +773,29 @@ memfile_free_mem( MEMFILE *f )
 
       /* Free up memory that was allocated for the memfile		*/
    bp = f->log_head;
+
+/******************************************************************
+ * The following was the original algorithm here.  This algorithm has a bug:
+ * the second loop references the physical blocks again after they have been
+ * freed.
+ ******************************************************************/
+
+#if 0		/**************** ****************/
+
+   if ( bp != NULL ) {
       /* Free the physical blocks that make up the compressed data	*/
-   pphys = (f->log_head)->phys_blk;
-   if( pphys->data_limit != NULL ) {
-      /* the data was compressed, free the chain of blocks		*/
-      while( pphys != NULL ) {
-	 tmpphys = pphys->link;
+     PHYS_MEMFILE_BLK *pphys = (f->log_head)->phys_blk;
+     if( pphys->data_limit != NULL ) {
+       /* the data was compressed, free the chain of blocks		*/
+       while( pphys != NULL ) {
+	 PHYS_MEMFILE_BLK *tmpphys = pphys->link;
+
 	 FREE(f, pphys, "memfile_free_mem(pphys)");
 	 pphys = tmpphys;
-      }
+       }
+     }
    }
+
       /* free the logical blocks 					*/
    while( bp != NULL ) {
 	/* if this logical block was not compressed, free the phys_blk	*/
@@ -726,22 +803,90 @@ memfile_free_mem( MEMFILE *f )
 	 FREE(f, bp->phys_blk, "memfile_free_mem(phys_blk)");
       }
       tmpbp = bp->link;
-      FREE(f, bp, "memfile_free_mem(bp)");
+      FREE(f, bp, "memfile_free_mem(log_blk)");
       bp = tmpbp;
    }
 
+#else		/**************** ****************/
+# if 1		/**************** ****************/
+
+/****************************************************************
+ * This algorithm is correct (we think).
+ ****************************************************************/
+
+   if ( bp != NULL ) {
+     /* Null out phys_blk pointers to compressed data. */
+     PHYS_MEMFILE_BLK *pphys = bp->phys_blk;
+     { for ( tmpbp = bp; tmpbp != NULL; tmpbp = tmpbp->link )
+         if ( tmpbp->phys_blk->data_limit != NULL )
+	   tmpbp->phys_blk = 0;
+     }
+      /* Free the physical blocks that make up the compressed data	*/
+     if( pphys->data_limit != NULL ) {
+       /* the data was compressed, free the chain of blocks		*/
+       while( pphys != NULL ) {
+	 PHYS_MEMFILE_BLK *tmpphys = pphys->link;
+
+	 FREE(f, pphys, "memfile_free_mem(pphys)");
+	 pphys = tmpphys;
+       }
+     }
+   }
+
+   /* Now free the logical blocks, and any uncompressed physical blocks. */
+   while( bp != NULL ) {
+      if( bp->phys_blk != NULL ) {
+	FREE(f, bp->phys_blk, "memfile_free_mem(phys_blk)");
+      }
+      tmpbp = bp->link;
+      FREE(f, bp, "memfile_free_mem(log_blk)");
+      bp = tmpbp;
+   }
+
+/***********************************************************************
+ * This algorithm appears to be both simpler and free of the bug that
+ * occasionally causes the older one to reference freed blocks; but in
+ * fact it can miss blocks, because the very last compressed logical block
+ * can have spill into a second physical block, which is not referenced by
+ * any logical block.
+ ***********************************************************************/
+
+# else		/**************** ****************/
+
+   { PHYS_MEMFILE_BLK *prev_phys = 0;
+
+     while ( bp != NULL ) {
+       PHYS_MEMFILE_BLK *phys = bp->phys_blk;
+       if ( phys != prev_phys ) {
+	 FREE(f, phys, "memfile_free_mem(phys_blk)");
+	 prev_phys = phys;
+       }
+      tmpbp = bp->link;
+      FREE(f, bp, "memfile_free_mem(log_blk)");
+      bp = tmpbp;
+     }
+   }
+
+# endif		/**************** ****************/
+#endif		/**************** ****************/
+
+   f->log_head = NULL;
+
      /* Free any internal compressor state. */
    if ( f->compressor_initialized ) {
-      if ( COMPRESS_TEMPLATE.release != 0 )
-        (*COMPRESS_TEMPLATE.release)((stream_state *)&f->filter_state);
+      if ( f->decompress_state->template->release != 0 )
+        (*f->decompress_state->template->release)(f->decompress_state);
+      if ( f->compress_state->template->release != 0 )
+        (*f->compress_state->template->release)(f->compress_state);
       f->compressor_initialized = false;
    }
 
       /* free the raw buffers						*/
    while( f->raw_head != NULL ) {
-      tmpraw = f->raw_head->fwd;
-      FREE(f, f->raw_head, "memfile_free_mem(raw)");
-      f->raw_head = tmpraw;
+     RAW_BUFFER *tmpraw = f->raw_head->fwd;
+
+     FREE(f, f->raw_head, "memfile_free_mem(raw)");
+     f->raw_head = tmpraw;
    }
 }
 
@@ -752,22 +897,12 @@ memfile_init_empty( MEMFILE *f )
 
    /* File empty - get a physical mem block (includes the buffer area)	*/
    f->phys_curr = NULL; 	/* flag as file not compressed    	*/
-   pphys = (PHYS_MEMFILE_BLK *) MALLOC(f, sizeof(PHYS_MEMFILE_BLK),
-				       "memfile pphys");
-   if( ! pphys ) {
-      eprintf1("memfile_init_empty(0x%lx): MALLOC for 'pphys' failed\n",
-	       (ulong)f);
-      return_error(gs_error_VMerror);
-   }
+   ALLOCATE(pphys, f, "memfile pphys",
+	    "memfile_init_empty: MALLOC for 'pphys' failed\n");
    pphys->data_limit = NULL;			/* raw data for now 	*/
 
-   f->log_curr_blk = (LOG_MEMFILE_BLK *) MALLOC(f, sizeof(LOG_MEMFILE_BLK),
-						"memfile log blk");
-   if( ! f->log_curr_blk ) {
-      eprintf1("memfile_init_empty(0x%lx): MALLOC for log_curr_blk failed\n",
-		(ulong)f);
-      return_error(gs_error_VMerror);
-   }
+   ALLOCATE(f->log_curr_blk, f, "memfile log blk",
+	    "memfile_init_empty: MALLOC for log_curr_blk failed\n");
    f->log_head = f->log_curr_blk;
    f->log_curr_blk->link = NULL;
    f->log_curr_blk->phys_blk = pphys;
@@ -785,6 +920,5 @@ memfile_init_empty( MEMFILE *f )
       /* Tag the compressor state as uninitialized			*/
    f->compressor_initialized = false;
 
-   need_to_compress = -10;		/*====<<<<>>>>====*/
    return 0;
 }
