@@ -1,4 +1,4 @@
-/* Copyright (C) 1995, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -32,6 +32,7 @@
 #include "gxdevmem.h"			/* must precede gxcldev.h */
 #include "gxcldev.h"
 #include "gxclpath.h"
+#include "gxcolor2.h"
 #include "gxpaint.h"			/* for gx_fill/stroke_params */
 #include "gzpath.h"
 #include "gzcpath.h"
@@ -44,9 +45,9 @@ ulong cmd_diffs[5];
 #endif
 
 /* Forward declarations */
-private int cmd_put_path(P7(gx_device_clist_writer *cldev,
+private int cmd_put_path(P8(gx_device_clist_writer *cldev,
   gx_clist_state *pcls, const gx_path *ppath, fixed ymin, fixed ymax, byte op,
-  bool implicit_close));
+  bool implicit_close, segment_notes keep_notes));
 /* Driver procedures */
 private dev_proc_fill_path(clist_fill_path);
 private dev_proc_stroke_path(clist_stroke_path);
@@ -91,35 +92,19 @@ cmd_put_drawing_color(gx_device_clist_writer *cldev, gx_clist_state *pcls,
 
 	if ( gx_dc_is_pure(pdcolor) )
 	  { gx_color_index color1 = gx_dc_pure_color(pdcolor);
-	    return (color1 == pcls->colors[1] ? 0 :
-		    cmd_set_color1(cldev, pcls, color1));
-	  }
-	else if ( gx_dc_is_binary_halftone(pdcolor) )
-	  { tile = gx_dc_binary_tile(pdcolor);
-	    color0 = gx_dc_binary_color0(pdcolor);
-	    color1 = gx_dc_binary_color1(pdcolor);
-	  }
-	else		/* should handle colored halftones */
-	  return -1;
-
-	/* Set up tile and colors as for clist_tile_rectangle. */
-	if ( !cls_has_tile_id(cldev, pcls, tile->id, offset_temp) )
-	  {	int depth =
-		  (color1 == gx_no_color_index &&
-		   color0 == gx_no_color_index ?
-		   cldev->color_info.depth : 1);
-		if ( tile->id == gx_no_bitmap_id ||
-		     clist_change_tile(cldev, pcls, tile, depth) < 0
-		   )
-		  return -1;		/* can't cache tile */
-	  }
-	if ( color1 != pcls->tile_colors[1] ||
-	     color0 != pcls->tile_colors[0]
-	   )
-	  {	int code = cmd_set_tile_colors(cldev, pcls, color0, color1);
-		if ( code < 0 )
+	    if ( color1 != pcls->colors[1] )
+	      { int code = cmd_set_color1(cldev, pcls, color1);
+	        if ( code < 0 )
 		  return code;
+	      }
+#ifdef FUTURE
+	    return cmd_dc_type_pure;
+#else
+	    return 0;
+#endif
 	  }
+#ifdef FUTURE
+	/* Any non-pure color will require the phase. */
 	{ int px = pdcolor->phase.x, py = pdcolor->phase.y;
 	  if ( px != pcls->tile_phase.x || py != pcls->tile_phase.y )
 	    {	int code = cmd_set_tile_phase(cldev, pcls, px, py);
@@ -127,6 +112,87 @@ cmd_put_drawing_color(gx_device_clist_writer *cldev, gx_clist_state *pcls,
 		  return code;
 	    }
 	}
+#endif
+	if ( gx_dc_is_binary_halftone(pdcolor) )
+	  { tile = gx_dc_binary_tile(pdcolor);
+	    color0 = gx_dc_binary_color0(pdcolor);
+	    color1 = gx_dc_binary_color1(pdcolor);
+	    /* Set up tile and colors as for clist_tile_rectangle. */
+	    if ( !cls_has_tile_id(cldev, pcls, tile->id, offset_temp) )
+	      { int depth =
+		  (color1 == gx_no_color_index &&
+		   color0 == gx_no_color_index ?
+		   cldev->color_info.depth : 1);
+		if ( tile->id == gx_no_bitmap_id ||
+		     clist_change_tile(cldev, pcls, tile, depth) < 0
+		   )
+		  return_error(-1);		/* can't cache tile */
+	      }
+	    if ( color1 != pcls->tile_colors[1] ||
+		 color0 != pcls->tile_colors[0]
+	       )
+	      { int code = cmd_set_tile_colors(cldev, pcls, color0, color1);
+		if ( code < 0 )
+		  return code;
+	      }
+#ifdef FUTURE
+	    return cmd_dc_type_ht;
+#endif
+	  }
+#ifdef FUTURE
+	else if ( gx_dc_is_colored_halftone(pdcolor) ) 
+	  { const gx_device_halftone *pdht = pdcolor->colors.colored.c_ht;
+	    int num_comp = pdht->num_comp;
+	    byte buf[4 + 4 * cmd_max_intsize(sizeof(pdcolor->colors.colored.c_level[0]))];
+	    byte *bp = buf;
+	    int i;
+	    uint short_bases = 0;
+	    ulong bases = 0;
+	    byte *dp;
+
+	    /****** HOW TO TELL IF COLOR IS ALREADY SET? ******/
+	    if ( pdht->id != cldev->device_halftone_id )
+	      { int code = cmd_put_halftone(cldev, pdht, pdht->type);
+	        if ( code < 0 )
+		  return code;
+		cldev->device_halftone_id = pdht->id;
+	      }
+	    for ( i = 0; i < num_comp; ++i )
+	      { uint base = pdcolor->colors.colored.c_base[i];
+	        if ( base > 31 )
+		  return_error(gs_error_rangecheck);
+		bases |= base << ((3 - i) * 5);
+		short_bases |= base << (3 - i);
+	      }
+	    if ( bases & 0xf7bde )
+	      { /* Some base value requires more than 1 bit. */
+		*bp++ = 0x10 + (byte)(bases >> 16);
+		*bp++ = (byte)(bases >> 8);
+		*bp++ = (byte)bases;
+	      }
+	    else
+	      { /* The bases all fit in 1 bit each. */
+		*bp++ = 0x00 + (byte)short_bases;
+	      }
+	    for ( i = 0; i < num_comp; ++i )
+	      bp = cmd_put_w((uint)pdcolor->colors.colored.c_level[i], bp);
+	    /****** IGNORE alpha ******/
+	    set_cmd_put_op(dp, cldev, pcls, cmd_opv_set_color, bp - buf + 1);
+	    memcpy(dp + 1, buf, bp - buf);
+	    return cmd_dc_type_color;
+	  }
+#endif
+	else
+	  return_error(-1);
+#ifndef FUTURE
+	{ int px = pdcolor->phase.x, py = pdcolor->phase.y;
+	  if ( px != pcls->tile_phase.x || py != pcls->tile_phase.y )
+	    {	int code = cmd_set_tile_phase(cldev, pcls, px, py);
+		if ( code < 0 )
+		  return code;
+	    }
+	}
+#endif
 	return 0;
 }
 
@@ -251,27 +317,38 @@ cmd_write_unknown(gx_device_clist_writer *cldev, gx_clist_state *pcls,
 	    memcpy(dp + 1, &cldev->imager_state.line_params.miter_limit, sizeof(float));
 	    pcls->known |= miter_limit_known;
 	  }
-	if ( unknown & misc_known )
+	if ( unknown & misc0_known )
 	  { byte *dp;
-	    set_cmd_put_op(dp, cldev, pcls, cmd_opv_set_misc, 2);
-	    dp[1] =
-	      (cldev->imager_state.overprint ? 0x80 : 0) +
-	      (cldev->imager_state.stroke_adjust ? 0x40 : 0) +
+	    set_cmd_put_op(dp, cldev, pcls, cmd_opv_set_misc2, 2);
+	    dp[1] = cmd_set_misc2_cap_join +
 	      (cldev->imager_state.line_params.cap << 3) +
 	      cldev->imager_state.line_params.join;
-	    pcls->known |= misc_known;
+	    pcls->known |= misc0_known;
+	  }
+	if ( unknown & misc1_known )
+	  { byte *dp;
+	    set_cmd_put_op(dp, cldev, pcls, cmd_opv_set_misc2, 2);
+	    dp[1] = cmd_set_misc2_ac_op_sa +
+	      (cldev->imager_state.accurate_curves ? 4 : 0) +
+	      (cldev->imager_state.overprint ? 2 : 0) +
+	      (cldev->imager_state.stroke_adjust ? 1 : 0);
+	    pcls->known |= misc1_known;
 	  }
 	if ( unknown & dash_known )
 	  { byte *dp;
 	    int n = cldev->imager_state.line_params.dash.pattern_size;
 
 	    set_cmd_put_op(dp, cldev, pcls, cmd_opv_set_dash,
-			   2 + (n + 1) * sizeof(float));
-	    dp[1] = n;
-	    memcpy(dp + 2, &cldev->imager_state.line_params.dash.offset,
+			   2 + (n + 2) * sizeof(float));
+	    dp[1] = n + (cldev->imager_state.line_params.dash.adapt ? 0x80 : 0) +
+	      (cldev->imager_state.line_params.dot_length_absolute ? 0x40 : 0);
+	    memcpy(dp + 2, &cldev->imager_state.line_params.dot_length,
+		   sizeof(float));
+	    memcpy(dp + 2 + sizeof(float),
+		   &cldev->imager_state.line_params.dash.offset,
 		   sizeof(float));
 	    if ( n != 0 )
-	      memcpy(dp + 2 + sizeof(float),
+	      memcpy(dp + 2 + sizeof(float) * 2,
 		     cldev->imager_state.line_params.dash.pattern,
 		     n * sizeof(float));
 	    pcls->known |= dash_known;
@@ -280,19 +357,35 @@ cmd_write_unknown(gx_device_clist_writer *cldev, gx_clist_state *pcls,
 	  { /* We can write out the clipping path either as rectangles */
 	    /* or as a real (filled) path. */
 	    const gx_clip_path *pcpath = cldev->clip_path;
-	    int ymin = (pcls - cldev->states) * cldev->band_height;
-	    int ymax = min(ymin + cldev->band_height, cldev->height);
+	    int band_height = cldev->page_band_height;
+	    int ymin = (pcls - cldev->states) * band_height;
+	    int ymax = min(ymin + band_height, cldev->height);
+	    gs_fixed_rect box;
 	    byte *dp;
 
 	    set_cmd_put_op(dp, cldev, pcls, cmd_opv_begin_clip, 1);
 	    if ( pcpath->segments_valid )
-	      { /* Write out the path. */
-		int code = cmd_put_path(cldev, pcls, &pcpath->path,
-					int2fixed(ymin) - fixed_half,
-					int2fixed(ymax) + fixed_half,
+	      { int code;
+
+	        if ( gx_path_is_rectangle(&pcpath->path, &box) &&
+		     fixed_is_int(box.p.x | box.p.y | box.q.x | box.q.y)
+		   )
+		  { /* Write the path as a rectangle. */
+		    code = cmd_write_rect_cmd(cldev, pcls, cmd_op_fill_rect,
+					      fixed2int_var(box.p.x),
+					      fixed2int_var(box.p.y),
+					      fixed2int(box.q.x - box.p.x),
+					      fixed2int(box.q.y - box.p.y));
+		  }
+	        else
+		  { /* Write the path. */
+		    code = cmd_put_path(cldev, pcls, &pcpath->path,
+					int2fixed(ymin - 1),
+					int2fixed(ymax + 1),
 					(pcpath->rule == gx_rule_even_odd ?
 					 cmd_opv_eofill : cmd_opv_fill),
-					true);
+					true, sn_not_first);
+		  }
 		if ( code < 0 )
 		  return code;
 	      }
@@ -314,7 +407,6 @@ cmd_write_unknown(gx_device_clist_writer *cldev, gx_clist_state *pcls,
 		      if ( code < 0 )
 			return code;
 		    }
-		      
 	      }
 	    set_cmd_put_op(dp, cldev, pcls, cmd_opv_end_clip, 2);
 	    dp[1] = (gx_cpath_is_outside(pcpath) ? 1 : 0);
@@ -325,9 +417,26 @@ cmd_write_unknown(gx_device_clist_writer *cldev, gx_clist_state *pcls,
 	  { byte *dp;
 
 	    if ( cldev->color_space & 8 )	/* indexed */
-	      { set_cmd_put_op(dp, cldev, pcls, cmd_opv_set_color_space,
-			       2 + cmd_sizew(cldev->indexed_hival));
-		cmd_put_w(cldev->indexed_hival, dp + 2);
+	      { uint num_values = (cldev->indexed_params.hival + 1) *
+		  gs_color_space_num_components(
+		    (const gs_color_space *)&cldev->indexed_params.base_space);
+	        bool use_proc = cldev->color_space & 4;
+		const void *map_data;
+		uint map_size;
+
+		if ( use_proc )
+		  { map_data = cldev->indexed_params.lookup.map->values;
+		    map_size = num_values *
+		      sizeof(cldev->indexed_params.lookup.map->values[0]);
+		  }
+		else
+		  { map_data = cldev->indexed_params.lookup.table.data;
+		    map_size = num_values;
+		  }
+	        set_cmd_put_op(dp, cldev, pcls, cmd_opv_set_color_space,
+			2 + cmd_sizew(cldev->indexed_params.hival) + map_size);
+	        memcpy(cmd_put_w(cldev->indexed_params.hival, dp + 2),
+		       map_data, map_size);
 	      }
 	    else
 	      { set_cmd_put_op(dp, cldev, pcls, cmd_opv_set_color_space, 2);
@@ -345,25 +454,35 @@ clist_fill_path(gx_device *dev, const gs_imager_state *pis, gx_path *ppath,
   const gx_fill_params *params, const gx_drawing_color *pdcolor,
   const gx_clip_path *pcpath)
 {	uint unknown = 0;
-	int y, height;
+	int y, height, y0, y1;
 	gs_logical_operation_t lop = pis->log_op;
 	byte op = (byte)
 	  (params->rule == gx_rule_even_odd ?
+#ifdef FUTURE
+	   cmd_opv_eofill : cmd_opv_fill
+#else
 	   (gx_dc_is_pure(pdcolor) ? cmd_opv_eofill : cmd_opv_hteofill) :
-	   (gx_dc_is_pure(pdcolor) ? cmd_opv_fill : cmd_opv_htfill));
+	   (gx_dc_is_pure(pdcolor) ? cmd_opv_fill : cmd_opv_htfill)
+#endif
+	   );
 	gs_fixed_point adjust;
 
+#ifdef DEBUG
+	if ( gs_debug_c(',') )		/* disable path-based banding */
+	  return gx_default_fill_path(dev, pis, ppath, params, pdcolor,
+				      pcpath);
+#endif
 	adjust = params->adjust;
 	{ gs_fixed_rect bbox;
 	  gx_path_bbox(ppath, &bbox);
-	  if ( params->fill_zero_width )
-	    gx_adjust_if_empty(&bbox, &adjust);
-	  y = fixed2int(bbox.p.y - adjust.y);
-	  height = fixed2int_ceiling(bbox.q.y + adjust.y) - y;
+	  y = fixed2int(bbox.p.y) - 1;
+	  height = fixed2int_ceiling(bbox.q.y) - y + 1;
 	  fit_fill_yh(dev, y, height);
 	  if ( height <= 0 )
 	    return 0;
 	}
+	y0 = y;
+	y1 = y + height;
 	if ( cdev->imager_state.flatness != params->flatness )
 	  { unknown |= flatness_known;
 	    cdev->imager_state.flatness = params->flatness;
@@ -390,12 +509,9 @@ clist_fill_path(gx_device *dev, const gs_imager_state *pis, gx_path *ppath,
 	  }
 	else
 	  { if ( lop != pcls->lop )
-	      { byte *dp;
-		set_cmd_put_op(dp, cdev, pcls,
-			       cmd_opv_set_lop, 1 + cmd_size_w(lop));
-		++dp;
-		cmd_put_w(lop, dp);
-		pcls->lop = lop;
+	      { code = cmd_set_lop(cdev, pcls, lop);
+	        if ( code < 0 )
+		  return code;
 	      }
 	    cmd_enable_lop(cdev, pcls);
 	  }
@@ -406,9 +522,14 @@ clist_fill_path(gx_device *dev, const gs_imager_state *pis, gx_path *ppath,
 					pcpath);
 	  }
 	code = cmd_put_path(cdev, pcls, ppath,
-			    int2fixed(y) - params->adjust.y,
-			    int2fixed(y + height) + params->adjust.y,
-			    op, true);
+			    int2fixed(max(y - 1, y0)),
+			    int2fixed(min(y + height + 1, y1)),
+#ifdef FUTURE
+			    op + code,	/* cmd_dc_type */
+#else
+			    op,
+#endif
+			    true, sn_none /* fill doesn't need the notes */);
 	if ( code < 0 )
 	  return code;
 	END_RECT
@@ -423,22 +544,38 @@ clist_stroke_path(gx_device *dev, const gs_imager_state *pis, gx_path *ppath,
 	uint unknown = 0;
 	gs_fixed_rect bbox;
 	gs_fixed_point expansion;
-	fixed adjust_y;
-	int y, height;
+	int adjust_y;
+	int y, height, y0, y1;
 	gs_logical_operation_t lop = pis->log_op;
+#ifndef FUTURE
 	byte op = (byte)
 	  (gx_dc_is_pure(pdcolor) ? cmd_opv_stroke : cmd_opv_htstroke);
+#endif
 
+#ifdef DEBUG
+	if ( gs_debug_c(',') )		/* disable path-based banding */
+	  return gx_default_stroke_path(dev, pis, ppath, params, pdcolor,
+					pcpath);
+#endif
 	gx_path_bbox(ppath, &bbox);
 	/* We must use the supplied imager state, not our saved one, */
 	/* for computing the stroke expansion. */
-	gx_stroke_expansion(pis, &expansion);
-	adjust_y = expansion.y + pis->fill_adjust.y;
-	y = fixed2int(bbox.p.y - adjust_y);
-	height = fixed2int_ceiling(bbox.q.y + adjust_y) - y;
-	fit_fill_yh(dev, y, height);
-	if ( height <= 0 )
-	  return 0;
+	if ( gx_stroke_path_expansion(pis, ppath, &expansion) < 0 )
+	  { /* Expansion is too large: use the entire page. */
+	    adjust_y = 0;
+	    y = 0;
+	    height = dev->height;
+	  }
+	else
+	  { adjust_y = fixed2int_ceiling(expansion.y) + 1;
+	    y = fixed2int(bbox.p.y) - adjust_y;
+	    height = fixed2int_ceiling(bbox.q.y) - y + adjust_y;
+	    fit_fill_yh(dev, y, height);
+	    if ( height <= 0 )
+	      return 0;
+	  }
+	y0 = y;
+	y1 = y + height;
 	/* Check the dash pattern, since we bail out if */
 	/* the pattern is too large. */
 	cdev->imager_state.line_params.dash.pattern = cdev->dash_pattern;
@@ -446,7 +583,15 @@ clist_stroke_path(gx_device *dev, const gs_imager_state *pis, gx_path *ppath,
 	     (pattern_size != 0 &&
 	      memcmp(cdev->imager_state.line_params.dash.pattern,
 		     pis->line_params.dash.pattern,
-		     pattern_size * sizeof(float)))
+		     pattern_size * sizeof(float))) ||
+	     cdev->imager_state.line_params.dash.offset !=
+	       pis->line_params.dash.offset ||
+	     cdev->imager_state.line_params.dash.adapt !=
+	       pis->line_params.dash.adapt ||
+	     cdev->imager_state.line_params.dot_length !=
+	       pis->line_params.dot_length ||
+	     cdev->imager_state.line_params.dot_length_absolute !=
+	       pis->line_params.dot_length_absolute
 	   )
 	  { /* Bail out if the dash pattern is too long. */
 	    if ( pattern_size > cmd_max_dash )
@@ -457,6 +602,11 @@ clist_stroke_path(gx_device *dev, const gs_imager_state *pis, gx_path *ppath,
 			pis->line_params.dash.pattern,
 			pis->line_params.dash.pattern_size,
 			pis->line_params.dash.offset, NULL);
+	    gx_set_dash_adapt(&cdev->imager_state.line_params.dash,
+			      pis->line_params.dash.adapt);
+	    gx_set_dot_length(&cdev->imager_state.line_params,
+			      pis->line_params.dot_length,
+			      pis->line_params.dot_length_absolute);
 	  }
 	if ( state_neq(flatness) )
 	  { unknown |= flatness_known;
@@ -484,12 +634,17 @@ clist_stroke_path(gx_device *dev, const gs_imager_state *pis, gx_path *ppath,
 	    gx_set_miter_limit(&cdev->imager_state.line_params,
 			       pis->line_params.miter_limit);
 	  }
-	if ( state_neq(line_params.cap) || state_neq(line_params.join) ||
-	     state_neq(stroke_adjust)
-	   )
-	  { unknown |= misc_known;
+	if ( state_neq(line_params.cap) || state_neq(line_params.join) )
+	  { unknown |= misc0_known;
 	    state_update(line_params.cap);
 	    state_update(line_params.join);
+	  }
+	if ( state_neq(accurate_curves) || state_neq(overprint) ||
+	     state_neq(stroke_adjust)
+	   )
+	  { unknown |= misc1_known;
+	    state_update(accurate_curves);
+	    state_update(overprint);
 	    state_update(stroke_adjust);
 	  }
 	if ( cmd_check_clip_path(cdev, pcpath) )
@@ -506,12 +661,9 @@ clist_stroke_path(gx_device *dev, const gs_imager_state *pis, gx_path *ppath,
 	  }
 	else
 	  { if ( lop != pcls->lop )
-	      { byte *dp;
-		set_cmd_put_op(dp, cdev, pcls,
-			       cmd_opv_set_lop, 1 + cmd_size_w(lop));
-		++dp;
-		cmd_put_w(lop, dp);
-		pcls->lop = lop;
+	      { code = cmd_set_lop(cdev, pcls, lop);
+	        if ( code < 0 )
+		  return code;
 	      }
 	    cmd_enable_lop(cdev, pcls);
 	  }
@@ -526,12 +678,18 @@ clist_stroke_path(gx_device *dev, const gs_imager_state *pis, gx_path *ppath,
 	  /* outside the clipping region, because that would throw off */
 	  /* the pattern. */
 	  if ( pattern_size == 0 )
-	    ymin = int2fixed(y) - adjust_y,
-	    ymax = int2fixed(y + height) + adjust_y;
+	    ymin = int2fixed(max(y - adjust_y, y0)),
+	    ymax = int2fixed(min(y + height + adjust_y, y1));
 	  else
 	    ymin = min_fixed,
 	    ymax = max_fixed;
-	  code = cmd_put_path(cdev, pcls, ppath, ymin, ymax, op, false);
+	  code = cmd_put_path(cdev, pcls, ppath, ymin, ymax,
+#ifdef FUTURE
+			      cmd_opv_stroke + code,	/* cmd_dc_type */
+#else
+			      op,
+#endif
+			      false, (segment_notes)~0);
 	  if ( code < 0 )
 	    return code;
 	}
@@ -547,6 +705,7 @@ typedef struct cmd_segment_writer_s {
   gx_device_clist_writer *cldev;
   gx_clist_state *pcls;
 	/* Updated dynamically */
+  segment_notes notes;
   byte *dp;
   int len;
   gs_fixed_point delta_first;
@@ -556,7 +715,7 @@ typedef struct cmd_segment_writer_s {
 /* Put out a path segment command. */
 private int near
 cmd_put_segment(cmd_segment_writer _ss *psw, byte op,
-  const fixed _ss *operands)
+  const fixed _ss *operands, segment_notes notes)
 {	const fixed _ss *optr = operands;
 	/* Fetch num_operands before possible command merging. */
 	int i = clist_segment_op_num_operands[op & 0xf];
@@ -565,7 +724,8 @@ cmd_put_segment(cmd_segment_writer _ss *psw, byte op,
 #ifdef DEBUG
 	if ( gs_debug_c('L') )
 	  { int j;
-	    dprintf1("[L]  %s:", cmd_segment_op_names[op & 0xf]);
+	    dprintf2("[L]  %s:%d:", cmd_segment_op_names[op & 0xf],
+		     (int)notes);
 	    for ( j = 0; j < i; ++j )
 	      dprintf1(" %g", fixed2float(operands[j]));
 	    dputs("\n");
@@ -590,11 +750,15 @@ merge:		  cmd_uncount_op(*psw->dp, psw->len);
 		  q += psw->len - 1;
 		  break;
 		case cmd_opv_rmlineto:
+		  if ( notes != psw->notes )
+		    break;
 		  op = cmd_opv_rm2lineto;
 		  goto merge;
 		case cmd_opv_rm2lineto:
+		  if ( notes != psw->notes )
+		    break;
 		  if ( operands[0] == -psw->delta_first.x &&
-		      operands[1] == -psw->delta_first.y
+		       operands[1] == -psw->delta_first.y
 		      )
 		    { cmd_uncount_op(cmd_opv_rm2lineto, psw->len);
 		      *psw->dp = cmd_count_op(cmd_opv_rm3lineto, psw->len);
@@ -617,7 +781,7 @@ merge:		  cmd_uncount_op(*psw->dp, psw->len);
 		  }
 		else if ( is_bits(d, 19) && i > 0 && is_bits(d2 = optr[1], 19) )
 		  { cmd_count_add1(cmd_diffs[0]);
-		    q[1] = (d >> 13) & 0x3f;
+		    q[1] = (byte)((d >> 13) & 0x3f);
 		    q[2] = (byte)(d >> 5);
 		    q[3] = (byte)((d << 3) + ((d2 >> 16) & 7));
 		    q[4] = (byte)(d2 >> 8);
@@ -628,12 +792,12 @@ merge:		  cmd_uncount_op(*psw->dp, psw->len);
 		  }
 		else if ( is_bits(d, 22) )
 		  { cmd_count_add1(cmd_diffs[1]);
-		    q[1] = ((d >> 16) & 0x3f) + 0x40;
+		    q[1] = (byte)(((d >> 16) & 0x3f) + 0x40);
 		    q += 3;
 		  }
 		else if ( is_bits(d, 30) )
 		  { cmd_count_add1(cmd_diffs[2]);
-		    q[1] = ((d >> 24) & 0x3f) + 0x80;
+		    q[1] = (byte)(((d >> 24) & 0x3f) + 0x80);
 		    q[2] = (byte)(d >> 16);
 		    q += 4;
 		  }
@@ -648,6 +812,13 @@ merge:		  cmd_uncount_op(*psw->dp, psw->len);
 		q[-1] = (byte)(d >> 8);
 		*q = (byte)d;
 	      }
+	if ( notes != psw->notes )
+	  { byte *dp;
+
+	    set_cmd_put_op(dp, psw->cldev, psw->pcls, cmd_opv_set_misc2, 2);
+	    dp[1] = cmd_set_misc2_notes + notes;
+	    psw->notes = notes;
+	  }
 	{ int len = q + 2 - psw->cmd;
 	  byte *dp;
 
@@ -660,9 +831,9 @@ merge:		  cmd_uncount_op(*psw->dp, psw->len);
 }
 /* Put out a line segment command. */
 #define cmd_put_rmoveto(psw, operands)\
-  cmd_put_segment(psw, cmd_opv_rmoveto, operands)
-#define cmd_put_rlineto(psw, operands)\
-  cmd_put_segment(psw, cmd_opv_rlineto, operands)
+  cmd_put_segment(psw, cmd_opv_rmoveto, operands, sn_none)
+#define cmd_put_rlineto(psw, operands, notes)\
+  cmd_put_segment(psw, cmd_opv_rlineto, operands, notes)
 
 /*
  * Write a path.  We go to a lot of trouble to omit segments that are
@@ -671,7 +842,7 @@ merge:		  cmd_uncount_op(*psw->dp, psw->len);
 private int
 cmd_put_path(gx_device_clist_writer *cldev, gx_clist_state *pcls,
   const gx_path *ppath, fixed ymin, fixed ymax, byte path_op,
-  bool implicit_close)
+  bool implicit_close, segment_notes keep_notes)
 {	gs_path_enum cenum;
 	cmd_segment_writer writer;
 	static byte initial_op = { cmd_opv_end_run };
@@ -706,6 +877,9 @@ cmd_put_path(gx_device_clist_writer *cldev, gx_clist_state *pcls,
 	int side;
 		/* The last point with side != 0: */
 	gs_fixed_point out;
+		/* If the last out-going segment was a lineto, */
+		/* its notes: */
+	segment_notes out_notes;
 
 	/*
 	 * The following track the emitted segments:
@@ -720,14 +894,17 @@ cmd_put_path(gx_device_clist_writer *cldev, gx_clist_state *pcls,
 	int open = 0;		/* -1 if last was moveto, 1 if line/curveto, */
 				/* 0 if newpath/closepath */
 
+
 	if_debug4('p', "[p]initial (%g,%g), clip [%g..%g)\n",
 		  fixed2float(px), fixed2float(py),
 		  fixed2float(ymin), fixed2float(ymax));
 	gx_path_enum_init(&cenum, ppath);
 	writer.cldev = cldev;
 	writer.pcls = pcls;
-	writer.dp = &initial_op;
-#define first_point() ((writer).dp == &initial_op)
+	writer.notes = sn_none;
+#define set_first_point() (writer.dp = &initial_op)
+#define first_point() (writer.dp == &initial_op)
+	set_first_point();
 	for ( ; ; )
 	  { fixed vs[6];
 #define A vs[0]
@@ -779,11 +956,15 @@ cmd_put_path(gx_device_clist_writer *cldev, gx_clist_state *pcls,
 		break;
 	      case gs_pe_lineto:
 		{ int next_side = which_side(B);
+		  segment_notes notes =
+		    gx_path_enum_notes(&cenum) & keep_notes;
+
 		  if ( next_side == side && side != 0 )
 		    { /* Skip a line completely outside the clip region. */
 		      if ( open < 0 )
 			start_skip = true;
 		      out.x = A, out.y = B;
+		      out_notes = notes;
 		      if_debug3('p', "[p]skip lineto (%g,%g) side %d\n",
 				fixed2float(out.x), fixed2float(out.y),
 				side);
@@ -797,7 +978,7 @@ cmd_put_path(gx_device_clist_writer *cldev, gx_clist_state *pcls,
 			  code = cmd_put_rmoveto(&writer, &C);
 			}
 		      else
-			code = cmd_put_rlineto(&writer, &C);
+			code = cmd_put_rlineto(&writer, &C, out_notes);
 		      if ( code < 0 )
 			return code;
 		      px = out.x, py = out.y;
@@ -809,11 +990,11 @@ cmd_put_path(gx_device_clist_writer *cldev, gx_clist_state *pcls,
 		    { /* Note a vertex going outside the clip region. */
 		      out.x = A, out.y = B;
 		    }
+		  C = A - px, D = B - py;
+		  px = A, py = B;
+		  open = 1;
+		  code = cmd_put_rlineto(&writer, &C, notes);
 		}
-		C = A - px, D = B - py;
-		px = A, py = B;
-		open = 1;
-		code = cmd_put_rlineto(&writer, &C);
 		if_debug3('p', "[p]lineto (%g,%g) side %d\n",
 			  fixed2float(px), fixed2float(py), side);
 		break;
@@ -839,7 +1020,7 @@ close:		if ( side != start_side )
 		  { /* If we skipped any segments, put out a moveto/lineto. */
 		    if ( side && (px != out.x || py != out.y || first_point()) )
 		      { C = out.x - px, D = out.y - py;
-			code = cmd_put_rlineto(&writer, &C);
+			code = cmd_put_rlineto(&writer, &C, out_notes);
 			if ( code < 0 )
 			  return code;
 			px = out.x, py = out.y;
@@ -849,7 +1030,7 @@ close:		if ( side != start_side )
 		    if ( open > 0 && start_skip )
 		      { /* Draw the closing line back to the start. */
 			C = start.x - px, D = start.y - py;
-			code = cmd_put_rlineto(&writer, &C);
+			code = cmd_put_rlineto(&writer, &C, sn_none);
 			if ( code < 0 )
 			  return code;
 			px = start.x, py = start.y;
@@ -865,14 +1046,22 @@ close:		if ( side != start_side )
 		 */
 		if ( implicit_close || open <= 0 )
 		  { open = 0;
+		    /*
+		     * Force writing an explicit moveto if the next subpath
+		     * starts with a moveto to the same point where this one
+		     * ends.
+		     */
+		    set_first_point();
 		    continue;
 		  }
 		open = 0;
 		px = first.x, py = first.y;
-		code = cmd_put_segment(&writer, cmd_opv_closepath, &A);
+		code = cmd_put_segment(&writer, cmd_opv_closepath, &A, sn_none);
 		if_debug0('p', "[p]close\n");
 		break;
 	      case gs_pe_curveto:
+		{ segment_notes notes =
+		    gx_path_enum_notes(&cenum) & keep_notes;
 		{ fixed bpy, bqy;
 		  int all_side, out_side;
 
@@ -888,6 +1077,7 @@ close:		if ( side != start_side )
 			  if ( open < 0 )
 			    start_skip = true;
 			  out.x = E, out.y = F;
+			  out_notes = notes;
 			  if_debug3('p', "[p]skip curveto (%g,%g) side %d\n",
 				    fixed2float(out.x), fixed2float(out.y),
 				    side);
@@ -906,7 +1096,7 @@ close:		if ( side != start_side )
 			  code = cmd_put_rmoveto(&writer, diff);
 			}
 		      else
-			code = cmd_put_rlineto(&writer, diff);
+			code = cmd_put_rlineto(&writer, diff, out_notes);
 		      if ( code < 0 )
 			return code;
 		      px = out.x, py = out.y;
@@ -931,9 +1121,23 @@ close:		if ( side != start_side )
 		  C -= A, D -= B;
 		  A -= px, B -= py;
 		  if ( B == 0 && E == 0 )
-		    B = A, E = F, optr++, op = cmd_opv_hvcurveto;
+		    { B = A, E = F, optr++, op = cmd_opv_hvcurveto;
+		      if ( (B ^ D) >= 0 )
+			{ if ( C == D && E == B )
+			    op = cmd_opv_hqcurveto;
+			}
+		      else if ( C == -D && E == -B )
+			C = D, op = cmd_opv_hqcurveto;
+		    }
 		  else if ( A == 0 && F == 0 )
-		    optr++, op = cmd_opv_vhcurveto;
+		    { optr++, op = cmd_opv_vhcurveto;
+		      if ( (B ^ C) >= 0 )
+			{ if ( D == C && E == B )
+			    op = cmd_opv_vqcurveto;
+			}
+		      else if ( D == -C && E == -B )
+			op = cmd_opv_vqcurveto;
+		    }
 		  else if ( A == 0 && B == 0 )
 		    optr += 2, op = cmd_opv_nrcurveto;
 		  else if ( E == 0 && F == 0 )
@@ -942,7 +1146,8 @@ close:		if ( side != start_side )
 		    op = cmd_opv_rrcurveto;
 		  px = nx, py = ny;
 		  open = 1;
-		  code = cmd_put_segment(&writer, op, optr);
+		  code = cmd_put_segment(&writer, op, optr, notes);
+		}
 		} break;
 	      default:
 		return_error(gs_error_rangecheck);
