@@ -1,19 +1,24 @@
 /* Copyright (C) 1991, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of Aladdin Ghostscript.
+  This file is part of GNU Ghostscript.
   
-  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
-  or distributor accepts any responsibility for the consequences of using it,
-  or for whether it serves any particular purpose or works at all, unless he
-  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
-  License (the "License") for full details.
+  GNU Ghostscript is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
+  anyone for the consequences of using it or for whether it serves any
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
-  Every copy of Aladdin Ghostscript must include a copy of the License,
-  normally in a plain ASCII text file named PUBLIC.  The License grants you
-  the right to copy, modify and redistribute Aladdin Ghostscript, but only
-  under certain conditions described in the License.  Among other things, the
-  License requires that the copyright notice and this notice be preserved on
-  all copies.
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gxclread.c */
@@ -1153,8 +1158,22 @@ stp:			state_tile.size.x = state_slot->width;
 			state_tile.shift = state_slot->shift;
 set_phase:		tile_phase.x =
 			  (state.tile_phase.x + x0) % state_tile.size.x;
-			tile_phase.y =
-			  (state.tile_phase.y + y0) % state_tile.size.y;
+			/*
+			 * The true tile height for shifted tiles is not
+			 * size.y: see gxbitmap.h for the computation.
+			 */
+			{ int full_height;
+
+			  if ( state_tile.shift == 0 )
+			    full_height = state_tile.size.y;
+			  else
+			    full_height = state_tile.rep_height *
+			      (state_tile.rep_width /
+			       igcd(state_tile.rep_shift,
+				    state_tile.rep_width));
+			  tile_phase.y =
+			    (state.tile_phase.y + y0) % full_height;
+			}
 			gx_imager_setscreenphase(&imager_state,
 						 -(state.tile_phase.x + x0),
 						 -(state.tile_phase.y + y0),
@@ -1531,7 +1550,7 @@ set_phase:		tile_phase.x =
 				    }
 				  else
 				    { uint bytes_per_plane, nbytes;
-				      byte *data;
+				      const byte *data;
 				      byte *data_on_heap = 0;
 				      const byte *planes[64];
 				      /****** DOESN'T HANDLE #PLANES YET *****/
