@@ -1,4 +1,4 @@
-/* Copyright (C) 1992, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1992, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -106,20 +106,52 @@ private cmap_proc_rgb(cmap_rgb_to_cmyk);
 private cmap_proc_cmyk(cmap_cmyk_direct);
 private cmap_proc_cmyk(cmap_cmyk_to_gray);
 private cmap_proc_cmyk(cmap_cmyk_to_rgb);
+#ifdef DPNEXT
+private cmap_proc_rgb_alpha(cmap_rgb_alpha_halftoned);
+private cmap_proc_rgb_alpha(cmap_rgb_alpha_direct);
+/* Procedure names are only guaranteed unique to 23 characters.... */
+private cmap_proc_rgb_alpha(cmap_rgb_alpha2gray_halftoned);
+private cmap_proc_rgb_alpha(cmap_rgb_alpha2gray_direct);
+private cmap_proc_rgb_alpha(cmap_rgb_alpha_to_cmyk);
+#endif
 
 private const gx_color_map_procs
   cmap_gray_few =
-   { cmap_gray_halftoned, cmap_rgb_to_gray_halftoned, cmap_cmyk_to_gray },
+   { cmap_gray_halftoned, cmap_rgb_to_gray_halftoned, cmap_cmyk_to_gray
+#ifdef DPNEXT
+     , cmap_rgb_alpha2gray_halftoned
+#endif
+   },
   cmap_gray_many =
-   { cmap_gray_direct, cmap_rgb_to_gray_direct, cmap_cmyk_to_gray },
+   { cmap_gray_direct, cmap_rgb_to_gray_direct, cmap_cmyk_to_gray
+#ifdef DPNEXT
+     , cmap_rgb_alpha2gray_direct
+#endif
+   },
   cmap_rgb_few =
-   { cmap_gray_to_rgb_halftoned, cmap_rgb_halftoned, cmap_cmyk_to_rgb },
+   { cmap_gray_to_rgb_halftoned, cmap_rgb_halftoned, cmap_cmyk_to_rgb
+#ifdef DPNEXT
+     , cmap_rgb_alpha_halftoned
+#endif
+   },
   cmap_rgb_many =
-   { cmap_gray_to_rgb_direct, cmap_rgb_direct, cmap_cmyk_to_rgb },
+   { cmap_gray_to_rgb_direct, cmap_rgb_direct, cmap_cmyk_to_rgb
+#ifdef DPNEXT
+     , cmap_rgb_alpha_direct
+#endif
+   },
   cmap_cmyk_few =
-   { cmap_gray_to_cmyk_halftoned, cmap_rgb_to_cmyk, cmap_cmyk_halftoned },
+   { cmap_gray_to_cmyk_halftoned, cmap_rgb_to_cmyk, cmap_cmyk_halftoned
+#ifdef DPNEXT
+     , cmap_rgb_alpha_to_cmyk
+#endif
+   },
   cmap_cmyk_many =
-   { cmap_gray_to_cmyk_direct, cmap_rgb_to_cmyk, cmap_cmyk_direct };
+   { cmap_gray_to_cmyk_direct, cmap_rgb_to_cmyk, cmap_cmyk_direct
+#ifdef DPNEXT
+     , cmap_rgb_alpha_to_cmyk
+#endif
+   };
 
 const gx_color_map_procs *cmap_procs_default = &cmap_gray_many;
 
@@ -141,33 +173,37 @@ gx_device_cmap_procs(const gx_device *dev)
 
 /* Set the color mapping procedures in the graphics state. */
 void
-gx_set_cmap_procs(gs_state *pgs)
-{	pgs->cmap_procs = gx_device_cmap_procs(pgs->device);
+gx_set_cmap_procs(gs_imager_state *pis, const gx_device *dev)
+{	pis->cmap_procs = gx_device_cmap_procs(dev);
 }
 
 /* Remap the color in the graphics state. */
 int
 gx_remap_color(gs_state *pgs)
 {	const gs_color_space *pcs = pgs->color_space;
-	return (*pcs->type->remap_color)(pgs->ccolor, pcs, pgs->dev_color, pgs);
+	/* The current color in the graphics state is always used for */
+	/* the texture, never for the source. */
+	return (*pcs->type->remap_color)(pgs->ccolor, pcs, pgs->dev_color,
+					 (gs_imager_state *)pgs, pgs->device,
+					 gs_color_select_texture);
 }
 
 /* Indicate that a color space has no underlying concrete space. */
 const gs_color_space *
-gx_no_concrete_space(const gs_color_space *pcs, const gs_state *pgs)
+gx_no_concrete_space(const gs_color_space *pcs, const gs_imager_state *pis)
 {	return NULL;
 }
 
 /* Indicate that a color space is concrete. */
 const gs_color_space *
-gx_same_concrete_space(const gs_color_space *pcs, const gs_state *pgs)
+gx_same_concrete_space(const gs_color_space *pcs, const gs_imager_state *pis)
 {	return pcs;
 }
 
 /* Indicate that a color cannot be concretized. */
 int
 gx_no_concretize_color(const gs_client_color *pcc, const gs_color_space *pcs,
-  frac *pconc, const gs_state *pgs)
+  frac *pconc, const gs_imager_state *pis)
 {	return_error(gs_error_rangecheck);
 }
 
@@ -175,14 +211,17 @@ gx_no_concretize_color(const gs_client_color *pcc, const gs_color_space *pcs,
 /* remapping the concrete color. */
 int
 gx_default_remap_color(const gs_client_color *pcc, const gs_color_space *pcs,
-  gx_device_color *pdc, const gs_state *pgs)
+  gx_device_color *pdc, const gs_imager_state *pis, gx_device *dev,
+  gs_color_select_t select)
 {	frac conc[4];
 	const gs_color_space *pconcs;
-	int code = (*pcs->type->concretize_color)(pcc, pcs, conc, pgs);
+	int code = (*pcs->type->concretize_color)(pcc, pcs, conc, pis);
+
 	if ( code < 0 )
 	  return code;
-	pconcs = cs_concrete_space(pcs, pgs);
-	return (*pconcs->type->remap_concrete_color)(conc, pdc, pgs);
+	pconcs = cs_concrete_space(pcs, pis);
+	return (*pconcs->type->remap_concrete_color)(conc, pdc, pis, dev,
+						     select);
 }
 
 /* Color remappers for the standard color spaces. */
@@ -196,32 +235,34 @@ gx_default_remap_color(const gs_client_color *pcc, const gs_color_space *pcs,
 /* DeviceGray */
 int
 gx_concretize_DeviceGray(const gs_client_color *pc, const gs_color_space *pcs,
-  frac *pconc, const gs_state *pgs)
+  frac *pconc, const gs_imager_state *pis)
 {	float ftemp;
 	pconc[0] = unit_frac(pc->paint.values[0], ftemp);
 	return 0;
 }
 int
 gx_remap_concrete_DGray(const frac *pconc,
-  gx_device_color *pdc, const gs_state *pgs)
-{	(*pgs->cmap_procs->map_gray)
-		(pconc[0], pdc, pgs);
+  gx_device_color *pdc, const gs_imager_state *pis, gx_device *dev,
+  gs_color_select_t select)
+{	(*pis->cmap_procs->map_gray)
+	  (pconc[0], pdc, pis, dev, select);
 	return 0;
 }
 int
 gx_remap_DeviceGray(const gs_client_color *pc, const gs_color_space *pcs,
-  gx_device_color *pdc, const gs_state *pgs)
+  gx_device_color *pdc, const gs_imager_state *pis, gx_device *dev,
+  gs_color_select_t select)
 {	float ftemp;
-	(*pgs->cmap_procs->map_gray)
-		(unit_frac(pc->paint.values[0], ftemp),
-		 pdc, pgs);
+	(*pis->cmap_procs->map_gray)
+	  (unit_frac(pc->paint.values[0], ftemp),
+	   pdc, pis, dev, select);
 	return 0;
 }
 
 /* DeviceRGB */
 int
 gx_concretize_DeviceRGB(const gs_client_color *pc, const gs_color_space *pcs,
-  frac *pconc, const gs_state *pgs)
+  frac *pconc, const gs_imager_state *pis)
 {	float ftemp;
 	pconc[0] = unit_frac(pc->paint.values[0], ftemp);
 	pconc[1] = unit_frac(pc->paint.values[1], ftemp);
@@ -230,25 +271,28 @@ gx_concretize_DeviceRGB(const gs_client_color *pc, const gs_color_space *pcs,
 }
 int
 gx_remap_concrete_DRGB(const frac *pconc,
-  gx_device_color *pdc, const gs_state *pgs)
-{	gx_remap_concrete_rgb(pconc[0], pconc[1], pconc[2], pdc, pgs);
+  gx_device_color *pdc, const gs_imager_state *pis, gx_device *dev,
+  gs_color_select_t select)
+{	gx_remap_concrete_rgb(pconc[0], pconc[1], pconc[2], pdc, pis, dev,
+			      select);
 	return 0;
 }
 int
 gx_remap_DeviceRGB(const gs_client_color *pc, const gs_color_space *pcs,
-  gx_device_color *pdc, const gs_state *pgs)
+  gx_device_color *pdc, const gs_imager_state *pis, gx_device *dev,
+  gs_color_select_t select)
 {	float ft0, ft1, ft2;
 	gx_remap_concrete_rgb(unit_frac(pc->paint.values[0], ft0),
 			      unit_frac(pc->paint.values[1], ft1),
 			      unit_frac(pc->paint.values[2], ft2),
-			      pdc, pgs);
+			      pdc, pis, dev, select);
 	return 0;
 }
 
 /* DeviceCMYK */
 int
 gx_concretize_DeviceCMYK(const gs_client_color *pc, const gs_color_space *pcs,
-  frac *pconc, const gs_state *pgs)
+  frac *pconc, const gs_imager_state *pis)
 {	float ftemp;
 	pconc[0] = unit_frac(pc->paint.values[0], ftemp);
 	pconc[1] = unit_frac(pc->paint.values[1], ftemp);
@@ -258,42 +302,46 @@ gx_concretize_DeviceCMYK(const gs_client_color *pc, const gs_color_space *pcs,
 }
 int
 gx_remap_concrete_DCMYK(const frac *pconc,
-  gx_device_color *pdc, const gs_state *pgs)
-{	gx_remap_concrete_cmyk(pconc[0], pconc[1], pconc[2], pconc[3], pdc, pgs);
+  gx_device_color *pdc, const gs_imager_state *pis, gx_device *dev,
+  gs_color_select_t select)
+{	gx_remap_concrete_cmyk(pconc[0], pconc[1], pconc[2], pconc[3], pdc,
+			       pis, dev, select);
 	return 0;
 }
 int
 gx_remap_DeviceCMYK(const gs_client_color *pc, const gs_color_space *pcs,
-  gx_device_color *pdc, const gs_state *pgs)
+  gx_device_color *pdc, const gs_imager_state *pis, gx_device *dev,
+  gs_color_select_t select)
 {	float ft0, ft1, ft2, ft3;
 	gx_remap_concrete_cmyk(unit_frac(pc->paint.values[0], ft0),
 			       unit_frac(pc->paint.values[1], ft1),
 			       unit_frac(pc->paint.values[2], ft2),
 			       unit_frac(pc->paint.values[3], ft3),
-			       pdc, pgs);
+			       pdc, pis, dev, select);
 	return 0;
 }
 
 /* ------ Render Gray color. ------ */
 
 private void
-cmap_gray_halftoned(frac gray, gx_device_color *pdc, const gs_state *pgs)
-{	if ( gx_render_gray(gx_map_color_frac(pgs, gray, effective_transfer.colored.gray), pdc, pgs) == 1 )
-	  gx_color_load(pdc, pgs);
+cmap_gray_halftoned(frac gray, gx_device_color *pdc,
+  const gs_imager_state *pis, gx_device *dev, gs_color_select_t select)
+{	if ( gx_render_gray(gx_map_color_frac(pis, gray, effective_transfer.colored.gray), pdc, pis, dev, select) == 1 )
+	  gx_color_load_select(pdc, pis, dev, select);
 }
 
 private void
-cmap_gray_direct(frac gray, gx_device_color *pdc, const gs_state *pgs)
-{	gx_device *dev = gs_currentdevice_inline(pgs);
-	frac mgray = gx_map_color_frac(pgs, gray, effective_transfer.colored.gray);
+cmap_gray_direct(frac gray, gx_device_color *pdc, const gs_imager_state *pis,
+  gx_device *dev, gs_color_select_t select)
+{	frac mgray = gx_map_color_frac(pis, gray, effective_transfer.colored.gray);
 	gx_color_value cv_gray = frac2cv(mgray);
 	gx_color_index color =
-	  (pgs->alpha == gx_max_color_value ?
+	  (pis->alpha == gx_max_color_value ?
 	   gx_map_rgb_color(dev, cv_gray, cv_gray, cv_gray) :
-	   gx_map_rgb_alpha_color(dev, cv_gray, cv_gray, cv_gray, pgs->alpha));
+	   gx_map_rgb_alpha_color(dev, cv_gray, cv_gray, cv_gray, pis->alpha));
 	if ( color == gx_no_color_index )
-	{	if ( gx_render_gray(mgray, pdc, pgs) == 1 )
-		  gx_color_load(pdc, pgs);
+	{	if ( gx_render_gray(mgray, pdc, pis, dev, select) == 1 )
+		  gx_color_load_select(pdc, pis, dev, select);
 		return;
 	}
 	color_set_pure(pdc, color);
@@ -301,39 +349,39 @@ cmap_gray_direct(frac gray, gx_device_color *pdc, const gs_state *pgs)
 
 private void
 cmap_gray_to_rgb_halftoned(frac gray, gx_device_color *pdc,
-  const gs_state *pgs)
-{	cmap_rgb_halftoned(gray, gray, gray, pdc, pgs);
+  const gs_imager_state *pis, gx_device *dev, gs_color_select_t select)
+{	cmap_rgb_halftoned(gray, gray, gray, pdc, pis, dev, select);
 }
 
 private void
-cmap_gray_to_rgb_direct(frac gray, gx_device_color *pdc, const gs_state *pgs)
-{	cmap_rgb_direct(gray, gray, gray, pdc, pgs);
+cmap_gray_to_rgb_direct(frac gray, gx_device_color *pdc,
+  const gs_imager_state *pis, gx_device *dev, gs_color_select_t select)
+{	cmap_rgb_direct(gray, gray, gray, pdc, pis, dev, select);
 }
 
 private void
 cmap_gray_to_cmyk_halftoned(frac gray, gx_device_color *pdc,
-  const gs_state *pgs)
+  const gs_imager_state *pis, gx_device *dev, gs_color_select_t select)
 {	/*
 	 * Per the last paragraph of section 6.3 (p. 309) of the
 	 * PostScript Language Reference Manual, 2nd Edition,
 	 * we must bypass the C, M, and Y transfer functions in this case.
 	 */
-	frac mgray = gx_map_color_frac(pgs, gray, effective_transfer.colored.gray);
-	if ( gx_render_gray(mgray, pdc, pgs) == 1 )
-	  gx_color_load(pdc, pgs);
+	frac mgray = gx_map_color_frac(pis, gray, effective_transfer.colored.gray);
+	if ( gx_render_gray(mgray, pdc, pis, dev, select) == 1 )
+	  gx_color_load_select(pdc, pis, dev, select);
 }
 
 private void
 cmap_gray_to_cmyk_direct(frac gray, gx_device_color *pdc,
-  const gs_state *pgs)
+  const gs_imager_state *pis, gx_device *dev, gs_color_select_t select)
 {	/*
 	 * Per the last paragraph of section 6.3 (p. 309) of the
 	 * PostScript Language Reference Manual, 2nd Edition,
 	 * we must bypass the C, M, and Y transfer functions in this case.
 	 */
-	frac mgray = gx_map_color_frac(pgs, gray, effective_transfer.colored.gray);
+	frac mgray = gx_map_color_frac(pis, gray, effective_transfer.colored.gray);
 	frac mblack = frac_1 - mgray;
-	gx_device *dev = gs_currentdevice_inline(pgs);
 	gx_color_index color =
 	  gx_map_cmyk_color(dev,
 			    frac2cv(frac_0), frac2cv(frac_0),
@@ -342,8 +390,8 @@ cmap_gray_to_cmyk_direct(frac gray, gx_device_color *pdc,
 	{	color_set_pure(pdc, color);
 		return;
 	}
-	if ( gx_render_gray(mgray, pdc, pgs) == 1 )
-	  gx_color_load(pdc, pgs);
+	if ( gx_render_gray(mgray, pdc, pis, dev, select) == 1 )
+	  gx_color_load_select(pdc, pis, dev, select);
 }
 
 /* ------ Render RGB color. ------ */
@@ -364,74 +412,73 @@ cmap_gray_to_cmyk_direct(frac gray, gx_device_color *pdc,
 
 private void
 cmap_rgb_halftoned(frac r, frac g, frac b, gx_device_color *pdc,
-  const gs_state *pgs)
-{	frac mred = gx_map_color_frac(pgs, r, effective_transfer.colored.red);
-	frac mgreen = gx_map_color_frac(pgs, g, effective_transfer.colored.green);
-	frac mblue = gx_map_color_frac(pgs, b, effective_transfer.colored.blue);
+  const gs_imager_state *pis, gx_device *dev, gs_color_select_t select)
+{	frac mred = gx_map_color_frac(pis, r, effective_transfer.colored.red);
+	frac mgreen = gx_map_color_frac(pis, g, effective_transfer.colored.green);
+	frac mblue = gx_map_color_frac(pis, b, effective_transfer.colored.blue);
 	if ( (mred == mgreen && mred == mblue ?	/* gray shade */
-	      gx_render_gray(mred, pdc, pgs) :
-	      gx_render_rgb(mred, mgreen, mblue, pdc, pgs)) == 1 )
-	  gx_color_load(pdc, pgs);
+	      gx_render_gray(mred, pdc, pis, dev, select) :
+	      gx_render_rgb(mred, mgreen, mblue, pdc, pis, dev, select)) == 1 )
+	  gx_color_load_select(pdc, pis, dev, select);
 }
 
 private void
 cmap_rgb_direct(frac r, frac g, frac b, gx_device_color *pdc,
-  const gs_state *pgs)
-{	gx_device *dev = gs_currentdevice_inline(pgs);
-	frac mred = gx_map_color_frac(pgs, r, effective_transfer.colored.red);
-	frac mgreen = gx_map_color_frac(pgs, g, effective_transfer.colored.green);
-	frac mblue = gx_map_color_frac(pgs, b, effective_transfer.colored.blue);
+  const gs_imager_state *pis, gx_device *dev, gs_color_select_t select)
+{	frac mred = gx_map_color_frac(pis, r, effective_transfer.colored.red);
+	frac mgreen = gx_map_color_frac(pis, g, effective_transfer.colored.green);
+	frac mblue = gx_map_color_frac(pis, b, effective_transfer.colored.blue);
 	gx_color_index color =
-	  (pgs->alpha == gx_max_color_value ?
+	  (pis->alpha == gx_max_color_value ?
 	   gx_map_rgb_color(dev, frac2cv(mred), frac2cv(mgreen),
 			    frac2cv(mblue)) :
 	   gx_map_rgb_alpha_color(dev, frac2cv(mred), frac2cv(mgreen),
-				  frac2cv(mblue), pgs->alpha));
+				  frac2cv(mblue), pis->alpha));
 	if ( color != gx_no_color_index )
 	  {	color_set_pure(pdc, color);
 		return;
 	  }
 	if ( (mred == mgreen && mred == mblue ?	/* gray shade */
-	      gx_render_gray(mred, pdc, pgs) :
-	      gx_render_rgb(mred, mgreen, mblue, pdc, pgs)) == 1 )
-	  gx_color_load(pdc, pgs);
+	      gx_render_gray(mred, pdc, pis, dev, select) :
+	      gx_render_rgb(mred, mgreen, mblue, pdc, pis, dev, select)) == 1 )
+	  gx_color_load_select(pdc, pis, dev, select);
 }
 
 private void
 cmap_rgb_to_gray_halftoned(frac r, frac g, frac b, gx_device_color *pdc,
-  const gs_state *pgs)
-{	cmap_gray_halftoned(color_rgb_to_gray(r, g, b, pgs), pdc, pgs);
+  const gs_imager_state *pis, gx_device *dev, gs_color_select_t select)
+{	cmap_gray_halftoned(color_rgb_to_gray(r, g, b, pis), pdc, pis, dev, select);
 }
 
 private void
 cmap_rgb_to_gray_direct(frac r, frac g, frac b, gx_device_color *pdc,
-  const gs_state *pgs)
-{	cmap_gray_direct(color_rgb_to_gray(r, g, b, pgs), pdc, pgs);
+  const gs_imager_state *pis, gx_device *dev, gs_color_select_t select)
+{	cmap_gray_direct(color_rgb_to_gray(r, g, b, pis), pdc, pis, dev, select);
 }
 
 private void
 cmap_rgb_to_cmyk(frac r, frac g, frac b, gx_device_color *pdc,
-  const gs_state *pgs)
+  const gs_imager_state *pis, gx_device *dev, gs_color_select_t select)
 {	frac cmyk[4];
-	color_rgb_to_cmyk(r, g, b, pgs, cmyk);
-	(*pgs->cmap_procs->map_cmyk)(cmyk[0], cmyk[1], cmyk[2], cmyk[3], pdc, pgs);
+	color_rgb_to_cmyk(r, g, b, pis, cmyk);
+	(*pis->cmap_procs->map_cmyk)(cmyk[0], cmyk[1], cmyk[2], cmyk[3], pdc, pis, dev, select);
 }
 
 /* ------ Render CMYK color. ------ */
 
 private void
-cmap_cmyk_to_gray(frac c, frac m, frac y, frac k, gx_device_color *pdc, const gs_state *pgs)
-{	(*pgs->cmap_procs->map_gray)(color_cmyk_to_gray(c, m, y, k, pgs), pdc, pgs);
+cmap_cmyk_to_gray(frac c, frac m, frac y, frac k, gx_device_color *pdc,
+  const gs_imager_state *pis, gx_device *dev, gs_color_select_t select)
+{	(*pis->cmap_procs->map_gray)(color_cmyk_to_gray(c, m, y, k, pis), pdc, pis, dev, select);
 }
 
 private void
 cmap_cmyk_direct(frac c, frac m, frac y, frac k, gx_device_color *pdc,
-  const gs_state *pgs)
-{	gx_device *dev = gs_currentdevice_inline(pgs);
-	frac mcyan = frac_1 - gx_map_color_frac(pgs, frac_1 - c, effective_transfer.colored.red);
-	frac mmagenta = frac_1 - gx_map_color_frac(pgs, frac_1 - m, effective_transfer.colored.green);
-	frac myellow = frac_1 - gx_map_color_frac(pgs, frac_1 - y, effective_transfer.colored.blue);
-	frac mblack = frac_1 - gx_map_color_frac(pgs, frac_1 - k, effective_transfer.colored.gray);
+  const gs_imager_state *pis, gx_device *dev, gs_color_select_t select)
+{	frac mcyan = frac_1 - gx_map_color_frac(pis, frac_1 - c, effective_transfer.colored.red);
+	frac mmagenta = frac_1 - gx_map_color_frac(pis, frac_1 - m, effective_transfer.colored.green);
+	frac myellow = frac_1 - gx_map_color_frac(pis, frac_1 - y, effective_transfer.colored.blue);
+	frac mblack = frac_1 - gx_map_color_frac(pis, frac_1 - k, effective_transfer.colored.gray);
 	/* We make a test for direct vs. halftoned, rather than */
 	/* duplicating most of the code of this procedure. */
 	if ( dev->color_info.max_color >= 31 )
@@ -446,19 +493,118 @@ cmap_cmyk_direct(frac c, frac m, frac y, frac k, gx_device_color *pdc,
 	}
 	/* Don't convert colors with C = M = Y to gray shades: */
 	/* on a CMYK device, this may produce quite different output. */
-	if ( gx_render_cmyk(mcyan, mmagenta, myellow, mblack, pdc, pgs) == 1 )
-	  gx_color_load(pdc, pgs);
+	if ( gx_render_cmyk(mcyan, mmagenta, myellow, mblack, pdc, pis, dev, select) == 1 )
+	  gx_color_load_select(pdc, pis, dev, select);
 }
 
 private void
-cmap_cmyk_to_rgb(frac c, frac m, frac y, frac k, gx_device_color *pdc, const gs_state *pgs)
+cmap_cmyk_to_rgb(frac c, frac m, frac y, frac k, gx_device_color *pdc,
+  const gs_imager_state *pis, gx_device *dev, gs_color_select_t select)
 {	frac rgb[3];
-	color_cmyk_to_rgb(c, m, y, k, pgs, rgb);
-	(*pgs->cmap_procs->map_rgb)(rgb[0], rgb[1], rgb[2], pdc, pgs);
+	color_cmyk_to_rgb(c, m, y, k, pis, rgb);
+	(*pis->cmap_procs->map_rgb)(rgb[0], rgb[1], rgb[2], pdc, pis, dev, select);
 }
 
+#ifdef DPNEXT
+
+/* ------ Render RGB+alpha color. ------ */
+
+private void
+cmap_rgb_alpha2gray_halftoned(frac r, frac g, frac b, frac alpha,
+  gx_device_color *pdc, const gs_imager_state *pis, gx_device *dev,
+  gs_color_select_t select)
+{	frac gray = color_rgb_to_gray(r, g, b, pis);
+
+	if ( gx_render_gray_alpha(gx_map_color_frac(pis, gray, effective_transfer.colored.gray), frac2cv(alpha), pdc, pis, dev, select) == 1 )
+	  gx_color_load_select(pdc, pis, dev, select);
+}
+
+private void
+cmap_rgb_alpha2gray_direct(frac r, frac g, frac b, frac alpha,
+  gx_device_color *pdc, const gs_imager_state *pis, gx_device *dev,
+  gs_color_select_t select)
+{	frac gray = color_rgb_to_gray(r, g, b, pis);
+	frac mgray =
+	  gx_map_color_frac(pis, gray, effective_transfer.colored.gray);
+	gx_color_value cv_gray = frac2cv(mgray);
+	gx_color_value cv_alpha = frac2cv(alpha);
+	gx_color_index color =
+	  (cv_alpha == gx_max_color_value ?
+	   gx_map_rgb_color(dev, cv_gray, cv_gray, cv_gray) :
+	   gx_map_rgb_alpha_color(dev, cv_gray, cv_gray, cv_gray, cv_alpha));
+
+	if ( color == gx_no_color_index )
+	  { if ( gx_render_gray_alpha(mgray, cv_alpha, pdc, pis, dev, select) == 1 )
+	      gx_color_load_select(pdc, pis, dev, select);
+	    return;
+	  }
+	color_set_pure(pdc, color);
+}
+
+private void
+cmap_rgb_alpha_halftoned(frac r, frac g, frac b, frac alpha,
+  gx_device_color *pdc, const gs_imager_state *pis, gx_device *dev,
+  gs_color_select_t select)
+{	frac mred = gx_map_color_frac(pis, r, effective_transfer.colored.red);
+	frac mgreen = gx_map_color_frac(pis, g, effective_transfer.colored.green);
+	frac mblue = gx_map_color_frac(pis, b, effective_transfer.colored.blue);
+	gx_color_value cv_alpha = frac2cv(alpha);
+
+	if ( (mred == mgreen && mred == mblue ?	/* gray shade */
+	      gx_render_gray_alpha(mred, cv_alpha, pdc, pis, dev, select) :
+	      gx_render_rgb_alpha(mred, mgreen, mblue, cv_alpha, pdc, pis, dev, select)) == 1 )
+	  gx_color_load_select(pdc, pis, dev, select);
+}
+
+private void
+cmap_rgb_alpha_direct(frac r, frac g, frac b, frac alpha, gx_device_color *pdc,
+  const gs_imager_state *pis, gx_device *dev, gs_color_select_t select)
+{	frac mred = gx_map_color_frac(pis, r, effective_transfer.colored.red);
+	frac mgreen = gx_map_color_frac(pis, g, effective_transfer.colored.green);
+	frac mblue = gx_map_color_frac(pis, b, effective_transfer.colored.blue);
+	gx_color_value cv_alpha = frac2cv(alpha);
+	gx_color_index color =
+	  (cv_alpha == gx_max_color_value ?
+	   gx_map_rgb_color(dev, frac2cv(mred), frac2cv(mgreen),
+			    frac2cv(mblue)) :
+	   gx_map_rgb_alpha_color(dev, frac2cv(mred), frac2cv(mgreen),
+				  frac2cv(mblue), cv_alpha));
+
+	if ( color != gx_no_color_index )
+	  {	color_set_pure(pdc, color);
+		return;
+	  }
+	if ( (mred == mgreen && mred == mblue ?	/* gray shade */
+	      gx_render_gray_alpha(mred, cv_alpha, pdc, pis, dev, select) :
+	      gx_render_rgb_alpha(mred, mgreen, mblue, cv_alpha, pdc, pis, dev, select)) == 1 )
+	  gx_color_load_select(pdc, pis, dev, select);
+}
+
+/* Currently CMYK devices can't handle alpha. */
+/* Just multiply the values towards white. */
+private void
+cmap_rgb_alpha_to_cmyk(frac r, frac g, frac b, frac alpha,
+  gx_device_color *pdc, const gs_imager_state *pis, gx_device *dev,
+  gs_color_select_t select)
+{	frac white = frac_1 - alpha;
+
+	cmap_rgb_to_cmyk((frac)((long)r * alpha / frac_1) + white,
+			 (frac)((long)g * alpha / frac_1) + white,
+			 (frac)((long)b * alpha / frac_1) + white,
+			 pdc, pis, dev, select);
+}
+
+#endif
+
 /* ------ Transfer function mapping ------ */
 
+/* Define the generic transfer function for the library layer. */
+/* This just returns what's already in the map. */
+float
+gs_mapped_transfer(floatp value, const gx_transfer_map *pmap)
+{	return gx_map_color_float(pmap, value);
+}
+
 #if FRAC_MAP_INTERPOLATE		/* NOTA BENE */
 
 /* Map a color fraction through a transfer map. */
@@ -529,11 +675,12 @@ gx_default_gray_map_rgb_color(gx_device *dev,
   gx_color_value r, gx_color_value g, gx_color_value b)
 {	/* We round the value rather than truncating it. */
 	gx_color_value gray =
-		((r * (ulong)lum_red_weight) +
-		 (g * (ulong)lum_green_weight) +
-		 (b * (ulong)lum_blue_weight) +
-		 (lum_all_weights / 2)) / lum_all_weights
-		* dev->color_info.max_gray / gx_max_color_value;
+	  (((r * (ulong)lum_red_weight) +
+	    (g * (ulong)lum_green_weight) +
+	    (b * (ulong)lum_blue_weight) +
+	    (lum_all_weights / 2)) / lum_all_weights
+	   * dev->color_info.max_gray +
+	   (gx_max_color_value / 2)) / gx_max_color_value;
 	return gray;
 }
 
@@ -617,10 +764,18 @@ gx_default_cmyk_map_cmyk_color(gx_device *dev,
 	return (color == gx_no_color_index ? color ^ 1 : color);
 }
 
-/* Default mapping from RGB-alpha to RGB. */
+/* Default mapping from RGB+alpha to RGB. */
 
 gx_color_index
 gx_default_map_rgb_alpha_color(gx_device *dev,
   gx_color_value r, gx_color_value g, gx_color_value b, gx_color_value alpha)
-{	return gx_map_rgb_color(dev, r, g, b);
+{	/* Premultiply the values towards white. */
+	gx_color_value white = gx_max_color_value - alpha;
+
+	if ( white == 0 )
+	  return gx_map_rgb_color(dev, r, g, b);
+	return gx_map_rgb_color(dev,
+	  (gx_color_value)((ulong)r * alpha / gx_max_color_value) + white,
+	  (gx_color_value)((ulong)g * alpha / gx_max_color_value) + white,
+	  (gx_color_value)((ulong)b * alpha / gx_max_color_value) + white);
 }
