@@ -1,19 +1,24 @@
 /* Copyright (C) 1991, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of Aladdin Ghostscript.
+  This file is part of GNU Ghostscript.
   
-  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
-  or distributor accepts any responsibility for the consequences of using it,
-  or for whether it serves any particular purpose or works at all, unless he
-  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
-  License (the "License") for full details.
+  GNU Ghostscript is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
+  anyone for the consequences of using it or for whether it serves any
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
-  Every copy of Aladdin Ghostscript must include a copy of the License,
-  normally in a plain ASCII text file named PUBLIC.  The License grants you
-  the right to copy, modify and redistribute Aladdin Ghostscript, but only
-  under certain conditions described in the License.  Among other things, the
-  License requires that the copyright notice and this notice be preserved on
-  all copies.
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gxcpath.c */
@@ -129,8 +134,12 @@ clip_list_validate(const gx_clip_list *clp)
 int
 gx_cpath_path(gx_clip_path *pcpath, gx_path *ppath)
 {	if ( !pcpath->segments_valid )
-	{	int code = gx_clip_list_add_to_path(&pcpath->list, &pcpath->path);
-		if ( code < 0 ) return code;
+	{	int code;
+
+		gx_path_reset(&pcpath->path);
+		code = gx_clip_list_add_to_path(&pcpath->list, &pcpath->path);
+		if ( code < 0 )
+		  return code;
 		pcpath->segments_valid = 1;
 	}
 	*ppath = pcpath->path;
@@ -167,7 +176,7 @@ gx_cpath_outer_box(const gx_clip_path *pcpath, gs_fixed_rect *pbox)
 
 /* Test if a clipping path includes a rectangle. */
 /* The rectangle need not be oriented correctly, i.e. x0 > x1 is OK. */
-int
+bool
 gx_cpath_includes_rectangle(register const gx_clip_path *pcpath,
   fixed x0, fixed y0, fixed x1, fixed y1)
 {	return
@@ -190,7 +199,7 @@ gx_cpath_set_outside(gx_clip_path *pcpath, bool outside)
 }
 
 /* Return the current outsideness of a clipping path. */
-int
+bool
 gx_cpath_is_outside(const gx_clip_path *pcpath)
 {	return pcpath->list.outside;
 }
