@@ -1,27 +1,46 @@
-/* Copyright (C) 1989, 1995 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gxdevice.h */
-/* Device description structure for Ghostscript library */
+/* Device description structure */
 
 #ifndef gxdevice_INCLUDED
 #  define gxdevice_INCLUDED
 
+#include "stdio_.h"		/* for FILE */
+#include "gsdcolor.h"
 #include "gsmatrix.h"
+#include "gsiparam.h"		/* requires gsmatrix.h */
+#include "gsropt.h"
+#include "gsstruct.h"
 #include "gsxfont.h"
 #include "gxbitmap.h"
-#include "gxdcolor.h"
+#include "gxcindex.h"
+#include "gxcvalue.h"
+#include "gxfixed.h"
 
 /* See drivers.doc for documentation of the driver interface. */
+
 #ifndef gx_device_DEFINED
 #  define gx_device_DEFINED
 typedef struct gx_device_s gx_device;
@@ -34,6 +53,32 @@ typedef struct gx_device_s gx_device;
 typedef struct gs_state_s gs_state;
 #endif
 
+/* We need abstract types for paths and fill/stroke parameters, */
+/* for the path-oriented device procedures. */
+#ifndef gx_path_DEFINED
+#  define gx_path_DEFINED
+typedef struct gx_path_s gx_path;
+#endif
+#ifndef gx_clip_path_DEFINED
+#  define gx_clip_path_DEFINED
+typedef struct gx_clip_path_s gx_clip_path;
+#endif
+#ifndef gx_fill_params_DEFINED
+#  define gx_fill_params_DEFINED
+typedef struct gx_fill_params_s gx_fill_params;
+#endif
+#ifndef gx_stroke_params_DEFINED
+#  define gx_stroke_params_DEFINED
+typedef struct gx_stroke_params_s gx_stroke_params;
+#endif
+#ifndef gs_imager_state_DEFINED
+#  define gs_imager_state_DEFINED
+typedef struct gs_imager_state_s gs_imager_state;
+#endif
+
+/* Define the type for colors passed to the higher-level procedures. */
+typedef gx_device_color gx_drawing_color;
+
 /* Define a type for telling get_alpha_bits what kind of object */
 /* is being rendered. */
 typedef enum {
@@ -41,17 +86,11 @@ typedef enum {
 	go_graphics
 } graphics_object_type;
 
-/* Define the type for gray or RGB values at the driver interface. */
-typedef unsigned short gx_color_value;
-#define arch_sizeof_gx_color_value arch_sizeof_short
-/* We might use less than the full range someday. */
-/* ...bits must lie between 8 and 16. */
-#define gx_color_value_bits (sizeof(gx_color_value) * 8)
-#define gx_max_color_value ((gx_color_value)((1L << gx_color_value_bits) - 1))
-#define gx_color_value_to_byte(cv)\
-  ((cv) >> (gx_color_value_bits - 8))
-#define gx_color_value_from_byte(cb)\
-  (((cb) << (gx_color_value_bits - 8)) + ((cb) >> (16 - gx_color_value_bits)))
+/* Define an edge of a trapezoid.  Requirement: end.y >= start.y. */
+typedef struct gs_fixed_edge_s {
+	gs_fixed_point start;
+	gs_fixed_point end;
+} gs_fixed_edge;
 
 /* Define the structure for device color capabilities. */
 typedef struct gx_device_color_info_s {
@@ -106,6 +145,22 @@ dev_page_proc_install(gx_default_install);
 dev_page_proc_begin_page(gx_default_begin_page);
 dev_page_proc_end_page(gx_default_end_page);
 
+/* Define default pages sizes. */
+/* U.S. letter paper (8.5" x 11"). */
+#define DEFAULT_WIDTH_10THS_US_LETTER 85
+#define DEFAULT_HEIGHT_10THS_US_LETTER 110
+/* A4 paper (210mm x 297mm).  The dimensions are off by a few mm.... */
+#define DEFAULT_WIDTH_10THS_A4 83
+#define DEFAULT_HEIGHT_10THS_A4 117
+/* Choose a default.  A4 may be set in the makefile. */
+#ifdef A4
+#  define DEFAULT_WIDTH_10THS DEFAULT_WIDTH_10THS_A4
+#  define DEFAULT_HEIGHT_10THS DEFAULT_HEIGHT_10THS_A4
+#else
+#  define DEFAULT_WIDTH_10THS DEFAULT_WIDTH_10THS_US_LETTER
+#  define DEFAULT_HEIGHT_10THS DEFAULT_HEIGHT_10THS_US_LETTER
+#endif
+
 /* ---------------- Device structure ---------------- */
 
 /*
@@ -129,50 +184,42 @@ dev_page_proc_end_page(gx_default_end_page);
  * specification in terms of a default device resolution rather than
  * 1/72" units, are due to Adobe.
  *
- * The xxx_set members are interpreted as follows:
- *	-1 : device does not support this parameter.
- *	0 : device supports the parameter, but no value has been specified.
- *	1 : a value has been specified for the parameter.
+ * ****** NOTE: If you define any subclasses of gx_device, you *must* define
+ * ****** the finalization procedure as gx_device_finalize.  Finalization
+ * ****** procedures are not automatically inherited.
  */
 #define gx_device_common\
-	int params_size;		/* OBSOLETE */\
+	int params_size;		/* OBSOLETE if stype != 0: */\
 					/* size of this structure */\
-	gx_device_procs *static_procs;	/* OBSOLETE */\
+	const gx_device_procs *static_procs;	/* OBSOLETE */\
 					/* pointer to std_procs */\
 	const char *dname;		/* the device name */\
-	gs_memory_t *memory;		/* (0 if statically allocated instance) */\
+	gs_memory_t *memory;		/* (0 iff static prototype) */\
+	gs_memory_type_ptr_t stype;	/* memory manager structure type, */\
+					/* 0 iff static prototype */\
 	bool is_open;			/* true if device has been opened */\
-	struct {\
-		gx_color_index white;	/* (cache) device white color */\
-		gx_color_index black;	/* (cache) device black color */\
-	} cached;\
+	int max_fill_band;		/* limit on band size for fill, */\
+					/* must be 0 or a power of 2 */\
+					/* (see gdevabuf.c for more info) */\
 	gx_device_color_info color_info;	/* color information */\
 	int width;			/* width in pixels */\
 	int height;			/* height in pixels */\
-	float PageSize[2];		/* page dimensions in points */\
+	float MediaSize[2];		/* media dimensions in points */\
 	float ImagingBBox[4];		/* imageable region in points */\
 	  bool ImagingBBox_set;\
 	float HWResolution[2];		/* resolution, dots per inch */\
-	float Margins_HWResolution[2];	/* resolution for Margins */\
-	float Margins[2];		/* offset of (0,0) from page corner, */\
-					/* in units given by Margins_HWResolution */\
+	float MarginsHWResolution[2];	/* resolution for Margins */\
+	float Margins[2];		/* offset of physical page corner */\
+					/* from device coordinate (0,0), */\
+					/* in units given by MarginsHWResolution */\
 	float HWMargins[4];		/* margins around imageable area, */\
-					/* in points */\
+					/* in default user units ("points") */\
 	long PageCount;			/* number of pages written */\
 	long ShowpageCount;		/* number of calls on showpage */\
-	int Orientation;\
-	  int Orientation_set;\
-	bool Duplex;\
-	  int Duplex_set;\
+	bool IgnoreNumCopies;		/* if true, force num_copies = 1 */\
 	gx_page_device_procs page_procs;	/* must be last */\
 		/* end of std_device_body */\
 	gx_device_procs std_procs	/* standard procedures */
-/* Define an enumerated type for the optional parameters. */
-typedef enum {
-	param_Orientation,
-	param_Duplex
-} param_enum;
-#define no_options_(p) (-1)
 #define x_pixels_per_inch HWResolution[0]
 #define y_pixels_per_inch HWResolution[1]
 #define offset_margin_values(x, y, left, bot, right, top)\
@@ -181,22 +228,23 @@ typedef enum {
   offset_margin_values(0, 0, left, bot, right, top)
 #define no_margins margin_values(0, 0, 0, 0)
 #define no_margins_() no_margins
-/* Define macros that give the page offset in inches. */
-#define dev_x_offset(dev) ((dev)->Margins[0] / (dev)->Margins_HWResolution[0])
-#define dev_y_offset(dev) ((dev)->Margins[1] / (dev)->Margins_HWResolution[1])
+/* Define macros that give the page offset ("Margins") in inches. */
+#define dev_x_offset(dev) ((dev)->Margins[0] / (dev)->MarginsHWResolution[0])
+#define dev_y_offset(dev) ((dev)->Margins[1] / (dev)->MarginsHWResolution[1])
+#define dev_y_offset_points(dev) (dev_y_offset(dev) * 72.0)
 /* Note that left/right/top/bottom are defined relative to */
 /* the physical paper, not the coordinate system. */
 /* For backward compatibility, we define macros that give */
 /* the margins in inches. */
 #define dev_l_margin(dev) ((dev)->HWMargins[0] / 72.0)
 #define dev_b_margin(dev) ((dev)->HWMargins[1] / 72.0)
+#define dev_b_margin_points(dev) ((dev)->HWMargins[1])
 #define dev_r_margin(dev) ((dev)->HWMargins[2] / 72.0)
 #define dev_t_margin(dev) ((dev)->HWMargins[3] / 72.0)
+#define dev_t_margin_points(dev) ((dev)->HWMargins[3])
 /* The extra () are to prevent premature expansion. */
-#define open_init_closed()\
-  0 /*false*/, {gx_no_color_index, gx_no_color_index}
-#define open_init_open()\
-  1 /*true*/, {1, 0}
+#define open_init_closed() 0 /*false*/, 0 /* max_fill_band */
+#define open_init_open() 1 /*true*/, 0 /* max_fill_band */
 /* Accessors for device procedures */
 #define dev_proc(dev, p) ((dev)->std_procs.p)
 #define set_dev_proc(dev, p, proc) ((dev)->std_procs.p = (proc))
@@ -214,87 +262,85 @@ typedef enum {
  * can't cast a computed float to an int.  That's why we specify
  * the page width and height in inches/10 instead of inches.
  *
- * In order to conditionally initialize an open-ended set of parameters,
- * we pass each optional parameter identifier (of type param_enum) to a
- * macro, which is defined as a conditional expression that returns either
- * a non-negative initial value or -1.
- *
  * Note that the macro is broken up so as to be usable for devices that
  * add further initialized state to the generic device.
  * Note also that the macro does not initialize std_procs, which is
  * the next element of the structure.
  */
-#define std_device_part1_(devtype, ptr_procs, dev_name, open_init)\
+#define std_device_part1_(devtype, ptr_procs, dev_name, stype, open_init)\
 	sizeof(devtype), ptr_procs, dev_name,\
-	0 /*memory*/, open_init() /*is_open, cached*/
+	0 /*memory*/, stype, open_init() /*stype, is_open, cached*/
 /* color_info goes here */
+/*
+ * The MetroWerks compiler has some bizarre bug that produces a spurious
+ * error message if the width and/or height are defined as 0 below,
+ * unless we use the +/- workaround in the next macro.
+ */
 #define std_device_part2_(width, height, x_dpi, y_dpi)\
 	width, height,\
-	{ (width) * 72.0 / (x_dpi), (height) * 72.0 / (y_dpi) },\
+	{ (((width) * 72.0 + 0.5) - 0.5) / (x_dpi),\
+	  (((height) * 72.0 + 0.5) - 0.5) / (y_dpi) },\
 	{ 0, 0, 0, 0 }, 0/*false*/, { x_dpi, y_dpi }, { x_dpi, y_dpi }
 /* offsets and margins go here */
-#define std_device_part3_(options_macro)\
-	0, 0,\
-	options_macro(param_Orientation),\
-	  (options_macro(param_Orientation) >= 0 ? 0 : -1),\
-	options_macro(param_Duplex) > 0,\
-	  (options_macro(param_Duplex) >= 0 ? 0 : -1),\
+#define std_device_part3_()\
+	0, 0, 0/*false*/,\
 	{ gx_default_install, gx_default_begin_page, gx_default_end_page }
 /*
  * We need a number of different variants of the std_device_ macro simply
- * because we can't pass the color_info, offsets/margins, or optional
- * parameters as macro arguments, which in turn is because of the early macro
+ * because we can't pass the color_info or offsets/margins
+ * as macro arguments, which in turn is because of the early macro
  * expansion issue noted in stdpre.h.  The basic variants are:
  *	...body_with_macros_, which uses 0-argument macros to supply
- *	  open_init, color_info, and offsets/margins, and a 1-argument macro
- *	  to supply the optional parameters;
- *	...full_[options_]body, which takes 12 values (6 for dci_values,
+ *	  open_init, color_info, and offsets/margins;
+ *	...full_body, which takes 12 values (6 for dci_values,
  *	  6 for offsets/margins);
  *	...color_full_body, which takes 9 values (3 for dci_color,
  *	  6 for margins/offset).
- *	...std_color_full_[options_]body, which takes 7 values (1 for
- *	  dci_std_color, 6 for margins/offset).
+ *	...std_color_full_body, which takes 7 values (1 for dci_std_color,
+ *	  6 for margins/offset).
  *	
  */ 
-#define std_device_body_with_macros_(dtype, pprocs, dname, w, h, xdpi, ydpi, open_init, dci_macro, margins_macro, options_macro)\
-	std_device_part1_(dtype, pprocs, dname, open_init),\
+#define std_device_body_with_macros_(dtype, pprocs, dname, stype, w, h, xdpi, ydpi, open_init, dci_macro, margins_macro)\
+	std_device_part1_(dtype, pprocs, dname, stype, open_init),\
 	dci_macro(),\
 	std_device_part2_(w, h, xdpi, ydpi),\
 	margins_macro(),\
-	std_device_part3_(options_macro)
+	std_device_part3_()
 
 #define std_device_std_body(dtype, pprocs, dname, w, h, xdpi, ydpi)\
-	std_device_body_with_macros_(dtype, pprocs, dname,\
+	std_device_body_with_macros_(dtype, pprocs, dname, 0,\
 	  w, h, xdpi, ydpi,\
-	  open_init_closed, dci_black_and_white_, no_margins_, no_options_)
+	  open_init_closed, dci_black_and_white_, no_margins_)
 
 #define std_device_std_body_open(dtype, pprocs, dname, w, h, xdpi, ydpi)\
-	std_device_body_with_macros_(dtype, pprocs, dname,\
+	std_device_body_with_macros_(dtype, pprocs, dname, 0,\
 	  w, h, xdpi, ydpi,\
-	  open_init_open, dci_black_and_white_, no_margins_, no_options_)
+	  open_init_open, dci_black_and_white_, no_margins_)
 
-#define std_device_full_options_body(dtype, pprocs, dname, w, h, xdpi, ydpi, ncomp, depth, mg, mc, dg, dc, xoff, yoff, lm, bm, rm, tm, opt)\
-	std_device_part1_(dtype, pprocs, dname, open_init_closed),\
+#define std_device_full_body(dtype, pprocs, dname, w, h, xdpi, ydpi, ncomp, depth, mg, mc, dg, dc, xoff, yoff, lm, bm, rm, tm)\
+	std_device_part1_(dtype, pprocs, dname, 0, open_init_closed),\
 	dci_values(ncomp, depth, mg, mc, dg, dc),\
 	std_device_part2_(w, h, xdpi, ydpi),\
 	offset_margin_values(xoff, yoff, lm, bm, rm, tm),\
-	std_device_part3_(opt)
+	std_device_part3_()
 
-#define std_device_full_body(dtype, pprocs, dname, w, h, xdpi, ydpi, ncomp, depth, mg, mc, dg, dc, xoff, yoff, lm, bm, rm, tm)\
-	std_device_full_options_body(dtype, pprocs, dname, w, h, xdpi, ydpi, ncomp, depth, mg, mc, dg, dc, xoff, yoff, lm, bm, rm, tm, no_options_)
+#define std_device_dci_type_body(dtype, pprocs, dname, stype, w, h, xdpi, ydpi, ncomp, depth, mg, mc, dg, dc)\
+	std_device_part1_(dtype, pprocs, dname, stype, open_init_closed),\
+	dci_values(ncomp, depth, mg, mc, dg, dc),\
+	std_device_part2_(w, h, xdpi, ydpi),\
+	offset_margin_values(0, 0, 0, 0, 0, 0),\
+	std_device_part3_()
 
 #define std_device_dci_body(dtype, pprocs, dname, w, h, xdpi, ydpi, ncomp, depth, mg, mc, dg, dc)\
-	std_device_full_body(dtype, pprocs, dname,\
-	  w, h, xdpi, ydpi,\
-	  ncomp, depth, mg, mc, dg, dc,\
-	  0, 0, 0, 0, 0, 0)
+	std_device_dci_type_body(dtype, pprocs, dname, 0,\
+	  w, h, xdpi, ydpi, ncomp, depth, mg, mc, dg, dc)
 
 #define std_device_color_full_body(dtype, pprocs, dname, w, h, xdpi, ydpi, depth, max_value, dither, xoff, yoff, lm, bm, rm, tm)\
-	std_device_part1_(dtype, pprocs, dname, open_init_closed),\
+	std_device_part1_(dtype, pprocs, dname, 0, open_init_closed),\
 	dci_color(depth, max_value, dither),\
 	std_device_part2_(w, h, xdpi, ydpi),\
 	offset_margin_values(xoff, yoff, lm, bm, rm, tm),\
-	std_device_part3_(no_options_)
+	std_device_part3_()
 
 #define std_device_color_body(dtype, pprocs, dname, w, h, xdpi, ydpi, depth, max_value, dither)\
 	std_device_color_full_body(dtype, pprocs, dname,\
@@ -302,15 +348,19 @@ typedef enum {
 	  depth, max_value, dither,\
 	  0, 0, 0, 0, 0, 0)
 
-#define std_device_std_color_full_options_body(dtype, pprocs, dname, w, h, xdpi, ydpi, depth, xoff, yoff, lm, bm, rm, tm, opt)\
-	std_device_part1_(dtype, pprocs, dname, open_init_closed),\
+#define std_device_color_stype_body(dtype, pprocs, dname, stype, w, h, xdpi, ydpi, depth, max_value, dither)\
+	std_device_part1_(dtype, pprocs, dname, stype, open_init_closed),\
+	dci_color(depth, max_value, dither),\
+	std_device_part2_(w, h, xdpi, ydpi),\
+	offset_margin_values(0, 0, 0, 0, 0, 0),\
+	std_device_part3_()
+
+#define std_device_std_color_full_body(dtype, pprocs, dname, w, h, xdpi, ydpi, depth, xoff, yoff, lm, bm, rm, tm)\
+	std_device_part1_(dtype, pprocs, dname, 0, open_init_closed),\
 	dci_std_color(depth),\
 	std_device_part2_(w, h, xdpi, ydpi),\
 	offset_margin_values(xoff, yoff, lm, bm, rm, tm),\
-	std_device_part3_(opt)
-
-#define std_device_std_color_full_body(dtype, pprocs, dname, w, h, xdpi, ydpi, depth, xoff, yoff, lm, bm, rm, tm)\
-	std_device_std_color_full_options_body(dtype, pprocs, dname, w, h, xdpi, ydpi, depth, xoff, yoff, lm, bm, rm, tm, no_options_)
+	std_device_part3_()
 
 /* ---------------- Device procedures ---------------- */
 
@@ -326,7 +376,7 @@ typedef struct gs_param_list_s gs_param_list;
  * because many drivers maintain dynamic state in the device structure.
  * Note also that the structure is defined as a template, so that
  * we can instantiate it with device subclasses.
- * Because C doesn't have real templates, we must do this with a macros.
+ * Because C doesn't have real templates, we must do this with macros.
  */
 
 /* Define macros for declaring device procedures. */
@@ -384,7 +434,7 @@ typedef struct gs_param_list_s gs_param_list;
 
 #define dev_t_proc_copy_mono(proc, dev_t)\
   int proc(P11(dev_t *dev,\
-    const unsigned char *data, int data_x, int raster, gx_bitmap_id id,\
+    const byte *data, int data_x, int raster, gx_bitmap_id id,\
     int x, int y, int width, int height,\
     gx_color_index color0, gx_color_index color1))
 #define dev_proc_copy_mono(proc)\
@@ -392,11 +442,13 @@ typedef struct gs_param_list_s gs_param_list;
 
 #define dev_t_proc_copy_color(proc, dev_t)\
   int proc(P9(dev_t *dev,\
-    const unsigned char *data, int data_x, int raster, gx_bitmap_id id,\
+    const byte *data, int data_x, int raster, gx_bitmap_id id,\
     int x, int y, int width, int height))
 #define dev_proc_copy_color(proc)\
   dev_t_proc_copy_color(proc, gx_device)
 
+		/* OBSOLETED in release 3.66 */
+
 #define dev_t_proc_draw_line(proc, dev_t)\
   int proc(P6(dev_t *dev,\
     int x0, int y0, int x1, int y1, gx_color_index color))
@@ -407,7 +459,7 @@ typedef struct gs_param_list_s gs_param_list;
 
 #define dev_t_proc_get_bits(proc, dev_t)\
   int proc(P4(dev_t *dev,\
-    int y, unsigned char *data, unsigned char **actual_data))
+    int y, byte *data, byte **actual_data))
 #define dev_proc_get_bits(proc)\
   dev_t_proc_get_bits(proc, gx_device)
 
@@ -469,12 +521,143 @@ typedef struct gs_param_list_s gs_param_list;
   dev_t_proc_get_alpha_bits(proc, gx_device)
 
 #define dev_t_proc_copy_alpha(proc, dev_t)\
-  int proc(P11(dev_t *dev, const unsigned char *data, int data_x,\
+  int proc(P11(dev_t *dev, const byte *data, int data_x,\
     int raster, gx_bitmap_id id, int x, int y, int width, int height,\
     gx_color_index color, int depth))
 #define dev_proc_copy_alpha(proc)\
   dev_t_proc_copy_alpha(proc, gx_device)
 
+		/* Added in release 3.38 */
+
+#define dev_t_proc_get_band(proc, dev_t)\
+  int proc(P3(dev_t *dev, int y, int *band_start))
+#define dev_proc_get_band(proc)\
+  dev_t_proc_get_band(proc, gx_device)
+
+		/* Added in release 3.44 */
+
+#define dev_t_proc_copy_rop(proc, dev_t)\
+  int proc(P15(dev_t *dev,\
+    const byte *sdata, int sourcex, uint sraster, gx_bitmap_id id,\
+    const gx_color_index *scolors,\
+    const gx_tile_bitmap *texture, const gx_color_index *tcolors,\
+    int x, int y, int width, int height,\
+    int phase_x, int phase_y, gs_logical_operation_t lop))
+#define dev_proc_copy_rop(proc)\
+  dev_t_proc_copy_rop(proc, gx_device)
+
+		/* Added in release 3.60, changed in release 3.68. */
+
+#define dev_t_proc_fill_path(proc, dev_t)\
+  int proc(P6(dev_t *dev,\
+    const gs_imager_state *pis, gx_path *ppath,\
+    const gx_fill_params *params,\
+    const gx_drawing_color *pdcolor, const gx_clip_path *pcpath))
+#define dev_proc_fill_path(proc)\
+  dev_t_proc_fill_path(proc, gx_device)
+
+#define dev_t_proc_stroke_path(proc, dev_t)\
+  int proc(P6(dev_t *dev,\
+    const gs_imager_state *pis, gx_path *ppath,\
+    const gx_stroke_params *params,\
+    const gx_drawing_color *pdcolor, const gx_clip_path *pcpath))
+#define dev_proc_stroke_path(proc)\
+  dev_t_proc_stroke_path(proc, gx_device)
+
+		/* Added in release 3.60 */
+
+#define dev_t_proc_fill_mask(proc, dev_t)\
+  int proc(P13(dev_t *dev,\
+    const byte *data, int data_x, int raster, gx_bitmap_id id,\
+    int x, int y, int width, int height,\
+    const gx_drawing_color *pdcolor, int depth,\
+    gs_logical_operation_t lop, const gx_clip_path *pcpath))
+#define dev_proc_fill_mask(proc)\
+  dev_t_proc_fill_mask(proc, gx_device)
+
+		/* Added in release 3.66, changed in 3.69 */
+
+#define dev_t_proc_fill_trapezoid(proc, dev_t)\
+  int proc(P8(dev_t *dev,\
+    const gs_fixed_edge *left, const gs_fixed_edge *right,\
+    fixed ybot, fixed ytop, bool swap_axes,\
+    const gx_drawing_color *pdcolor, gs_logical_operation_t lop))
+#define dev_proc_fill_trapezoid(proc)\
+  dev_t_proc_fill_trapezoid(proc, gx_device)
+
+#define dev_t_proc_fill_parallelogram(proc, dev_t)\
+  int proc(P9(dev_t *dev,\
+    fixed px, fixed py, fixed ax, fixed ay, fixed bx, fixed by,\
+    const gx_drawing_color *pdcolor, gs_logical_operation_t lop))
+#define dev_proc_fill_parallelogram(proc)\
+  dev_t_proc_fill_parallelogram(proc, gx_device)
+
+#define dev_t_proc_fill_triangle(proc, dev_t)\
+  int proc(P9(dev_t *dev,\
+    fixed px, fixed py, fixed ax, fixed ay, fixed bx, fixed by,\
+    const gx_drawing_color *pdcolor, gs_logical_operation_t lop))
+#define dev_proc_fill_triangle(proc)\
+  dev_t_proc_fill_triangle(proc, gx_device)
+
+#define dev_t_proc_draw_thin_line(proc, dev_t)\
+  int proc(P7(dev_t *dev,\
+    fixed fx0, fixed fy0, fixed fx1, fixed fy1,\
+    const gx_drawing_color *pdcolor, gs_logical_operation_t lop))
+#define dev_proc_draw_thin_line(proc)\
+  dev_t_proc_draw_thin_line(proc, gx_device)
+
+		/* Added in release 3.66 (as stubs); */
+		/* changed in 3.68; */
+		/* begin_image and image_data changed in 4.30 */
+
+#define dev_t_proc_begin_image(proc, dev_t)\
+  int proc(P9(dev_t *dev,\
+    const gs_imager_state *pis, const gs_image_t *pim,\
+    gs_image_format_t format, const gs_int_rect *prect,\
+    const gx_drawing_color *pdcolor, const gx_clip_path *pcpath,\
+    gs_memory_t *memory, void **pinfo))
+#define dev_proc_begin_image(proc)\
+  dev_t_proc_begin_image(proc, gx_device)
+
+#define dev_t_proc_image_data(proc, dev_t)\
+  int proc(P6(dev_t *dev,\
+    void *info, const byte **planes, int data_x, uint raster, int height))
+#define dev_proc_image_data(proc)\
+  dev_t_proc_image_data(proc, gx_device)
+
+#define dev_t_proc_end_image(proc, dev_t)\
+  int proc(P3(dev_t *dev,\
+    void *info, bool draw_last))
+#define dev_proc_end_image(proc)\
+  dev_t_proc_end_image(proc, gx_device)
+
+		/* Added in release 3.68 */
+
+#define dev_t_proc_strip_tile_rectangle(proc, dev_t)\
+  int proc(P10(dev_t *dev,\
+    const gx_strip_bitmap *tiles, int x, int y, int width, int height,\
+    gx_color_index color0, gx_color_index color1,\
+    int phase_x, int phase_y))
+#define dev_proc_strip_tile_rectangle(proc)\
+  dev_t_proc_strip_tile_rectangle(proc, gx_device)
+
+#define dev_t_proc_strip_copy_rop(proc, dev_t)\
+  int proc(P15(dev_t *dev,\
+    const byte *sdata, int sourcex, uint sraster, gx_bitmap_id id,\
+    const gx_color_index *scolors,\
+    const gx_strip_bitmap *textures, const gx_color_index *tcolors,\
+    int x, int y, int width, int height,\
+    int phase_x, int phase_y, gs_logical_operation_t lop))
+#define dev_proc_strip_copy_rop(proc)\
+  dev_t_proc_strip_copy_rop(proc, gx_device)
+
+		/* Added in release 4.20 */
+
+#define dev_t_proc_get_clipping_box(proc, dev_t)\
+  void proc(P2(dev_t *dev, gs_fixed_rect *pbox))
+#define dev_proc_get_clipping_box(proc)\
+  dev_t_proc_get_clipping_box(proc, gx_device)
+
 /* Define the device procedure vector template proper. */
 
 #define gx_device_proc_struct(dev_t)\
@@ -500,18 +683,70 @@ typedef struct gs_param_list_s gs_param_list;
 	dev_t_proc_get_page_device((*get_page_device), dev_t);\
 	dev_t_proc_get_alpha_bits((*get_alpha_bits), dev_t);\
 	dev_t_proc_copy_alpha((*copy_alpha), dev_t);\
+	dev_t_proc_get_band((*get_band), dev_t);\
+	dev_t_proc_copy_rop((*copy_rop), dev_t);\
+	dev_t_proc_fill_path((*fill_path), dev_t);\
+	dev_t_proc_stroke_path((*stroke_path), dev_t);\
+	dev_t_proc_fill_mask((*fill_mask), dev_t);\
+	dev_t_proc_fill_trapezoid((*fill_trapezoid), dev_t);\
+	dev_t_proc_fill_parallelogram((*fill_parallelogram), dev_t);\
+	dev_t_proc_fill_triangle((*fill_triangle), dev_t);\
+	dev_t_proc_draw_thin_line((*draw_thin_line), dev_t);\
+	dev_t_proc_begin_image((*begin_image), dev_t);\
+	dev_t_proc_image_data((*image_data), dev_t);\
+	dev_t_proc_end_image((*end_image), dev_t);\
+	dev_t_proc_strip_tile_rectangle((*strip_tile_rectangle), dev_t);\
+	dev_t_proc_strip_copy_rop((*strip_copy_rop), dev_t);\
+	dev_t_proc_get_clipping_box((*get_clipping_box), dev_t);\
 }
 
 /* A generic device procedure record. */
 struct gx_device_procs_s gx_device_proc_struct(gx_device);
 
+/*
+ * Define a procedure for setting up a memory device for buffering output
+ * for a given device.  This is only used by band devices, but we define it
+ * here for convenience.  The default implementation just calls
+ * gs_make_mem_device.  Possibly this should be a generic device
+ * procedure....
+ */
+#ifndef gx_device_memory_DEFINED
+#  define gx_device_memory_DEFINED
+typedef struct gx_device_memory_s gx_device_memory;
+#endif
+#define dev_proc_make_buffer_device(proc)\
+  int proc(P4(gx_device_memory *, gx_device *, gs_memory_t *, bool))
+dev_proc_make_buffer_device(gx_default_make_buffer_device);
+
+/*
+ * Define unaligned analogues of the copy_xxx procedures.
+ * These are slower than the standard procedures, which require
+ * aligned bitmaps, and also are not portable to non-byte-addressed machines.
+ *
+ * We allow both unaligned data and unaligned scan line widths;
+ * however, we do require that both of these be aligned modulo the largest
+ * power of 2 bytes that divides the data depth, i.e.:
+ *	depth	alignment
+ *	<= 8	1
+ *	16	2
+ *	24	1
+ *	32	4
+ */
+dev_proc_copy_mono(gx_copy_mono_unaligned);
+dev_proc_copy_color(gx_copy_color_unaligned);
+dev_proc_copy_alpha(gx_copy_alpha_unaligned);
+
 /* A generic device */
 struct gx_device_s {
 	gx_device_common;
 };
 extern_st(st_device);
+struct_proc_finalize(gx_device_finalize);	/* public for subclasses */
+/* We use vacuous enum/reloc procedures, rather than 0, so that */
+/* gx_device can have subclasses. */
 #define public_st_device()	/* in gsdevice.c */\
-  gs_public_st_simple(st_device, gx_device, "gx_device")
+  gs_public_st_complex_only(st_device, gx_device, "gx_device",\
+    0, gs_no_struct_enum_ptrs, gs_no_struct_reloc_ptrs, gx_device_finalize)
 #define st_device_max_ptrs 0
 
 /* Enumerate or relocate a pointer to a device. */
@@ -536,18 +771,25 @@ typedef struct gx_device_forward_s {
 } gx_device_forward;
 extern_st(st_device_forward);
 #define public_st_device_forward()	/* in gsdevice.c */\
-  gs_public_st_composite(st_device_forward, gx_device_forward,\
-    "gx_device_forward", device_forward_enum_ptrs, device_forward_reloc_ptrs)
+  gs_public_st_complex_only(st_device_forward, gx_device_forward,\
+    "gx_device_forward", 0, device_forward_enum_ptrs,\
+    device_forward_reloc_ptrs, gx_device_finalize)
 #define st_device_forward_max_ptrs (st_device_max_ptrs + 1)
 
 /* A null device.  This is used to temporarily disable output. */
-typedef struct gx_device_null_s {
+#ifndef gx_device_null_DEFINED
+#  define gx_device_null_DEFINED
+typedef struct gx_device_null_s gx_device_null;
+#endif
+struct gx_device_null_s {
 	gx_device_forward_common;
-} gx_device_null;
+};
+extern gx_device_null far_data gs_null_device;	/* (should be const) */
 extern_st(st_device_null);
 #define public_st_device_null()	/* in gsdevice.c */\
-  gs_public_st_suffix_add0(st_device_null, gx_device_null, "gx_device_null",\
-    device_forward_enum_ptrs, device_forward_reloc_ptrs)
+  gs_public_st_complex_only(st_device_null, gx_device_null,\
+    "gx_device_null", 0, device_forward_enum_ptrs,\
+    device_forward_reloc_ptrs, gx_device_finalize)
 #define st_device_null_max_ptrs st_device_forward_max_ptrs
 /* Make a null device. */
 /* The gs_memory_t argument is 0 if the device is temporary and local, */
@@ -556,7 +798,7 @@ void	gs_make_null_device(P2(gx_device_null *, gs_memory_t *));
 
 /* Calculate the raster (number of bytes in a scan line), */
 /* with byte or word padding. */
-uint	gx_device_raster(P2(const gx_device *dev, int pad_to_word));
+uint	gx_device_raster(P2(const gx_device *dev, bool pad_to_word));
 
 /* Adjust the resolution for devices that only have a fixed set of */
 /* geometries, so that the apparent size in inches remains constant. */
@@ -569,12 +811,15 @@ int	gx_device_adjust_resolution(P4(gx_device *dev, int actual_width, int actual_
 void	gx_device_set_margins(P3(gx_device *dev, const float *margins /*[4]*/,
   bool move_origin));
 
-/* Set the width and height, updating PageSize. */
+/* Set the width and height (in pixels), updating MediaSize. */
 void gx_device_set_width_height(P3(gx_device *dev, int width, int height));
-/* Set the resolution, updating width and height. */
+/* Set the resolution (in pixels per inch), updating width and height. */
 void gx_device_set_resolution(P3(gx_device *dev, floatp x_dpi, floatp y_dpi));
-/* Set the PageSize, updating width and height. */
-void gx_device_set_page_size(P3(gx_device *dev, floatp page_width, floatp page_height));
+/* Set the MediaSize (in 1/72" units), updating width and height. */
+void gx_device_set_media_size(P3(gx_device *dev, floatp media_width, floatp media_height));
+/****** BACKWARD COMPATIBILITY ******/
+#define gx_device_set_page_size(dev, w, h)\
+  gx_device_set_media_size(dev, w, h)
 
 /*
  * Macros to help the drawing procedures clip coordinates to
@@ -582,20 +827,30 @@ void gx_device_set_page_size(P3(gx_device *dev, floatp page_width, floatp page_h
  * x, y, w, h, data, data_x, and id.
  */
 
-/* Macro for fill_rectangle and tile_rectangle. */
-#define fit_fill_xw(dev, x, y, w, h)\
+/* Macros for fill_rectangle and [strip_]tile_rectangle. */
+#define fit_fill_xy(dev, x, y, w, h)\
 	if ( (x | y) < 0 )\
 	{	if ( x < 0 ) w += x, x = 0;\
 		if ( y < 0 ) h += y, y = 0;\
-	}\
+	}
+#define fit_fill_xyw(dev, x, y, w, h)\
+	fit_fill_xy(dev, x, y, w, h);\
 	if ( x > dev->width - w ) w = dev->width - x
-#define fit_fill_xwh(dev, x, y, w, h)\
-	fit_fill_xw(dev, x, y, w, h);\
-	if ( w <= 0 || h <= 0 ) return 0
+#define fit_fill_y(dev, y, h)\
+	if ( y < 0 ) h += y, y = 0
+#define fit_fill_h(dev, y, h)\
+	if ( y > dev->height - h ) h = dev->height - y
+#define fit_fill_yh(dev, y, h)\
+	fit_fill_y(dev, y, h);\
+	fit_fill_h(dev, y, h)
+#define fit_fill_xywh(dev, x, y, w, h)\
+	fit_fill_xyw(dev, x, y, w, h);\
+	fit_fill_h(dev, y, h)
+#define if_fit_fill_empty(dev, x, y, w, h)\
+	fit_fill_xywh(dev, x, y, w, h);\
+	if ( w <= 0 || h <= 0 )
 #define fit_fill(dev, x, y, w, h)\
-	fit_fill_xw(dev, x, y, w, h);\
-	if ( y > dev->height - h ) h = dev->height - y;\
-	if ( w <= 0 || h <= 0 ) return 0
+	if_fit_fill_empty(dev, x, y, w, h) return 0
 
 /* Macro for copy_mono and copy_color. */
 #define fit_copy_xw(dev, data, data_x, raster, id, x, y, w, h)\
@@ -625,6 +880,7 @@ dev_proc_map_color_rgb(gx_default_w_b_map_color_rgb);
 #define gx_default_map_rgb_color gx_default_w_b_map_rgb_color
 #define gx_default_map_color_rgb gx_default_w_b_map_color_rgb
 dev_proc_tile_rectangle(gx_default_tile_rectangle);
+dev_proc_copy_mono(gx_default_copy_mono);
 dev_proc_copy_color(gx_default_copy_color);
 dev_proc_draw_line(gx_default_draw_line);
 dev_proc_get_bits(gx_default_get_bits);
@@ -637,7 +893,28 @@ dev_proc_map_rgb_alpha_color(gx_default_map_rgb_alpha_color);
 dev_proc_get_page_device(gx_default_get_page_device);	/* returns NULL */
 dev_proc_get_page_device(gx_page_device_get_page_device);  /* returns dev */
 dev_proc_get_alpha_bits(gx_default_get_alpha_bits);
+dev_proc_copy_alpha(gx_no_copy_alpha);		/* gives error */
 dev_proc_copy_alpha(gx_default_copy_alpha);
+dev_proc_get_band(gx_default_get_band);
+dev_proc_copy_rop(gx_no_copy_rop);		/* gives error */
+extern dev_proc_copy_rop((*gx_default_copy_rop_proc));
+dev_proc_copy_rop(gx_default_copy_rop);		/* calls ...proc */
+dev_proc_fill_path(gx_default_fill_path);
+dev_proc_stroke_path(gx_default_stroke_path);
+dev_proc_fill_mask(gx_default_fill_mask);
+dev_proc_fill_trapezoid(gx_default_fill_trapezoid);
+dev_proc_fill_parallelogram(gx_default_fill_parallelogram);
+dev_proc_fill_triangle(gx_default_fill_triangle);
+dev_proc_draw_thin_line(gx_default_draw_thin_line);
+dev_proc_begin_image(gx_default_begin_image);
+dev_proc_image_data(gx_default_image_data);
+dev_proc_end_image(gx_default_end_image);
+dev_proc_strip_tile_rectangle(gx_default_strip_tile_rectangle);
+dev_proc_strip_copy_rop(gx_no_strip_copy_rop);		/* gives error */
+extern dev_proc_strip_copy_rop((*gx_default_strip_copy_rop_proc));
+dev_proc_strip_copy_rop(gx_default_strip_copy_rop);	/* calls ...proc */
+dev_proc_get_clipping_box(gx_default_get_clipping_box);
+dev_proc_get_clipping_box(gx_get_largest_clipping_box);
 
 /* Color mapping routines for black-on-white, gray scale, true RGB, */
 /* and true CMYK color. */
@@ -651,8 +928,12 @@ dev_proc_map_cmyk_color(gx_default_cmyk_map_cmyk_color);
 
 /* Default implementations for forwarding devices */
 dev_proc_get_initial_matrix(gx_forward_get_initial_matrix);
+dev_proc_sync_output(gx_forward_sync_output);
+dev_proc_output_page(gx_forward_output_page);
 dev_proc_map_rgb_color(gx_forward_map_rgb_color);
 dev_proc_map_color_rgb(gx_forward_map_color_rgb);
+dev_proc_tile_rectangle(gx_forward_tile_rectangle);
+dev_proc_get_bits(gx_forward_get_bits);
 dev_proc_get_params(gx_forward_get_params);
 dev_proc_put_params(gx_forward_put_params);
 dev_proc_map_cmyk_color(gx_forward_map_cmyk_color);
@@ -661,6 +942,21 @@ dev_proc_get_xfont_device(gx_forward_get_xfont_device);
 dev_proc_map_rgb_alpha_color(gx_forward_map_rgb_alpha_color);
 dev_proc_get_page_device(gx_forward_get_page_device);
 dev_proc_get_alpha_bits(gx_forward_get_alpha_bits);
+dev_proc_get_band(gx_forward_get_band);
+extern dev_proc_copy_rop((*gx_forward_copy_rop_proc));
+dev_proc_fill_path(gx_forward_fill_path);
+dev_proc_stroke_path(gx_forward_stroke_path);
+dev_proc_fill_mask(gx_forward_fill_mask);
+dev_proc_fill_trapezoid(gx_forward_fill_trapezoid);
+dev_proc_fill_parallelogram(gx_forward_fill_parallelogram);
+dev_proc_fill_triangle(gx_forward_fill_triangle);
+dev_proc_draw_thin_line(gx_forward_draw_thin_line);
+dev_proc_begin_image(gx_forward_begin_image);
+dev_proc_image_data(gx_forward_image_data);
+dev_proc_end_image(gx_forward_end_image);
+dev_proc_strip_tile_rectangle(gx_forward_strip_tile_rectangle);
+extern dev_proc_strip_copy_rop((*gx_forward_strip_copy_rop_proc));
+dev_proc_get_clipping_box(gx_forward_get_clipping_box);
 
 /* Convert the device procedures to the proper form (see above). */
 void	gx_device_set_procs(P1(gx_device *));
@@ -677,6 +973,11 @@ void	gx_device_forward_color_procs(P1(gx_device_forward *));
 void gx_device_no_output(P1(gs_state *));
 void gx_set_device_only(P2(gs_state *, gx_device *));
 
+/* Open the output file for a device. */
+int gx_device_open_output_file(P5(const gx_device *dev, const char *fname,
+				  bool binary, bool positionable,
+				  FILE **pfile));
+
 /* Close a device. */
 int gs_closedevice(P1(gx_device *));
 
