@@ -1,4 +1,4 @@
-/* Copyright (C) 1989, 1995 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -48,10 +48,6 @@ int gs_point_transform2fixed(P4(const gs_matrix_fixed *, floatp, floatp,
 int gs_distance_transform2fixed(P4(const gs_matrix_fixed *, floatp, floatp,
 				   gs_fixed_point *));
 
-/* Macro for testing whether matrix coefficients are zero, */
-/* for shortcuts when the matrix has no skew. */
-#define is_skewed(pmat) !is_fzero2((pmat)->xy, (pmat)->yx)
-
 /*
  * Define the fixed-point coefficient structure for avoiding
  * floating point in coordinate transformations.
