@@ -1,19 +1,24 @@
 /* Copyright (C) 1989, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of Aladdin Ghostscript.
+  This file is part of GNU Ghostscript.
   
-  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
-  or distributor accepts any responsibility for the consequences of using it,
-  or for whether it serves any particular purpose or works at all, unless he
-  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
-  License (the "License") for full details.
+  GNU Ghostscript is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
+  anyone for the consequences of using it or for whether it serves any
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
-  Every copy of Aladdin Ghostscript must include a copy of the License,
-  normally in a plain ASCII text file named PUBLIC.  The License grants you
-  the right to copy, modify and redistribute Aladdin Ghostscript, but only
-  under certain conditions described in the License.  Among other things, the
-  License requires that the copyright notice and this notice be preserved on
-  all copies.
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* gxstroke.c */
@@ -154,9 +159,10 @@ typedef partial_line _ss *pl_ptr;
 /* Other forward declarations */
 private bool width_is_thin(P1(pl_ptr));
 private void near adjust_stroke(P3(pl_ptr, const gs_imager_state *, bool));
-private int near line_join_points(P4(const gx_line_params *pgs_lp,
+private int near line_join_points(P5(const gx_line_params *pgs_lp,
 				     pl_ptr plp, pl_ptr nplp,
-				     gs_fixed_point _ss *join_points));
+				     gs_fixed_point _ss *join_points,
+				     const gs_matrix *pmat));
 private void near compute_caps(P1(pl_ptr));
 private int near add_points(P4(gx_path *, const gs_fixed_point _ss *,
 			       int, bool));
@@ -194,9 +200,9 @@ if(to_path==&stroke_path_body && !gx_path_is_void_inline(&stroke_path_body))\
  * with an appropriate cap.
  */
 #define stroke_line_proc(proc)\
-  int near proc(P9(gx_path *, int, pl_ptr, pl_ptr, const gx_device_color *,\
-		   gx_device *, const gs_imager_state *,\
-		   const gx_stroke_params *, const gs_fixed_rect *))
+  int near proc(P10(gx_path *, int, pl_ptr, pl_ptr, const gx_device_color *,\
+		    gx_device *, const gs_imager_state *,\
+		    const gx_stroke_params *, const gs_fixed_rect *, int))
 typedef stroke_line_proc((*stroke_line_proc_t));
 
 private stroke_line_proc(stroke_add);
@@ -209,8 +215,14 @@ typedef enum {
   orient_landscape		/* [0 xy yx 0 tx ty] */
 } orientation;
 
-/* Stroke a path.  If to_path != 0, append the stroke outline to it; */
-/* if to_path == 0, draw the strokes on dev. */
+/*
+ * Stroke a path.  If to_path != 0, append the stroke outline to it;
+ * if to_path == 0, draw the strokes on dev.
+ *
+ * Note that gx_stroke_path_only with to_path != NULL may clip the path to
+ * the clipping path, as for to_path == NULL.  This is almost never
+ * what is wanted.
+ */
 int
 gx_stroke_path_only(gx_path *ppath, gx_path *to_path, gx_device *pdev,
   const gs_imager_state *pis, const gx_stroke_params *params,
@@ -247,6 +259,8 @@ gx_stroke_path_only(gx_path *ppath, gx_path *to_path, gx_device *pdev,
 		(uniform = (xy == yx ? -1 : xy == -yx ? 1 : 0),
 		 reflected = (uniform ? uniform < 0 : (xy < 0) == (yx < 0)),
 		 orient_landscape) :
+	   /* We should optimize uniform rotated coordinate systems */
+	   /* here as well, but we don't. */
 #endif
 	   (uniform = 0,
 	    reflected = xy * yx > xx * yy,
@@ -305,8 +319,12 @@ if ( gs_debug_c('o') )
 	/* Check the expanded bounding box against the clipping regions. */
 	if ( pcpath )
 	  gx_cpath_inner_box(pcpath, &cbox);
-	else
+	else if ( pdevc )
 	  (*dev_proc(dev, get_clipping_box))(dev, &cbox);
+	else
+	  { /* This is strokepath, not stroke.  Don't clip. */
+	    cbox = ibox;
+	  }
 	if ( !rect_within(ibox, cbox) )
 	  {	/* Intersect the path box and the clip bounding box. */
 		/* If the intersection is empty, this call is a no-op. */
@@ -374,10 +392,12 @@ sat:		    if ( xa < 0 ) xa = -xa;
 		    break;
 		  default:
 		    { /* The check is more complicated, but it's worth it. */
-		      float xsq = xx * xx + xy * xy;
-		      float ysq = yx * yx + yy * yy;
-		      float cross = xx * yx + xy * yy;
-		      if ( cross < 0 ) cross = 0;
+		      double xsq = xx * xx + xy * xy;
+		      double ysq = yx * yx + yy * yy;
+		      double cross = xx * yx + xy * yy;
+
+		      if ( cross < 0 )
+			cross = 0;
 		      always_thin =
 			((max(xsq, ysq) + cross) * line_width * line_width
 			 < 0.25);
@@ -596,7 +616,8 @@ d:		pl.e.p.x = sx, pl.e.p.y = sy;
 			    lptr = &pl;
 			  }
 			code = (*line_proc)(to_path, first, &pl_prev, lptr,
-					    pdevc, dev, pis, params, &cbox);
+					    pdevc, dev, pis, params, &cbox,
+					    uniform);
 			if ( code < 0 )
 			  goto exit;
 			fill_stroke_path(always_thin);
@@ -618,7 +639,7 @@ d:		pl.e.p.x = sx, pl.e.p.y = sy;
 		   (pl_ptr)0 : (pl_ptr)&pl_first);
 
 		code = (*line_proc)(to_path, index - 1, &pl_prev, lptr,
-				    pdevc, dev, pis, params, &cbox);
+				    pdevc, dev, pis, params, &cbox, uniform);
 		if ( code < 0 )
 		  goto exit;
 		fill_stroke_path(always_thin);
@@ -799,7 +820,8 @@ set_thin_widths(register pl_ptr plp)
 private int near
 stroke_fill(gx_path *ppath, int first, register pl_ptr plp, pl_ptr nplp,
   const gx_device_color *pdevc, gx_device *dev, const gs_imager_state *pis,
-  const gx_stroke_params *params, const gs_fixed_rect *pbbox)
+  const gx_stroke_params *params, const gs_fixed_rect *pbbox,
+  int uniform)
 {	if ( plp->thin )
 	   {	/* Minimum-width line, don't have to be careful. */
 		/* We do have to check for the entire line being */
@@ -850,7 +872,9 @@ stroke_fill(gx_path *ppath, int first, register pl_ptr plp, pl_ptr nplp,
 	      if ( nplp == 0 )
 		code = cap_points(cap, &plp->e, points + npoints);
 	      else
-		code = line_join_points(pgs_lp, plp, nplp, points + npoints);
+		code = line_join_points(pgs_lp, plp, nplp, points + npoints,
+					(uniform ? (gs_matrix *)0 :
+					 &ctm_only(pis)));
 	      if ( code < 0 )
 		return code;
 	      if ( nplp != 0 )
@@ -863,7 +887,8 @@ stroke_fill(gx_path *ppath, int first, register pl_ptr plp, pl_ptr nplp,
 			 )
 			goto fill;
 		    }
-		  { gs_fixed_point *bevel = points + 2;
+		  { const gs_fixed_point *bevel = points + 2;
+
 		    /* Identify which 3 points define the bevel triangle. */
 		    if ( points[3].x == nplp->o.p.x &&
 			 points[3].y == nplp->o.p.y
@@ -897,7 +922,7 @@ fill:	      code = add_points(ppath, points, npoints + code, true);
 	}
 	/* General case: construct a path for the fill algorithm. */
 	return stroke_add(ppath, first, plp, nplp, pdevc, dev, pis, params,
-			  pbbox);
+			  pbbox, uniform);
 }
 
 #undef lix
@@ -909,7 +934,8 @@ fill:	      code = add_points(ppath, points, npoints + code, true);
 private int near
 stroke_add(gx_path *ppath, int first, register pl_ptr plp, pl_ptr nplp,
   const gx_device_color *pdevc, gx_device *dev, const gs_imager_state *pis,
-  const gx_stroke_params *params, const gs_fixed_rect *ignore_pbbox)
+  const gx_stroke_params *params, const gs_fixed_rect *ignore_pbbox,
+  int uniform)
 {	const gx_line_params *pgs_lp = gs_currentlineparams_inline(pis);
 	gs_fixed_point points[8];
 	int npoints;
@@ -959,7 +985,8 @@ stroke_add(gx_path *ppath, int first, register pl_ptr plp, pl_ptr nplp,
 	else if ( nplp->thin )		/* no join */
 	  code = cap_points(gs_cap_butt, &plp->e, points + npoints);
 	else				/* non-round join */
-	  code = line_join_points(pgs_lp, plp, nplp, points + npoints);
+	  code = line_join_points(pgs_lp, plp, nplp, points + npoints,
+				  (uniform ? (gs_matrix *)0 : &ctm_only(pis)));
 	if ( code < 0 )
 	  return code;
 	code = add_points(ppath, points, npoints + code, moveto_first);
@@ -986,9 +1013,11 @@ add_points(gx_path *ppath, const gs_fixed_point _ss *points, int npoints,
 
 /* Compute the points for a bevel, miter, or triangle join. */
 /* Treat no join the same as a bevel join. */
+/* If pmat != 0, we must inverse-transform the distances for */
+/* the miter check. */
 private int near
 line_join_points(const gx_line_params *pgs_lp, pl_ptr plp, pl_ptr nplp,
-  gs_fixed_point _ss *join_points)
+  gs_fixed_point _ss *join_points, const gs_matrix *pmat)
 {	gs_line_join join =  pgs_lp->join;
 #define jp1 join_points[0]
 #define np1 join_points[1]
@@ -1013,11 +1042,15 @@ line_join_points(const gx_line_params *pgs_lp, pl_ptr plp, pl_ptr nplp,
 		 * In the present instance, x1,y1 are plp->width,
 		 * x2,y2 are nplp->width, and the origin is
 		 * their common point (plp->e.p, nplp->o.p).
+		 * ccw will be true iff nplp.o.co (nplp.o.p + width) is
+		 * counter-clockwise from plp.e.ce (plp.e.p + width),
+		 * in which case we want tan(a-b) rather than tan(b-a).
+		 *
+		 * We make the test using double arithmetic only because
+		 * the !@#&^*% C language doesn't give us access to
+		 * the double-width-result multiplication operation
+		 * that almost all CPUs provide!
 		 */
-		/* We make the test using double arithmetic only because */
-		/* the !@#&^*% C language doesn't give us access to */
-		/* the double-width-result multiplication operation */
-		/* that almost all CPUs provide! */
 		bool ccw =
 		  (double)(plp->width.x) *	/* x1 */
 		    (nplp->width.y) >		/* y2 */
@@ -1078,21 +1111,58 @@ line_join_points(const gx_line_params *pgs_lp, pl_ptr plp, pl_ptr nplp,
 		     * by using the following formula:
 		     * If tan(a)=u1/v1 and tan(b)=u2/v2, then
 		     * tan(a-b) = (u1*v2 - u2*v1) / (u1*u2 + v1*v2).
+		     *
 		     * We can do all the computations unscaled,
 		     * because we're only concerned with ratios.
+		     * However, if we have a non-uniform coordinate
+		     * system (indicated by pmat != 0), we must do the
+		     * computations in user space.
 		     */
-		    float u1 = plp->e.cdelta.y, v1 = plp->e.cdelta.x;
-		    float u2 = nplp->o.cdelta.y, v2 = nplp->o.cdelta.x;
-		    double num = u1 * v2 - u2 * v1;
-		    double denom = u1 * u2 + v1 * v2;
 		    float check = pgs_lp->miter_check;
+		    double u1 = plp->e.cdelta.y, v1 = plp->e.cdelta.x;
+		    double u2 = nplp->o.cdelta.y, v2 = nplp->o.cdelta.x;
+		    double num, denom;
+
+		    if ( pmat )
+		      { gs_point pt;
+
+		        gs_distance_transform_inverse(v1, u1, pmat, &pt);
+			v1 = pt.x, u1 = pt.y;
+		        gs_distance_transform_inverse(v2, u2, pmat, &pt);
+			v2 = pt.x, u2 = pt.y;
+			/*
+			 * We need to recompute ccw according to the
+			 * relative positions of the lines in user space.
+			 * We repeat the computation described above,
+			 * using the cdelta values instead of the widths.
+			 * Because the definition of ccw above is inverted
+			 * from the intuitive one (for historical reasons),
+			 * we actually have to do the test backwards.
+			 */
+			ccw = v1 * u2 < v2 * u1;
+#ifdef DEBUG
+			{ double a1 = atan2(u1, v1), a2 = atan2(u2, v2),
+			    dif = a1 - a2;
+
+			  if ( dif < 0 )
+			    dif += 2 * M_PI;
+			  else if ( dif >= 2 * M_PI )
+			    dif -= 2 * M_PI;
+			  if ( (dif < M_PI) != ccw )
+			    lprintf8("ccw wrong: tan(a1=%g)=%g/%g, tan(a2=%g)=%g,%g, dif=%g, ccw=%d\n",
+				     a1, u1, v1, a2, u2, v2, dif, ccw);
+			}
+#endif
+		      }
+		    num = u1 * v2 - u2 * v1;
+		    denom = u1 * u2 + v1 * v2;
 		    /*
 		     * We will want either tan(a-b) or tan(b-a)
 		     * depending on the orientations of the lines.
 		     * Fortunately we know the relative orientations already.
 		     */
 		    if ( !ccw )		/* have plp - nplp, want vice versa */
-			num = -num;
+		      num = -num;
 #ifdef DEBUG
 if ( gs_debug_c('O') )
                    {	dprintf4("[o]Miter check: u1/v1=%f/%f, u2/v2=%f/%f,\n",
