@@ -1,4 +1,4 @@
-/* Copyright (C) 1989, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -51,10 +51,11 @@
  *	maxlength - a t_integer whose value gives the client's view of
  *	the capacity (C).  C may be less than M (see below).
  *
- * C < M is possible because on large-memory systems, we round up M so that
- * M is a power of 2; this allows us to use masking rather than division
- * for computing the initial hash probe.  However, C is always the
- * maxlength specified by the client, so clients get a consistent story.
+ * C < M is possible because on large-memory systems, we usually round up M
+ * so that M is a power of 2 (see idict.h for details); this allows us to
+ * use masking rather than division for computing the initial hash probe.
+ * However, C is always the maxlength specified by the client, so clients
+ * get a consistent story.
  *
  * As noted above, the keys may be either in packed or unpacked form.
  * The markers for unused and deleted entries are different in the two forms.
@@ -74,8 +75,7 @@
  * for newly created dictionaries.
  *
  * Note that name keys with indices above packed_name_max_index require using
- * the unpacked form.
- */
+ * the unpacked form.  */
 #define dict_is_packed(dct) r_has_type(&(dct)->keys, t_shortarray)
 #define packed_key_empty (pt_tag(pt_integer) + 0)
 #define packed_key_deleted (pt_tag(pt_integer) + 1)
@@ -135,7 +135,7 @@ dict_find(const ref *pdref, const ref *pkey, ref **ppvalue)
 	if ( r_has_type(pkey, t_name) && dict_is_packed(pdict) )
 	{	uint nidx = name_index(pkey);
 		uint hash =
-		  hash_mod(dict_name_index_hash(nidx), npairs(pdict)) + 1;
+		  dict_hash_mod(dict_name_index_hash(nidx), npairs(pdict)) + 1;
 		if (  pdict->keys.value.packed[hash] ==
 		        pt_tag(pt_literal_name) + nidx
 		   )
@@ -161,7 +161,7 @@ dict_find_name_by_index(uint nidx)
 	dn_lookups++;
 	if ( dict_is_packed(pdict) )
 	{	uint hash =
-		  hash_mod(dict_name_index_hash(nidx), npairs(pdict)) + 1;
+		  dict_hash_mod(dict_name_index_hash(nidx), npairs(pdict)) + 1;
 		if (  pdict->keys.value.packed[hash] ==
 		        pt_tag(pt_literal_name) + nidx
 		   )
@@ -180,6 +180,23 @@ dict_find_name_by_index(uint nidx)
 #define dict_find_name_by_index real_dict_find_name_by_index
 #endif
 
+/* Round up the size of a dictionary.  Return 0 if too large. */
+uint
+dict_round_size_small(uint rsize)
+{	return (rsize > dict_max_size ? 0 : rsize);
+}
+uint
+dict_round_size_large(uint rsize)
+{	/* Round up to a power of 2 if not huge. */
+	/* If the addition overflows, the new rsize will be zero, */
+	/* which will (correctly) be interpreted as a limitcheck. */
+	if ( rsize > dict_max_non_huge )
+	  return (rsize > dict_max_size ? 0 : rsize);
+	while ( rsize & (rsize - 1) )
+	  rsize = (rsize | (rsize - 1)) + 1;
+	return (rsize <= dict_max_size ? rsize : dict_max_non_huge);
+}
+
 /* Create a dictionary in the current VM space. */
 int
 dict_create(uint size, ref *pdref)
@@ -223,13 +240,12 @@ dict_create_unpacked_keys(uint asize, const ref *pdref)
 private int
 dict_create_contents(uint size, const ref *pdref, bool pack)
 {	dict *pdict = pdref->value.pdict;
-	uint asize = (size == 0 ? 1 : size);
+	uint asize = dict_round_size((size == 0 ? 1 : size));
 	int code;
 	register uint i;
 
-	/* If appropriate, round up the actual allocated size to the next */
-	/* higher power of 2, so we can use & instead of %. */
-	dict_round_size(asize);
+	if ( asize == 0 || asize > max_array_size - 1 )	/* too large */
+	  return_error(e_limitcheck);
 	asize++;		/* allow room for wraparound entry */
 	code = ialloc_ref_array(&pdict->values, a_all, asize,
 				"dict_create(values)");
@@ -288,6 +304,8 @@ dict_unpack(ref *pdref)
 		    { packed_get(okp, nkp);
 		      ref_mark_new(nkp);
 		    }
+		  else if ( *okp == packed_key_deleted )
+		    r_set_attrs(nkp, a_executable);
 		if ( !ref_must_save(&old_keys) )
 		  ifree_ref_array(&old_keys, "dict_unpack(old keys)");
 		dict_set_top();	/* just in case */
@@ -307,11 +325,12 @@ dict_unpack(ref *pdref)
  * We break the macro into two to avoid overflowing some preprocessors.
  */
 /* packed_search_body also uses kp and kbot as free variables. */
-#define packed_search_body(del,pre,post,miss)\
+#define packed_search_value_pointer (pdict->values.value.refs + (kp - kbot))
+#define packed_search_body(found1,found2,del,miss)\
     { if_debug2('D', "[D]probe 0x%lx: 0x%x\n", (ulong)kp, *kp);\
       if ( *kp == kpack )\
-       { pre (pdict->values.value.refs + (kp - kbot));\
-	 post;\
+       { found1;\
+	 found2;\
        }\
       else if ( !r_packed_is_name(kp) )\
        { /* Empty, deleted, or wraparound. Figure out which. */\
@@ -320,14 +339,14 @@ dict_unpack(ref *pdref)
 	 else { del; }\
        }\
     }
-#define packed_search_1(del,pre,post,miss)\
+#define packed_search_1(found1,found2,del,miss)\
    const ref_packed *kbot = pdict->keys.value.packed;\
    register const ref_packed *kp;\
-   for ( kp = kbot + hash_mod(hash, size) + 1; ; kp-- )\
-     packed_search_body(del,pre,post,miss)
-#define packed_search_2(del,pre,post,miss)\
+   for ( kp = kbot + dict_hash_mod(hash, size) + 1; ; kp-- )\
+     packed_search_body(found1,found2,del,miss)
+#define packed_search_2(found1,found2,del,miss)\
    for ( kp += size; ; kp-- )\
-     packed_search_body(del,pre,post,miss)
+     packed_search_body(found1,found2,del,miss)
 
 /*
  * Look up a key in a dictionary.  Store a pointer to the value slot
@@ -383,10 +402,12 @@ nh:		hash = dict_name_index_hash(nidx);
 	/* Search the dictionary */
 	if ( dict_is_packed(pdict) )
 	{	const ref_packed *pslot = 0;
-		packed_search_1(if ( pslot == 0 ) pslot = kp,
-				*ppvalue =, return 1, goto miss);
-		packed_search_2(if ( pslot == 0 ) pslot = kp,
-				*ppvalue =, return 1, goto miss);
+		packed_search_1(*ppvalue = packed_search_value_pointer,
+				return 1,
+				if ( pslot == 0 ) pslot = kp, goto miss);
+		packed_search_2(*ppvalue = packed_search_value_pointer,
+				return 1,
+				if ( pslot == 0 ) pslot = kp, goto miss);
 		/*
 		 * Double wraparound, dict is full.
 		 * Note that even if there was an empty slot (pslot != 0),
@@ -409,7 +430,7 @@ miss:		/* Key is missing, not double wrap.  See above re dictfull. */
 		register ref *kp;
 		ref *pslot = 0;
 		int wrap = 0;
-		for ( kp = kbot + hash_mod(hash, size) + 2; ; )
+		for ( kp = kbot + dict_hash_mod(hash, size) + 2; ; )
 		{	--kp;
 			if ( (etype = r_type(kp)) == ktype )
 			{	/* Fast comparison if both keys are names */
@@ -512,9 +533,11 @@ dict_find_name_by_index(uint nidx)
 		}
 #endif
 		if ( dict_is_packed(pdict) )
-		   {	packed_search_1(DO_NOTHING, return,
+		   {	packed_search_1(DO_NOTHING,
+					return packed_search_value_pointer,
 					DO_NOTHING, goto miss);
-			packed_search_2(DO_NOTHING, return,
+			packed_search_2(DO_NOTHING,
+					return packed_search_value_pointer,
 					DO_NOTHING, break);
  miss:			;
 		   }
@@ -523,7 +546,7 @@ dict_find_name_by_index(uint nidx)
 			register ref *kp;
 			int wrap = 0;
 			/* Search the dictionary */
-			for ( kp = kbot + hash_mod(hash, size) + 2; ; )
+			for ( kp = kbot + dict_hash_mod(hash, size) + 2; ; )
 			   {	--kp;
 				if ( r_has_type(kp, t_name) )
 				   {	if ( name_index(kp) == nidx )
@@ -756,6 +779,12 @@ dict_maxlength(const ref *pdref	/* t_dictionary */)
 {	return d_maxlength(pdref->value.pdict);
 }
 
+/* Return the maximum index of a slot within a dictionary. */
+uint
+dict_max_index(const ref *pdref	/* t_dictionary */)
+{	return npairs(pdref->value.pdict) - 1;
+}
+
 /* Copy one dictionary into another. */
 /* If new_only is true, only copy entries whose keys */
 /* aren't already present in the destination. */
@@ -862,23 +891,33 @@ int
 dict_grow(ref *pdref)
 {	dict *pdict = pdref->value.pdict;
 	/* We might have maxlength < npairs, if */
-	/* dict_round_size is true. */
-	ulong new_size = d_maxlength(pdict) * 3 / 2 + 2;
-	if ( new_size > dict_max_size )
-	{	if ( d_maxlength(pdict) == dict_max_size )
-		  return_error(e_dictfull);
-		new_size = dict_max_size;
-	}
+	/* dict_round_size increased the size. */
+	ulong new_size = (ulong)d_maxlength(pdict) * 3 / 2 + 2;
+
+#if arch_sizeof_int < arch_sizeof_long
+	if ( new_size > max_uint )
+	  new_size = max_uint;
+#endif
 	if ( new_size > npairs(pdict) )
 	{	int code = dict_resize(pdref, (uint)new_size);
-		if ( code < 0 )
+		if ( code >= 0 )
 		  return code;
+		/* new_size was too big. */
+		if ( npairs(pdict) < dict_max_size )
+		  { code = dict_resize(pdref, dict_max_size);
+		    if ( code >= 0 )
+		      return code;
+		  }
+		if ( npairs(pdict) == d_maxlength(pdict) )
+		  { /* Can't do it. */
+		    return code;
+		  }
+		/* We can't grow to new_size, but we can grow to npairs. */
+		new_size = npairs(pdict);
 	}
-	else
-	{	/* maxlength < npairs, we can grow in place */
-		ref_save(pdref, &pdict->maxlength, "dict_put(maxlength)");
-		d_set_maxlength(pdict, new_size);
-	}
+	/* maxlength < npairs, we can grow in place */
+	ref_save(pdref, &pdict->maxlength, "dict_put(maxlength)");
+	d_set_maxlength(pdict, new_size);
 	return 0;
 }
 
