@@ -1,13 +1,24 @@
-/* Copyright (C) 1994, 1995 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1994, 1995, 1996 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* igcref.c */
@@ -15,6 +26,7 @@
 #include "memory_.h"
 #include "ghost.h"
 #include "gsexit.h"
+#include "gsstruct.h"			/* for gxalloc.h */
 #include "iname.h"
 #include "iastate.h"
 #include "idebug.h"
@@ -22,6 +34,13 @@
 #include "ipacked.h"
 #include "store.h"			/* for ref_assign_inline */
 
+/* Define whether to trace every step of relocating ref pointers. */
+#if 0
+#  define rputc(c) dputc(c)
+#else
+#  define rputc(c) DO_NOTHING
+#endif
+
 /*
  * Define the 'structure' type descriptor for refs.
  * This is special because it has different shared procs.
@@ -64,6 +83,27 @@ RELOC_PTRS_BEGIN(ref_struct_reloc_ptrs) {
 	ref_struct_clear_marks(vptr, size);
 } RELOC_PTRS_END
 
+/* Define a fast lookup table for ref types that use the size field. */
+private byte types_using_size[1 << r_type_bits];
+
+/* ------ Initialization ------ */
+void
+gs_igcref_init(gs_memory_t *mem)
+{	int i;
+	for ( i = 0; i < t_next_index; ++i )
+	  switch ( i )
+	    {
+	    default:
+	      types_using_size[i] = 0;
+	      break;
+	    case t_null:	/* see relocation planning phase below */
+	    case_types_with_size:
+	      types_using_size[i] = 1;
+	    }
+	memset(types_using_size + t_next_index, 1,
+	       countof(types_using_size) - t_next_index);
+}
+
 /* ------ Unmarking phase ------ */
 
 /* Unmark a single ref. */
@@ -78,17 +118,9 @@ ptr_ref_unmark(void *vptr, gc_state_t *ignored)
 /* Unmarking routine for ref objects. */
 private void
 refs_clear_marks(void /*obj_header_t*/ *vptr, uint size)
-{	gs_mark_refs((ref_packed *)vptr,
-		     (ref *)((byte *)vptr + size),
-		     false);
-}
-/* Mark or unmark a block of refs. */
-/* The last ref must be full-size, and is never marked. */
-void
-gs_mark_refs(ref_packed *from, ref *to, bool mark)
-{	ref_packed *rp = from;
-	ushort pmark = (mark ? lp_mark : 0);
-	ushort rmark = (mark ? l_mark : 0);
+{	ref_packed *rp = (ref_packed *)vptr;
+	ref_packed *end = (ref_packed *)((byte *)vptr + size);
+
 	/* Since the last ref is full-size, we only need to check for */
 	/* the end of the block when we see one of those. */
 	for ( ; ; )
@@ -101,7 +133,7 @@ if ( gs_debug_c('8') )
 			dprintf("\n");
 }
 #endif
-			r_store_pmark(rp, pmark);
+			r_clear_pmark(rp);
 			rp++;
 		}
 		else			/* full-size ref */
@@ -113,13 +145,10 @@ if ( gs_debug_c('8') )
 			dprintf("\n");
 }
 #endif
-			r_store_attrs((ref *)rp, l_mark, rmark);
+			r_clear_attrs((ref *)rp, l_mark);
 			rp += packed_per_ref;
-			if ( rp >= (ref_packed *)to )
-			{	/* Ensure the last ref is not marked. */
-				r_clear_attrs((ref *)rp - 1, l_mark);
-				break;
-			}
+			if ( rp >= (ref_packed *)end )
+			  break;
 		}
 	}
 }
@@ -161,20 +190,15 @@ refs_clear_reloc(obj_header_t *hdr, uint size)
 	ref_packed *end = (ref_packed *)((byte *)rp + size);
 	while ( rp < end )
 	{	if ( r_is_packed(rp) )
-			rp++;
+		  rp++;
 		else			/* full-size ref */
-		  {	uint type;
-			/* Store the relocation here if possible. */
-			switch ( type = r_type((ref *)rp) )
-			{	case_types_using_size:
-					break;
-				default:
-					if ( type >= t_next_index )
-					  break;
-					if_debug1('8', "  [8]clearing reloc at 0x%lx\n",
-						  (ulong)rp);
-					r_set_size((ref *)rp, 0);
-			}
+		  {	/* Store the relocation here if possible. */
+			if ( !types_using_size[r_type((ref *)rp)] )
+			  {	if_debug1('8',
+					  "  [8]clearing reloc at 0x%lx\n",
+					  (ulong)rp);
+				r_set_size((ref *)rp, 0);
+			  }
 			rp += packed_per_ref;
 		}
 	}
@@ -186,6 +210,7 @@ refs_set_reloc(obj_header_t *hdr, uint reloc, uint size)
 {	ref_packed *rp = (ref_packed *)(hdr + 1);
 	ref_packed *end = (ref_packed *)((byte *)rp + size);
 	uint freed = 0;
+
 	/*
 	 * We have to be careful to keep refs aligned properly.
 	 * For the moment, we do this by either keeping or discarding
@@ -199,33 +224,75 @@ refs_set_reloc(obj_header_t *hdr, uint reloc, uint size)
 		{
 #if align_packed_per_ref == 1
 			if ( r_has_pmark(rp) )
-			{	if_debug1('8',
+			  {	if_debug1('8',
 					  "  [8]packed ref 0x%lx is marked\n",
 					  (ulong)rp);
 				rp++;
-			}
+			  }
+			else
+			  {
 #else
-			uint marked = *rp;
 			int i;
+			/*
+			 * Note: align_packed_per_ref is typically
+			 * 2 or 4 for 32-bit processors.
+			 */
+#define all_marked (align_packed_per_ref * lp_mark)
+# if align_packed_per_ref == 2
+#  if arch_sizeof_long == arch_sizeof_short * 2
+#    undef all_marked
+#    define all_marked ( ((long)lp_mark << (sizeof(short) * 8)) + lp_mark )
+#    define marked (*(long *)rp & all_marked)
+#  else
+#    define marked ((*rp & lp_mark) + (rp[1] & lp_mark))
+#  endif
+# else
+#  if align_packed_per_ref == 4
+#    define marked ((*rp & lp_mark) + (rp[1] & lp_mark) +\
+		    (rp[2] & lp_mark) + (rp[3] & lp_mark))
+#  else
+			uint marked = *rp & lp_mark;
 			for ( i = 1; i < align_packed_per_ref; i++ )
-				marked |= rp[i];
-			if ( marked & lp_mark )
-			{	/* At least one packed_ref in the block */
+			  marked += rp[i] & lp_mark;
+#  endif
+# endif
+			/*
+			 * Now marked is lp_mark * the number of marked
+			 * packed refs in the aligned block, except for
+			 * a couple of special cases above.
+			 */
+			switch ( marked )
+			  {
+			  case all_marked:
+				if_debug2('8',
+					  "  [8]packed refs 0x%lx..0x%lx are marked\n",
+					  (ulong)rp,
+					  (ulong)(rp + (align_packed_per_ref - 1)));
+				rp += align_packed_per_ref;
+				break;
+			  default:
+				/* At least one packed ref in the block */
 				/* is marked: Keep the whole block. */
-				for ( ; i; i--, rp++ )
+				for ( i = align_packed_per_ref; i--; rp++ )
 				{	r_set_pmark(rp);
 					if_debug1('8',
 					  "  [8]packed ref 0x%lx is marked\n",
 					  (ulong)rp);
 				}
-			}
+				break;
+			  case 0:
 #endif
-			else
-			{	if_debug2('8', "  [8]%d packed ref(s) at 0x%lx are unmarked\n",
+				if_debug2('8', "  [8]%d packed ref(s) at 0x%lx are unmarked\n",
 					  align_packed_per_ref, (ulong)rp);
+				{ uint rel = reloc + freed;
+				  /* Change this to an integer so we can */
+				  /* store the relocation here. */
+				  *rp = pt_tag(pt_integer) +
+				    min(rel, packed_max_value);
+				}
 				rp += align_packed_per_ref;
 				freed += sizeof(ref_packed) * align_packed_per_ref;
-			}
+			  }
 		}
 		else			/* full-size ref */
 		{	uint rel = reloc + freed;
@@ -243,18 +310,11 @@ refs_set_reloc(obj_header_t *hdr, uint reloc, uint size)
 				freed += sizeof(ref);
 			}
 			else
-			{	uint type;
-				if_debug1('8', "  [8]ref 0x%lx is marked\n",
+			{	if_debug1('8', "  [8]ref 0x%lx is marked\n",
 					  (ulong)pref);
 				/* Store the relocation here if possible. */
-				switch ( type = r_type(pref) )
-				  {
-				  case_types_using_size:
-					break;
-				  default:
-					if ( type >= t_next_index )
-					  break;
-					if_debug2('8', "  [8]storing reloc %u at 0x%lx\n",
+				if ( !types_using_size[r_type(pref)] )
+				  {	if_debug2('8', "  [8]storing reloc %u at 0x%lx\n",
 						  rel, (ulong)pref);
 					r_set_size(pref, rel);
 				  }
@@ -298,16 +358,8 @@ refs_set_reloc(obj_header_t *hdr, uint reloc, uint size)
 				r_set_size(pref, reloc);
 			  }
 			else
-			  {	uint type;
-				switch ( type = r_type(pref) )
-				  {
-				  case_types_using_size:
-					break;
-				  default:
-					if ( type >= t_next_index )
-					  break;
-					r_set_size(pref, reloc);
-				  }
+			  {	if ( !types_using_size[r_type(pref)] )
+				  r_set_size(pref, reloc);
 			  }
 			rp += packed_per_ref;
 		  }
@@ -328,9 +380,14 @@ refs_do_reloc(void /*obj_header_t*/ *vptr, uint size, gc_state_t *gcst)
 		      gcst);
 }
 /* Relocate the contents of a block of refs. */
+/* If gcst->relocating_untraced is true, we are relocating pointers from an */
+/* untraced space, so relocate all refs, not just marked ones. */
 void
 gs_reloc_refs(ref_packed *from, ref_packed *to, gc_state_t *gcst)
-{	ref_packed *rp = from;
+{	int min_trace = gcst->min_collect;
+	ref_packed *rp = from;
+	bool do_all = gcst->relocating_untraced;
+
 	while ( rp < to )
 	{	ref *pref;
 		if ( r_is_packed(rp) )
@@ -343,7 +400,9 @@ gs_reloc_refs(ref_packed *from, ref_packed *to, gc_state_t *gcst)
 		if_debug3('8', "  [8]relocating %s %d ref at 0x%lx\n",
 			  (r_has_attr(pref, l_mark) ? "marked" : "unmarked"),
 			  r_btype(pref), (ulong)pref);
-		if ( r_has_attr(pref, l_mark) && !r_is_foreign(pref) )
+		if ( (r_has_attr(pref, l_mark) || do_all) &&
+		     r_space(pref) >= min_trace
+		   )
 		  switch ( r_type(pref) )
 		{
 		/* Struct cases */
@@ -361,19 +420,76 @@ gs_reloc_refs(ref_packed *from, ref_packed *to, gc_state_t *gcst)
 #undef ref_case
 		/* Non-trivial non-struct cases */
 		case t_dictionary:
+			rputc('d');
 			pref->value.pdict =
 			  (dict *)gs_reloc_ref_ptr((ref_packed *)pref->value.pdict, gcst);
 			break;
 		case t_array:
+		  {	uint size = r_size(pref);
+			if ( size != 0 ) /* value.refs might be NULL */
+			  { /*
+			     * If the array is large, we allocated it in its
+			     * own object (at least originally -- this might
+			     * be a pointer to a subarray.)  In this case,
+			     * we know it is the only object in its
+			     * containing st_refs object, so we know that
+			     * the mark containing the relocation appears
+			     * just after it.
+			     */
+			    if ( size < max_size_st_refs / sizeof(ref) )
+			      { rputc('a');
+			        pref->value.refs =
+				  (ref *)gs_reloc_ref_ptr(
+				    (ref_packed *)pref->value.refs, gcst);
+			      }
+			    else
+			      { rputc('A');
+			        /*
+				 * See the t_shortarray case below for why we
+				 * decrement size.
+				 */
+			        --size;
+			        pref->value.refs =
+				  (ref *)gs_reloc_ref_ptr(
+				    (ref_packed *)(pref->value.refs + size),
+				    gcst) - size;
+			      }
+			  }
+		  }	break;
+		case t_mixedarray:
 			if ( r_size(pref) != 0 ) /* value.refs might be NULL */
-			  pref->value.refs =
-			    (ref *)gs_reloc_ref_ptr((ref_packed *)pref->value.refs, gcst);
-			break;
-		case t_mixedarray: case t_shortarray:
-			if ( r_size(pref) != 0 ) /* value.refs might be NULL */
-			  pref->value.packed =
-			    (const ref_packed *)gs_reloc_ref_ptr((ref_packed *)pref->value.packed, gcst); /* discard const */
+			  { rputc('m');
+			    pref->value.packed =
+			      gs_reloc_ref_ptr(pref->value.packed, gcst);
+			  }
 			break;
+		case t_shortarray:
+		  {	uint size = r_size(pref);
+			/*
+			 * Since we know that gs_reloc_ref_ptr works by
+			 * scanning forward, and we know that all the
+			 * elements of this array itself are marked, we can
+			 * save some scanning time by relocating the pointer
+			 * to the end of the array rather than the
+			 * beginning.
+			 */
+			if ( size != 0 ) /* value.refs might be NULL */
+			  { rputc('s');
+			    /*
+			     * gs_reloc_ref_ptr has to be able to determine
+			     * whether the pointer points into a space that
+			     * isn't being collected.  It does this by
+			     * checking whether the referent of the pointer
+			     * is marked.  For this reason, we have to pass
+			     * a pointer to the last real element of the
+			     * array, rather than just beyond it.
+			     */
+			    --size;
+			    pref->value.packed =
+			      gs_reloc_ref_ptr(pref->value.packed + size,
+					       gcst) - size;
+			  }
+		  }	break;
 		case t_name:
 		  {	void *psub = name_ref_sub_table(pref);
 			void *rsub = gs_reloc_struct_ptr(psub, gcst);
@@ -388,8 +504,9 @@ gs_reloc_refs(ref_packed *from, ref_packed *to, gc_state_t *gcst)
 			pref->value.bytes = str.data;
 		  }	break;
 		case t_oparray:
+			rputc('o');
 			pref->value.const_refs =
-			  (const ref *)gs_reloc_ref_ptr((ref_packed *)pref->value.const_refs, gcst);
+			  (const ref *)gs_reloc_ref_ptr((const ref_packed *)pref->value.const_refs, gcst);
 			break;
 		}
 		rp += packed_per_ref;
@@ -397,49 +514,72 @@ gs_reloc_refs(ref_packed *from, ref_packed *to, gc_state_t *gcst)
 }
 
 /* Relocate a pointer to a ref. */
+/* See gsmemory.h for why the argument is const and the result is not. */
 ref_packed *
-gs_reloc_ref_ptr(ref_packed *prp, gc_state_t *ignored)
-{	/* Search forward for relocation. */
-	/* This algorithm is intrinsically very inefficient; */
-	/* we hope eventually to replace it with a better one. */
-	register ref_packed *rp = prp;
+gs_reloc_ref_ptr(const ref_packed *prp, gc_state_t *ignored)
+{	/*
+	 * Search forward for relocation.  This algorithm is intrinsically
+	 * very inefficient; we hope eventually to replace it with a better
+	 * one.
+	 */
+	register const ref_packed *rp = prp;
 	uint dec = 0;
+
+	/*
+	 * Iff this pointer points into a space that wasn't traced,
+	 * the referent won't be marked.  In this case, we shouldn't
+	 * do any relocation.  Check for this first.
+	 */
+	if ( r_is_packed(rp) )
+	  { if ( !r_has_pmark(rp) )
+	      return (ref_packed *)rp;
+	  }
+	else
+	  { if ( !r_has_attr((const ref *)rp, l_mark) )
+	      return (ref_packed *)rp;
+	  }
 	for ( ; ; )
-	{	uint type;
-		if ( r_is_packed(rp) )
-		{	/* For each unmarked packed ref we pass over, */
-			/* we have to decrement the final relocation. */
-			if ( r_is_packed(rp + 1) )
-			{	/* Almost all packed refs are marked, */
-				/* so test both at the same time. */
-				if ( !(*rp & rp[1] & lp_mark) )
-				{	if ( (*rp | rp[1]) & lp_mark )
-						dec += sizeof(ref_packed);
-					else
-						dec += sizeof(ref_packed) * 2;
-				}
-				rp += 2;
-				continue;
-			}
-			else if ( !r_has_pmark(rp) )
-				dec += sizeof(ref_packed);
-			rp++;		/* fall through */
-		}
-		switch ( type = r_type((ref *)rp) )
-		{
-		default:		/* reloc is in r_size */
-			if ( type < t_next_index )
-			  {	/* These refs might be in a space */
-				/* that isn't being compacted.  If so, */
-				/* the relocation value here will be zero. */
+	{	if ( r_is_packed(rp) )
+		{	/*
+			 * Normally, an unmarked packed ref will be an
+			 * integer whose value is the amount of relocation.
+			 * However, the relocation value might have been
+			 * too large to fit.  If this is the case, for
+			 * each such unmarked packed ref we pass over,
+			 * we have to decrement the final relocation.
+			 */
+			rputc((*rp & lp_mark ? '1' : '0'));
+			if ( !(*rp & lp_mark) )
+			  { if ( *rp != pt_tag(pt_integer) + packed_max_value )
+			      { /* This is a stored relocation value. */
+				rputc('\n');
 				return print_reloc(prp, "ref",
-					(r_size((ref *)rp) == 0 ? prp :
-					 (ref_packed *)((char *)prp - r_size((ref *)rp) + dec)));
+				  (ref_packed *)
+				    ((const char *)prp -
+				     (*rp & packed_value_mask) + dec));
+			      }
+			    /*
+			     * We know this is the first of an aligned block
+			     * of packed refs.  Skip over the entire block,
+			     * decrementing the final relocation.
+			     */
+			    dec += sizeof(ref_packed) * align_packed_per_ref;
+			    rp += align_packed_per_ref;
 			  }
-			/* falls through */
-		case_types_using_size:
-			rp += packed_per_ref;
+			else
+			  rp++;
+			continue;
 		}
+		if ( !types_using_size[r_type((const ref *)rp)] )
+		  {	/* reloc is in r_size */
+			rputc('\n');
+			return print_reloc(prp, "ref",
+				(ref_packed *)
+				  (r_size((const ref *)rp) == 0 ? prp :
+				   (const ref_packed *)((const char *)prp - r_size((const ref *)rp) + dec)));
+		  }
+		rputc('u');
+		rp += packed_per_ref;
 	}
 }
 
