@@ -1,18 +1,30 @@
-/* Copyright (C) 1989, 1995 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* interp.c */
 /* Ghostscript language interpreter */
 #include "memory_.h"
+#include "string_.h"
 #include "ghost.h"
 #include "gsstruct.h"			/* for iastruct.h */
 #include "stream.h"
@@ -20,7 +32,7 @@
 #include "estack.h"
 #include "ialloc.h"
 #include "iastruct.h"
-#include "iname.h"
+#include "inamedef.h"
 #include "interp.h"
 #include "ipacked.h"
 #include "ostack.h"		/* must precede iscan.h */
@@ -37,14 +49,28 @@
 #include "oper.h"
 #include "store.h"
 
+/*
+ * We may or may not optimize the handling of the special fast operators
+ * in packed arrays.  If we do this, they run much faster when packed, but
+ * slightly slower when not packed.
+ */
 #define PACKED_SPECIAL_OPS 1
 
+/*
+ * Pseudo-operators (procedures of type t_oparray) record
+ * the operand and dictionary stack pointers, and restore them if an error
+ * occurs during the execution of the procedure and if the procedure hasn't
+ * (net) decreased the depth of the stack.  While this obviously doesn't
+ * do all the work of restoring the state if a pseudo-operator gets an
+ * error, it's a big help.  The only downside is that pseudo-operators run
+ * a little slower.
+ */
+
 /* Imported operator procedures */
 extern int zop_add(P1(os_ptr));
 extern int zop_def(P1(os_ptr));
 extern int zop_sub(P1(os_ptr));
 /* Other imported procedures */
-extern int in_stopped(P0());
 extern int ztokenexec_continue(P1(os_ptr));
 
 /* 
@@ -72,49 +98,79 @@ int (*gs_interp_time_slice_proc)(P0()) = no_time_slice_proc;
  */
 int gs_interp_time_slice_ticks = 0x7fff;
 
+/*
+ * Apply an operator.  We route all operator calls through a procedure when
+ * debugging.
+ */
+#ifdef DEBUG
+private int
+call_operator(int (*op_proc)(P1(os_ptr)), os_ptr op)
+{	int code = (*op_proc)(op);
+	return code;
+}
+#else
+#  define call_operator(proc, op) ((*(proc))(op))
+#endif
+
 /* Forward references */
 private int estack_underflow(P1(os_ptr));
 private int interp(P2(ref *, ref *));
 private int interp_exit(P1(os_ptr));
 private void set_gc_signal(P2(int *, int));
 private int copy_stack(P2(const ref_stack *, ref *));
+private int oparray_pop(P1(os_ptr));
+private int oparray_cleanup(P1(os_ptr));
 
 /* Stack sizes */
 
 /* The maximum stack sizes may all be set in the makefile. */
 
 /*
- * The minimum value of block_ostack is the larger of:
+ * Define the initial maximum size of the operand stack (MaxOpStack
+ * user parameter).  Currently this is also the block size for extending
+ * the operand stack, but this isn't guaranteed in the future.
+ */
+#ifndef MAX_OSTACK
+#  define MAX_OSTACK 800
+#endif
+/*
+ * The minimum block size for extending the operand stack is the larger of:
  *	- the maximum number of parameters to an operator
  *	(currently setcolorscreen, with 12 parameters);
  *	- the maximum number of values pushed by an operator
  *	(currently setcolortransfer, which calls zcolor_remap_one 4 times
  *	and therefore pushes 16 values).
  */
-#ifndef MAX_OSTACK
-#  define MAX_OSTACK 800
-#endif
 #define min_block_ostack 16
 const int gs_interp_max_op_num_args = min_block_ostack;	/* for iinit.c */
 
 /*
- * The minimum value of block_estack is the largest size of a contiguous
- * block surrounding an e_stack mark, currently ???.  At least, that's
- * what the minimum value would be if we supported multi-block estacks,
- * which we currently don't.
+ * Define the initial maximum size of the execution stack (MaxExecStack
+ * user parameter).
  */
 #ifndef MAX_ESTACK
 #  define MAX_ESTACK 250
 #endif
+/*
+ * The minimum block size for extending the execution stack is the largest
+ * size of a contiguous block surrounding an e-stack mark, currently ???.
+ * At least, that's what the minimum value would be if we supported
+ * multi-block estacks, which we currently don't.
+ */
 #define min_block_estack MAX_ESTACK
 
 /*
- * The minimum value of block_dstack is the number of permanent entries
- * on the dictionary stack, currently 3.
+ * Define the initial maximum size of the dictionary stack (MaxDictStack
+ * user parameter).  Again, this is also currently the block size for
+ * extending the d-stack.
  */
 #ifndef MAX_DSTACK
 #  define MAX_DSTACK 20
 #endif
+/*
+ * The minimum block size for extending the dictionary stack is the number
+ * of permanent entries on the dictionary stack, currently 3.
+ */
 #define min_block_dstack 3
 uint min_dstack_size;		/* set by iinit.c */
 
@@ -199,6 +255,7 @@ const int tx_next_index = tx_next_op;
 static ref null_proc;
 
 /* Initialize the interpreter */
+#ifndef DPNEXT
 void
 gs_interp_init(void)
 {	/* Initialize the stacks. */
@@ -261,6 +318,86 @@ gs_interp_init(void)
 	d_stack.overflow_error = e_dictstackoverflow;
 	ref_stack_set_max_count(&d_stack, MAX_DSTACK);
 }
+#else				/* DPNEXT */
+void
+gs_interp_init(void)
+{	/* Initialize the stacks. */
+	gs_ref_memory_t *smem = iimemory_system;
+	int code =
+	  gs_interp_create_stacks(smem, &ref_ref_stacks[0],
+				  &ref_ref_stacks[1], &ref_ref_stacks[2]);
+	/****** FATAL ERROR IF code < 0 ******/
+
+	d_stack = *r_ptr(&ref_ref_stacks[0], ref_stack);
+	e_stack = *r_ptr(&ref_ref_stacks[1], ref_stack);
+	o_stack = *r_ptr(&ref_ref_stacks[2], ref_stack);
+	esfile_clear_cache();
+}
+int
+gs_interp_create_stacks(gs_ref_memory_t *smem, ref *prds, ref *pres,
+  ref *pros)
+{	ref sdata, stk;
+
+#define refs_size_ostack os_refs_size(MAX_OSTACK)
+#define refs_size_estack es_refs_size(MAX_ESTACK)
+#define refs_size_dstack ds_refs_size(MAX_DSTACK)
+	gs_alloc_ref_array(smem, &sdata, 0,
+			   refs_size_ostack + refs_size_estack +
+			   refs_size_dstack, "gs_interp_create_stacks");
+
+	
+	make_struct(pros, avm_system,
+		    gs_alloc_struct((gs_memory_t *)smem, ref_stack,
+				    &st_ref_stack,
+				    "gs_interp_create_stacks(ostack)"));
+	make_array(&stk, avm_system, refs_size_ostack, sdata.value.refs);
+	{ ref_stack *pos = r_ptr(pros, ref_stack);
+	  ref_stack_init(pos, &stk, os_guard_under, os_guard_over, NULL,
+			 smem);
+	  pos->underflow_error = e_stackunderflow;
+	  pos->overflow_error = e_stackoverflow;
+	  ref_stack_set_max_count(pos, MAX_OSTACK);
+	}
+
+	make_struct(pres, avm_system,
+		    gs_alloc_struct((gs_memory_t *)smem, ref_stack,
+				    &st_ref_stack,
+				    "gs_interp_create_stacks(estack)"));
+	stk.value.refs += refs_size_ostack;
+	r_set_size(&stk, refs_size_estack);
+	{ ref_stack *pes = r_ptr(pres, ref_stack);
+	  ref euop;
+
+	  make_oper(&euop, 0, estack_underflow);
+	  ref_stack_init(pes, &stk, es_guard_under, es_guard_over, &euop,
+			 smem);
+	  pes->underflow_error = e_ExecStackUnderflow;
+	  pes->overflow_error = e_execstackoverflow;
+	  /**************** E-STACK EXPANSION IS NYI. ****************/
+	  pes->allow_expansion = false;
+	  ref_stack_set_max_count(pes, MAX_ESTACK);
+	}
+
+	make_struct(prds, avm_system,
+		    gs_alloc_struct((gs_memory_t *)smem, ref_stack,
+				    &st_ref_stack,
+				    "gs_interp_create_stacks(dstack)"));
+	stk.value.refs += refs_size_estack;
+	r_set_size(&stk, refs_size_dstack);
+	{ ref_stack *pds = r_ptr(prds, ref_stack);
+
+	  ref_stack_init(pds, &stk, 0, 0, NULL, smem);
+	  pds->underflow_error = e_dictstackunderflow;
+	  pds->overflow_error = e_dictstackoverflow;
+	  ref_stack_set_max_count(pds, MAX_DSTACK);
+	}
+
+#undef refs_size_ostack
+#undef refs_size_estack
+#undef refs_size_dstack
+	return 0;
+}
+#endif
 void
 gs_interp_reset(void)
 {	/* Reset the stacks. */
@@ -315,6 +452,8 @@ gs_interpret(ref *pref, int user_errors, int *pexit_code, ref *perror_object)
 	make_null(perror_object);
 	code = gs_call_interp(pref, user_errors, pexit_code, perror_object);
 	gs_unregister_root(imemory_system, &error_root, "gs_interpret");
+	/* Avoid a dangling reference to a stack-allocated GC signal. */
+	set_gc_signal(NULL, 0);
 	return code;
 }
 private int
@@ -332,8 +471,17 @@ gs_call_interp(ref *pref, int user_errors, int *pexit_code, ref *perror_object)
 again:	o_stack.requested = e_stack.requested = d_stack.requested = 0;
 	while ( gc_signal )
 	  {	/* Some routine below triggered a GC. */
+		gs_gc_root_t epref_root;
+
 		gc_signal = 0;
+		/* Make sure that doref will get relocated properly if */
+		/* a garbage collection happens with epref == &doref. */
+		gs_register_ref_root(imemory_system, &epref_root,
+				     (void **)&epref,
+				     "gs_call_interpret(epref)");
 		code = (*idmemory->reclaim)(idmemory, -1);
+		gs_unregister_root(imemory_system, &epref_root,
+				   "gs_call_interpret(epref)");
 		if ( code < 0 )
 		  return code;
 	  }
@@ -353,12 +501,9 @@ again:	o_stack.requested = e_stack.requested = d_stack.requested = 0;
 		*perror_object = osp[-1];
 		*pexit_code = code = osp->value.intval;
 		osp -= 2;
-		if ( code == 0 )
-		  return e_Quit;
-		else if ( code > 0 )
-		  return e_Fatal;
-		else
-		  return code;
+		return
+		  (code == 0 ? e_Quit :
+		   code < 0 && code > -100 ? code : e_Fatal);
 	case e_InterpreterExit:
 		return 0;
 	case e_ExecStackUnderflow:
@@ -376,6 +521,8 @@ again:	o_stack.requested = e_stack.requested = d_stack.requested = 0;
 		make_oper(&doref, 0, zpop);
 		epref = &doref;
 		goto again;
+	case e_NeedInput:
+		return code;
 	}
 	/* Adjust osp in case of operand stack underflow */
 	if ( osp < osbot - 1 )
@@ -458,7 +605,7 @@ again:	o_stack.requested = e_stack.requested = d_stack.requested = 0;
 	if ( user_errors < 0 )
 	  return code;
 	if ( gs_errorname(code, &error_name) < 0 )
-	  return code;		/* something is very wrong! */
+	  return code;		/* out-of-range error code! */
 	if ( dict_find_string(systemdict, "errordict", &perrordict) <= 0 ||
 	     dict_find(perrordict, &error_name, &epref) <= 0
 	   )
@@ -540,14 +687,18 @@ gs_errorinfo_put_string(const char *str)
 /* and return a (negative) error code. */
 private int
 interp(ref *pref /* object to interpret */, ref *perror_object)
-{	register ref *iref = pref;
+{	/*
+	 * Note that iref is declared as a ref *, but it may actually be
+	 * a ref_packed *.
+	 */
+	register const ref *iref = pref;
 	register int icount = 0;	/* # of consecutive tokens at iref */
 	register os_ptr iosp = osp;	/* private copy of osp */
 	register es_ptr iesp = esp;	/* private copy of esp */
 	int code;
 	ref token;		/* token read from file or string, */
 				/* must be declared in this scope */
-	register ref *pvalue;
+	register const ref *pvalue;
 	os_ptr whichp;
 	/*
 	 * We have to make the error information into a struct;
@@ -560,7 +711,7 @@ interp(ref *pref /* object to interpret */, ref *perror_object)
 	struct interp_error_s {
 		int code;
 		int line;
-		ref *obj;
+		const ref *obj;
 		ref full;
 	} ierror;
 	/*
@@ -578,6 +729,10 @@ interp(ref *pref /* object to interpret */, ref *perror_object)
   { ierror.line = __LINE__; goto rweci; }
 #define return_with_error_code_op(nargs)\
   return_with_code_iref()
+#define return_with_stackoverflow(objp)\
+  { o_stack.requested = 1; return_with_error(e_stackoverflow, objp); }
+#define return_with_stackoverflow_iref()\
+  { o_stack.requested = 1; return_with_error_iref(e_stackoverflow); }
 	int ticks_left = gs_interp_time_slice_ticks;
 
 	/*
@@ -592,14 +747,14 @@ interp(ref *pref /* object to interpret */, ref *perror_object)
 	 * to the top entry on the execution stack: icount is the count
 	 * of sequential entries remaining AFTER the current one.
 	 */
-#define add1_short(pref) (ref *)((ushort *)(pref) + 1)
+#define add1_short(pref) (const ref *)((const ushort *)(pref) + 1)
 #define add1_either(pref) (r_is_packed(pref) ? add1_short(pref) : (pref) + 1)
 #define store_state(ep)\
-  ( icount > 0 ? (ep->value.refs = iref + 1, r_set_size(ep, icount)) : 0 )
+  ( icount > 0 ? (ep->value.const_refs = iref + 1, r_set_size(ep, icount)) : 0 )
 #define store_state_short(ep)\
-  ( icount > 0 ? (ep->value.refs = add1_short(iref), r_set_size(ep, icount)) : 0 )
+  ( icount > 0 ? (ep->value.const_refs = add1_short(iref), r_set_size(ep, icount)) : 0 )
 #define store_state_either(ep)\
-  ( icount > 0 ? (ep->value.refs = add1_either(iref), r_set_size(ep, icount)) : 0 )
+  ( icount > 0 ? (ep->value.const_refs = add1_either(iref), r_set_size(ep, icount)) : 0 )
 #define next()\
   if ( --icount > 0 ) { iref++; goto top; } else goto out
 #define next_short()\
@@ -633,17 +788,19 @@ top:	/*
 	 */
 #ifdef DEBUG
 	/* Do a little validation on the top o-stack entry. */
-	if ( iosp >= osbot && r_type(iosp) == t__invalid )
+	if ( iosp >= osbot &&
+	     (r_type(iosp) == t__invalid || r_type(iosp) >= tx_next_op)
+	   )
 	  {	lprintf("Invalid value on o-stack!\n");
 		return_with_error_iref(e_Fatal);
 	  }
 if ( gs_debug['I'] ||
      (gs_debug['i'] &&
       (r_is_packed(iref) ?
-       r_packed_is_name((ref_packed *)iref) :
+       r_packed_is_name((const ref_packed *)iref) :
        r_has_type(iref, t_name)))
    )
-   {	void debug_print_ref(P1(ref *));
+   {	void debug_print_ref(P1(const ref *));
 	os_ptr save_osp = osp;		/* avoid side-effects */
 	es_ptr save_esp = esp;
 	int edepth;
@@ -750,6 +907,8 @@ x_def:		if ( (code = zop_def(iosp)) < 0 )
 	case plain_exec(tx_op_dup):
 x_dup:		if ( iosp < osbot )
 		  return_with_error_iref(e_stackunderflow);
+		if ( iosp >= ostop )
+		  return_with_stackoverflow_iref();
 		iosp++;
 		ref_assign_inline(iosp, iosp - 1);
 		next_either();
@@ -764,6 +923,8 @@ x_exch:		if ( iosp <= osbot )
 x_if:		if ( !r_has_type(iosp - 1, t_boolean) )
 		  return_with_error_iref((iosp <= osbot ?
 					  e_stackunderflow : e_typecheck));
+		if ( !r_is_proc(iosp) )
+		  return_with_error_iref(check_proc_failed(iosp));
 		if ( !iosp[-1].value.boolval )
 		  { iosp -= 2;
 		    next_either();
@@ -778,24 +939,17 @@ x_if:		if ( !r_has_type(iosp - 1, t_boolean) )
 x_ifelse:	if ( !r_has_type(iosp - 2, t_boolean) )
 		  return_with_error_iref((iosp < osbot + 2 ?
 					  e_stackunderflow : e_typecheck));
+		if ( !r_is_proc(iosp - 1) )
+		  return_with_error_iref(check_proc_failed(iosp - 1));
+		if ( !r_is_proc(iosp) )
+		  return_with_error_iref(check_proc_failed(iosp));
 		if ( iesp >= estop )
 		  return_with_error_iref(e_execstackoverflow);
 		store_state_either(iesp);
 		whichp = (iosp[-2].value.boolval ? iosp - 1 : iosp);
 		iosp -= 3;
 		/* Open code "up" for the array case(s) */
-ifup:		if ( !r_is_proc(whichp) )
-		   {	/* This is an unusual enough case that we go ahead */
-			/* and clear the currentfile cache without */
-			/* checking whether we have an exec(t_file), */
-			/* just to avoid another case test. */
-			esfile_clear_cache();
-			ref_assign_inline(iesp + 1, whichp);
-			iref = iesp + 1;
-			icount = 0;
-			goto top;
-		   }
-		if ( (icount = r_size(whichp) - 1) <= 0 )
+ifup:		if ( (icount = r_size(whichp) - 1) <= 0 )
 		   {	if ( icount < 0 ) goto up;	/* 0-element proc */
 			iref = whichp->value.refs;	/* 1-element proc */
 			if ( --ticks_left > 0 )
@@ -834,7 +988,19 @@ x_sub:		if ( (code = zop_sub(iosp)) < 0 )
 		goto bot;
 	case plain_exec(t_oparray):
 		/* Replace with the definition and go again. */
-		pvalue = (ref *)iref->value.const_refs;	/* discard const */
+		pvalue = (const ref *)iref->value.const_refs;
+opst:		/* Prepare to call a t_oparray procedure in *pvalue. */
+		store_state(iesp);
+oppr:		/* Record the stack depths in case of failure. */
+		if ( iesp >= estop - 3 )
+		  return_with_error_iref(e_execstackoverflow);
+		iesp += 4;
+		osp = iosp;	/* ref_stack_count_inline needs this */
+		make_mark_estack(iesp - 3, es_other, oparray_cleanup);
+		make_int(iesp - 2, ref_stack_count_inline(&o_stack));
+		make_int(iesp - 1, ref_stack_count_inline(&d_stack));
+		make_op_estack(iesp, oparray_pop);
+		goto pr;
 prst:		/* Prepare to call the procedure (array) in *pvalue. */
 		store_state(iesp);
 pr:		/* Call the array in *pvalue.  State has been stored. */
@@ -875,7 +1041,7 @@ pr:		/* Call the array in *pvalue.  State has been stored. */
 		/* Note that each case must set iosp = osp: */
 		/* this is so we can switch on code without having to */
 		/* store it and reload it (for dumb compilers). */
-		switch ( code = (*real_opproc(iref))(iosp) )
+		switch ( code = call_operator(real_opproc(iref), iosp) )
 		   {
 		case 0:			/* normal case */
 		case 1:			/* alternative success case */
@@ -929,7 +1095,7 @@ oeinsert:		ref_assign_inline(iesp + 1, iref);
 		cases_lit_5():
 			/* Just push the value */
 			if ( iosp >= ostop )
-			  return_with_error(e_stackoverflow, pvalue);
+			  return_with_stackoverflow(pvalue);
 			++iosp;
 			ref_assign_inline(iosp, pvalue);
 			next();
@@ -951,8 +1117,8 @@ oeinsert:		ref_assign_inline(iesp + 1, iref);
 		case plain_exec(t_null):
 			goto bot;
 		case plain_exec(t_oparray):
-			pvalue = (ref *)pvalue->value.const_refs; /* discard const */
-			goto prst;
+			pvalue = (const ref *)pvalue->value.const_refs;
+			goto opst;
 		case plain_exec(t_operator):
 		   {	/* Shortcut for operators. */
 			/* See above for the logic. */
@@ -963,7 +1129,7 @@ oeinsert:		ref_assign_inline(iesp + 1, iref);
 			  }
 			esp = iesp;
 			osp = iosp;
-			switch ( code = (*real_opproc(pvalue))(iosp) )
+			switch ( code = call_operator(real_opproc(pvalue), iosp) )
 			   {
 			case 0:			/* normal case */
 			case 1:			/* alternative success case */
@@ -999,10 +1165,9 @@ oeinsert:		ref_assign_inline(iesp + 1, iref);
 	   {	/* Executable file.  Read the next token and interpret it. */
 	   	stream *s;
 		scanner_state sstate;
-		check_read_known_file(s, iref, return_with_error_iref,
-				      s = invalid_file_entry);
+		check_read_known_file(s, iref, return_with_error_iref);
 rt:		if ( iosp >= ostop )	/* check early */
-		  return_with_error_iref(e_stackoverflow);
+		  return_with_stackoverflow_iref();
 		osp = iosp;		/* scan_token uses ostack */
 		scanner_state_init(&sstate, false);
 again:		code = scan_token(s, &token, &sstate);
@@ -1061,7 +1226,7 @@ again:		code = scan_token(s, &token, &sstate);
 			ref_assign_inline(iesp, &token);
 			esp = iesp;
 			osp = iosp;
-			code = scan_handle_refill(&token, &sstate, true,
+			code = scan_handle_refill(&token, &sstate, true, true,
 						  ztokenexec_continue);
 			iosp = osp;
 			iesp = esp;
@@ -1071,12 +1236,7 @@ again:		code = scan_token(s, &token, &sstate);
 				iesp--;		/* don't push the file */
 				goto again;	/* stacks are unchanged */
 			  case o_push_estack:
-				/* See above for why it is safe to push */
-				/* the file onto the o-stack */
-				/* (for ztokenexec_continue). */
 				esfile_clear_cache();
-				iosp++;
-				ref_assign_inline(iosp, &token);
 				if ( --ticks_left > 0 )
 					goto up;
 				goto slice;
@@ -1101,13 +1261,18 @@ again:		code = scan_token(s, &token, &sstate);
 		case 0:			/* read a token */
 		case scan_BOS:		/* binary object sequence */
 		    store_state(iesp);
-		    /* Push the updated string back on the e-stack */
-		    if ( iesp >= estop )
-		      return_with_error_iref(e_execstackoverflow);
-		    ++iesp;
-		    iesp->tas.type_attrs = iref->tas.type_attrs;
-		    iesp->value.const_bytes = sbufptr(&ss);
-		    r_set_size(iesp, sbufavailable(&ss));
+		    /* If the updated string isn't empty, push it back */
+		    /* on the e-stack. */
+		    { uint size = sbufavailable(&ss);
+		      if ( size )
+			{ if ( iesp >= estop )
+			    return_with_error_iref(e_execstackoverflow);
+			  ++iesp;
+			  iesp->tas.type_attrs = iref->tas.type_attrs;
+			  iesp->value.const_bytes = sbufptr(&ss);
+			  r_set_size(iesp, size);
+			}
+		    }
 		    if ( code == 0 )
 		    {	iref = &token;
 			icount = 0;
@@ -1128,12 +1293,12 @@ again:		code = scan_token(s, &token, &sstate);
 	/* This also picks up some anomalous cases of non-packed arrays. */
 	default:
 	  {	uint index;
-		switch ( *(ushort *)iref >> r_packed_type_shift )
+		switch ( *(const ushort *)iref >> r_packed_type_shift )
 		{
 		case pt_full_ref:
 		case pt_full_ref+1:
 			if ( iosp >= ostop )
-			  return_with_error_iref(e_stackoverflow);
+			  return_with_stackoverflow_iref();
 			/* We know this can't be an executable object */
 			/* requiring special handling, so we just push it. */
 			++iosp;
@@ -1142,7 +1307,7 @@ again:		code = scan_token(s, &token, &sstate);
 			ref_assign_inline(iosp, iref);
 			next();
 		case pt_executable_operator:
-			index = *(ushort *)iref & packed_value_mask;
+			index = *(const ushort *)iref & packed_value_mask;
 			if ( --ticks_left <= 0 )
 			  {	/* The following doesn't work, */
 				/* and I can't figure out why. */
@@ -1152,13 +1317,13 @@ again:		code = scan_token(s, &token, &sstate);
 			{	store_state_short(iesp);
 				/* Call the operator procedure. */
 				index -= op_def_count;
-				pvalue = (ref *)	/* discard const */
+				pvalue = (const ref *)
 				  (index < r_size(&op_array_table_global.table) ?
 				   op_array_table_global.table.value.const_refs +
 				     index :
 				   op_array_table_local.table.value.const_refs +
 				     (index - r_size(&op_array_table_global.table)));
-				goto pr;
+				goto oppr;
 			}
 			/* See the main plain_exec(t_operator) case */
 			/* for details of what happens here. */
@@ -1186,7 +1351,7 @@ again:		code = scan_token(s, &token, &sstate);
 #endif
 			esp = iesp;
 			osp = iosp;
-			switch ( code = (*op_index_proc(index))(iosp) )
+			switch ( code = call_operator(op_index_proc(index), iosp) )
 			   {
 			case 0:
 			case 1:
@@ -1217,21 +1382,23 @@ again:		code = scan_token(s, &token, &sstate);
 			return_with_code_iref();
 		case pt_integer:
 			if ( iosp >= ostop )
-			  return_with_error_iref(e_stackoverflow);
+			  return_with_stackoverflow_iref();
 			++iosp;
-			make_int(iosp, (*(short *)iref & packed_int_mask) +
-					packed_min_intval);
+			make_int(iosp,
+				 (*(const short *)iref & packed_int_mask) +
+				 packed_min_intval);
 			next_short();
 		case pt_literal_name:
-		  {	uint nidx = *(ushort *)iref & packed_value_mask;
+		  {	uint nidx = *(const ushort *)iref & packed_value_mask;
 			if ( iosp >= ostop )
-			  return_with_error_iref(e_stackoverflow);
+			  return_with_stackoverflow_iref();
 			++iosp;
 			name_index_ref_inline(int_nt, nidx, iosp);
 			next_short();
 		  }
 		case pt_executable_name:
-		  {	uint nidx = (uint)*(ushort *)iref & packed_value_mask;
+		  {	uint nidx =
+			  (uint)*(const ushort *)iref & packed_value_mask;
 			pvalue = name_index_ptr_inline(int_nt, nidx)->pvalue;
 			if ( !pv_valid(pvalue) )
 			   {	uint htemp;
@@ -1243,7 +1410,7 @@ again:		code = scan_token(s, &token, &sstate);
 			if ( r_has_masked_attrs(pvalue, a_execute, a_execute + a_executable) )
 			  {	/* Literal, push it. */
 				if ( iosp >= ostop )
-				  return_with_error_iref(e_stackoverflow);
+				  return_with_stackoverflow_iref();
 				++iosp;
 				ref_assign_inline(iosp, pvalue);
 				next_short();
@@ -1266,7 +1433,7 @@ again:		code = scan_token(s, &token, &sstate);
 	}
 	/* Literal type, just push it. */
 	if ( iosp >= ostop )
-	  return_with_error_iref(e_stackoverflow);
+	  return_with_stackoverflow_iref();
 	++iosp;
 	ref_assign_inline(iosp, iref);
 bot:	next();
@@ -1369,9 +1536,35 @@ error_exit:
 
 }
 
+/* Pop the bookkeeping information for a normal exit from a t_oparray. */
+private int
+oparray_pop(os_ptr op)
+{	esp -= 3;
+	return o_pop_estack;
+}
+
+/* Restore the stack pointers after an error inside a t_oparray procedure. */
+/* This procedure is called only from pop_estack. */
+private int
+oparray_cleanup(os_ptr op)
+{	/* esp points just below the cleanup procedure. */
+	es_ptr ep = esp;
+	uint ocount_old = (uint)ep[2].value.intval;
+	uint dcount_old = (uint)ep[3].value.intval;
+	uint ocount = ref_stack_count(&o_stack);
+	uint dcount = ref_stack_count(&d_stack);
+
+	if ( ocount > ocount_old )
+	  ref_stack_pop(&o_stack, ocount - ocount_old);
+	if ( dcount > dcount_old )
+	  ref_stack_pop(&d_stack, dcount - dcount_old);
+	return 0;
+}
+
 /* ------ Initialization procedure ------ */
 
 BEGIN_OP_DEFS(interp_op_defs) {
 		/* Internal operators */
 	{"0%interp_exit", interp_exit},
+	{"0%oparray_pop", oparray_pop},
 END_OP_DEFS(0) }
