@@ -1,19 +1,24 @@
-/* Copyright (C) 1993, 1995 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1993, 1995, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of Aladdin Ghostscript.
+  This file is part of GNU Ghostscript.
   
-  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
-  or distributor accepts any responsibility for the consequences of using it,
-  or for whether it serves any particular purpose or works at all, unless he
-  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
-  License (the "License") for full details.
+  GNU Ghostscript is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
+  anyone for the consequences of using it or for whether it serves any
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
-  Every copy of Aladdin Ghostscript must include a copy of the License,
-  normally in a plain ASCII text file named PUBLIC.  The License grants you
-  the right to copy, modify and redistribute Aladdin Ghostscript, but only
-  under certain conditions described in the License.  Among other things, the
-  License requires that the copyright notice and this notice be preserved on
-  all copies.
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* iparam.c */
@@ -501,7 +506,7 @@ ref_param_read_int_array(gs_param_list *plist, gs_param_name pkey,
 	int code = ref_param_read_array(iplist, pkey, &loc);
 	int *piv;
 	uint size;
-	uint i;
+	long i;
 
 	if ( code != 0 )
 	  return code;
@@ -511,14 +516,16 @@ ref_param_read_int_array(gs_param_list *plist, gs_param_name pkey,
 	if ( piv == 0 )
 	  return_error(e_VMerror);
 	for ( i = 0; i < size; i++ )
-	{	const ref *pe = loc.pvalue->value.const_refs + i;
-		if ( !r_has_type(pe, t_integer) )
+	{	ref elt;
+
+		array_get(loc.pvalue, i, &elt);
+		if ( !r_has_type(&elt, t_integer) )
 		  { code = gs_note_error(e_typecheck); break; }
 #if arch_sizeof_int < arch_sizeof_long
-		if ( pe->value.intval != (int)pe->value.intval )
+		if ( elt.value.intval != (int)elt.value.intval )
 		  { code = gs_note_error(e_rangecheck); break; }
 #endif
-		piv[i] = (int)pe->value.intval;
+		piv[i] = (int)elt.value.intval;
 	}
 	if ( code < 0 )
 	  {	ifree_object(piv, "ref_param_read_int_array");
@@ -533,9 +540,11 @@ private int
 ref_param_read_float_array(gs_param_list *plist, gs_param_name pkey,
   gs_param_float_array *pvalue)
 {	iparam_loc loc;
+	ref aref, elt;
 	int code = ref_param_read_array(iplist, pkey, &loc);
 	float *pfv;
 	uint size;
+	long i;
 
 	if ( code != 0 )
 	  return code;
@@ -544,7 +553,12 @@ ref_param_read_float_array(gs_param_list *plist, gs_param_name pkey,
 					 "ref_param_read_float_array");
 	if ( pfv == 0 )
 	  return_error(e_VMerror);
-	code = num_params(loc.pvalue->value.const_refs + size - 1, size, pfv);
+	aref = *loc.pvalue;
+	loc.pvalue = &elt;
+	for ( i = 0; code >= 0 && i < size; i++ )
+	{	array_get(&aref, i, &elt);
+		code = float_param(&elt, pfv + i);
+	}
 	if ( code < 0 )
 	{	ifree_object(pfv, "ref_read_float_array_param");
 		return (*loc.presult = code);
@@ -558,22 +572,24 @@ private int
 ref_param_read_string_array(gs_param_list *plist, gs_param_name pkey,
   gs_param_string_array *pvalue)
 {	iparam_loc loc;
+	ref aref, elt;
 	int code = ref_param_read_array(iplist, pkey, &loc);
 	gs_param_string *psv;
-	ref *prefs;
 	uint size;
-	uint i;
+	long i;
 
 	if ( code != 0 )
 	  return code;
-	prefs = loc.pvalue->value.refs;
 	size = r_size(loc.pvalue);
-	psv = (gs_param_string *)ialloc_byte_array(size, sizeof(gs_param_string),
-						   "ref_param_read_string_array");
+	psv =
+	  (gs_param_string *)ialloc_byte_array(size, sizeof(gs_param_string),
+					       "ref_param_read_string_array");
 	if ( psv == 0 )
 	  return_error(e_VMerror);
+	aref = *loc.pvalue;
+	loc.pvalue = &elt;
 	for ( i = 0; code >= 0 && i < size; i++ )
-	{	loc.pvalue = prefs + i;
+	{	array_get(&aref, i, &elt);
 		code = ref_param_read_string_value(&loc, psv + i);
 	}
 	if ( code < 0 )
@@ -614,12 +630,11 @@ ref_param_end_read_dict(gs_param_list *plist, gs_param_name pkey,
 private int
 ref_param_read_get_policy(gs_param_list *plist, gs_param_name pkey)
 {	ref kname;
-	int code;
 	ref *pvalue;
 	/* We can't use dict_find_string directly here, because */
 	/* pkey might not be a _ds string. */
 	if ( !(r_has_type(&iplist->u.r.policies, t_dictionary) &&
-	       (code = name_ref((const byte *)pkey, strlen(pkey), &kname, -1)) >= 0 &&
+	       name_ref((const byte *)pkey, strlen(pkey), &kname, -1) >= 0 &&
 	       dict_find(&iplist->u.r.policies, &kname, &pvalue) > 0 &&
 	       r_has_type(pvalue, t_integer))
 	   )
@@ -680,12 +695,15 @@ s:		pvalue->data = pref->value.const_bytes;
 	return 0;
 }
 
-/* Read an array parameter. */
+/* Read an array (or packed array) parameter. */
 private int near
 ref_param_read_array(iparam_list *plist, gs_param_name pkey, iparam_loc *ploc)
-{	int code = ref_param_read(plist, pkey, ploc, t_array);
+{	int code = ref_param_read(plist, pkey, ploc, -1);
+
 	if ( code != 0 )
 	  return code;
+	if ( !r_is_array(ploc->pvalue) )
+	  iparam_return_error(*ploc, e_typecheck);
 	iparam_check_read(*ploc);
 	return 0;
 }
