@@ -1,4 +1,4 @@
-/* Copyright (C) 1993, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1993, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -50,9 +50,10 @@ private_st_alloc_save();
 /* see below for details. */
 private const long max_repeated_scan = 100000;
 
-#define print_save(str, space, sav)\
+/* Some compilers try to substitute macro args in string literals! */
+#define print_save(str, spacen, sav)\
   if_debug5('u', "[u]%s space %u 0x%lx: cdata = 0x%lx, id = %lu\n",\
-    str, space, (ulong)(sav), (ulong)(sav)->client_data, (ulong)(sav)->id);
+    str, spacen, (ulong)(sav), (ulong)(sav)->client_data, (ulong)(sav)->id);
 
 /*
  * The logic for saving and restoring the state is complex.
@@ -862,7 +863,9 @@ combine_space(gs_ref_memory_t *mem)
 	mem->clast = omem->clast;
 	mem->allocated += omem->allocated;
 	mem->gc_allocated += omem->allocated;
-	mem->freed_lost += omem->freed_lost;
+	mem->lost.objects += omem->lost.objects;
+	mem->lost.refs += omem->lost.refs;
+	mem->lost.strings += omem->lost.strings;
 	mem->saved = omem->saved;
 	mem->previous_status = omem->previous_status;
 	  {	/* Concatenate free lists. */
