@@ -1,19 +1,24 @@
 /* Copyright (C) 1989, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of Aladdin Ghostscript.
+  This file is part of GNU Ghostscript.
   
-  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
-  or distributor accepts any responsibility for the consequences of using it,
-  or for whether it serves any particular purpose or works at all, unless he
-  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
-  License (the "License") for full details.
+  GNU Ghostscript is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
+  anyone for the consequences of using it or for whether it serves any
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
-  Every copy of Aladdin Ghostscript must include a copy of the License,
-  normally in a plain ASCII text file named PUBLIC.  The License grants you
-  the right to copy, modify and redistribute Aladdin Ghostscript, but only
-  under certain conditions described in the License.  Among other things, the
-  License requires that the copyright notice and this notice be preserved on
-  all copies.
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* iscan.c */
@@ -386,7 +391,7 @@ scan_token(register stream *s, ref *pref, scanner_state *pstate)
   if ( osp >= osbot ) osp--;\
   else ref_stack_pop(&o_stack, 1)
 	int max_name_ctype =
-		(recognize_btokens() ? ctype_name : ctype_btoken);
+	  (recognize_btokens() ? ctype_name : ctype_btoken);
 #define scan_sign(sign, ptr)\
   switch ( *ptr ) {\
     case '-': sign = -1; ptr++; break;\
@@ -399,6 +404,7 @@ scan_token(register stream *s, ref *pref, scanner_state *pstate)
 	byte s1[2];
 	register const byte _ds *decoder = scan_char_decoder;
 	int status;
+	int sign;
 	scanner_state sstate;
 #define pstack sstate.s_pstack
 #define pdepth sstate.s_pdepth
@@ -406,6 +412,7 @@ scan_token(register stream *s, ref *pref, scanner_state *pstate)
 #define da sstate.s_da
 #define name_type sstate.s_ss.s_name.s_name_type
 #define try_number sstate.s_ss.s_name.s_try_number
+
 	if ( pstate->s_pstack != 0 )
 	  {	if_not_spush1()
 		  return retcode;
@@ -457,10 +464,8 @@ top:	c = scan_getc();
 	{
 	case ' ': case '\f': case '\t':
 	case char_CR: case char_EOL:
-	case char_NULL: case char_VT:
+	case char_NULL:
 		goto top;
-	case char_EOT:
-	case char_DOS_EOF:
 	case '[':
 	case ']':
 		s1[0] = (byte)c;
@@ -766,11 +771,18 @@ try_funny_name:
 	case '0': case '1': case '2': case '3': case '4':
 	case '5': case '6': case '7': case '8': case '9':
 	case '.':
-		/* We must stop the scan 1 character early, to be sure that */
-		/* we can test for CR+LF within the buffer.  That's why */
-		/* we pass endptr rather than endptr + 1. */
-		retcode = scan_number(sptr, endptr, 0, myref, &newptr);
-nr:		if ( retcode == 1 && decoder[newptr[-1]] == ctype_space )
+		sign = 0;
+nr:		/*
+		 * Skip a leading sign, if any, by conditionally passing
+		 * sptr + 1 rather than sptr.  Also, if the last character
+		 * in the buffer is a CR, we must stop the scan 1 character
+		 * early, to be sure that we can test for CR+LF within the
+		 * buffer, by passing endptr rather than endptr + 1.
+		 */
+		retcode = scan_number(sptr + (sign & 1),
+				endptr /*(*endptr == char_CR ? endptr : endptr + 1)*/,
+				sign, myref, &newptr);
+		if ( retcode == 1 && decoder[newptr[-1]] == ctype_space )
 		{	sptr = newptr - 1;
 			if ( *sptr == char_CR && sptr[1] == char_EOL )
 				sptr++;
@@ -781,10 +793,10 @@ nr:		if ( retcode == 1 && decoder[newptr[-1]] == ctype_space )
 		try_number = true;
 		goto do_name;
 	case '+':
-		retcode = scan_number(sptr + 1, endptr, 1, myref, &newptr);
+		sign = 1;
 		goto nr;
 	case '-':
-		retcode = scan_number(sptr + 1, endptr, -1, myref, &newptr);
+		sign = -1;
 		goto nr;
 
 	/* Check for a binary object */
@@ -917,8 +929,8 @@ nx:		switch ( decoder[c] )
 		  }
 		/* Check for a number */
 		if ( try_number )
-		{	int sign;
-			const byte *base = da.base;
+		{	const byte *base = da.base;
+
 			scan_sign(sign, base);
 			retcode = scan_number(base, daptr, sign, myref, &newptr);
 			if ( retcode == 1 )
