@@ -5,9 +5,20 @@
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* iscanbin.c */
@@ -267,11 +278,12 @@ scan_bin_num_array_continue(stream *s, ref *pref, scanner_state *pstate)
 			pstate->s_scan_type = scanning_binary;
 			return scan_Refill;
 		  }
-		code = sget_encoded_number(s, pbs->num_format, np);
+		code = sdecode_number(sbufptr(s), pbs->num_format, np);
 		switch ( code )
 		  {
 		case t_integer: case t_real:
 			r_set_type(np, code);
+			sbufskip(s, wanted);
 			break;
 		case t_null:
 			return_error(e_syntaxerror);
@@ -363,10 +375,16 @@ scan_bos_continue(register stream *s, ref *pref, scanner_state *pstate)
 		case bs_string:
 			swap_size_also_value();
 			sattrs = (ob.tx < 128 ? a_all : a_all + a_executable);
-str:			if ( ob.value.w < max_array_index * sizeof_bin_seq_obj ||
+str:			if ( ob.size.w == 0 )
+			  {	/* For zero-length strings, the offset */
+				/* doesn't matter, and may be zero. */
+				make_empty_string(op, sattrs);
+				break;
+			  }
+			if ( ob.value.w < max_array_index * sizeof_bin_seq_obj ||
 			    ob.value.w + ob.size.w > size
 			   )
-				return_error(e_syntaxerror);
+			  return_error(e_syntaxerror);
 			if ( ob.value.w < min_string_index )
 			  {	/* We have to (re)allocate the strings. */
 				uint str_size = size - ob.value.w;
@@ -468,7 +486,7 @@ scan_bos_resize(scanner_state *pstate, uint new_size, uint index)
 	/* it's quite possible that new and old addresses are the same. */
 	if ( relocated_base != old_base )
 	  for ( i = index; i != 0; i--, aptr++ )
-	    if ( r_has_type(aptr, t_string) )
+	    if ( r_has_type(aptr, t_string) && r_size(aptr) != 0 )
 	      aptr->value.bytes =
 		aptr->value.bytes - old_base + relocated_base;
 	return new_base;
@@ -524,7 +542,7 @@ scan_bos_string_continue(register stream *s, ref *pref, scanner_state *pstate)
 				  }
 			  }
 			else
-			  {	int code = dict_create(count >> 1, &rdict);
+			  {	code = dict_create(count >> 1, &rdict);
 				if ( code < 0 )
 				  return code;
 				while ( count )
