@@ -173,20 +173,16 @@ ref_stack_counttomark(const ref_stack *pstack)
 	return 0;
 }
 
-/* Store the top elements of a stack into an array, */
-/* with or without store/undo checking. */
-/* May return e_rangecheck or e_invalidaccess. */
+/* Do the store check for storing elements of a stack into an array. */
+/* May return e_invalidaccess. */
 int
-ref_stack_store(const ref_stack *pstack, ref *parray, uint count, uint skip,
-  int age, bool check, client_name_t cname)
-{	uint left, pass;
-	ref *to;
-	uint space = r_space(parray);
-	if ( count > ref_stack_count(pstack) || count > r_size(parray) )
-	  return_error(e_rangecheck);
-	if ( check && space != avm_local )
-	  {	/* Pre-check the elements being stored. */
-		left = count, pass = skip;
+ref_stack_store_check(const ref_stack *pstack, ref *parray, uint count,
+  uint skip)
+{	uint space = r_space(parray);
+
+	if ( space != avm_local )
+	  {	uint left = count, pass = skip;
+
 		STACK_LOOP_BEGIN(pstack, ptr, size)
 		  if ( size <= pass )
 		    pass -= size;
@@ -206,6 +202,25 @@ ref_stack_store(const ref_stack *pstack, ref *parray, uint count, uint skip,
 		    }
 		STACK_LOOP_END(ptr, size)
 	  }
+	return 0;
+}
+
+/* Store the top elements of a stack into an array, */
+/* with or without store/undo checking. */
+/* May return e_rangecheck or e_invalidaccess. */
+int
+ref_stack_store(const ref_stack *pstack, ref *parray, uint count, uint skip,
+  int age, bool check, client_name_t cname)
+{	uint left, pass;
+	ref *to;
+
+	if ( count > ref_stack_count(pstack) || count > r_size(parray) )
+	  return_error(e_rangecheck);
+	if ( check )
+	  { int code = ref_stack_store_check(pstack, parray, count, skip);
+	    if ( code < 0 )
+	      return code;
+	  }
 	to = parray->value.refs + count;
 	left = count, pass = skip;
 	STACK_LOOP_BEGIN(pstack, from, size)
