@@ -1,4 +1,4 @@
-/* Copyright (C) 1989, 1995, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -36,8 +36,10 @@
 #include "ivmspace.h"
 #include "oper.h"
 #include "store.h"
+#include "gsccode.h"			/* for gxfont.h */
 #include "gsmatrix.h"
 #include "gsutil.h"
+#include "gxfont.h"
 
 /* ------ Object utilities ------ */
 
@@ -92,9 +94,6 @@ obj_eq(register const ref *pref1, register const ref *pref2)
 			name_string_ref(pref1, &nref);
 			pref1 = &nref;
 			break;
-		case t_save:
-			return (r_has_type(pref2, t_save) &&
-				pref2->value.saveid == pref1->value.saveid);
 		case t_string:
 			if ( !r_has_type(pref2, t_name) )
 			  return false;
@@ -137,16 +136,31 @@ obj_eq(register const ref *pref1, register const ref *pref2)
 		return (op_index(pref1) == op_index(pref2));
 	case t_real:
 		return (pref1->value.realval == pref2->value.realval);
+	case t_save:
+		return (pref2->value.saveid == pref1->value.saveid);
 	case t_string:
 		return (!bytes_compare(pref1->value.bytes, r_size(pref1),
 				       pref2->value.bytes, r_size(pref2)));
 	case t_device:
 		return (pref1->value.pdevice == pref2->value.pdevice);
-	case t_fontID:
 	case t_struct:
 	case t_astruct:
 		return (pref1->value.pstruct == pref2->value.pstruct);
-	   }
+	case t_fontID:
+	  {	/*
+		 * In the Adobe implementations, different scalings of a
+		 * font have "equal" FIDs, so we do the same.
+		 */
+		const gs_font *pfont1 = r_ptr(pref1, gs_font);
+		const gs_font *pfont2 = r_ptr(pref2, gs_font);
+
+		while ( pfont1->base != pfont1 )
+		  pfont1 = pfont1->base;
+		while ( pfont2->base != pfont2 )
+		  pfont2 = pfont2->base;
+		return (pfont1 == pfont2);
+	  }
+	  }
 	return false;			/* shouldn't happen! */
 }
 
@@ -425,8 +439,9 @@ ref_to_string(const ref *pref, gs_memory_t *mem, client_name_t cname)
 /* The stack underflow check (check for t__invalid) is harmless */
 /* if the operands come from somewhere other than the stack. */
 int
-num_params(const ref *op, int count, float *pval)
+num_params(const ref *op, int count, double *pval)
 {	int mask = 0;
+
 	pval += count;
 	while ( --count >= 0 )
 	   {	mask <<= 1;
@@ -450,25 +465,53 @@ num_params(const ref *op, int count, float *pval)
 	/* In this case we clearly don't care about the value of mask. */
 	return (mask < 0 ? 0 : mask);
 }
+/* float_params doesn't bother to keep track of the mask. */
+int
+float_params(const ref *op, int count, float *pval)
+{	for ( pval += count; --count >= 0; --op )
+	  switch ( r_type(op) )
+	    {
+	    case t_real:
+		*--pval = op->value.realval;
+		break;
+	    case t_integer:
+		*--pval = op->value.intval;
+		break;
+	    case t__invalid:
+		return_error(e_stackunderflow);
+	    default:
+		return_error(e_typecheck);
+	    }
+	return 0;
+}
 
 /* Get a single real parameter. */
 /* The only possible error is e_typecheck. */
 /* If an error is returned, the return value is not updated. */
 int
-real_param(const ref *op, float *pparam)
+real_param(const ref *op, double *pparam)
 {	switch ( r_type(op) )
 	   {
-	case t_integer:
+	   case t_integer:
 		*pparam = op->value.intval;
 		break;
-	case t_real:
+	   case t_real:
 		*pparam = op->value.realval;
 		break;
-	default:
+	   default:
 		return_error(e_typecheck);
 	   }
 	return 0;
 }
+int
+float_param(const ref *op, float *pparam)
+{	double dval;
+	int code = real_param(op, &dval);
+
+	if ( code >= 0 )
+	  *pparam = (float)dval; /* can't overflow */
+	return code;
+}
 
 /* Get an integer parameter in a given range. */
 int
@@ -479,10 +522,20 @@ int_param(const ref *op, int max_value, int *pparam)
 }
 
 /* Make real values on the operand stack. */
-void
-make_reals(ref *op, const float *pval, int count)
-{	for ( ; count--; op++, pval++ )
+int
+make_reals(ref *op, const double *pval, int count)
+{	/* This should return e_limitcheck if any real is too large */
+	/* to fit into a float on the stack. */
+	for ( ; count--; op++, pval++ )
 	  make_real(op, *pval);
+	return 0;
+}
+int
+make_floats(ref *op, const float *pval, int count)
+{	/* This should return e_limitcheck for infinities. */
+	for ( ; count--; op++, pval++ )
+	  make_real(op, *pval);
+	return 0;
 }
 
 /* Compute the error code when check_proc fails. */
@@ -509,10 +562,24 @@ check_type_failed(const ref *op)
 int
 read_matrix(const ref *op, gs_matrix *pmat)
 {	int code;
-	check_read_type(*op, t_array);
+	ref values[6];
+	const ref *pvalues;
+
+	if ( r_has_type(op, t_array) )
+	  pvalues = op->value.refs;
+	else
+	  { int i;
+	    for ( i = 0; i < 6; ++i )
+	      { code = array_get(op, (long)i, &values[i]);
+	        if ( code < 0 )
+		  return code;
+	      }
+	    pvalues = values;
+	  }
+	check_read(*op);
 	if ( r_size(op) != 6 )
-		return_error(e_rangecheck);
-	code = num_params(op->value.refs + 5, 6, (float *)pmat);
+	  return_error(e_rangecheck);
+	code = float_params(pvalues + 5, 6, (float *)pmat);
 	return (code < 0 ? code : 0);
 }
 
@@ -523,14 +590,15 @@ write_matrix(register ref *op, const gs_matrix *pmat)
 {	ref *aptr;
 	const float *pel;
 	int i;
+
 	check_write_type(*op, t_array);
 	if ( r_size(op) != 6 )
-		return_error(e_rangecheck);
+	  return_error(e_rangecheck);
 	aptr = op->value.refs;
 	pel = (const float *)pmat;
 	for ( i = 5; i >= 0; i--, aptr++, pel++ )
-	{	ref_save(op, aptr, "write_matrix");
+	  {	ref_save(op, aptr, "write_matrix");
 		make_real_new(aptr, *pel);
-	}
+	  }
 	return 0;
 }
