@@ -53,6 +53,9 @@ int gs_jpeg_set_colorspace (P2(stream_DCT_state *st,
 int gs_jpeg_set_linear_quality (P3(stream_DCT_state *st,
 				   int scale_factor,
 				   boolean force_baseline));
+int gs_jpeg_set_quality (P3(stream_DCT_state *st,
+			    int quality,
+			    boolean force_baseline));
 int gs_jpeg_start_compress (P2(stream_DCT_state *st,
 			       boolean write_all_tables));
 int gs_jpeg_write_scanlines (P3(stream_DCT_state *st,
