@@ -198,6 +198,8 @@ int alloc_save_level(P1(const gs_dual_memory_t *));
 
 #define make_oper(pref,opidx,proc)\
   make_tasv(pref, t_operator, a_executable, opidx, opproc, proc)
+#define make_oper_new(pref,opidx,proc)\
+  make_tasv_new(pref, t_operator, a_executable, opidx, opproc, proc)
 
 #define make_real(pref,rval)\
   make_tv(pref, t_real, realval, rval)
