@@ -1,19 +1,24 @@
 /* Copyright (C) 1989, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of Aladdin Ghostscript.
+  This file is part of GNU Ghostscript.
   
-  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
-  or distributor accepts any responsibility for the consequences of using it,
-  or for whether it serves any particular purpose or works at all, unless he
-  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
-  License (the "License") for full details.
+  GNU Ghostscript is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
+  anyone for the consequences of using it or for whether it serves any
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
-  Every copy of Aladdin Ghostscript must include a copy of the License,
-  normally in a plain ASCII text file named PUBLIC.  The License grants you
-  the right to copy, modify and redistribute Aladdin Ghostscript, but only
-  under certain conditions described in the License.  Among other things, the
-  License requires that the copyright notice and this notice be preserved on
-  all copies.
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* zchar.c */
@@ -47,6 +52,7 @@ private bool map_glyph_to_char(P3(const ref *, const ref *, ref *));
 private int finish_show(P1(os_ptr));
 private int finish_stringwidth(P1(os_ptr));
 private int op_show_cleanup(P1(os_ptr));
+private int op_show_return_width(P3(os_ptr, uint, double *));
 
 /* <string> show - */
 private int
@@ -68,7 +74,8 @@ private int
 zashow(register os_ptr op)
 {	gs_show_enum *penum;
 	int code;
-	float axy[2];
+	double axy[2];
+
 	if (	(code = num_params(op - 1, 2, axy)) < 0 ||
 		(code = op_show_setup(op, &penum)) != 0
 	   )
@@ -86,7 +93,8 @@ private int
 zwidthshow(register os_ptr op)
 {	gs_show_enum *penum;
 	int code;
-	float cxy[2];
+	double cxy[2];
+
 	check_type(op[-1], t_integer);
 	if ( (gs_char)(op[-1].value.intval) != op[-1].value.intval )
 		return_error(e_rangecheck);
@@ -110,7 +118,8 @@ private int
 zawidthshow(register os_ptr op)
 {	gs_show_enum *penum;
 	int code;
-	float cxy[2], axy[2];
+	double cxy[2], axy[2];
+
 	check_type(op[-3], t_integer);
 	if ( (gs_char)(op[-3].value.intval) != op[-3].value.intval )
 		return_error(e_rangecheck);
@@ -212,7 +221,8 @@ zcharboxpath(register os_ptr op)
 /* <wx> <wy> <llx> <lly> <urx> <ury> setcachedevice - */
 int
 zsetcachedevice(register os_ptr op)
-{	float wbox[6];
+{	double wbox[6];
+
 	gs_show_enum *penum = op_show_find();
 	int code = num_params(op, 6, wbox);
 	if ( penum == 0 )
@@ -221,7 +231,7 @@ zsetcachedevice(register os_ptr op)
 	  return code;
 	if ( gs_show_width_only(penum) )
 	  return op_show_return_width(op, 6, &wbox[0]);
-	code = gs_setcachedevice(penum, igs, wbox);
+	code = gs_setcachedevice_double(penum, igs, wbox);
 	if ( code < 0 )
 	  return code;
 	pop(6);
@@ -233,9 +243,10 @@ zsetcachedevice(register os_ptr op)
 /* <w0x> <w0y> <llx> <lly> <urx> <ury> <w1x> <w1y> <vx> <vy> setcachedevice2 - */
 int
 zsetcachedevice2(os_ptr op)
-{	float wbox[10];
+{	double wbox[10];
 	gs_show_enum *penum = op_show_find();
 	int code = num_params(op, 10, wbox);
+
 	if ( penum == 0 )
 	  return_error(e_undefined);
 	if ( code < 0 )
@@ -244,7 +255,7 @@ zsetcachedevice2(os_ptr op)
 	  return op_show_return_width(op, 10,
 				      (gs_rootfont(igs)->WMode ?
 				       &wbox[6] : &wbox[0]));
-	code = gs_setcachedevice2(penum, igs, wbox);
+	code = gs_setcachedevice2_double(penum, igs, wbox);
 	if ( code < 0 )
 	  return code;
 	pop(10);
@@ -256,9 +267,10 @@ zsetcachedevice2(os_ptr op)
 /* <wx> <wy> setcharwidth - */
 private int
 zsetcharwidth(register os_ptr op)
-{	float width[2];
+{	double width[2];
 	gs_show_enum *penum = op_show_find();
 	int code = num_params(op, 2, width);
+
 	if ( penum == 0 )
 	  return_error(e_undefined);
 	if ( code < 0 )
@@ -276,7 +288,7 @@ zsetcharwidth(register os_ptr op)
 /* <dict> .fontbbox -false- */
 private int
 zfontbbox(register os_ptr op)
-{	float bbox[4];
+{	double bbox[4];
 	int code;
 
 	check_type(*op, t_dictionary);
@@ -524,8 +536,8 @@ op_show_find(void)
 /* of the setcharwidth or the setcachedevice[2] if we are in */
 /* a stringwidth or cshow, or if we are only collecting the scalable */
 /* width for an xfont character. */
-int
-op_show_return_width(os_ptr op, uint npop, float *pwidth)
+private int
+op_show_return_width(os_ptr op, uint npop, double *pwidth)
 {	uint index = op_show_find_index();
 	es_ptr ep = (es_ptr)ref_stack_index(&e_stack, index - (snumpush - 1));
 	int code = gs_setcharwidth(esenum(ep), igs, pwidth[0], pwidth[1]);
@@ -576,6 +588,14 @@ op_show_restore(bool for_error)
 	  { /* stringwidth does an extra gsave */
 	    --saved_level;
 	  }
+	/*
+	 * We might have been inside a cshow, in which case currentfont
+	 * was reset temporarily, as though we were inside a BuildChar/
+	 * BuildGlyph procedure.  If this is the case, set currentfont
+	 * back to its normal state, the root font.
+	 */
+	if ( penum->fstack.depth >= 0 )
+	  gs_set_currentfont(igs, penum->fstack.items[0].font);
 	while ( igs->level > saved_level && code >= 0 )
 	  { if ( igs->saved == 0 || igs->saved->saved == 0 )
 	      { /*
@@ -606,7 +626,7 @@ op_show_free(int code)
 
 /* Get a FontBBox parameter from a font dictionary. */
 int
-font_bbox_param(const ref *pfdict, float bbox[4])
+font_bbox_param(const ref *pfdict, double bbox[4])
 {	ref *pbbox;
 	/*
 	 * Pre-clear the bbox in case it's invalid.  The Red Books say that
