@@ -1,19 +1,24 @@
 /* Copyright (C) 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of Aladdin Ghostscript.
+  This file is part of GNU Ghostscript.
   
-  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
-  or distributor accepts any responsibility for the consequences of using it,
-  or for whether it serves any particular purpose or works at all, unless he
-  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
-  License (the "License") for full details.
+  GNU Ghostscript is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
+  anyone for the consequences of using it or for whether it serves any
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
-  Every copy of Aladdin Ghostscript must include a copy of the License,
-  normally in a plain ASCII text file named PUBLIC.  The License grants you
-  the right to copy, modify and redistribute Aladdin Ghostscript, but only
-  under certain conditions described in the License.  Among other things, the
-  License requires that the copyright notice and this notice be preserved on
-  all copies.
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* zchar42.c */
@@ -57,7 +62,7 @@ ztype42execchar(register os_ptr op)
 	int code = font_param(op - 3, &pfont);
 	gs_show_enum *penum = op_show_find();
 	int present;
-	float sbw[4];
+	double sbw[4];
 
 	if ( code < 0 )
 	  return code;
@@ -95,9 +100,14 @@ ztype42execchar(register os_ptr op)
 	  return code;
 	/* Get the metrics and set the cache device. */
 	if ( present == metricsNone )
-	  { code = gs_type42_get_metrics(pfont42, (uint)op->value.intval, sbw);
+	  { float sbw42[4];
+	    int i;
+
+	    code = gs_type42_get_metrics(pfont42, (uint)op->value.intval, sbw42);
 	    if ( code < 0 )
 	      return code;
+	    for ( i = 0; i < 4; ++i )
+	      sbw[i] = sbw42[i];
 	  }
 	return zchar_set_cache(op, pbfont, op - 1,
 			       (present == metricsSideBearingAndWidth ?
@@ -126,7 +136,7 @@ type42_finish(os_ptr op, int (*cont)(P1(gs_state *)))
 #define pfont42 ((gs_font_type42 *)pfont)
 	int code;
 	gs_show_enum *penum = op_show_find();
-	float sbxy[2];
+	double sbxy[2];
 	gs_point sbpt;
 	gs_point *psbpt = 0;
 	os_ptr opc = op;
