@@ -1,19 +1,24 @@
-/* Copyright (C) 1989, 1992, 1993, 1994, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1992, 1993, 1994, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of Aladdin Ghostscript.
+  This file is part of GNU Ghostscript.
   
-  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
-  or distributor accepts any responsibility for the consequences of using it,
-  or for whether it serves any particular purpose or works at all, unless he
-  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
-  License (the "License") for full details.
+  GNU Ghostscript is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
+  anyone for the consequences of using it or for whether it serves any
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
-  Every copy of Aladdin Ghostscript must include a copy of the License,
-  normally in a plain ASCII text file named PUBLIC.  The License grants you
-  the right to copy, modify and redistribute Aladdin Ghostscript, but only
-  under certain conditions described in the License.  Among other things, the
-  License requires that the copyright notice and this notice be preserved on
-  all copies.
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* zcolor.c */
@@ -64,9 +69,10 @@ zcurrentgray(register os_ptr op)
 private int
 zcurrentrgbcolor(register os_ptr op)
 {	float par[3];
+
 	gs_currentrgbcolor(igs, par);
 	push(3);
-	make_reals(op - 2, par, 3);
+	make_floats(op - 2, par, 3);
 	return 0;
 }
 
@@ -91,8 +97,9 @@ zprocesscolors(register os_ptr op)
 /* <alpha> setalpha - */
 private int
 zsetalpha(register os_ptr op)
-{	float alpha;
+{	double alpha;
 	int code;
+
 	if ( real_param(op, &alpha) < 0 )
 	  return_op_typecheck(op);
 	if ( (code = gs_setalpha(igs, alpha)) < 0 )
@@ -104,8 +111,9 @@ zsetalpha(register os_ptr op)
 /* <gray> setgray - */
 private int
 zsetgray(register os_ptr op)
-{	float gray;
+{	double gray;
 	int code;
+
 	if ( real_param(op, &gray) < 0 )
 	  return_op_typecheck(op);
 	if ( (code = gs_setgray(igs, gray)) < 0 )
@@ -118,10 +126,11 @@ zsetgray(register os_ptr op)
 /* <red> <green> <blue> setrgbcolor - */
 private int
 zsetrgbcolor(register os_ptr op)
-{	float par[3];
+{	double par[3];
 	int code;
-	if (	(code = num_params(op, 3, par)) < 0 ||
-		(code = gs_setrgbcolor(igs, par[0], par[1], par[2])) < 0
+
+	if ( (code = num_params(op, 3, par)) < 0 ||
+	     (code = gs_setrgbcolor(igs, par[0], par[1], par[2])) < 0
 	   )
 		return code;
 	make_null(&istate->colorspace.array);
@@ -179,16 +188,18 @@ zcolor_remap_one_store(os_ptr op, floatp min_value)
 	if ( ref_stack_count(&o_stack) < transfer_map_size )
 	  return_error(e_stackunderflow);
 	for ( i = 0; i < transfer_map_size; i++ )
-	{	float v;
+	{	double v;
 		int code =
 		  real_param(ref_stack_index(&o_stack,
 					     transfer_map_size - 1 - i),
 			     &v);
-		if ( code < 0 ) return code;
+
+		if ( code < 0 )
+		  return code;
 		pmap->values[i] =
-			(v < min_value ? float2frac(min_value) :
-			 v >= 1.0 ? frac_1 :
-			 float2frac(v));
+		  (v < min_value ? float2frac(min_value) :
+		   v >= 1.0 ? frac_1 :
+		   float2frac(v));
 	}
 	ref_stack_pop(&o_stack, transfer_map_size);
 	esp--;				/* pop pointer to transfer map */
