@@ -1,22 +1,35 @@
-/* Copyright (C) 1989, 1992, 1993, 1994 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* zcontrol.c */
 /* Control operators */
 #include "string_.h"
 #include "ghost.h"
+#include "stream.h"
 #include "errors.h"
 #include "oper.h"
 #include "estack.h"
+#include "files.h"
 #include "ipacked.h"
 #include "iutil.h"
 #include "store.h"
@@ -25,8 +38,9 @@
 extern void make_invalid_file(P1(ref *)); /* in zfile.c */
 
 /* Forward references */
-int no_cleanup(P1(os_ptr));
-private uint count_to_stopped(P0());
+private int no_cleanup(P1(os_ptr));
+private uint count_exec_stack(P1(os_ptr));
+private uint count_to_stopped(P1(long));
 
 /* See the comment in opdef.h for an invariant which allows */
 /* more efficient implementation of for, loop, and repeat. */
@@ -105,10 +119,57 @@ zexec(register os_ptr op)
 	return o_push_estack;
 }
 
+/* <obj1> ... <objn> <n> .execn - */
+int
+zexecn(register os_ptr op)
+{	uint n, i;
+	es_ptr esp_orig;
+
+	check_int_leu(*op, max_uint - 1);
+	n = (uint)op->value.intval;
+	check_op(n + 1);
+	check_estack(n);
+	esp_orig = esp;
+	for ( i = 0; i < n; ++i )
+	  { const ref *rp = ref_stack_index(&o_stack, (long)(i + 1));
+
+	    /* Make sure this object is legal to execute. */
+	    switch ( r_type(rp) )
+	      {
+	      case_types_with_access:
+		if ( !r_has_attr(rp, a_execute) &&
+		     r_has_attr(rp, a_executable)
+		   )
+		  { esp = esp_orig;
+		    return_error(e_invalidaccess);
+		  }
+	      default:
+		DO_NOTHING;
+	      }
+	    /* Executable nulls have a special meaning on the e-stack, */
+	    /* so since they are no-ops, don't push them. */
+	    if ( !r_has_type_attrs(rp, t_null, a_executable) )
+	      { ++esp;
+	        ref_assign(esp, rp);
+	      }
+	  }
+	esfile_check_cache();
+	pop(n + 1);
+	return o_push_estack;
+}
+
+/* <obj> superexec - */
+/* THIS IS NOT REALLY IMPLEMENTED YET. */
+private int
+zsuperexec(os_ptr op)
+{	return zexec(op);
+}
+
 /* <bool> <proc> if - */
 int
 zif(register os_ptr op)
 {	check_type(op[-1], t_boolean);
+	check_proc(*op);
 	if ( op[-1].value.boolval )
 	   {	check_estack(1);
 		++esp;
@@ -123,6 +184,8 @@ zif(register os_ptr op)
 int
 zifelse(register os_ptr op)
 {	check_type(op[-2], t_boolean);
+	check_proc(op[-1]);
+	check_proc(*op);
 	check_estack(1);
 	++esp;
 	if ( op[-2].value.boolval )
@@ -174,7 +237,8 @@ zfor(register os_ptr op)
 	else
 	  {	float params[3];
 		int code;
-		if ( (code = num_params(op - 1, 3, params)) < 0 )
+
+		if ( (code = float_params(op - 1, 3, params)) < 0 )
 		  return code;
 		make_real(ep - 4, params[0]);
 		make_real(ep - 3, params[1]);
@@ -350,125 +414,239 @@ zexit(register os_ptr op)
 	return_error(e_invalidexit);
 }
 
-/* <result> .stop - */
+/*
+ * .stopped pushes the following on the e-stack:
+ *	- A mark with type = es_stopped and procedure = no_cleanup.
+ *	- The result to be pushed on a normal return.
+ *	- The signal mask for .stop.
+ *	- The procedure %stopped_push, to handle the normal return case.
+ */
+
+/* In the normal (no-error) case, pop the mask from the e-stack, */
+/* and move the result to the o-stack. */
+private int
+stopped_push(register os_ptr op)
+{	push(1);
+	*op = esp[-1];
+	esp -= 3;
+	return o_pop_estack;
+}
+
+/* - stop - */
+/* Equivalent to true 1 .stop. */
+/* This is implemented in C because if were a pseudo-operator, */
+/* the stacks would get restored in case of an error. */
 private int
 zstop(register os_ptr op)
-{	uint count = count_to_stopped();
+{	uint count = count_to_stopped(1L);
+
 	if ( count )
-	{	pop_estack((uint)count);
-		return o_pop_estack;
-	}
+	  { /*
+	     * If there are any t_oparrays on the e-stack, they will pop
+	     * any new items from the o-stack.  Wait to push the 'true'
+	     * until we have run all the unwind procedures.
+	     */
+	    check_ostack(2);
+	    pop_estack(count);
+	    op = osp;
+	    push(1);
+	    make_true(op);
+	    return o_pop_estack;
+	  }
 	/* Return e_invalidexit if there is no mark at all. */
 	/* This is different from PostScript, which aborts. */
 	/* It shouldn't matter in practice. */
 	return_error(e_invalidexit);
 }
 
-/* <obj> <result> .stopped <result> */
+/* <result> <mask> .stop - */
+private int
+zzstop(register os_ptr op)
+{	uint count;
+
+	check_type(*op, t_integer);
+	count = count_to_stopped(op->value.intval);
+	if ( count )
+	  { /*
+	     * If there are any t_oparrays on the e-stack, they will pop
+	     * any new items from the o-stack.  Wait to push the result
+	     * until we have run all the unwind procedures.
+	     */
+	    ref save_result;
+
+	    check_op(2);
+	    save_result = op[-1];
+	    pop(2);
+	    pop_estack(count);
+	    op = osp;
+	    push(1);
+	    *op = save_result;
+	    return o_pop_estack;
+	  }
+	/* Return e_invalidexit if there is no mark at all. */
+	/* This is different from PostScript, which aborts. */
+	/* It shouldn't matter in practice. */
+	return_error(e_invalidexit);
+}
+
+/* <obj> stopped <stopped> */
+/* Equivalent to false 1 .stopped. */
+/* This is implemented in C because if were a pseudo-operator, */
+/* the stacks would get restored in case of an error. */
 private int
 zstopped(register os_ptr op)
-{	check_op(2);
+{	check_op(1);
 	/* Mark the execution stack, and push the default result */
 	/* in case control returns normally. */
-	check_estack(3);
+	check_estack(5);
 	push_mark_estack(es_stopped, no_cleanup);
-	*++esp = *op;			/* save the result */
-	*++esp = op[-1];		/* execute the operand */
+	++esp;
+	make_false(esp);		/* save the result */
+	++esp;
+	make_int(esp, 1);		/* save the signal mask */
+	push_op_estack(stopped_push);
+	*++esp = *op;			/* execute the operand */
 	esfile_check_cache();
-	pop(2);
+	pop(1);
 	return o_push_estack;
 }
 
-/* - .instopped false */
-/* - .instopped <result> true */
+/* <obj> <result> <mask> .stopped <result> */
+private int
+zzstopped(register os_ptr op)
+{	check_type(*op, t_integer);
+	check_op(3);
+	/* Mark the execution stack, and push the default result */
+	/* in case control returns normally. */
+	check_estack(5);
+	push_mark_estack(es_stopped, no_cleanup);
+	*++esp = op[-1];		/* save the result */
+	*++esp = *op;			/* save the signal mask */
+	push_op_estack(stopped_push);
+	*++esp = op[-2];		/* execute the operand */
+	esfile_check_cache();
+	pop(3);
+	return o_push_estack;
+}
+
+/* <mask> .instopped false */
+/* <mask> .instopped <result> true */
 private int
 zinstopped(register os_ptr op)
-{	uint count = count_to_stopped();
+{	uint count;
+
+	check_type(*op, t_integer);
+	count = count_to_stopped(op->value.intval);
 	if ( count )
-	{	push(2);
-		op[-1] = *ref_stack_index(&e_stack, count - 2);	/* default result */
-		make_true(op);
-	}
+	  { push(1);
+	    op[-1] = *ref_stack_index(&e_stack, count - 2);	/* default result */
+	    make_true(op);
+	  }
 	else
-	{	push(1);
-		make_false(op);
-	}
+	  make_false(op);
 	return 0;
 }
 
-/* - countexecstack <int> */
+/* <include_marks> .countexecstack <int> */
 private int
 zcountexecstack(register os_ptr op)
-{	push(1);
-	make_int(op, ref_stack_count(&e_stack));
+{	check_type(*op, t_boolean);
+	make_int(op, count_exec_stack(op));
 	return 0;
 }
 
-/* <array> execstack <subarray> */
+/* <array> <include_marks> .execstack <subarray> */
 private int execstack_continue(P1(os_ptr));
 private int
 zexecstack(register os_ptr op)
-{	/*
+{	os_ptr op1 = op - 1;
+	uint size;
+	/*
 	 * We can't do this directly, because the interpreter
 	 * might have cached some state.  To force the interpreter
 	 * to update the stored state, we push a continuation on
 	 * the exec stack; the continuation is executed immediately,
 	 * and does the actual transfer.
 	 */
-	uint depth = ref_stack_count(&e_stack);
+	uint depth;
 
-	check_write_type(*op, t_array);
-	if ( depth > r_size(op) )
+	check_type(*op, t_boolean);
+	check_write_type(*op1, t_array);
+	size = r_size(op1);
+	depth = count_exec_stack(op);
+	if ( depth > size )
 	  return_error(e_rangecheck);
+	{ int code = ref_stack_store_check(&e_stack, op1, size, 0);
+	  if ( code < 0 )
+	    return code;
+	}
 	check_estack(1);
-	r_set_size(op, (uint)depth);
+	r_set_size(op1, depth);
 	push_op_estack(execstack_continue);
 	return o_push_estack;
 }
 /* Continuation operator to do the actual transfer. */
-/* r_size(op) was set just above. */
+/* r_size(op - 1) was set just above. */
 private int
 execstack_continue(register os_ptr op)
-{	int code =
-	  ref_stack_store(&e_stack, op, r_size(op), 0, 0, true, "execstack");
-	uint asize = r_size(op);
+{	os_ptr op1 = op - 1;
+	ref *arefs = op1->value.refs;
+	uint asize = r_size(op1);
 	uint i;
-	ref *rp;
+	ref *rq;
 
-	if ( code < 0 )
-	  return code;
 	/*
+	 * Copy elements from the stack to the array,
+	 * optionally skipping executable nulls.
 	 * Clear the executable bit in any internal operators, and
 	 * convert t_structs and t_astructs (which can only appear
 	 * in connection with stack marks, which means that they will
 	 * probably be freed when unwinding) to something harmless.
 	 */
-	for ( i = 0, rp = op->value.refs; i < asize; i++, rp++ )
-	  switch ( r_type(rp) )
-	    {
-	    case t_operator:
-	      {	uint opidx = op_index(rp);
-		if ( opidx == 0 || op_def_is_internal(op_def_table[opidx]) )
-		  r_clear_attrs(rp, a_executable);
+	for ( i = 0, rq = arefs + asize; rq != arefs; ++i )
+	  { const ref *rp = ref_stack_index(&e_stack, (long)i);
+
+	    if ( r_has_type_attrs(rp, t_null, a_executable) &&
+		 !op->value.boolval
+	       )
+	      continue;
+	    --rq;
+	    ref_assign_old(op1, rq, rp, "execstack");
+	    switch ( r_type(rq) )
+	      {
+	      case t_operator:
+		{ uint opidx = op_index(rq);
+		  if ( opidx == 0 || op_def_is_internal(op_def_table[opidx]) )
+		    r_clear_attrs(rq, a_executable);
+		}
 		break;
-	      }
-	    case t_struct:
-	    case t_astruct:
-	      {	const char *tname =
-		  gs_struct_type_name_string(gs_object_type(imemory,
-							  rp->value.pstruct));
-		make_const_string(rp, a_readonly | avm_foreign,
-				  strlen(tname), (const byte *)tname);
+	      case t_struct:
+	      case t_astruct:
+		{ const char *tname =
+		    gs_struct_type_name_string(
+			gs_object_type(imemory, rq->value.pstruct));
+		  make_const_string(rq, a_readonly | avm_foreign,
+				    strlen(tname), (const byte *)tname);
+		}
 		break;
+	      default:
+		;
 	      }
-	    }
+	  }
+	pop(1);
 	return 0;
 }
 
+/* - .needinput - */
+private int
+zneedinput(register os_ptr op)
+{	return e_NeedInput;	/* interpreter will exit to caller */
+}
+
 /* <obj> <int> .quit - */
 private int
 zquit(register os_ptr op)
-{	check_ostack(2);
+{	check_op(2);
 	check_type(*op, t_integer);
 	return e_Quit;		/* Interpreter will do the exit */
 }
@@ -483,7 +661,7 @@ zcurrentfile(register os_ptr op)
 	if ( esfile != 0 )
 	{
 #ifdef DEBUG
-		/* Check that esfile is valid range. */
+		/* Check that esfile is valid. */
 		ref *efp = zget_current_file();
 		if ( esfile != efp )
 		  { lprintf2("currentfile: esfile=0x%lx, efp=0x%lx\n",
@@ -522,36 +700,31 @@ zget_current_file(void)
 	return 0;
 }
 
-/* ------ Non-operator routines ------ */
-
-/* Test whether we are inside a `stopped'. */
-/* The top level of the interpreter uses this. */
-int
-in_stopped(void)
-{	return count_to_stopped() != 0;
-}
-
 /* ------ Initialization procedure ------ */
 
 BEGIN_OP_DEFS(zcontrol_op_defs) {
 	{"1.cond", zcond},
-	{"0countexecstack", zcountexecstack},
+	{"1.countexecstack", zcountexecstack},
 	{"0currentfile", zcurrentfile},
 	{"1exec", zexec},
-	{"0execstack", zexecstack},
+	{"1.execn", zexecn},
+	{"2.execstack", zexecstack},
 	{"0exit", zexit},
 	{"2if", zif},
 	{"3ifelse", zifelse},
 	{"0.instopped", zinstopped},
+	{"0.needinput", zneedinput},
 	{"4for", zfor},
 	{"1loop", zloop},
 	{"2.quit", zquit},
 	{"2repeat", zrepeat},
-	{"1.stop", zstop},
-	{"2.stopped", zstopped},
+	{"0stop", zstop},
+	{"1.stop", zzstop},
+	{"1stopped", zstopped},
+	{"2.stopped", zzstopped},
 		/* Internal operators */
 	{"1%cond_continue", cond_continue},
-	{"0%execstack_continue", execstack_continue},
+	{"2%execstack_continue", execstack_continue},
 	{"0%for_pos_int_continue", for_pos_int_continue},
 	{"0%for_neg_int_continue", for_neg_int_continue},
 	{"0%for_real_continue", for_real_continue},
@@ -559,27 +732,53 @@ BEGIN_OP_DEFS(zcontrol_op_defs) {
 	{"0%for_fraction_continue", for_fraction_continue},
 	{"0%loop_continue", loop_continue},
 	{"0%repeat_continue", repeat_continue},
+	{"0%stopped_push", stopped_push},
+		/* Operators defined in internaldict */
+		op_def_begin_dict("internaldict"),
+	{"1superexec", zsuperexec},
 END_OP_DEFS(0) }
 
 /* ------ Internal routines ------ */
 
 /* Vacuous cleanup routine */
-int
+private int
 no_cleanup(os_ptr op)
 {	return 0;
 }
 
-/* Count the number of elements down to and including the first 'stopped' */
-/* mark on the e-stack.  Return 0 if there is no 'stopped' mark. */
+/*
+ * Count the number of elements on the exec stack, with or without
+ * the normally invisible elements (*op is a Boolean that indicates this).
+ */
 private uint
-count_to_stopped(void)
+count_exec_stack(os_ptr op)
+{	uint count = ref_stack_count(&e_stack);
+
+	if ( !op->value.boolval )
+	  { uint i;
+	    for ( i = count; i--; )
+	      if ( r_has_type_attrs(ref_stack_index(&e_stack, (long)i),
+				    t_null, a_executable) )
+		--count;
+	  }
+	return count;
+}
+
+/*
+ * Count the number of elements down to and including the first 'stopped'
+ * mark on the e-stack with a given mask.  Return 0 if there is no 'stopped'
+ * mark.
+ */
+private uint
+count_to_stopped(long mask)
 {	uint scanned = 0;
 	STACK_LOOP_BEGIN(&e_stack, ep, used)
 	{	uint count = used;
 		ep += used - 1;
 		for ( ; count; count--, ep-- )
 		  if ( r_is_estack_mark(ep) &&
-		       estack_mark_index(ep) == es_stopped
+		       estack_mark_index(ep) == es_stopped &&
+		       (ep[2].value.intval & mask) != 0
 		     )
 			return scanned + (used - count + 1);
 		scanned += used;
@@ -595,14 +794,16 @@ void
 pop_estack(uint count)
 {	uint idx = 0;
 	uint popped = 0;
+
 	esfile_clear_cache();
 	for ( ; idx < count; idx++ )
-	{	ref *ep = ref_stack_index(&e_stack, idx - popped);
-		if ( r_is_estack_mark(ep) )
-		{	ref_stack_pop(&e_stack, idx + 1 - popped);
-			popped = idx + 1;
-			(*real_opproc(ep))(osp);
-		}
-	}
+	  { ref *ep = ref_stack_index(&e_stack, idx - popped);
+
+	    if ( r_is_estack_mark(ep) )
+	      { ref_stack_pop(&e_stack, idx + 1 - popped);
+	        popped = idx + 1;
+		(*real_opproc(ep))(osp);
+	      }
+	  }
 	ref_stack_pop(&e_stack, count - popped);
 }
