@@ -1,19 +1,24 @@
 /* Copyright (C) 1990, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of Aladdin Ghostscript.
+  This file is part of GNU Ghostscript.
   
-  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
-  or distributor accepts any responsibility for the consequences of using it,
-  or for whether it serves any particular purpose or works at all, unless he
-  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
-  License (the "License") for full details.
+  GNU Ghostscript is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
+  anyone for the consequences of using it or for whether it serves any
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
-  Every copy of Aladdin Ghostscript must include a copy of the License,
-  normally in a plain ASCII text file named PUBLIC.  The License grants you
-  the right to copy, modify and redistribute Aladdin Ghostscript, but only
-  under certain conditions described in the License.  Among other things, the
-  License requires that the copyright notice and this notice be preserved on
-  all copies.
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* zdps1.c */
@@ -289,7 +294,7 @@ rect_get(local_rects *plr, os_ptr op)
 {	int format, code;
 	uint n, count;
 	gs_rect *pr;
-	float rv[4];
+	double rv[4];
 
 	switch ( r_type(op) )
 	   {
@@ -363,7 +368,8 @@ rect_release(local_rects *plr)
 /* <llx> <lly> <urx> <ury> setbbox - */
 int
 zsetbbox(register os_ptr op)
-{	float box[4];
+{	double box[4];
+
 	int code = num_params(op, 4, box);
 	if ( code < 0 ) return code;
 	if ( (code = gs_setbbox(igs, box[0], box[1], box[2], box[3])) < 0 )
