@@ -1,13 +1,24 @@
-/* Copyright (C) 1989, 1995 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
   anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU Ghostscript General Public License for full details.
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* zfile.c */
@@ -19,7 +30,7 @@
 #include "gsstruct.h"			/* for registering root */
 #include "errors.h"
 #include "oper.h"
-#include "estack.h"			/* for filenameforall, .runexec */
+#include "estack.h"			/* for filenameforall, .execfile */
 #include "ialloc.h"
 #include "ilevel.h"			/* %names only work in Level 2 */
 #include "interp.h"			/* gs_errorinfo_put_string prototype */
@@ -32,6 +43,7 @@
 					/* and before files.h */
 #include "files.h"			/* ditto */
 #include "fname.h"			/* ditto */
+#include "main.h"			/* for gs_lib_paths */
 #include "store.h"
 
 /* Import the file_open routine for %os%, which is the default. */
@@ -41,14 +53,10 @@ extern iodev_proc_open_file(iodev_os_open_file);
 int file_open(P6(const byte *, uint, const char *, uint, ref *, stream **));
 
 /* Forward references: other. */
-private int runexec_finish(P1(os_ptr));
-private int runexec_cleanup(P1(os_ptr));
+private int execfile_finish(P1(os_ptr));
+private int execfile_cleanup(P1(os_ptr));
 private stream_proc_report_error(filter_report_error);
 
-/* Imported from gs.c */
-extern const char **gs_lib_paths;	/* search path list, */
-					/* terminated by a null pointer */
-
 /*
  * Since there can be many file objects referring to the same file/stream,
  * we can't simply free a stream when we close it.  On the other hand,
@@ -293,8 +301,9 @@ zstatus(register os_ptr op)
 {	switch ( r_type(op) )
 	{
 	case t_file:
-		make_bool(op, (s_is_valid(fptr(op)) ? 1 : 0));
-		return 0;
+	  {	stream *s;
+		make_bool(op, (file_is_valid(s, op) ? 1 : 0));
+	  }	return 0;
 	case t_string:
 	{	parsed_file_name pname;
 		struct stat fstat;
@@ -330,6 +339,67 @@ zstatus(register os_ptr op)
 
 /* ------ Non-standard extensions ------ */
 
+/* <executable_file> .execfile - */
+private int
+zexecfile(register os_ptr op)
+{	check_type_access(*op, t_file, a_executable | a_read | a_execute);
+	check_estack(4);	/* cleanup, file, finish, file */
+	push_mark_estack(es_other, execfile_cleanup);
+	*++esp = *op;
+	push_op_estack(execfile_finish);
+	return zexec(op);
+}
+/* Finish normally. */
+private int
+execfile_finish(os_ptr op)
+{	check_ostack(1);
+	esp -= 2;
+	execfile_cleanup(op);
+	return o_pop_estack;
+}
+/* Clean up by closing the file. */
+private int
+execfile_cleanup(os_ptr op)
+{	check_ostack(1);
+	*++osp = esp[2];
+	return zclosefile(osp);
+}
+
+/* <dir> <name> .filenamedirseparator <string> */
+int
+zfilenamedirseparator(os_ptr op)
+{	const char *sepr;
+
+	check_read_type(*op, t_string);
+	check_read_type(op[-1], t_string);
+	sepr =
+	  gp_file_name_concat_string((const char *)op[-1].value.const_bytes,
+				     r_size(op - 1),
+				     (const char *)op->value.const_bytes,
+				     r_size(op));
+	make_const_string(op - 1, avm_foreign | a_readonly,
+			  strlen(sepr), (const byte *)sepr);
+	pop(1);
+	return 0;
+}
+
+/* - .filenamelistseparator <string> */
+int
+zfilenamelistseparator(os_ptr op)
+{	push(1);
+	make_const_string(op, avm_foreign | a_readonly, 1,
+			  (const byte *)&gp_file_name_list_separator);
+	return 0;
+}
+
+/* <name> .filenamesplit <dir> <base> <extension> */
+int
+zfilenamesplit(os_ptr op)
+{	check_read_type(*op, t_string);
+	/****** NOT IMPLEMENTED YET ******/
+	return_error(e_undefined);
+}
+
 /* <string> findlibfile <found_string> <file> true */
 /* <string> findlibfile <string> false */
 int
@@ -348,9 +418,13 @@ zfindlibfile(register os_ptr op)
 		pname.iodev = iodev_default;
 	if ( pname.iodev != iodev_default )
 	{	/* Non-OS devices don't have search paths (yet). */
-		code = (*pname.iodev->procs.open_file)(pname.iodev,
-					pname.fname, pname.len, "r", &s,
-					imemory);
+		code =
+		  (pname.fname == NULL ?
+		   (*pname.iodev->procs.open_device)(pname.iodev, "r",
+						     &s, imemory) :
+		   (*pname.iodev->procs.open_file)(pname.iodev,
+						   pname.fname, pname.len, "r",
+						   &s, imemory));
 		if ( code < 0 )
 		{	push(1);
 			make_false(op);
@@ -380,45 +454,22 @@ zfindlibfile(register os_ptr op)
 	return 0;
 }
 
-/* <executable_file> .runexec - */
-private int
-zrunexec(register os_ptr op)
-{	check_type_access(*op, t_file, a_executable | a_read | a_execute);
-	check_estack(4);	/* cleanup, file, finish, file */
-	push_mark_estack(es_other, runexec_cleanup);
-	*++esp = *op;
-	push_op_estack(runexec_finish);
-	return zexec(op);
-}
-/* Finish normally. */
-private int
-runexec_finish(os_ptr op)
-{	check_ostack(1);
-	esp -= 2;
-	runexec_cleanup(op);
-	return o_pop_estack;
-}
-/* Clean up by closing the file. */
-private int
-runexec_cleanup(os_ptr op)
-{	check_ostack(1);
-	*++osp = esp[2];
-	return zclosefile(osp);
-}
-
 /* ------ Initialization procedure ------ */
 
 BEGIN_OP_DEFS(zfile_op_defs) {
 	{"1deletefile", zdeletefile},
+	{"1.execfile", zexecfile},
 	{"2file", zfile},
 	{"3filenameforall", zfilenameforall},
+	{"2.filenamedirseparator", zfilenamedirseparator},
+	{"0.filenamelistseparator", zfilenamelistseparator},
+	{"1.filenamesplit", zfilenamesplit},
 	{"1findlibfile", zfindlibfile},
 	{"2renamefile", zrenamefile},
-	{"1.runexec", zrunexec},
 	{"1status", zstatus},
 		/* Internal operators */
 	{"0%file_continue", file_continue},
-	{"0%runexec_finish", runexec_finish},
+	{"0%execfile_finish", execfile_finish},
 END_OP_DEFS(zfile_init) }
 
 /* ------ Stream opening ------ */
@@ -447,7 +498,8 @@ lib_file_fopen(gx_io_device *iodev, const char *bname,
   const char *ignore_access, FILE **pfile, char *rfname, uint rnamelen)
 {	char fmode[3];			/* r, [b], null */
 	int len = strlen(bname);
-	const char **ppath;
+	const gs_file_path *pfpath = &gs_lib_path;
+	uint pi;
 
 	strcpy(fmode, "r");
 	strcat(fmode, gp_fmode_binary_suffix);
@@ -455,45 +507,35 @@ lib_file_fopen(gx_io_device *iodev, const char *bname,
 	  return (*iodev->procs.fopen)(iodev, bname, fmode, pfile,
 				       rfname, rnamelen);
 	/* Go through the list of search paths */
-	for ( ppath = gs_lib_paths; *ppath != 0; ppath++ )
-	   {	const char *path = *ppath;
+	for ( pi = 0; pi < r_size(&pfpath->list); ++pi )
+	   {	const ref *prdir = pfpath->list.value.refs + pi;
+		const char *pstr = (const char *)prdir->value.const_bytes;
+		uint plen = r_size(prdir);
+		const char *cstr =
+		  gp_file_name_concat_string(pstr, plen, bname, len);
+		int up, i;
+		int code;
 
-		for ( ; ; )
-		   {	/* Find the end of the next path */
-			const char *npath = path;
-			uint plen;
-			const char *cstr;
-			int code;
-			int up, i;
-
-			while ( *npath != 0 && *npath != gp_file_name_list_separator )
-			  npath++;
-			plen = npath - path;
-			cstr = gp_file_name_concat_string(path, plen, bname,
-							  len);
-			/* Concatenate the prefix, combiner, and file name. */
-			/* Do this carefully in case rfname is the same */
-			/* as fname.  (We don't worry about the case */
-			/* where rfname only overlaps fname.) */
-			up = plen + strlen(cstr);
-			if ( up + len + 1 > rnamelen )
-			  return_error(e_limitcheck);
-			for ( i = len + 1; --i >= 0; )
-			  rfname[i + up] = bname[i];
-			memcpy(rfname, (byte *)path, plen);
-			memcpy(rfname + plen, cstr, strlen(cstr));
-			code = (*iodev->procs.fopen)(iodev, rfname, fmode,
-						     pfile, rfname, rnamelen);
-			if ( code >= 0 )
-			  return code;
-			/* strcpy isn't guaranteed to work for overlapping */
-			/* source and destination, so: */
-			if ( rfname == bname )
-			  for ( i = 0; (rfname[i] = rfname[i + up]) != 0; i++ )
-			    ;
-			if ( !*npath ) break;
-			path = npath + 1;
-		   }
+		/* Concatenate the prefix, combiner, and file name. */
+		/* Do this carefully in case rfname is the same */
+		/* as fname.  (We don't worry about the case */
+		/* where rfname only overlaps fname.) */
+		up = plen + strlen(cstr);
+		if ( up + len + 1 > rnamelen )
+		  return_error(e_limitcheck);
+		for ( i = len + 1; --i >= 0; )
+		  rfname[i + up] = bname[i];
+		memcpy(rfname, pstr, plen);
+		memcpy(rfname + plen, cstr, strlen(cstr));
+		code = (*iodev->procs.fopen)(iodev, rfname, fmode,
+					     pfile, rfname, rnamelen);
+		if ( code >= 0 )
+		  return code;
+		/* strcpy isn't guaranteed to work for overlapping */
+		/* source and destination, so: */
+		if ( rfname == bname )
+		  for ( i = 0; (rfname[i] = rfname[i + up]) != 0; i++ )
+		    ;
 	   }
 	return_error(e_undefinedfilename);
 }
@@ -538,25 +580,16 @@ lib_file_open(const char *fname, uint len, byte *cname, uint max_clen,
 }
 
 /* Open a file stream that reads a string. */
-/* (This is currently used only by gs_run_string and the ccinit feature.) */
+/* (This is currently used only by the ccinit feature.) */
 /* The string must be allocated in non-garbage-collectable (foreign) space. */
 int
-file_read_string(const byte *str, uint len, stream *s, ref *pfile)
-{	int space;
+file_read_string(const byte *str, uint len, ref *pfile)
+{	stream *s = file_alloc_stream(imemory, "file_read_string");
+	int space;
+
 	if ( s == 0 )
-	  {	s = file_alloc_stream(imemory, "file_read_string");
-		if ( s == 0 )
-		  return_error(e_VMerror);
-		space = icurrent_space;
-	  }
-	else
-	  {	s->memory = 0;
-		s->report_error = s_no_report_error;
-		s_init_ids(s);
-		s->is_temp = 0;
-		s->next = s->prev = 0;
-		space = avm_foreign;
-	  }
+	  return_error(e_VMerror);
+	space = icurrent_space;
 	sread_string(s, str, len);
 	s->foreign = 1;
 	s->write_id = 0;
@@ -636,11 +669,12 @@ file_open_stream(const char *fname, uint len, const char *file_access,
 int
 filter_open(const char *file_access, uint buffer_size, ref *pfile,
   const stream_procs _ds *procs, const stream_template *template,
-  const stream_state *st, stream **ps)
+  const stream_state *st)
 {	stream *s;
 	uint ssize = gs_struct_type_size(template->stype);
 	stream_state *sst = 0;
 	int code;
+
 	if ( template->stype != &st_stream_state )
 	{	sst = s_alloc_state(imemory, template->stype,
 				    "filter_open(stream_state)");
@@ -655,7 +689,6 @@ filter_open(const char *file_access, uint buffer_size, ref *pfile,
 	}
 	s_std_init(s, s->cbuf, s->bsize, procs,
 		   (*file_access == 'r' ? s_mode_read : s_mode_write));
-	make_stream_file(pfile, s, file_access);
 	s->procs.process = template->process;
 	s->save_close = s->procs.close;
 	s->procs.close = file_close_file;
@@ -665,7 +698,7 @@ filter_open(const char *file_access, uint buffer_size, ref *pfile,
 		sst = (stream_state *)s;
 	}
 	else if ( st != 0 )	/* might not have client parameters */
-		memcpy(sst, st, ssize);
+	  memcpy(sst, st, ssize);
 	s->state = sst;
 	sst->template = template;
 	sst->memory = imemory;
@@ -678,7 +711,7 @@ filter_open(const char *file_access, uint buffer_size, ref *pfile,
 			return code;
 		}
 	}
-	*ps = s;
+	make_stream_file(pfile, s, file_access);
 	return 0;
 }
 
@@ -811,21 +844,20 @@ file_save(void)
 /* Update the file list before a restore. */
 void
 file_restore(const alloc_save_t *save, const gs_memory_t *mem)
-{	stream *s;
+{	stream *prev = 0;
+	stream *s;
 	stream **ps;
 	/* We must be careful to unlink only those streams which */
 	/* were allocated in the VM space being restored. */
 	for ( ps = &file_list; (s = *ps) != 0 && s->save_count == 0; )
 	  if ( s->memory == mem )
-	    *ps = s->next;
+	    { if ( (*ps = s->next) != 0 )
+		(*ps)->prev = prev;
+	    }
 	  else
-	    ps = &s->next;
-	if ( s != 0 )
-	  {	if ( s->prev )
-		  s->prev->next = 0;
-		s->prev = 0;
-		s->save_count--;
-	  }
+	    prev = s, ps = &s->next;
+	if ( s != 0 )			/* i.e., s->save_count != 0 */
+	  s->save_count--;
 	if_debug2('u', "[u]file_restore 0x%lx for 0x%lx\n",
 		  (ulong)file_list, (ulong)save);
 }
