@@ -1,4 +1,4 @@
-/* Copyright (C) 1989, 1995, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
   This file is part of GNU Ghostscript.
   
@@ -80,7 +80,7 @@ zbuildfont3(os_ptr op)
 	if ( code < 0 )
 	  return code;
 	code = build_gs_simple_font(op, &pfont, ft_user_defined,
-				    &st_gs_font_base, &build);
+				    &st_gs_font_base, &build, bf_options_none);
 	if ( code < 0 )
 	  return code;
 	return define_gs_font((gs_font *)pfont);
@@ -110,7 +110,6 @@ zregisterencoding(register os_ptr op)
 }
 
 /* Encode a character. */
-/* (This is very inefficient right now; we can speed it up later.) */
 private gs_glyph
 zfont_encode_char(gs_show_enum *penum, gs_font *pfont, gs_char *pchr)
 {	const ref *pencoding = &pfont_data(pfont)->Encoding;
@@ -144,7 +143,20 @@ typedef const char *const_chars;
 private const_chars
 zfont_glyph_name(gs_glyph index, uint *plen)
 {	ref nref, sref;
-	name_index_ref(index, &nref);
+
+	if ( index >= gs_min_cid_glyph )
+	  { /* Fabricate a numeric name. */
+	    char cid_name[sizeof(gs_glyph) * 3 + 1];
+	    int code;
+
+	    sprintf(cid_name, "%lu", (ulong)index);
+	    code = name_ref((const byte *)cid_name, strlen(cid_name),
+			    &nref, 1);
+	    if ( code < 0 )
+	      return 0;		/* What can we possibly do here? */
+	  }
+	else
+	  name_index_ref(index, &nref);
 	name_string_ref(&nref, &sref);
 	*plen = r_size(&sref);
 	return (const char *)sref.value.const_bytes;
@@ -164,12 +176,21 @@ int
 build_proc_name_refs(build_proc_refs *pbuild,
   const char _ds *bcstr, const char _ds *bgstr)
 {	int code;
-	if ( (code = name_ref((const byte *)bcstr, strlen(bcstr), &pbuild->BuildChar, 0)) < 0 ||
-	     (code = name_ref((const byte *)bgstr, strlen(bgstr), &pbuild->BuildGlyph, 0)) < 0
-	   )
-		return code;
-	r_set_attrs(&pbuild->BuildChar, a_executable);
-	r_set_attrs(&pbuild->BuildGlyph, a_executable);
+
+	if ( !bcstr )
+	  make_null(&pbuild->BuildChar);
+	else
+	  { if ( (code = name_ref((const byte *)bcstr, strlen(bcstr), &pbuild->BuildChar, 0)) < 0 )
+	      return code;
+	    r_set_attrs(&pbuild->BuildChar, a_executable);
+	  }
+	if ( !bgstr )
+	  make_null(&pbuild->BuildGlyph);
+	else
+	  { if ( (code = name_ref((const byte *)bgstr, strlen(bgstr), &pbuild->BuildGlyph, 0)) < 0 )
+	      return code;
+	    r_set_attrs(&pbuild->BuildGlyph, a_executable);
+	  }
 	return 0;
 }
 
@@ -206,10 +227,11 @@ build_gs_font_procs(os_ptr op, build_proc_refs *pbuild)
 /* The caller guarantees that *op is a dictionary. */
 int
 build_gs_primitive_font(os_ptr op, gs_font_base **ppfont, font_type ftype,
-  gs_memory_type_ptr_t pstype, const build_proc_refs *pbuild)
+  gs_memory_type_ptr_t pstype, const build_proc_refs *pbuild,
+  build_font_options_t options)
 {	int painttype;
 	float strokewidth;
-	ref *pcharstrings;
+	ref *pcharstrings = 0;
 	gs_font_base *pfont;
 	font_data *pdata;
 	int code;
@@ -220,17 +242,30 @@ build_gs_primitive_font(os_ptr op, gs_font_base **ppfont, font_type ftype,
 	code = dict_float_param(op, "StrokeWidth", 0.0, &strokewidth);
 	if ( code < 0 )
 	  return code;
-	if ( dict_find_string(op, "CharStrings", &pcharstrings) <= 0 ||
-	    !r_has_type(pcharstrings, t_dictionary)
-	   )
-	  return_error(e_invalidfont);
-	code = build_gs_simple_font(op, &pfont, ftype, pstype, pbuild);
+	if ( dict_find_string(op, "CharStrings", &pcharstrings) <= 0 )
+	  { if ( !(options & bf_CharStrings_optional) )
+	      return_error(e_invalidfont);
+	  }
+	else
+	  { ref *ignore;
+
+	    if ( !r_has_type(pcharstrings, t_dictionary) )
+	      return_error(e_invalidfont);
+	    if ( (options & bf_notdef_required) != 0 &&
+		 dict_find_string(pcharstrings, ".notdef", &ignore) <= 0
+	       )
+	      return_error(e_invalidfont);
+	  }
+	code = build_gs_simple_font(op, &pfont, ftype, pstype, pbuild, options);
 	if ( code != 0 )
 	  return code;
 	pfont->PaintType = painttype;
 	pfont->StrokeWidth = strokewidth;
 	pdata = pfont_data(pfont);
-	ref_assign(&pdata->CharStrings, pcharstrings);
+	if ( pcharstrings )
+	  ref_assign(&pdata->CharStrings, pcharstrings);
+	else
+	  make_null(&pdata->CharStrings);
 	/* Check that the UniqueIDs match.  This is part of the */
 	/* Adobe protection scheme, but we may as well emulate it. */
 	if ( uid_is_valid(&pfont->UID) &&
@@ -245,18 +280,29 @@ build_gs_primitive_font(os_ptr op, gs_font_base **ppfont, font_type ftype,
 /* The caller guarantees that *op is a dictionary. */
 int
 build_gs_simple_font(os_ptr op, gs_font_base **ppfont, font_type ftype,
-  gs_memory_type_ptr_t pstype, const build_proc_refs *pbuild)
-{	float bbox[4];
+  gs_memory_type_ptr_t pstype, const build_proc_refs *pbuild,
+  build_font_options_t options)
+{	double bbox[4];
 	gs_uid uid;
 	int code;
 	gs_font_base *pfont;
 
 	code = font_bbox_param(op, bbox);
-	if ( code < 0 ) return code;
+	if ( code < 0 )
+	  return code;
+	if ( (options & bf_FontBBox_required) &&
+	     bbox[0] == 0 && bbox[1] == 0 && bbox[2] == 0 && bbox[3] == 0
+	   )
+	  return_error(e_invalidfont);
 	code = dict_uid_param(op, &uid, 0, imemory);
-	if ( code < 0 ) return code;
-	code = build_gs_font(op, (gs_font **)ppfont, ftype, pstype, pbuild);
-	if ( code != 0 ) return code;	/* invalid or scaled font */
+	if ( code < 0 )
+	  return code;
+	if ( (options & bf_UniqueID_ignored) && uid_is_UniqueID(&uid) )
+	  uid_set_invalid(&uid);
+	code = build_gs_font(op, (gs_font **)ppfont, ftype, pstype, pbuild,
+			     options);
+	if ( code != 0 )	/* invalid or scaled font */
+	  return code;
 	pfont = *ppfont;
 	pfont->procs.init_fstack = gs_default_init_fstack;
 	pfont->procs.next_char = gs_default_next_char;
@@ -332,13 +378,14 @@ private void get_font_name(P2(ref *, const ref *));
 private void copy_font_name(P2(gs_font_name *, const ref *));
 int
 build_gs_font(os_ptr op, gs_font **ppfont, font_type ftype,
-  gs_memory_type_ptr_t pstype, const build_proc_refs *pbuild)
+  gs_memory_type_ptr_t pstype, const build_proc_refs *pbuild,
+  build_font_options_t options)
 {	ref kname, fname;		/* t_string */
 	ref *pftype;
 	ref *pfontname;
 	ref *pmatrix;
 	gs_matrix mat;
-	ref *pencoding;
+	ref *pencoding = 0;
 	bool bitmapwidths;
 	int exactsize, inbetweensize, transformedchar;
 	int wmode;
@@ -346,16 +393,23 @@ build_gs_font(os_ptr op, gs_font **ppfont, font_type ftype,
 	gs_font *pfont;
 	ref *pfid;
 	ref *aop = dict_access_ref(op);
+
 	get_font_name(&kname, op - 1);
 	if ( dict_find_string(op, "FontType", &pftype) <= 0 ||
-	    !r_has_type(pftype, t_integer) ||
-	    pftype->value.intval != (int)ftype ||
-	    dict_find_string(op, "FontMatrix", &pmatrix) <= 0 ||
-	    read_matrix(pmatrix, &mat) < 0 ||
-	    dict_find_string(op, "Encoding", &pencoding) <= 0 ||
-	    !r_is_array(pencoding)
+	     !r_has_type(pftype, t_integer) ||
+	     pftype->value.intval != (int)ftype ||
+	     dict_find_string(op, "FontMatrix", &pmatrix) <= 0 ||
+	     read_matrix(pmatrix, &mat) < 0
 	   )
-		return_error(e_invalidfont);
+	  return_error(e_invalidfont);
+	if ( dict_find_string(op, "Encoding", &pencoding) <= 0 )
+	  { if ( !(options & bf_Encoding_optional) )
+	      return_error(e_invalidfont);
+	  }
+	else
+	  { if ( !r_is_array(pencoding) )
+	      return_error(e_invalidfont);
+	  }
 	if ( dict_find_string(op, "FontName", &pfontname) > 0 )
 		get_font_name(&fname, pfontname);
 	else
@@ -385,7 +439,7 @@ build_gs_font(os_ptr op, gs_font **ppfont, font_type ftype,
 			  return_error(e_invalidfont);
 			if ( obj_eq(pfont_dict(pfont), op) )
 			  {	*ppfont = pfont;
-				return 0;
+				return 1;
 			  }
 			/*
 			 * This is a re-encoded font, or some other
@@ -407,6 +461,7 @@ build_gs_font(os_ptr op, gs_font **ppfont, font_type ftype,
 		/* Make sure that we allocate the font data */
 		/* in the same VM as the font dictionary. */
 		uint space = ialloc_space(idmemory);
+
 		ialloc_set_space(idmemory, r_space(op));
 		pfont = ialloc_struct(gs_font, pstype,
 				      "buildfont(font)");
@@ -426,17 +481,21 @@ build_gs_font(os_ptr op, gs_font **ppfont, font_type ftype,
 		ref_assign_new(&pdata->dict, op);
 		ref_assign_new(&pdata->BuildChar, &pbuild->BuildChar);
 		ref_assign_new(&pdata->BuildGlyph, &pbuild->BuildGlyph);
-		ref_assign_new(&pdata->Encoding, pencoding);
-		pfont->base = pfont;
+		if ( pencoding )
+		  ref_assign_new(&pdata->Encoding, pencoding);
+		/* Clear the chain pointers so as not to confuse the memory */
+		/* manager if we bail out after returning from here. */
+		pfont->next = pfont->prev = 0;
 		pfont->memory = imemory;
 		pfont->dir = 0;
+		pfont->base = pfont;
 		pfont->client_data = pdata;
 		pfont->FontType = ftype;
 		pfont->FontMatrix = mat;
 		pfont->BitmapWidths = bitmapwidths;
-		pfont->ExactSize = exactsize;
-		pfont->InBetweenSize = inbetweensize;
-		pfont->TransformedChar = transformedchar;
+		pfont->ExactSize = (fbit_type)exactsize;
+		pfont->InBetweenSize = (fbit_type)inbetweensize;
+		pfont->TransformedChar = (fbit_type)transformedchar;
 		pfont->WMode = wmode;
 		pfont->PaintType = 0;
 		pfont->StrokeWidth = 0.0;
