@@ -1,19 +1,24 @@
 /* Copyright (C) 1989, 1991, 1993, 1994, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of Aladdin Ghostscript.
+  This file is part of GNU Ghostscript.
   
-  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
-  or distributor accepts any responsibility for the consequences of using it,
-  or for whether it serves any particular purpose or works at all, unless he
-  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
-  License (the "License") for full details.
+  GNU Ghostscript is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
+  anyone for the consequences of using it or for whether it serves any
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
-  Every copy of Aladdin Ghostscript must include a copy of the License,
-  normally in a plain ASCII text file named PUBLIC.  The License grants you
-  the right to copy, modify and redistribute Aladdin Ghostscript, but only
-  under certain conditions described in the License.  Among other things, the
-  License requires that the copyright notice and this notice be preserved on
-  all copies.
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* zht.c */
@@ -187,8 +192,9 @@ screen_sample(register os_ptr op)
 /* Continuation procedure for processing sampled pixels. */
 private int
 set_screen_continue(register os_ptr op)
-{	float value;
-	int code = num_params(op, 1, &value);
+{	double value;
+	int code = real_param(op, &value);
+
 	if ( code < 0 ) return code;
 	code = gs_screen_next(senum, value);
 	if ( code < 0 ) return code;
@@ -218,10 +224,11 @@ screen_cleanup(os_ptr op)
 /* Get parameters for a single screen. */
 int
 zscreen_params(os_ptr op, gs_screen_halftone *phs)
-{	float fa[2];
+{	double fa[2];
 	int code = num_params(op - 1, 2, fa);
+
 	if ( code < 0 )
-		return code;
+	  return code;
 	check_proc(*op);
 	phs->frequency = fa[0];
 	phs->angle = fa[1];
