@@ -1,19 +1,24 @@
 /* Copyright (C) 1990, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of Aladdin Ghostscript.
+  This file is part of GNU Ghostscript.
   
-  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
-  or distributor accepts any responsibility for the consequences of using it,
-  or for whether it serves any particular purpose or works at all, unless he
-  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
-  License (the "License") for full details.
+  GNU Ghostscript is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
+  anyone for the consequences of using it or for whether it serves any
+  particular purpose or works at all, unless he says so in writing.  Refer to
+  the GNU General Public License for full details.
   
-  Every copy of Aladdin Ghostscript must include a copy of the License,
-  normally in a plain ASCII text file named PUBLIC.  The License grants you
-  the right to copy, modify and redistribute Aladdin Ghostscript, but only
-  under certain conditions described in the License.  Among other things, the
-  License requires that the copyright notice and this notice be preserved on
-  all copies.
+  Everyone is granted permission to copy, modify and redistribute GNU
+  Ghostscript, but only under the conditions described in the GNU General
+  Public License.  A copy of this license is supposed to have been given to
+  you along with GNU Ghostscript so you can know your rights and
+  responsibilities.  It should be in a file named COPYING.  Among other
+  things, the copyright notice and this notice must be preserved on all
+  copies.
+  
+  Aladdin Enterprises is not affiliated with the Free Software Foundation or
+  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
+  does not depend on any other GNU software.
 */
 
 /* zupath.c */
@@ -100,12 +105,15 @@ zinustroke(os_ptr op)
 	int spop, npop;
 	gx_device hdev;
 
-	if ( code < 0 ) return code;
-	if ( (spop = upath_stroke(op)) < 0 ||
-	     (npop = in_path(op - spop, op, &hdev)) < 0
-	   )
+	if ( code < 0 )
+	  return code;
+	if ( (spop = upath_stroke(op)) < 0 )
 	   {	gs_grestore(igs);
-		return code;
+		return spop;
+	   }
+	if ( (npop = in_path(op - spop, op, &hdev)) < 0 )
+	   {	gs_grestore(igs);
+		return npop;
 	   }
 	code = gs_stroke(igs);
 	return in_upath_result(op, npop + spop, code);
@@ -183,7 +191,7 @@ private int
 in_path(os_ptr oppath, os_ptr op, gx_device *phdev)
 {	int code = gs_gsave(igs);
 	int npop;
-	float uxy[2];
+	double uxy[2];
 
 	if ( code < 0 )
 	  return code;
@@ -192,6 +200,7 @@ in_path(os_ptr oppath, os_ptr op, gx_device *phdev)
 	   {	/* Aperture is a single pixel. */
 		gs_point dxy;
 		gs_fixed_rect fr;
+
 		gs_transform(igs, uxy[0], uxy[1], &dxy);
 		fr.p.x = fixed_floor(float2fixed(dxy.x));
 		fr.p.y = fixed_floor(float2fixed(dxy.y));
@@ -222,6 +231,7 @@ in_path(os_ptr oppath, os_ptr op, gx_device *phdev)
 	/* Install the hit detection device. */
 	gx_set_device_color_1(igs);
 	*phdev = hit_device;
+	phdev->width = phdev->height = max_int;
 	gx_device_fill_in_procs(phdev);
 	gx_set_device_only(igs, phdev);
 	return npop;
@@ -385,16 +395,20 @@ zustroke(register os_ptr op)
 /* <userpath> <matrix> ustrokepath - */
 private int
 zustrokepath(register os_ptr op)
-{	int code = gs_gsave(igs);
-	int npop;
-	if ( code < 0 ) return code;
+{	gx_path save;
+	int code, npop;
+
+	/* Save and reset the path. */
+	save = *igs->path;
+	gx_path_reset(igs->path);
 	if ( (code = npop = upath_stroke(op)) < 0 ||
-	     (code = gs_strokepath(igs)) < 0 ||
-	     (code = gs_upmergepath(igs)) < 0
+	     (code = gs_strokepath(igs)) < 0
 	   )
-		DO_NOTHING;
-	gs_grestore(igs);
-	if ( code < 0 ) return code;
+	  { gs_newpath(igs);	/* release partial path */
+	    *igs->path = save;
+	    return code;
+	  }
+	gx_path_release(&save);
 	pop(npop);
 	return 0;
 }
